/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_CLICAPPLICATION_HPP_
#define _NMM_CLICAPPLICATION_HPP_

#include "nmm/utils/NamedObject.hpp"

#include "nmm/base/registry/GraphDescription.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/resourcemgr/RMConfigManager.hpp"

#include "nmm/utils/gdparse/GDParse.hpp"

#include "nmm/utils/net/NetReactor.hpp"

#include "nmm/misc/timer/Timer.hpp"

#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"
#include "nmm/base/graph/CompositeNode.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"

#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"

namespace NMM {

class ISeekable;

class ClicApplication  : public NamedObject, public NetHandler, private TimerListener {
public:
  ClicApplication(int argc, char** argv);

  virtual ~ClicApplication();

  /**
   * Sets up graph description
   * (1) setup of out-of-band communication, 
   * (2,3) sets input and output files
   * (4) triggers calling of out-of-band events
   * (5) realizes graph (mainly connection setup)
   * (6) setup of synchronization (if enabled)
   * (7,8) activate and start graph 
   * (7) register listener (FIXME: should this be done before starting the graph?)
   */
  void setupGraph();

        GraphDescription& getGraphDescription()       { return __gd; }
  const GraphDescription& getGraphDescription() const { return __gd; }

        NMMApplication*   getNMMApplication()         { return __app; }
  const NMMApplication*   getNMMApplication()   const { return __app; }

        void waitUntilKeyPressed(const char c);

  virtual NetResult receiveEvent(int socket, short event, short revents);
  virtual bool requireThread() const { return false; }
protected:
  void parseArguments(int argc, char**argv);
  Result eofHandler();
  Result finishedHandler();

  void usage(const string& appname);
  
  Result keyReleased(string& key);

  void nextInputFile();

  void printResourceStatistics();
  void printCPUInfo();
  
  void buildGraph();
  void destroyGraph();
  
  void seekForward();
  void seekBackward();
  void pause();
  void play();
  void togglePause();
  void toggleFullscreen();
  void quit();
  
  void enableDisplayEvents();
  void disableDisplayEvents();
  
private:
  virtual void timeout(Timer* );
  
private:
  GraphDescription  __gd;                 ///< stores the GraphDescription
  NMMApplication*   __app;                ///< Pointer to the NMMApplication
  string            __gd_file;            ///< filename for the grahdescription file.
  string            __graph_url;          ///< string that holds url for GraphURL object and GraphBuilder2
  CompositeNode*    __composite;          ///< composite node for graph builder
  ISeekable*        __seekable;           ///< Interface for seeking
  INode*            __playback;           ///< playback node for graph builder
  string            __playback_location;  ///< location for playback node
  INode*            __display;            ///< display node for graph builder
  string            __display_location;   ///< location for display node
  MultiAudioVideoSynchronizer* __av_sync; ///< synchronizer for graph builder
  IMultiAudioVideoSynchronizer* __synchronizer; ///< synchronizer interface
  vector<string>    __input_files;        ///< input-filename which is set to source nodes with IFilename
  size_t            __curr_input_file;    ///< number of the current input file
  string            __output_file;        ///< output-filename which is set to sink nodes with IFilename
  bool              __sync;               ///< enables/disables synchronization
  bool              __prefetching;        ///< enables/disables prefetching
  bool              __resource;           ///< enables/disables resource measurements
  int               __eof_counter;        ///< counts how many eof events occurs
  int               __sink_nodes;         ///< stores how many SinkeNodes are used

  IVideoDisplay_var __video_display;      ///< Video display
  bool              __fullscreen;         ///< Fullscreen mode
  unsigned int      __delay;              ///< specifies the delay in seconds between different input files (e.g. for slideshows)
  bool              __exit_flag;          ///< is true if the clic should be terminate
  bool              __infinite_loop;      ///< is true if clic should terminate after playing all files

  ThreadMutex       __cond_mutex;         ///< thread mutex for sync. access to the __condition
  ThreadCondition   __condition;          ///< condition to wait for an input from stdin.

  int               __timeout;            ///< If timeout (in sec) is > 0, clic automatically exit after this time. 
  Timer             __timer;              ///< This timer is used to terminate clic. The attribute __timout must be > 0.

  IResourceManager_var  __resource_mgr;   ///< resource manager, only set if __resource is true
  bool              __enable_registry;    ///< if this flag is true, enable server registry
  int               __port;               ///< port used for server registry
  
  bool              __paused;             ///< Playback is paused
  bool              __quit;               ///< Quit application
  EDObject*         __x_listener;         ///< event listener for x events
};
}

#endif 


