/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include "cd.hpp"

// avoid forward declarations
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"

// used interfaces
#include "nmm/interfaces/general/ICopy.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/file/ITrack.hpp"

#include "nmm/base/connect.hpp" // we want do connect :-)
#include "nmm/utils/utils.hpp"  // for string + int

// we want to be in:
namespace NMM {

  CdState::CdState(string id, MMBoxApplication* _app) 
    : MMBoxState(id, _app),
      state(STOPPED),
      actual_track_name("Playback stopped"),
      current_progress(0),
      can_seek_time(false),
      startTrack_listener(this, &CdState::startTrack),
      setProgress_listener(this, &CdState::setProgress),
      cdda(NULL),
      copy(NULL),
      icdda(NULL),
      icddb(NULL),
      iseek(NULL),
      iprogress(NULL),
      text(NULL),
      plane(NULL),
      border(NULL),
      scroll(NULL),
      header(NULL),
      win(NULL)
  {
    // we initialize ALL members explicit and reasonable via colon
    // after the super class:
    // some variables (e.g. Mutexes, ThreadConditions or EDObjects)
    // must be initialized like this, so to be uniform, we initialize
    // all members like this!
    // Attention with reordering warnings from gcc!!!!
  }

  CdState::~CdState() {
    // nothing to do here, all allocated resourced are freed in doKill!
  }


  /******************** MMBoxState interface ********************/
  Result CdState::doCreate(string) {
    GLOBAL_DEBUG_STREAM("CdState::doCreate()" << endl);

    // init the background
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        GLOBAL_ERROR_STREAM("Could not create default Graph!" << endl);
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        GLOBAL_ERROR_STREAM("Could not start default Graph!" << endl);
        return FAILURE;
      }
    }

    // set up osd as early as possible, to avoid empty windows
    // by using the Playback-Control symbols first of all osd widgets,
    // we force them to be in the topmost layer above everything else!
    app->getOsdManager()->displaySymbol("STOP");
    createWindow();

    // read config values
    bool do_cddb_query(true);
    if (!app->getMMBoxConfig()->getValue("cddb", do_cddb_query)) {
      do_cddb_query = true;
    }
    string cddb_server("freedb.freedb.org");
    if (!app->getMMBoxConfig()->getValue("cddbserver",cddb_server)) {
      cddb_server = "freedb.freedb.org";
    }

    try {
      /*Begin> Create requests for needed CDDANode */
      NodeDescription cdda_req;
      // we request a Node, capable of reading AudioCDs
      cdda_req.addInterface(ICDDA::ICDDA_type);
      /*End< */
   
      /*Begin> Get the ClientRegistry and request Nodes */
      ClientRegistry& registry = app->getNMMApplication()->getRegistry();
      {
        RegistryLock lock(registry);
     
        list<Response> cdda_response = registry.initRequest(cdda_req);
        list<Response> copy_response =
          // request CopyNode by name
          registry.initRequest(NodeDescription("CopyNode"));

        if (cdda_response.empty() ||
            copy_response.empty()) {
          GLOBAL_ERROR_STREAM("Size of cdda_response " << cdda_response.size() << endl);
          GLOBAL_ERROR_STREAM("Size of copy_response " << copy_response.size() << endl);
          // a detailed Exception
          throw Exception(string("A required node is not available\n") +
                          string("Size of cdda_response ") + cdda_response.size() + string("\n") +
                          string("Size of copy_response ") + copy_response.size());
        }
        cdda = registry.requestNode(cdda_response.front());
        copy = registry.requestNode(copy_response.front());
      }
      /*End< */
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("CdState::create(): requesting failed!" << endl);
      GLOBAL_ERROR_STREAM(e << endl);

      // cleanup allocated stuff
      cleanup();
      app->getOsdManager()->displayMessage("Error", "Cannot start CDPlayer.", "", "OK");
      changeState("OSDInputState");
      app->getOsdManager()->update();
      return SUCCESS;
    } catch (...) {
      cerr << "Something strang went wrong while requesting in CdState::doCreate()" << endl;
      cleanup();
      return FAILURE;
    }

    //configure the nodes
    icdda = cdda->getParentObject()->getCheckedInterface<ICDDA>();
    icdda->setRepeat(true);
    icddb = cdda->getParentObject()->getInterface<ICDDB>();

    if (icddb) {
      // we might got a Source not capable of CDDB
      icddb->setCDDBServer(cddb_server,8880);
      icddb->setCDDB(do_cddb_query);
    }
    
    // store interfaces for later use
    iprogress = cdda->getParentObject()->getInterface<IProgress>();
    iseek = cdda->getParentObject()->getInterface<ISeekable>();
    can_seek_time = (iseek && iseek->hasTimeSeek());

    // we only need the interface once: store it locally and let the
    // auto-ptr delete it for us
    ICopy_var icopy(copy->getParentObject()->getCheckedInterface<ICopy>());
    icopy->setNumberOfOutputs(2);
    /*End< */
 
    // just now for watching if someone does not comply with our
    // convenience
    if (!app->getOsdNode()->isStarted()) {
      cerr << "OSDNode has unexpected state: " << app->getOsdNode()->getState() << endl;
    }
  
    // just now for watching if someone does not comply with our
    // convenience
    if (!app->getAudioNode()->isInitialized()) {
      cerr << "AudioNode has unexpected state: " << app->getOsdNode()->getState() << endl;
      app->getAudioNode()->reachInitialized();
    }

    // initialize & connect the nodes
    try {
      cdda->init();
      copy->init();

      cdda->initOutput();
      connect(cdda, copy);
      cdda->activate();

      copy->initOutput();
      // we want to connect to the OSDNode, so bring it to the right
      // state
      app->getOsdNode()->reachInitialized();
      connect(copy, string("0"), app->getAudioNode());
      connect(copy, string("1"), app->getOsdNode(), string("stream2"));
      copy->activate();

      app->getOsdNode()->initOutput();
      app->getOsdNode()->activate();

      app->getAudioNode()->initOutput();
      app->getAudioNode()->activate();
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("CdState::doCreate(): connecting failed!" << endl);
      GLOBAL_ERROR_STREAM(e << endl);

      // cleanup allocated stuff
      cleanup();
      app->getOsdManager()->displayMessage("Error", "Cannot start CDPlayer.", "", "OK");
      changeState("OSDInputState");
      app->getOsdManager()->update();
      return SUCCESS;
    } catch (...) {
      cerr << "Something strang went wrong while connecting in CdState::doCreate()" << endl;
      cleanup();
      return FAILURE;
    }

    // register event listener
    app->getDispatcher()->registerEventListener(ITrack::startTrack_event, &startTrack_listener);
    if (iprogress != NULL) {
      iprogress->sendProgressInformation(true);
      app->getAudioNode()->getParentObject()->registerEventListener(IProgressListener::setProgress_event, &setProgress_listener);
    }

    // fill up osd
    text->deleteAllLines();
    if (icddb) {
      header->setText(icddb->getDiscArtist() + " - " + icddb->getDiscTitle());
      actual_track_name = icddb->getTrackName(1);
      for(int i=1; i <= icdda->getTotalTracks(); i++) {
        text->addLine(icddb->getTrackName(i));
      }
    } else {
      header->setText("unknown artist - unknown title");
      actual_track_name = string("Track 1");
      for(int i=1; i <= icdda->getTotalTracks(); i++) {
        text->addLine(string("Track ") + i);
      }
    }
    text->selectLine(1);
    win->update();
    app->getOsdManager()->displaySymbol("PLAY");
    app->getOsdManager()->displayProgressBar(actual_track_name);
    app->getOsdManager()->displayTime();
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->update();

    cdda->start();
    copy->start();
    app->getOsdNode()->start(); 
    app->getAudioNode()->start();
    state = PLAYING; // by setting the state always correctly, we
                     // don't have to test the states of the running
                     // Nodes!

    GLOBAL_DEBUG_STREAM("CdState::doCreate() == SUCCESS" << endl);
    return SUCCESS;
  }

  Result CdState::doKill() {
    GLOBAL_DEBUG_STREAM("CdState::doKill()" << endl);

    if (!cdda) {
      // when we do not have a CDDANode here, we must have had
      // problems and have displayed a error-message, so we only have
      // to disconnect the default graph
      GLOBAL_DEBUG_STREAM("Recovering from Error-Message!" << endl);
    } else {
      app->getOsdManager()->hideProgressBar();
      app->getOsdManager()->hideTime();
      app->getOsdManager()->hideSymbol();

      GLOBAL_DEBUG_STREAM("deregistering start track event" << endl);
      app->getDispatcher()->removeEventListener(&startTrack_listener);
      if (iprogress != NULL) {
        iprogress->sendProgressInformation(false);
        app->getAudioNode()->getParentObject()->removeEventListener(&setProgress_listener);
      }
    }

    // in all cases: cleanup & disconnect the default graph
    cleanup();
    // this leaves the osd in INITIALIZED where the other MMBoxStates
    // expect it.
    if (isBackgroundFree()) {
      app->disconnectDefaultGraph();
    }

    GLOBAL_DEBUG_STREAM("CdState::doKill() == SUCCESS" << endl);

    return SUCCESS;
  }

  Result CdState::doInitialize(string) {
    GLOBAL_DEBUG_STREAM("CdState::doInitialize()" << endl);

    if (!cdda) {
      // when we do not have a CDDANode here, we must have had
      // problems and have displayed a error-message, so we have
      // nothing to do init as we will receive a message (with
      // receiveMessage()) and will deactive ourselve then!
      GLOBAL_DEBUG_STREAM("Displaying Error-Message, doing nothing!" << endl);
      return SUCCESS;
    }

    // init the background
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph( background )) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }    
    /* stop and flush nodes */
    copy->reachOutputInitialized();
    app->getOsdNode()->reachInitialized();
    connect(copy, string("1"), app->getOsdNode(), string("stream2"));
    
    /* reactivate copy node */
    copy->activate();
    if (state == PLAYING) {
      // only start copy node (again) if we are in state PLAYING!
      copy->start();
    }
    app->getOsdNode()->reachStarted();
    
    win->setVisible(true); 
    win->update();
    app->getOsdManager()->displayTime();
    app->getOsdManager()->displayProgressBar(actual_track_name);
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    app->getOsdManager()->updateProgress();
    // we choose the right symbol upon our stored state
    switch (state) {
    case PLAYING:
      app->getOsdManager()->displaySymbol("PLAY");
      break;
    case PAUSED:
      app->getOsdManager()->displaySymbol("PAUSE");
      break;
    case STOPPED:
      app->getOsdManager()->displaySymbol("STOP");
      break;
    }
    app->getOsdManager()->update();
    
    GLOBAL_DEBUG_STREAM("CdState::doInitialize() == SUCCESS" << endl);
    
    return SUCCESS;
  }

  Result CdState::doUninitialize() {
    GLOBAL_DEBUG_STREAM("CdState::doUninitialize()" << endl);

    if (!cdda) {
      // when we do not have a CDDANode here, we must have had
      // problems and display a error-message, so we have
      // nothing to do here!
      GLOBAL_DEBUG_STREAM("Displaying Error-Message, doing nothing!" << endl);
      return SUCCESS;
    }

    // hide osd
    app->getOsdManager()->hideTime();
    app->getOsdManager()->hideProgressBar();
    app->getOsdManager()->hideSymbol();
    win->setVisible(false);
  
    /* stop and flush nodes */
    copy->reachOutputInitialized();
    app->getOsdNode()->reachInitialized();
  
    IInputJack_var osdin_jack(app->getOsdNode()->getInputStream("stream2"));
    if (osdin_jack.get() && osdin_jack->isConnected()) {
      disconnect_input(app->getOsdNode(),string("stream2"));
    }

    
    /* reactivate copy node */
    copy->activate();
    if (state == PLAYING) {
      // only start copy node (again) if we are in state PLAYING!
      copy->start();
    }
    app->getOsdNode()->reachStarted();
    if (isBackgroundFree()) { 
      // we go in the background so:
      app->disconnectVideoGraph();
    }
    //}
    
    GLOBAL_DEBUG_STREAM("CdState::doUninitialize() == SUCCESS" << endl);
    
    return SUCCESS;
  }

  /* Switch to last state */
  Result CdState::Back() {
    GLOBAL_DEBUG_STREAM("CdState::Back()" << endl);
    GLOBAL_DEBUG_STREAM("Switch to stateid: " << lastid << endl);

    return goToLastState();
  }

  Result CdState::Pause() {
    GLOBAL_DEBUG_STREAM("CdState::Pause()" << endl);

    switch (state) {
    case PLAYING:
      // pause when playing
      cdda->stop();
      copy->stop();
      app->getAudioNode()->stop();
      state = PAUSED;

      app->getOsdManager()->displaySymbol("PAUSE");
      app->getOsdManager()->update();
      break;

    case PAUSED:
      // play when pausing
      cdda->start();
      copy->start();
      app->getAudioNode()->start();
      state = PLAYING;

      app->getOsdManager()->displaySymbol("PLAY");
      app->getOsdManager()->update();
      break;

    case STOPPED:
      // nothing to be done here!
      break;
    }

    GLOBAL_DEBUG_STREAM("CdState::Pause() == SUCCESS" << endl);

    return SUCCESS;
  }

  Result CdState::Up() {
    GLOBAL_DEBUG_STREAM("CdState::Up()" << endl);

    text->selectUp();
    win->update();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result CdState::Down() {
    GLOBAL_DEBUG_STREAM("CdState::Down()" << endl);

    text->selectDown();
    win->update();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result CdState::Return() {
    GLOBAL_DEBUG_STREAM("CdState::Return()" << endl);

    switch (state) {
    case PLAYING:
    case STOPPED:
      // select track when PLAYING or STOPPED
      setTrack(text->getSelectedLine());
      break;
    case PAUSED:
      // "Return" is the Play-Button on IR. If you pause and later
      // press play, you don't want to play the selected song but to
      // continue the paused playback!
      // calling Pause() will do the right stuff!
      Pause();
      break;
    }

    return SUCCESS;
  }

  Result CdState::Rewind() {
    GLOBAL_DEBUG_STREAM("CdState::Rewind()" << endl);

    if (state == PLAYING) {
      app->getOsdManager()->displaySymbol("REW");
      app->getOsdManager()->update();

      if (iseek) {
        if (can_seek_time) {
          // a little bit more to compensate the data in the queue
          iseek->seekTime(Interval(-32, 0));
        } else {
          // seekingn in percent should work always
          iseek->seekPercent(Rational(-7,100));
        }
      }
      // PLAY symbol will be displayed either from startTrack() or
      // setProgress(). Like this, FF is displayed as long as Graph is
      // seeking (CD might take longer to seek!)
    }

    return SUCCESS;
  }

  Result CdState::Forward() {
    GLOBAL_DEBUG_STREAM("CdState::Forward()" << endl);

    if (state == PLAYING) {
      app->getOsdManager()->displaySymbol("FF");
      app->getOsdManager()->update();

      if (iseek) {
        if (can_seek_time) {
          iseek->seekTime(Interval(30, 0));
        } else {
          // seeking in percent should work always
          iseek->seekPercent(Rational(5,100));
        }
      }
      // PLAY symbol will be displayed either from startTrack() or
      // setProgress(). Like this, FF is displayed as long as Graph is
      // seeking (CD might take longer to seek!)
    }

    return SUCCESS;
  }

  Result CdState::Stop() {
    GLOBAL_DEBUG_STREAM("CdState::Stop()" << endl);

    if (state == PLAYING) {
      app->getAudioNode()->stop();
      copy->stop();
      cdda->stop();
    }

    app->getAudioNode()->flush();
    copy->flush();
    cdda->flush();
    state = STOPPED; // don't forget to update our internal state
    current_progress = Rational(0);

    actual_track_name = "Playback stopped";
    app->getOsdManager()->resetTime();
    app->getOsdManager()->displayProgressBar(actual_track_name);
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->hideSymbol();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result CdState::receiveMessage(string message) {
    GLOBAL_DEBUG_STREAM("CdState::receiveMessage(" << message << ")" << endl);

    app->getOsdManager()->deleteMessage();

    deactivate();

    goToLastState();

    GLOBAL_DEBUG_STREAM("CdState::receiveMessage(" << message << ") == SUCCESS" << endl);

    return SUCCESS;
  }
  /******************** MMBoxState interface ********************/


  /******************** EventListener methods ********************/
  Result CdState::setProgress(u_int64_t& val, u_int64_t& max) {
    current_progress = Rational(val, max);
    GLOBAL_DEBUG_STREAM("CdState::setProgress(" << current_progress << ")" << endl);

    if (hasFocus()) {
      // no osd updates if we don't have the focus
      app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
      app->getOsdManager()->setProgress(current_progress.getNumerator());
      app->getOsdManager()->updateProgress();
      app->getOsdManager()->displaySymbol("PLAY");
      app->getOsdManager()->update();
    }

    return SUCCESS;
  }

  Result CdState::startTrack(string& name) {
    GLOBAL_DEBUG_STREAM("CdState::startTrack(" << name << ")" << endl);

    actual_track_name = name;
    // when a new track is started we can reset the progress
    current_progress = Rational(0);

    if (hasFocus()) {
      // no osd updates if we don't have the focus
      app->getOsdManager()->displayProgressBar(actual_track_name);
      app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
      app->getOsdManager()->setProgress(current_progress.getNumerator());
      app->getOsdManager()->updateProgress();
      app->getOsdManager()->displaySymbol("PLAY");
      app->getOsdManager()->update();
      app->syncReset();
    }

    return SUCCESS;
  }
  /******************** EventListener methods ********************/


  /******************** private helpers ********************/
  Result CdState::setTrack(unsigned int no) {
    GLOBAL_DEBUG_STREAM("CdState::setTrack(" << no << ")" << endl);

    if (hasFocus()) {
      app->getOsdManager()->hideSymbol();
      app->getOsdManager()->update();
    }

    if (1 > no || no > (unsigned int) icdda->getTotalTracks() ) {
      GLOBAL_ERROR_STREAM("not setting new track because wrong number"<<endl);
      return FAILURE;
    }

    if (state == PLAYING) {
      app->getAudioNode()->stop();
      copy->stop();
      cdda->stop();
    }

    app->getAudioNode()->flush();
    copy->flush();
    cdda->flush();

    Result switchtrack = FAILURE;
    try {
      switchtrack = icdda->setTrack(no);
    }
    catch(...) {
      GLOBAL_ERROR_STREAM("Could not switch to Track #" << no << endl);
    }

    cdda->start();
    copy->start();
    app->getAudioNode()->start();
    state = PLAYING;
    current_progress = Rational(0);

    if (FAILURE == switchtrack) {
      return FAILURE;
    }

    if (hasFocus()) {
      if (icddb) {
        actual_track_name = icddb->getTrackName(no);
      } else {
        actual_track_name = string("Track ") + no;
      }
      app->getOsdManager()->displayProgressBar(actual_track_name);
      app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
      app->getOsdManager()->setProgress(current_progress.getNumerator());
      app->getOsdManager()->updateProgress();
      app->getOsdManager()->displaySymbol("PLAY");
      app->getOsdManager()->update();
      app->getOsdManager()->createTime(true);
    }

    return SUCCESS;
  }

  void CdState::cleanup() {
    GLOBAL_DEBUG_STREAM("CdState::cleanup()" << endl);

    deleteWindow();

    // make sure, AudioNode is flushed!
    if (app->getAudioNode()->isStarted()) {
      if (copy && copy->isStarted()) {
        copy->stop();
      }
      app->getAudioNode()->stop();
      app->getAudioNode()->flush();
    }

    // bring AudioNode in the right state for disconnecting
    // this is also the in which a MMBoxState should leave it after having
    // used it!
    app->getAudioNode()->reachInitialized();

    if (copy) {
      bool copy0_connected(copy->isOutputConnected("0"));
      bool copy1_connected(copy->isOutputConnected("1"));
      if (copy0_connected || copy1_connected) {
        copy->reachOutputInitialized();
        if (copy0_connected) {
          disconnect_output(copy,"0");
        }
        if (copy1_connected) {
          // we want to disconnect from OSDNode, so bring it to the
          // right state
          app->getOsdNode()->reachInitialized();
          disconnect_output(copy,"1");
        }
      }
      copy->reachInitialized();
    }
  
    if (cdda && cdda->isOutputConnected()) {
      cdda->reachOutputInitialized();
      disconnect_output(cdda);
    }
  
    if (copy) {
      copy->reachConstructed();
      app->getNMMApplication()->getRegistry().releaseNode(*copy);
      delete copy;
      copy = NULL;
    }

    if (cdda) {
      cdda->reachConstructed();
      app->getNMMApplication()->getRegistry().releaseNode(*cdda);
      delete cdda;
      cdda = NULL;
    
      if (iprogress != NULL) {
        delete iprogress;
        iprogress = NULL;
      }
    
      if (iseek != NULL) {
        delete iseek;
        icdda = NULL;
      }

      if (icdda != NULL) {
        delete icdda;
        icdda = NULL;
      }

      if (icddb != NULL) {
        delete icddb;
        icddb = NULL;
      }
    }
  
    // bring OSD back to started as the default graph is still
    // connected
    app->getOsdNode()->reachStarted();

    GLOBAL_DEBUG_STREAM("CdState::cleanup() return" << endl);
  }

  void CdState::createWindow() {
    GLOBAL_DEBUG_STREAM("CdState::createWindow()" << endl);

    deleteWindow();

    text = app->getOSDWidgetFactory().createTextView();
    text->setFont("medium");
    text->setSelectionColor(0xff0080ff);
  
    scroll = app->getOSDWidgetFactory().createScrollDecorator(*text);
  
    plane = app->getOSDWidgetFactory().createPlaneDecorator(*scroll);
    plane->setPlaneColor(0x80ff0000);
  
    border = app->getOSDWidgetFactory().createBorderDecorator(*plane);
  
    header = app->getOSDWidgetFactory().createHeaderDecorator(*border);
    header->setFont("medium");
    header->setText("Accessing CDRom...");

    win = app->getOsdManager()->createWindow();
    win->setDimension(620, 350);
    win->setPosition(50, 50);
    win->setContent(*header);
  
    win->update();
    app->getOsdManager()->update();

    GLOBAL_DEBUG_STREAM("CdState::createWindow() return" << endl);
  }

  void CdState::deleteWindow() {
    GLOBAL_DEBUG_STREAM("CdState::deleteWindow()" << endl);

    if (win) {
      app->getOsdManager()->deleteWindow(*win);
      delete win;
      win = NULL;
    }
  
    if (text) {
      app->getOsdManager()->deleteWidget(*text);
      delete text;
      text = NULL;
    }

    if (plane) {
      app->getOsdManager()->deleteWidget(*plane);
      delete plane;
      plane = NULL;
    }

    if (border) {
      app->getOsdManager()->deleteWidget(*border);
      delete border;
      border = NULL;
    }

    if (scroll) {
      app->getOsdManager()->deleteWidget(*scroll);
      delete scroll;
      scroll = NULL;
    }

    if (header) {
      app->getOsdManager()->deleteWidget(*header);
      delete header;
      header = NULL;
    }

    GLOBAL_DEBUG_STREAM("CdState::deleteWindow() return" << endl);
  }
  /******************** private helpers ********************/

} // namespace NMM
