/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include"config.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/ProxyObject.hpp"

using namespace NMM;

ConfigState::ConfigState( string id, MMBoxApplication* _app) : 
  MMBoxState(id, _app),
  manager(_app->getOsdManager()),
  factory(manager->getParentObject()->getCheckedInterface<IWidgetFactory>()),
  dlistunit(NULL) {

  config   = app->getMMBoxConfig();
}

ConfigState::~ConfigState(){
  if (factory) {
    delete factory;
    factory = NULL;
  }

  if ( dlistunit ) {
    delete dlistunit;
  }
}

void ConfigState::deleteWindow() {
  dlistunit->hide();
}

void ConfigState::createWindow() {
  deleteWindow();

  dlistunit->setHeader( "MMBox Configuration" );

  dlistunit->update();
  dlistunit->show();

  manager->update();
}

Result ConfigState::showOption( string id, string shortform ) {
  shown_options[shortform] = id;
  return SUCCESS;
}

Result ConfigState::doCreate( string ) {
  GLOBAL_DEBUG_STREAM("ConfigState: create: " << stateid << endl);

  if ( isBackgroundFree() ) {
    if (FAILURE == app->createDefaultGraph( background )) {
      return FAILURE;
    }

    if (FAILURE == app->startDefaultGraph()) {
      return FAILURE;
    }
  }

  dlistunit = new DoubleListUnit(manager, factory, "medium");
  dlistunit->setFirstColumnWidth(80);

  dlistunit->setDimension(620, 450);
  dlistunit->setPosition(50, 50);
  dlistunit->setColors(0x80ff0000,
                       0x80ff0000,
                       0x800000ff,
                       0x800000ff,
                       0x800000ff);
  dlistunit->selectWindow();

  /* fill up dlistunit with added values */
  for (map<string,string>::const_iterator s = shown_options.begin();
       s != shown_options.end();
       ++s) {
    const MMBoxConfig::Setting* setting;
    if (!config->getSetting(s->first, setting)) {
      GLOBAL_ERROR_STREAM("Unknown setting: " << s->first << endl);
      continue;
    }
    string val(s->second.empty() ? setting->getDescription() : s->second);
    string shortform(s->first);
    const list<string>& values(setting->getValues());
    
    DoubleListUnit::line_t line;
    
    /* add all values */
    string default_value;
    for (list<string>::const_iterator i = values.begin();
         i != values.end();
         ++i) {
      string value = *i;
      line.push_back( value );
      if (default_value.empty()) {
        default_value = value;
      }
    }
    
    /* add human readable configuration text to list */
    dlistunit->addLine( val, line, shortform );
  }
  
  initList();
  createWindow();

  dlistunit->update();
  manager->update();

  return( SUCCESS );
}

Result ConfigState::doInitialize( string ) {
  GLOBAL_DEBUG_STREAM("ConfigState: initialize: " << stateid << endl);
  if (dlistunit) dlistunit->show();

  if ( isBackgroundFree() ) {
    if (FAILURE == app->createDefaultGraph( background )) {
      return FAILURE;
    }

    if (FAILURE == app->startDefaultGraph()) {
      return FAILURE;
    }
  }

  initList();
  manager->update();

  return( SUCCESS );
}

Result ConfigState::doKill() {

  deleteWindow();
  if ( isBackgroundFree() ) app->disconnectVideoGraph();

  return( SUCCESS );
}

Result ConfigState::doUninitialize() {
  GLOBAL_DEBUG_STREAM("ConfigState: initialize: " << stateid << endl);
  if (dlistunit) dlistunit->hide();
  if ( isBackgroundFree() ) app->disconnectVideoGraph();

  return( SUCCESS );
}

Result ConfigState::receiveMessage( string /*message*/ ) {
  manager->deleteMessage();
  manager->update();
  return SUCCESS;
}

void ConfigState::initList() {
  unsigned int size = dlistunit->getNumLines();

  for( unsigned int i=1; i<=size; i++ ) {
    string val = dlistunit->getLineUrl( i );
    string value;
    config->getValue( val, value );
    if(val == "configfile") {
      unsigned int pos = value.rfind(".xml");
      if(pos != string::npos) {
	value = value.substr(0, pos);
      }
    }

    dlistunit->chooseLineContent( i, value );
  }

  dlistunit->update();
}

Result ConfigState::Down() {
  GLOBAL_DEBUG_STREAM( "ConfigState: Down" << endl );

  dlistunit->selectDown();
  dlistunit->update();

  manager->update();

  return(SUCCESS);
}

Result ConfigState::Up() {
  GLOBAL_DEBUG_STREAM( "ConfigState: Up" << endl );

  dlistunit->selectUp();
  dlistunit->update();

  manager->update();

  return(SUCCESS);
}

Result ConfigState::Left() {
  GLOBAL_DEBUG_STREAM( "ConfigState: Left" << endl );

  dlistunit->selectLeft();
  dlistunit->update();

  manager->update();

  return(SUCCESS);
}

Result ConfigState::Right() {
  GLOBAL_DEBUG_STREAM( "ConfigState: Right" << endl );

  dlistunit->selectRight();
  dlistunit->update();

  manager->update();

  return(SUCCESS);
}

Result ConfigState::Back() {
  GLOBAL_DEBUG_STREAM( "ConfigState: Back" << endl );

  deactivate();

  return( goToLastState() );
}

Result ConfigState::Return() {
  GLOBAL_DEBUG_STREAM( "ConfigState: Return" << endl );

  unsigned int size = dlistunit->getNumLines();
  for( unsigned int i=1; i<=size; i++ ) {
    string val   = dlistunit->getLineUrl( i );
    string value = dlistunit->getLineContent( i );

    config->setValue( val, value );
  }
  config->saveFile();

  manager->displayMessage( "Values saved", "New values were stored" , "OK" );
  manager->update();
  changeState("OSDInputState");

  return(SUCCESS);
}



