/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA7
 */


#include <iostream>
#include <sstream>
#include <unistd.h>

#include "dvb.hpp"

// avoid forward declarations
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"

// for remote registries
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"
#include "nmm/base/registry/RegistryStructs.hpp"

#include "nmm/interfaces/base/IReferenceCount.hpp"
#include "nmm/interfaces/base/sync/ISwitch.hpp"
#include "nmm/interfaces/base/INode.hpp"

// used interfaces
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/IBufferSize.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/sync/ISwitch.hpp"



#include "nmm/base/connect.hpp"    // we want do connect :-)
#include "nmm/utils/utils.hpp"     // for string + int
#include <unistd.h>                // for usleep()

#include "nmm/interfaces/device/IDevice.hpp"

#include "../tvtimer/tvtimer.hpp"
#include "../epg/search_result.hpp"

#include "nmm/base/utils/GraphURL.hpp"

#include "nmm/base/sync/LocalTime.hpp"

// we want to be in:
namespace NMM {

  DvbState::DvbState(string id, MMBoxApplication* _app)
    : MMBoxState(id, _app),
      state_mode(LIVETV),
      overview_mode(BACK),
      timer_id(0),
      switchingCards(false),
      fast_forward_speed(0),
      rewind_speed(0),
      osd_enabled(true),
      setProgress_listener(this, &DvbState::setProgress),
      live_listener(this, &DvbState::live),
      timeshiftingBegin_listener(this, &DvbState::timeshiftingBegin),
      setChannel_listener(this, &DvbState::setChannel),
      switchingCard_listener(this, &DvbState::switchingCard),
      sharedCard_listener(this, &DvbState::sharedCard),
      skippingChannel_listener(this, &DvbState::skippingChannel),
      video_mpeg_nd(NULL),
      ts_nd(NULL),  
      mpa_dec_nd(NULL),
      last_video_nd(NULL),
      itimeshift(NULL),
      text(NULL),
      win(NULL),
      epgWin(NULL),
      detailWin(NULL),
      overviewTable(NULL),
      overviewDetails(NULL),
      sink_synchronizer(NULL),
      ts_offset(0),
      currentShow(),
      tvmanager(NULL),
      currentChannel(0),
      epg(NULL),
      isMaster(true),
      channelSkipped(false),
      overview_table_selected(true),
      deinterlace(false),
      eventMutex()
  {
    // we initialize ALL members explicit and reasonable via colon
    // after the super class:
    // some variables (e.g. Mutexes, ThreadConditions or EDObjects)
    // must be initialized like this, so to be uniform, we initialize
    // all members like this!
    // Attention with reordering warnings from gcc!!!!



    string channelFilename;
    if (app->getMMBoxConfig()->getValue("tvChannelFile", channelFilename)) {
      tvmanager = new TVManager(this->getInterface<IUsageNotify>(), IUsage::LIVETV, channelFilename);
    }else{
      tvmanager = new TVManager(this->getInterface<IUsageNotify>(), IUsage::LIVETV);
    }
    
    if(tvmanager){
      tvmanager->setApplication(app->getNMMApplication());
    }
  }
  

  DvbState::~DvbState() {
  }

  /******************** MMBoxState interface ********************/

  Result DvbState::doCreate(string session_location){


    if(!session_location.empty()){
      state_mode = SHAREDTV;
    }

    string useDeinterlacer;
    if (app->getMMBoxConfig()->getValue("deinterlace", useDeinterlacer)) {
      if(useDeinterlacer == "1" || useDeinterlacer == "yes"){
        deinterlace=true;
      }else{
        deinterlace=false;
      }
    }else{
      deinterlace=false;
    }

    isMaster = true;
    channelSkipped = false;
    graph.clear();
    GLOBAL_DEBUG_STREAM("DvbState::doCreate()" << endl);
    // init the background
    if (FAILURE == app->createDefaultGraph(background)) {
      GLOBAL_ERROR_STREAM("Could not create default Graph!" << endl);
      return FAILURE;
    }

    if (FAILURE == app->startDefaultGraph()) {
      GLOBAL_ERROR_STREAM("Could not start default Graph!" << endl);
      return FAILURE;
    }

    if (session_location != "") {
      ClientRegistry& reg = app->getNMMApplication()->getRegistry();
      reg.addRegistry(TCPAddress(session_location, ClientRegistry::default_port));
    }

    createBootWindow();

    /* Begin> Read timeshifting buffer location */
    string tsbuffer;
    if (!app->getMMBoxConfig()->getValue("tsbuffer", tsbuffer)) {
      tsbuffer = string("/tmp/tsbuffer.mpeg");
    }
    /* End< */

    app->getOsdManager()->setProgress(25);
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->update();

    cerr << endl << endl << endl << endl << endl;

    
    ts_nd = new NodeDescription("MPEGTimeShiftingNode");
    ts_nd->setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  
    NodeDescription demux_nd("MPEGDemuxNode");
    demux_nd.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  
    mpa_dec_nd = new NodeDescription("MPEGAudioDecodeNode");
    mpa_dec_nd->setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);

    

    if(deinterlace){
      last_video_nd = new NodeDescription("YUVDeInterlaceNode");
      last_video_nd->setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);

    }else{
      last_video_nd = new NodeDescription("MPEGVideoDecodeNode");
      last_video_nd->setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
    }
    
    /* End< */
    
    ts_nd->addNodeEvent(Event(ITimeshifting::setFilename_event, new TInValue<string>(tsbuffer)), INode::CONSTRUCTED);
    ts_nd->addNodeEvent(Event(ITimeshifting::setFileBufferThreshhold_event, new TInValue<unsigned int>(32000)), INode::CONSTRUCTED);
    ts_nd->addNodeEvent(Event(ITimeshifting::enableMemBuf_event, new TInValue<unsigned int>(60)), INode::CONSTRUCTED);

    last_video_nd->addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
    last_video_nd->addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);

    mpa_dec_nd->addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
    mpa_dec_nd->addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);
    

    /*Begin> setup the GraphDescription */
    graph.addEdge(*ts_nd, demux_nd);
    graph.addEdge(demux_nd, "mpeg_audio0", *mpa_dec_nd);

    if(deinterlace){
      NodeDescription mpg2yuv_nd("MPEGVideoDecodeNode");
      mpg2yuv_nd.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);

      graph.addEdge(demux_nd, "mpeg_video0", mpg2yuv_nd);
      graph.addEdge(mpg2yuv_nd, *last_video_nd);
    }else{
      graph.addEdge(demux_nd, "mpeg_video0", *last_video_nd);
    }

    app->getOsdManager()->setProgress(50);
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->update();

    //set pointer to the GraphDescription used and the first node after the sink 
    tvmanager->setGraphDescription(&graph, ts_nd);
    //set pointers to the sink, so the TVManager can disconnect them
    tvmanager->setSinkNodes(app->getAudioNode(), app->getOsdNode());
    /* End< */

    tvmanager->registerEventListener(ITVManager::sharedSource_event, &sharedCard_listener);


    app->getOsdManager()->setProgress(75);
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->update();

    try{
      app->disconnectDefaultGraph();
      
    } catch (const Exception& e) {
      cerr << "Error disconnecting defaultGraph " << e << endl;
      return FAILURE;
    }

    switch(state_mode){
    case SHAREDTV:
      video_mpeg_nd = new NodeDescription();
      video_mpeg_nd->disableLocation();
      video_mpeg_nd->setSharingType(NodeDescription::SHARED);

      if(tvmanager->realizeFixGraph(video_mpeg_nd, session_location) == FAILURE){
        cleanup();
        GLOBAL_ERROR_STREAM("Can't establish SHAREDTV with " << session_location << endl);
        return FAILURE;
      }

      isMaster = false;
      break;
    case TIMERPLAYBACK:
      video_mpeg_nd->setSharingType(NodeDescription::SHARED);
      if(tvmanager->realizeFixGraph(video_mpeg_nd) == FAILURE){
        cleanup();
        return FAILURE;
      }
      break;
    default:
      {
        Result res = tvmanager->getInterface<ITVManager>()->setChannel(currentChannel);

        if(res == FAILURE){
          //try to set the frist channel. if this fails too, search a channel
          if(tvmanager->getInterface<ITVManager>()->setChannel(0) == FAILURE){
            if(tvmanager->getInterface<ITVManager>()->channelUp() == FAILURE){
              cleanup();
              return FAILURE;
            }
          }
        }
    
        //register lsitener to react on channel-switch of a master (only with LiveTV)
        //if(tvmanager->getTVCard())
        //  tvmanager->getTVCard()->getParentObject()->registerEventListener(ITVCard::setChannel_event, &setChannel_listener);    
    
        break;
      }
    }

    app->getOsdManager()->setPause(false);
    app->setSynchronized(true);

    //disconnect the synchronizer from sinks and connect the session-synchronizer
    disconnectSinkSynchronizer(app->getSynchronizer());
    synchronize();

    //connect our Sinks to the graph
    connectGlobalSinks();

    /* Disconnect the application's synchronizer from auio and video sink */


    //Setting ITimeshifting-Interface
    INode *ts_node = graph.getINode(*ts_nd);
    itimeshift = ts_node->getParentObject()->getInterface<ITimeshifting>();

 
    // register event listener
    app->getAudioNode()->getParentObject()->registerEventListener(ITimeshiftingListener::live_event, app->getDispatcher());
    app->getDispatcher()->registerEventListener(ITimeshiftingListener::live_event, &live_listener);
    app->getAudioNode()->getParentObject()->registerEventListener(ITimeshiftingListener::timeshiftingBegin_event, app->getDispatcher());
    app->getDispatcher()->registerEventListener(ITimeshiftingListener::timeshiftingBegin_event, &timeshiftingBegin_listener);
    app->getAudioNode()->getParentObject()->registerEventListener(ITimeshiftingListener::setProgress_event, &setProgress_listener);

    tvmanager->registerEventListener(ITVManager::switchingSource_event, &switchingCard_listener);    
    tvmanager->registerEventListener(ITVManager::skippingChannel_event, &skippingChannel_listener);    

    deleteBootWindow();

    // fill up osd
    createWindow();
    app->getOsdManager()->displaySymbol("PLAY");
    app->getOsdManager()->displayProgressBar("Timeshifting-Buffer");
    // we initialize the ProgressBar here, but hide it again as it
    // should only appear in Timeshifting mode
    app->getOsdManager()->hideProgressBar();


    epg = new DvbStateEPG(app, this);

    /*
    cerr << "sleep 10" << endl;
    sleep(10);
    */

    try{

      graph.activateGraph();
      app->getOsdNode()->reachStarted();
      app->getDisplayNode()->reachStarted();
      app->getAudioNode()->reachStarted();
      graph.startGraph();
      
    } catch (const Exception &e) {
      GLOBAL_ERROR_STREAM("DvbState: Error starting graph: " << e << endl);
      
      cleanup();
      return FAILURE;
    
    } catch (...) {
      cerr << "Something strange went wrong while connecting in DvbState::doCreate()" << endl;
      
      cleanup();
      return FAILURE;
    }

    fast_forward_speed = rewind_speed = 0;

    /* specify that graph is the currently used graph description */
    setCurrentGraphDescription(graph, tvmanager->getInterface<ITVManager>()->getTVCard()->getParentInfo().getHostname());

    updateOSD();
    return SUCCESS;
  }


  string DvbState::getSessionLocation(){
    if(tvmanager) {
      return tvmanager->getInterface<ITVManager>()->getTVCard()->getParentInfo().getHostname();
    }
    return string("");
  }

  Result DvbState::doKill() {
    GLOBAL_DEBUG_STREAM("DvbState::doKill()" << endl);

    state_mode = LIVETV;
    
    try{
      app->disconnectDefaultGraph();
    } catch (const Exception& e) {
      cerr << "Error disconnecting defaultGraph " << e << endl;
    }

    app->getOsdManager()->deleteMessage();
    app->getOsdManager()->update();

    app->getOsdManager()->hideColorBar();

    app->getOsdManager()->hideProgressBar();
    app->getOsdManager()->hideSymbol();

    graph.stopGraph();

    app->getAudioNode()->getParentObject()->removeEventListener(&setProgress_listener);
    app->getAudioNode()->getParentObject()->removeEventListener(ITimeshiftingListener::live_event, app->getDispatcher());
    app->getDispatcher()->removeEventListener(&live_listener);
    app->getAudioNode()->getParentObject()->removeEventListener(ITimeshiftingListener::timeshiftingBegin_event, app->getDispatcher());
    app->getDispatcher()->removeEventListener(&timeshiftingBegin_listener);

    tvmanager->removeEventListener(&switchingCard_listener);
    tvmanager->removeEventListener(&sharedCard_listener);
    tvmanager->removeEventListener(&skippingChannel_listener);
    if(tvmanager->getInterface<ITVManager>()->getTVCard())
      tvmanager->getInterface<ITVManager>()->getTVCard()->getParentObject()->removeEventListener(&setChannel_listener);

    cleanup();

    graph.clear();

    return SUCCESS;
  }


  Result DvbState::doInitialize(string) {
    GLOBAL_DEBUG_STREAM("DvbState::doInitialize()" << endl);

   
    if (win) {
      win->setVisible(osd_enabled);
      if ((itimeshift->getMode() == ITimeshifting::LIVE) ||
          (itimeshift->getPlayMode() == ITimeshifting::FB_LIVE)) {
        app->getOsdManager()->displaySymbol("PLAY");
      } else {
        app->getOsdManager()->displaySymbol("PLAY");
      }

      updateOSD();
    }
    
    return SUCCESS;
  }


  Result DvbState::doUninitialize() {
    //app->getOsdManager()->deleteMessage();
  
    if (win) {
      //osd_enabled = win->getVisible();
      win->setVisible(false);
      win->update();
      app->getOsdManager()->hideSymbol();
      app->getOsdManager()->hideProgressBar();
    }
    
    if (epgWin) {
        epgWin->hide();
        epgWin->update();
      }
    
    if (detailWin) {
        detailWin->hide();
        detailWin->update();
      }
    
    if (overviewTable) {
      overviewTable->hide();
      overviewTable->update();
    }

    if (overviewDetails) {
      overviewDetails->hide();
      overviewDetails->update();
    }

    app->getOsdManager()->hideColorBar();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  /* Switch to last state */
  Result DvbState::Back() {
    GLOBAL_DEBUG_STREAM("DvbState::Back()" << endl);
    GLOBAL_DEBUG_STREAM("Switch to stateid: " << lastid << endl);

    //commented out because of don't know
    //until timeshifting works
    //if(itimeshift) itimeshift->setModeLive(true);

    app->getOsdManager()->deleteMessage();

    return goToLastState();
  }

  Result DvbState::Up() {
    GLOBAL_DEBUG_STREAM("DvbState::Up()" << endl);
    //channel-switch when watching a recording leads to livetv
    if(tvmanager->getInterface<ITVManager>()->getTVCard())
      tvmanager->getInterface<ITVManager>()->getTVCard()->getParentObject()->removeEventListener(&setChannel_listener);    


    if(state_mode == TIMERPLAYBACK){       
      state_mode = LIVETV;
      Result res = tvmanager->getInterface<ITVManager>()->setChannel(currentChannel + 1);
      if(res == FAILURE){
        if(tvmanager->getInterface<ITVManager>()->channelUp() == FAILURE){
          return goToLastState();
        }
      }

      restartGraph();
      updateOSD();
      return SUCCESS;
    }

    if(overview_mode != BACK){
      if(overview_table_selected){
        int sel = overviewTable->getSelectedLine();
        sel--;
        overviewTable->selectLine(sel);
        overviewTable->update();
        epg->updateOverviewDescription(sel);
      }else{
        int sel = overviewDetails->getSelectedLine();
        sel--;
        overviewDetails->selectLine(sel);
        overviewDetails->update();
      }
      app->getOsdManager()->update();
      return SUCCESS;
    }

    isMaster = true;
    state_mode = LIVETV;

    cerr << endl << endl << endl << endl << endl;

    if(tvmanager->getInterface<ITVManager>()->channelUp() == FAILURE){
      app->getOsdManager()->deleteMessage();
      app->getOsdManager()->displayMessage("No Channel Available", "There are currently no other channels available", "OK");
      changeState("OSDInput");
    }else{
      if(hasFocus() && switchingCards){
        app->getOsdManager()->deleteMessage();
      }
    }

    restartGraph();
    updateOSD();
    return SUCCESS;
  }
  
  Result DvbState::Down() {
    GLOBAL_DEBUG_STREAM("DvbState::Down()" << endl);
    
    if(tvmanager->getInterface<ITVManager>()->getTVCard())
      tvmanager->getInterface<ITVManager>()->getTVCard()->getParentObject()->removeEventListener(&setChannel_listener);    

    if(state_mode == TIMERPLAYBACK){       
      state_mode = LIVETV;
      Result res = tvmanager->getInterface<ITVManager>()->setChannel(currentChannel - 1);
      if(res == FAILURE){
        if(tvmanager->getInterface<ITVManager>()->channelDown() == FAILURE){
          return goToLastState();
        }
      }

      restartGraph();
      updateOSD();
      return SUCCESS;
    }



    if(overview_mode != BACK){
      if(overview_table_selected){
        int sel = overviewTable->getSelectedLine();
        sel++;
        overviewTable->selectLine(sel);
        overviewTable->update();
        epg->updateOverviewDescription(sel);
      }else{
        int sel = overviewDetails->getSelectedLine();
        sel++;
        overviewDetails->selectLine(sel);
        overviewDetails->update();
      }
      app->getOsdManager()->update();
      return SUCCESS;
    }
    
    isMaster = true;
    state_mode = LIVETV;

    cerr << endl << endl << endl << endl << endl;

    if(tvmanager->getInterface<ITVManager>()->channelDown() == FAILURE){
      app->getOsdManager()->deleteMessage();
      app->getOsdManager()->displayMessage("No Channel Available", "There are currently no other channels available", "OK");
      changeState("OSDInput");
    }else{
      if(hasFocus() && switchingCards){
        app->getOsdManager()->deleteMessage();
      }
    }

    restartGraph();
    updateOSD();
    return SUCCESS;
  }

  Result DvbState::Left(){
    if(overview_mode != BACK){
      if(overview_table_selected){
        overviewTable->deSelectWindow();
        overviewDetails->selectWindow();
        overview_table_selected=false;
      }else{
        overviewTable->selectWindow();
        overviewDetails->deSelectWindow();
        overviewDetails->deSelectLine();
        overview_table_selected=true;
      }
      overviewTable->update();
      overviewDetails->update();
      app->getOsdManager()->update();
    }
    return SUCCESS;
  }

  Result DvbState::Right(){
    if(overview_mode != BACK){
      if(overview_table_selected){
        overviewTable->deSelectWindow();
        overviewDetails->selectWindow();
        overview_table_selected=false;
      }else{
        overviewTable->selectWindow();
        overviewDetails->deSelectWindow();
        overviewDetails->deSelectLine();
        overview_table_selected=true;
      }
      overviewTable->update();
      overviewDetails->update();
      app->getOsdManager()->update();
    }
    return SUCCESS;
  }


  Result DvbState::Pause() {
    GLOBAL_DEBUG_STREAM("DvbState::Pause()" << endl);

    app->getOsdManager()->hideColorBar();

    if (itimeshift->getMode() == ITimeshifting::LIVE) {
      GLOBAL_DEBUG_STREAM("DvbState::Pause() LIVE -> PAUSE" << endl);
      GLOBAL_DEBUG_STREAM("DvbState::Pause() pausing live video" << endl);
      app->getOsdManager()->displaySymbol("PAUSE");
      app->getOsdManager()->update();
      app->waitForSymbolChange("PAUSE");
      itimeshift->setModePause();

      sink_synchronizer->pause();
      
      ts_offset = getTime() - itimeshift->getLivePosition();

      updateOSD();

      return SUCCESS;
    }

    ITimeshifting::PlayMode ts_play_mode =
      itimeshift->getPlayMode();

    switch (ts_play_mode) {
    case ITimeshifting::PLAY:
    case ITimeshifting::FB_LIVE:
      GLOBAL_DEBUG_STREAM("DvbState::Pause(): PLAY/FB_LIVE -> PAUSE" << endl);
      app->getOsdManager()->displaySymbol("PAUSE");
      app->getOsdManager()->update();
      app->waitForSymbolChange("PAUSE");
      itimeshift->setModePause();
      sink_synchronizer->pause();
      break;

    case ITimeshifting::PAUSE:
      GLOBAL_DEBUG_STREAM("DvbState::Pause(): PAUSE -> PLAY" << endl);
      fast_forward_speed  = 0;
      sink_synchronizer->wakeup();
      itimeshift->setModePlay(false);
      app->getOsdManager()->displaySymbol("PLAY");
      updateOSD();
      break;

    case ITimeshifting::FAST_FORWARD:
    case ITimeshifting::REWIND:
      // do nothing here!
      break;
    }

    return SUCCESS;
  }

  Result DvbState::Forward() {
    GLOBAL_DEBUG_STREAM("DvbState::Forward()" << endl);

    ITimeshifting::Mode ts_mode =
      itimeshift->getMode();

    ITimeshifting::PlayMode ts_play_mode = 
      itimeshift->getPlayMode();

    if ((ts_mode == ITimeshifting::LIVE) ||
        ((ts_mode == ITimeshifting::FILEBUF) && (ts_play_mode == ITimeshifting::FB_LIVE))) {
      GLOBAL_DEBUG_STREAM("DvbState::Forward(): Unable to fast-forward a live picture!" << endl);
      return FAILURE;
    }

    switch (ts_play_mode) {
    case ITimeshifting::PAUSE:
      GLOBAL_DEBUG_STREAM( "DvbState::Forward(): PAUSED!" << endl);
      itimeshift->setModePlay(false);
      sink_synchronizer->wakeup();
      fast_forward_speed = 0;
      // fallthrough here!

    case ITimeshifting::PLAY:
    case ITimeshifting::FAST_FORWARD:
    case ITimeshifting::REWIND:
      GLOBAL_DEBUG_STREAM( "DvbState::Forward(): * -> FF!" << endl);
      if (fast_forward_speed == 0) {
        fast_forward_speed = 2;
      } else {
        fast_forward_speed *= 2;
      }
      GLOBAL_DEBUG_STREAM("DvbState::Forward(): FAST_FORWARD (" << fast_forward_speed << "x)" << endl);
      itimeshift->setModeFastForward(fast_forward_speed);
      break;

    case ITimeshifting::FB_LIVE:
      // do nothing here, case handled above
      break;
    }

    app->getOsdManager()->displaySymbol("FF");
    updateOSD();

    return SUCCESS;
  }

  Result DvbState::Return() { // PLAY
    GLOBAL_DEBUG_STREAM("DvbState::Return()" << endl);
    
    if(overview_mode != BACK){
      int sel = overviewTable->getSelectedLine();
      string s = overviewTable->getLineURL(sel);
      switchChannelTo(atoi(s.c_str()));
      switchOverview(BACK);
      updateOSD();
      return SUCCESS;
    }

    app->getOsdManager()->deleteMessage();
    app->getOsdManager()->update();

    ITimeshifting::Mode ts_mode = 
      itimeshift->getMode();

    ITimeshifting::PlayMode ts_play_mode =
      itimeshift->getPlayMode();

    if (ts_mode == ITimeshifting::LIVE) {
      GLOBAL_DEBUG_STREAM("DvbState::Return(): \"Play\" while playing -> ignored! " << endl);
      return SUCCESS;
    }

    switch (ts_play_mode) {
    case ITimeshifting::PAUSE:
      GLOBAL_DEBUG_STREAM("DvbState::Return(): PAUSE -> PLAY" << endl);
      GLOBAL_DEBUG_STREAM("DvbState::Return(): simply calling DvbState::Pause()" << endl);
      Pause();
      break;

    case ITimeshifting::PLAY:
      GLOBAL_DEBUG_STREAM("DvbState::Return(): PLAY" << endl);
      // errror 20030701: pressed "Play" and is playing, what the hell
      //                  do we syncReset?
      /*
        app->syncReset();
      */
      break;

    case ITimeshifting::FB_LIVE:
      GLOBAL_DEBUG_STREAM("DvbState::Return(): FB_LIVE -> PLAY!" << endl);
      itimeshift->setModePlay(true);
      break;

    case ITimeshifting::FAST_FORWARD:
    case ITimeshifting::REWIND:
      GLOBAL_DEBUG_STREAM("DvbState::Return(): FF/REW -> PLAY" << endl);
      // errror 20030701: "if" handled above!
      /*
        if (ts_mode != ITimeshifting::LIVE) {
      */
      fast_forward_speed = 0;
      itimeshift->setModePlay(true);
      /*
        }
      */
    }
    app->getOsdManager()->displaySymbol("PLAY");
    updateOSD();

    return SUCCESS;
  }

  Result DvbState::Rewind() {
    GLOBAL_DEBUG_STREAM("DvbState::Rewind()" << endl);

    app->getOsdManager()->hideColorBar();

    ITimeshifting::Mode ts_mode = 
      itimeshift->getMode();

    ITimeshifting::PlayMode ts_play_mode = 
      itimeshift->getPlayMode();

    if (ts_mode == ITimeshifting::LIVE) {
      GLOBAL_DEBUG_STREAM("DvbState::Rewind(): INSTANT-REPLAY" << endl);
      fast_forward_speed = 0;
      itimeshift->disableMemBuf(true);
      updateOSD();
      return SUCCESS;
    }

    switch (ts_play_mode) {
    case ITimeshifting::PLAY:
    case ITimeshifting::FAST_FORWARD:
    case ITimeshifting::FB_LIVE:
      GLOBAL_DEBUG_STREAM("DvbState::Rewind(): PLAY/FF -> REW!" << endl);
      fast_forward_speed = 0;
      rewind_speed = -1;
      itimeshift->setModeRewind(rewind_speed);
      break;

    case ITimeshifting::REWIND:
      GLOBAL_DEBUG_STREAM("DvbState::Rewind(): REW -> 2*REW!" << endl);
      rewind_speed *= 2;
      itimeshift->setModeRewind(rewind_speed);
      break;

    case ITimeshifting::PAUSE:
      GLOBAL_DEBUG_STREAM("DvbState::Rewind(): PAUSED!" << endl);
      GLOBAL_DEBUG_STREAM("DvbState::Rewind(): PAUSE -> REW!" << endl);
      rewind_speed = -1;
      sink_synchronizer->wakeup();
      itimeshift->setModeRewind(rewind_speed);
      break;
    }
    GLOBAL_DEBUG_STREAM("DvbState::Rewind(): REWIND (" << rewind_speed << "x)" << endl);

    app->getOsdManager()->displaySymbol("REW");
    app->getOsdManager()->hideColorBar();
    updateOSD();

    return SUCCESS;
  }

  Result DvbState::Stop() {
    GLOBAL_DEBUG_STREAM("DvbState::Stop()" << endl);

    ITimeshifting::Mode ts_mode =
      itimeshift->getMode();

    ITimeshifting::PlayMode ts_play_mode =
      itimeshift->getPlayMode();

    if (ts_mode == ITimeshifting::FILEBUF) {
      if (ts_play_mode == ITimeshifting::FB_LIVE) {
        GLOBAL_DEBUG_STREAM("DvbState::Stop(): FB_LIVE -> LIVE!" << endl);
        app->getOsdManager()->hideProgressBar();
        itimeshift->setModeLive(true, true);
      } else {
        GLOBAL_DEBUG_STREAM("DvbState::Stop(): PLAY -> FB_LIVE!" << endl);
        app->getOsdManager()->hideProgressBar();
        itimeshift->setModeFBLive();
        sink_synchronizer->wakeup();
      }
    }

    updateOSD();

    return SUCCESS;
  }

  Result DvbState::Record() {
    GLOBAL_DEBUG_STREAM("DvbState::Record()" << endl);

    if (itimeshift->isRecording()) {
      itimeshift->stopRecord();
    } else {
      itimeshift->startRecord();
    }
    updateOSD();

    return SUCCESS;
  }


  Result DvbState::OSDKey() {
    GLOBAL_DEBUG_STREAM("DvbState::OSDKey()" << endl);

    if(win){
    
      if (!win->getVisible()) {        
        win->setVisible(true);
        osd_enabled = true;
        if((itimeshift->getMode() == ITimeshifting::LIVE)
           ||(itimeshift->getPlayMode() == ITimeshifting::FB_LIVE)){
          //showShortEPGInfo();
          
        }
      } else {
        win->setVisible(false);
        osd_enabled = false;
        app->getOsdManager()->hideSymbol();
        app->getOsdManager()->hideProgressBar();
        
        detailWin->hide();
        detailWin->update();
        epgWin->hide();
        epgWin->update();
        
        app->getOsdManager()->hideColorBar();
        app->getOsdManager()->update();
      }
      
      updateOSD();
    }
    return SUCCESS;
  }

  Result DvbState::specialKey01(){
    GLOBAL_DEBUG_STREAM("DvbState::specialKey01()" << endl);    
    return SUCCESS;
  }

  Result DvbState::numKey0(){
    return switchChannelTo(0);
  }
  Result DvbState::numKey1(){
    return switchChannelTo(1);
  }
  Result DvbState::numKey2(){
    return switchChannelTo(2);
  }
  Result DvbState::numKey3(){
    return switchChannelTo(3);
  }
  Result DvbState::numKey4(){
    return switchChannelTo(4);
  }
  Result DvbState::numKey5(){
    return switchChannelTo(5);
  }
  Result DvbState::numKey6(){
    return switchChannelTo(6);
  }
  Result DvbState::numKey7(){
    return switchChannelTo(7);
  }
  Result DvbState::numKey8(){
    return switchChannelTo(8);
  }
  Result DvbState::numKey9(){
    return switchChannelTo(9);
  }
  
  Result DvbState::colorKeyRed(){
    if(state_mode == TIMERPLAYBACK) {
      return SUCCESS;
    }else{}
    
    TvTimer::TvTimer* timer = dynamic_cast<TvTimer*>(app->getState("TV-Timer"));
    ProgramDescription show;
    
    if(overview_mode == BACK){
      cerr << "program from short" << endl;
      show = epg->getCurrentShow();
    }else{
      cerr << "program from windows" << endl;
      show = epg->getShow(overviewTable->getSelectedLine());
    }


    if(!show.empty()){
      cerr << "program" << endl;
      try{
        int res = timer->programTimer(show);
        if(res >= 0){
          cerr << "timer-programming SUCCESSFUL" << endl;
          
        }else{
          app->getOsdManager()->displayMessage("Conflict" , "Timer Not programmed because of Conflicts!", "OK" );
          app->getOsdManager()->update();
          changeState("OSDInputState");
        }
        
        return SUCCESS;
      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("Something went wrong during timer-programming: " << e << endl);
        app->getOsdManager()->displayMessage("ERROR" , "Timer Not programmed!", "OK" );
        app->getOsdManager()->update();
        changeState("OSDInputState");
        return FAILURE;
      } 
    }else{
      cerr << "program unlimited" << endl;
      try{
        time_t now = time(NULL);
        TVTimerEntry *e = new TVTimerEntry(UserInfo().getUsername());
        e->setStartTime(now);
        cerr << "getChannel" << endl;
        TVChannel c = tvmanager->getCurrentChannel();
        e->setChannelName(c.getName());
        if(!show.empty()){
          e->setTimerInfo(show);
        }

        int res = timer->programTimer(*e);
        if(res >= 0){
          cerr << "timer-programming SUCCESSFUL" << endl;
          
        }else{
          app->getOsdManager()->displayMessage("Conflict" , "Timer Not programmed because of Conflicts!", "OK" );
          app->getOsdManager()->update();
          changeState("OSDInputState");
        }
        if(e){
          delete e;
          e = NULL;
        }
      }catch(const Exception &e){
        app->getOsdManager()->displayMessage("ERROR" , "Timer Not programmed!", "OK" );
        app->getOsdManager()->update();
        changeState("OSDInputState");
        
        return FAILURE;
      }
    }
    return SUCCESS;
  }

  Result DvbState::colorKeyGreen(){

    if(!app->getOsdManager()->getGreenColorBarText().empty()){
      if(overview_mode != BACK){
        int sel = overviewTable->getSelectedLine();
        string s = overviewTable->getLineURL(sel);
        switchChannelTo(atoi(s.c_str()));
        switchOverview(BACK);
        return SUCCESS;
      }
      
      if(!detailWin->isHidden()){
        detailWin->hide();
        if(osd_enabled) win->setVisible(true);
        epgWin->deSelectLine();
        updateOSD();
      }else{
        detailWin->show();
        //hide main-window that information does not overlap
        win->setVisible(false);
        app->getOsdManager()->setGreenColorBarText("Close");
      }
      win->update();
      detailWin->update();
      app->getOsdManager()->update();
      
    }

    return SUCCESS;
  }

  Result DvbState::colorKeyYellow(){
    if(state_mode == TIMERPLAYBACK) return SUCCESS;

    if(!app->getOsdManager()->getYellowColorBarText().empty())
      switchOverview(NOW);
    return SUCCESS;

  }
  Result DvbState::colorKeyBlue(){
    if(state_mode == TIMERPLAYBACK) return SUCCESS;

    if(!app->getOsdManager()->getBlueColorBarText().empty())
    switchOverview(NEXT);

    return SUCCESS;
  }
  

  Result DvbState::receiveMessage( string message ) {
    if(message == "UP"){
      if(!(state_mode == TIMERPLAYBACK)){
        if(tvmanager->getInterface<ITVManager>()->channelUp() == FAILURE){
          app->getOsdManager()->deleteMessage();
          app->getOsdManager()->displayMessage("No Channel Available", 
                                               "There are currently no other channels available", 
                                               "OK");
          changeState("OSDInput");
        }else{
          if(hasFocus() && switchingCards){
            app->getOsdManager()->deleteMessage();
          }
        }

        restartGraph();
        updateOSD();
      }else{
        state_mode = LIVETV;
        return Up();
      }
    }else if(message == "Down"){
      state_mode = LIVETV;
      return Down();
    }
      
    return SUCCESS;
  }
  /******************** MMBoxState interface ********************/

  Result DvbState::becomeMaster(){
    cerr << "I'AM THE NEW MASTER !!!!!!!!!!!!!!!!!!!!!!!!" << endl;
    isMaster = true;
    updateOSD();
    return SUCCESS;
  }

  Result DvbState::becomeSlave(){
    cerr << "Heul, becomed Slave :(((((((((((((" << endl;
    isMaster = false;
    updateOSD();
    return SUCCESS;
  }


  /******************** EventListener methods ********************/
  Result DvbState::live() {
    GLOBAL_DEBUG_STREAM("DvbState::live()" << endl);
    
    if ( win->getVisible() ) {
      app->getOsdManager()->hideProgressBar();
      app->getOsdManager()->displaySymbol("PLAY");
    }
    
    return SUCCESS;
  }

  Result DvbState::timeshiftingBegin() {
    GLOBAL_DEBUG_STREAM("DvbState::timeshiftingBegin()" << endl);
    
    if ( win->getVisible() ) {
      app->getOsdManager()->displaySymbol("PLAY");
      app->getOsdManager()->update();
    }
    
    return SUCCESS;
  }

  // small helper for setProgress
  static string getPrettyTime(const Time& ugly_time) {
    LocalTime local_time(ugly_time, true);
    return local_time.toString("%R");
  }

  Result DvbState::setProgress(Time& ts_time) {
    GLOBAL_DEBUG_STREAM("DvbState::setProgress("<< ts_time << ")" << endl);
  
    if (itimeshift->getMode() == ITimeshifting::LIVE) {
      GLOBAL_DEBUG_STREAM("DvbState::setProgress(): ignoring progress in LIVE mode." << endl);
      app->getOsdManager()->hideProgressBar();
      return SUCCESS;
    }

    if (itimeshift->getPlayMode() == ITimeshifting::FB_LIVE) {
      GLOBAL_DEBUG_STREAM("DvbState::setProgress(): ignoring progress in FB_LIVE mode." << endl);
      app->getOsdManager()->hideProgressBar();
      return SUCCESS;
    }

    if (ts_time == Time(0)) {
      GLOBAL_DEBUG_STREAM("DvbState::setProgress(): ts_time=0 ignoring." << endl);
      return SUCCESS;
    }

    Time ts_end(itimeshift->getLivePosition()); /// the live position
    Time ts_begin(itimeshift->getStartTime()); // at this time, the timeshifting buffer started recording

    
    // compute progress values
    u_int64_t progress((ts_time.sec*1000 + ts_time.nsec/(nano_per_sec/1000)) - 
                       (ts_begin.sec*1000 + ts_begin.nsec/(nano_per_sec/1000)));
    u_int64_t max((ts_end.sec*1000 + ts_end.nsec/(nano_per_sec/1000)) - 
                  (ts_begin.sec*1000 + ts_begin.nsec/(nano_per_sec/1000)));
    
    // add timeshifting offset
    ts_begin += ts_offset;
    ts_end += ts_offset;
    ts_time += ts_offset;
    
    
    if (win->getVisible()) {
      app->getOsdManager()->displayProgressBar(getPrettyTime(ts_begin) + 
                                               string("                 ") + 
                                               getPrettyTime(ts_time) + 
                                               string("                 ") + 
                                               getPrettyTime(ts_end));
    }
    GLOBAL_DEBUG_STREAM("DvbState::setProgress(): \"" <<
                        getPrettyTime(ts_begin) + 
                        string("             ") + 
                        getPrettyTime(ts_time) + 
                        string("             ") +
                        getPrettyTime(ts_end) <<
                        "\": " <<
                        progress << " / " << max << endl);
    app->getOsdManager()->setMaxProgress(max);
    app->getOsdManager()->setProgress(progress);
    app->getOsdManager()->updateProgress();
    if (win->getVisible()) {
      app->getOsdManager()->update();
    }
    
    return SUCCESS;
  }

  void DvbState::run(){
    try{

      MutexGuard mg(eventMutex);

      if(tvmanager->getInterface<ITVManager>()->getTVCard())
        tvmanager->getInterface<ITVManager>()->getTVCard()->getParentObject()->removeEventListener(&setChannel_listener);    

      isMaster = true;

      if(tvmanager->getInterface<ITVManager>()->setChannel(currentChannel) == FAILURE){
        if(tvmanager->getInterface<ITVManager>()->channelUp() == FAILURE){
          GLOBAL_ERROR_STREAM("DvbState: setting channel FAILED" << endl);
        } 
      }else{
      }
      
      restartGraph();
      app->getOsdManager()->deleteMessage();

      updateOSD();

    }catch(const Exception &e){}catch(...){}
    deactivateThread();
  }

  Result DvbState::setChannel(int &ch){
    cerr << "Catched Event: setChannel" << flush;
    int c = tvmanager->getCurrentChannelOnCurrentTVCard();
    cerr << " to " << c << " " << ch <<  endl;
    if(c != ch && hasFocus()) activateThread();
    return SUCCESS;
  }

  Result DvbState::switchingCard(){
    
    MutexGuard mg(eventMutex);

    cerr << "Catched Event: switchingCard" << endl;
    switchingCards = true;
    if(hasFocus()){
      app->getOsdManager()->displayMessage("Switching TVCards", "Searching for new TVCard","","","");
      app->getOsdManager()->update();
    }
    return SUCCESS;
  }
  Result DvbState::sharedCard(){
    cerr << "Catched Event: sharedCard" << endl;
    if(hasFocus()){
      app->getOsdManager()->deleteMessage();
      app->getOsdManager()->displayMessage("Shared Card", "You are not Master on this TV-Card", "OK");
      changeState("OSDInput");
    }
    isMaster = false;
    return SUCCESS;
  }

  Result DvbState::skippingChannel(){

    if(hasFocus()){
      cerr << "skipping channel ***************************************** " << endl;
      channelSkipped = true;
      //app->getOsdManager()->deleteMessage();
      //app->getOsdManager()->update();
      //app->getOsdManager()->displayMessage("Skipped Channels", "Unable to tune next channel. Searching next available channel", "OK");
      //app->getOsdManager()->update(); 
   }
    return SUCCESS;
  }


  /******************** EventListener methods ********************/


  /******************** private helpers ********************/
  void DvbState::cleanup() {
    /* if we had already request the graph with a synchronizer we must reconnect
     * the application's synchronizer */

    if (sink_synchronizer.get()) {
      disconnectSinkSynchronizer( sink_synchronizer.get() );
      connectSinkSynchronizer( app->getSynchronizer() ); 
      sink_synchronizer.reset(NULL);
    }

    app->setUseExternalSynchronizer(false);

    GLOBAL_DEBUG_STREAM("DvbState::cleanup()" << endl);
    deleteBootWindow();
    deleteWindow();

    if (app->getDisplayNode()->isStarted()) {
      app->getDisplayNode()->stop();
      app->getDisplayNode()->flush();

    }
    if (app->getAudioNode()->isStarted()) {
      app->getAudioNode()->stop();
      app->getAudioNode()->flush();
      app->getAudioNode()->reachInitialized();
      disconnect_input(app->getAudioNode());

    }

    if (app->getOsdNode()->isInputConnected("stream1")) {
      app->getOsdNode()->reachInitialized();
      disconnect_input(app->getOsdNode(), "stream1");
    }

    /* Delete all interfaces */
    if (itimeshift) {
      delete itimeshift;
      itimeshift = NULL;
    }
  

    if(video_mpeg_nd){
      delete video_mpeg_nd;
      video_mpeg_nd = NULL;
    }

    if(ts_nd){
      delete ts_nd;
      ts_nd = NULL;
    }

    if(last_video_nd){
      delete last_video_nd;
      last_video_nd = NULL;
    }
    if(mpa_dec_nd){
      delete mpa_dec_nd;
      mpa_dec_nd = NULL;
    }

    /* Release TV source graph */
    try {
      tvmanager->releaseGraphDescription();

      app->getNMMApplication()->getRegistry().releaseGraph(graph);


    } catch(const Exception& e){
      GLOBAL_ERROR_STREAM("Exception while releasing graph" << e << endl);
    }

    // just to be sure:

    if (app->getOsdNode()->isInputConnected("stream1")) {
      // if OSD's video input is still connected after cleanup, the
      // default graph must be connected
      app->disconnectDefaultGraph();
      // this leaves the osd in INITIALIZED where the other MMBoxStates
      // expect it.
    }

    GLOBAL_DEBUG_STREAM("DvbState::cleanup() return" << endl);
  }


  void DvbState::updateOSD() {
    currentChannel = tvmanager->getCurrentChannelNumber();
    if(win){
      if (win->getVisible()) {
        TVChannel c = tvmanager->getCurrentChannel();
        text->deleteAllLines(); 
        text->addLine(itoa(currentChannel));
        text->addLine(c.getName());
        
        //display the color-bar with default-buttons
        app->getOsdManager()->displayColorBar();
        app->getOsdManager()->setColorBarText("Record","","Now","Next");

        switch(state_mode){
        case LIVETV:
          {
            ITimeshifting::Mode ts__mode = itimeshift->getMode();
            switch(ts__mode){
            case ITimeshifting::LIVE:
              text->addLine("LiveTV");
              break;
            case ITimeshifting::FILEBUF:
              {
                ITimeshifting::PlayMode ts_play_mode = itimeshift->getPlayMode();
                if(ts_play_mode == ITimeshifting::FB_LIVE){
                  text->addLine("FB-LiveTV");
                }else{
                  text->addLine("Timeshift");
                  app->getOsdManager()->hideColorBar();
                }
                break;
              }
            default:
              text->addLine("LiveTV");
              break;
            }
            break;
          }
        case TIMERPLAYBACK:
          {
            text->addLine("TvTimer-Playback");
            app->getOsdManager()->setColorBarText("","","Now","Next");
            break;
          }
        case SHAREDTV:
          {
            text->addLine("FollowMe-TV");
            break;
          }        
        default:
          text->addLine("statemode " + itoa((int)state_mode));
          break;
        }
        

        if(tvmanager->getInterface<ITVManager>()->getTVCard()){
          string s = tvmanager->getInterface<ITVManager>()->getTVCard()->getParentInfo().getHostname();
          if(isMaster){
            s += "(Master)";
          }else{
            s += "(Slave)";
          }
          text->addLine(s);
        }
        win->update();
                
        showShortEPGInfo();
        
      }
      
      app->getOsdManager()->update();
    }
  }
  
  void DvbState::createWindow() {
    GLOBAL_DEBUG_STREAM("DvbState::createWindow()" << endl);

    deleteWindow();

    text = app->getOSDWidgetFactory().createTextView();
    text->setFont("medium");
    text->setSelectionColor(0xff0080ff);

    win = app->getOsdManager()->createWindow();
    win->setDimension(300, 150);
    win->setPosition(50, 40);
    win->setContent(*text);
    win->setVisible(osd_enabled);
    win->update();


    epgWin = new TableUnit(app->getOsdManager(), &(app->getOSDWidgetFactory()), "medium");
    epgWin->setScrollbarEnabled(false);

    epgWin->addColumn(15);
    
    epgWin->setColors( 0x80ff0000,
                       0x80808080,
                       0xff00ff80,
                       0xff0080ff,
                       0xff808080
                       );
    epgWin->setDimension(560, 70);
    epgWin->setPosition(80, 410);
    epgWin->selectWindow();
    epgWin->hide();
    epgWin->update();

    //extended EPG-Info

    detailWin = new ListUnit(app->getOsdManager(), &(app->getOSDWidgetFactory()), "medium");

    detailWin->setHeader("Extended Information");
    detailWin->setColors( 0x80ff0000,
                          0x80808080,
                          0xff00ff80,
                          0xff0080ff,
                          0xff808080
                          );
    detailWin->setDimension(510, 380);
    detailWin->setPosition(80, 40);
    detailWin->selectWindow();
    detailWin->hide();
    detailWin->update();
    
    overviewTable = new TableUnit(app->getOsdManager(), &(app->getOSDWidgetFactory()), "medium");
    overviewTable->setScrollbarEnabled(true);

    overviewTable->addColumn(30);
    overviewTable->addColumn(15);
    
    overviewTable->setColors( 0x80ff0000,
                       0x80808080,
                       0xff00ff80,
                       0xff0080ff,
                       0xff808080
                       );
    overviewTable->setDimension(510,216);
    overviewTable->setPosition(80, 50);
    overviewTable->selectWindow();
    overviewTable->hide();
    overviewTable->update();

    overviewDetails = new ListUnit(app->getOsdManager(), &(app->getOSDWidgetFactory()), "medium");

    overviewDetails->setColors( 0x80ff0000,
                          0x80808080,
                          0xff00ff80,
                          0xff0080ff,
                          0xff808080
                          );
    overviewDetails->setDimension(510, 200);
    overviewDetails->setPosition(80, 274);
    overviewDetails->deSelectWindow();
    overviewDetails->hide();
    overviewDetails->update();

    app->getOsdManager()->update();
  }
  
  void DvbState::deleteWindow() {
    GLOBAL_DEBUG_STREAM("DvbState::deleteWindow()" << endl);
    if (win) {
      app->getOsdManager()->deleteWindow(*win);
      delete win;
      win = NULL;
    }
    if(epgWin){
      delete epgWin;
      epgWin = NULL;
    }
    if(detailWin){
      delete detailWin;
      detailWin = NULL;
    }

    if(overviewTable){
      delete overviewTable;
      overviewTable = NULL;
    }
    if(overviewDetails){
      delete overviewDetails;
      overviewDetails = NULL;
    }
  }

  void DvbState::createBootWindow() {
    app->getOsdManager()->displayProgressBar("Accessing TV-Device");
    app->getOsdManager()->setMaxProgress(100);
    app->getOsdManager()->setProgress(0);
    app->getOsdManager()->updateProgress();

    app->getOsdManager()->update();
  }

  void DvbState::deleteBootWindow() {
    app->getOsdManager()->hideProgressBar();
    app->getOsdManager()->update();
  }

  TVChannelList* DvbState::getChannelConfig(){
    return tvmanager->getChannelConfig();
  }

  /****************************************************************/
  /*  methods for switching beween hardware- and software-decoder */
  /****************************************************************/
  Result DvbState::switchChannelTo(int c){
    GLOBAL_DEBUG_STREAM("DvbState::switchChannelTo()" << endl);
    
    if(tvmanager->getInterface<ITVManager>()->getTVCard())
      tvmanager->getInterface<ITVManager>()->getTVCard()->getParentObject()->removeEventListener(&setChannel_listener);    

    isMaster = true;

    if(tvmanager->getInterface<ITVManager>()->setChannel(c) == SUCCESS){
      currentChannel = tvmanager->getCurrentChannelNumber();
      //undisplay Message
      if(hasFocus() && switchingCards){
        app->getOsdManager()->deleteMessage();
      }
    }else{
      app->getOsdManager()->deleteMessage();
      app->getOsdManager()->displayMessage( "Channel not Available", "Search next available channel?" , "UP", "NO", "", 1 );
      app->getOsdManager()->update();
      changeState(string("OSDInputState"));
    }

    restartGraph();
    updateOSD();

    return SUCCESS;
  }

  void DvbState::connectSinkSynchronizer(ISinkSynchronizer * tmp_sync) {
    IMultiAudioSinkSynchronizer_var ia_sync(tmp_sync->getParentObject()->getInterface<IMultiAudioSinkSynchronizer>());
    IMultiVideoSinkSynchronizer_var iv_sync(tmp_sync->getParentObject()->getInterface<IMultiVideoSinkSynchronizer>());
    
    /* Begin> Add the controller to the synchronizer */
    ISynchronizedSink_var av(app->getAudioNode()->getParentObject()->getInterface<ISynchronizedSink>());
    if (av.get()) {
      ISinkController_var controller (av->getController());
       av->setSynchronized(true);
       if(ia_sync.get()) { // add audio controller ...
         ia_sync->addAudioSinkController(controller.get());
       }
    }
    
    av.reset(app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>());
    if(iv_sync.get()) { // ... or video (unbuffered)
       ISinkController_var controller (av->getController());
       av->setSynchronized(true);
       iv_sync->addVideoSinkController(controller.get());
    }
  }
  
  void DvbState::synchronize(){
    try {     
      if(sink_synchronizer.get() == NULL){
      // Disconnect the application's synchronizer from auio and video sink 
        cerr << "USING GRAPH SYNC" << endl;
        // Use the given synchronizer from graph description ... 
        sink_synchronizer.reset(graph.getISession()->getSinkSynchronizer() );
        // and connect it with our audio and video sink 
        connectSinkSynchronizer( sink_synchronizer.get() );
      }else{
        cerr << "already synced" << endl;
        // Disconnect the old synchronizer from auio and video sink 
        disconnectSinkSynchronizer(sink_synchronizer.get());
        sink_synchronizer.reset(graph.getISession()->getSinkSynchronizer() );
        // and connect it with our audio and video sink 
        connectSinkSynchronizer( sink_synchronizer.get() );
      }
     
    } catch (const Exception& e){
      GLOBAL_ERROR_STREAM("DvbState::synchronize() : Can't connect synchronizer : " << e << endl);
      cerr << "DvbState::synchronize() : Can't connect synchronizer : " << e << endl;
    }
  }
  
  void DvbState::disconnectSinkSynchronizer(ISinkSynchronizer* tmp_sync) {
    ISinkSynchronizer_var isync(tmp_sync->getParentObject()->getCheckedInterface<ISinkSynchronizer>());
    if (isync.get()) {
      ISynchronizedSink_var av(app->getAudioNode()->getParentObject()->getInterface<ISynchronizedSink>());
      if (av.get()) {
        ISinkController_var controller (av->getController());
        isync->removeSinkController(controller.get());
      }
      av.reset( app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>() );
      if (av.get()) {
        ISinkController_var controller (av->getController());
        isync->removeSinkController(controller.get());
      }
    }
  } 

  time_t DvbState::stringToTime(string s){
    
    try{
      struct tm * abs_time = new struct tm();
      
      int h    = atoi( (s.substr(8,2)).c_str() );
      int min  = atoi( (s.substr(10,2)).c_str() );    
      int day  = atoi( (s.substr(6,2)).c_str() );
      int mon  = atoi( (s.substr(4,2)).c_str() );
      int year = atoi( (s.substr(0,4)).c_str() );
      
      abs_time->tm_sec  = 0;
      abs_time->tm_min  = min;
      abs_time->tm_hour = h;
      abs_time->tm_mday = day;
      abs_time->tm_mon  = mon-1;
      abs_time->tm_year = year-1900;
      abs_time->tm_isdst= 1;

      return( mktime( abs_time ) - timezone);

    }catch(const Exception &e){
      //cerr << e << endl;
      return 0;
    }
  }

  Result DvbState::setTvChannel(int &c){
    
    //      tvmanager->setChannel(c);
    currentChannel = c;// tvmanager->getCurrentChannelNumber();
    
    return SUCCESS;
  }
  
  Result DvbState::setTVTimerPlayback(TVTimerEntry &entry){
    state_mode = TIMERPLAYBACK;
    if(video_mpeg_nd){
      *video_mpeg_nd = entry.getTVCard();
    }else{
      video_mpeg_nd = new NodeDescription(entry.getTVCard());
    }
    currentShow = entry.getTimerInfo();
    timer_id = entry.getID();
    isMaster = false;
    return SUCCESS;
  }

  Result DvbState::setLiveTV(){
    state_mode = LIVETV;
    return SUCCESS;
  }

  /***************************************/
  /* methods for showing EPG-Information */
  /***************************************/
  Result DvbState::showShortEPGInfo() {
    switch(state_mode){
    case LIVETV:
      if( (itimeshift->getMode()        == ITimeshifting::LIVE 
           || itimeshift->getPlayMode() == ITimeshifting::FB_LIVE)
          && win->getVisible()){
        if(epg && tvmanager){
          epg->updateShort(tvmanager->getCurrentChannel().getChannelName());
        }
      }else{
        epgWin->hide();
        detailWin->hide();
        overviewTable->hide();
        overviewDetails->hide();
      }
      break;
    case TIMERPLAYBACK:
      epg->updateShort(currentShow);
      break;
    default:
      break;
    }

    return SUCCESS;
  }  
  void DvbState::hideShortEPGInfo(){
    if(epgWin){
      epgWin->hide();
      app->getOsdManager()->hideColorBar();
      updateOSD();
    }
  }
  
  void DvbState::switchOverview(epgDisplayMode_t mode){

    app->getOsdManager()->setYellowColorBarText("Now");
    app->getOsdManager()->setBlueColorBarText("Next");

    overview_table_selected = true;
    overviewTable->selectWindow();
    overviewDetails->deSelectWindow();


    switch(mode){
      
    case BACK:
      overview_mode = BACK;
      if(osd_enabled){
        win->setVisible(true);
        win->update();
      }

      overviewTable->hide();
      overviewTable->update();
      
      overviewDetails->hide();
      overviewDetails->update();
      break;
    case NOW:
      {
        if(overview_mode == NOW){
          overview_mode = BACK;

          if(osd_enabled){
            win->setVisible(true);
            win->update();
          }

          overviewTable->hide();
          overviewTable->update();
          
          overviewDetails->hide();
          overviewDetails->update();
          
        }else{

          overview_mode = NOW;;
          win->setVisible(false);
          win->update();

          epgWin->hide();
          epgWin->update();
          
          detailWin->hide();
          detailWin->update();
          
          //fill Overview with content
          overviewTable->deSelectLine();
          overviewTable->deleteAllLines();
          TVChannelList *cc = tvmanager->getChannelConfig();

          int counter = 0;
          list<TVChannel> l = cc->getChannels();
          for(list<TVChannel>::iterator i = l.begin(); i != l.end(); i++){
            list<string> lineContent;
            lineContent.push_back(i->getChannelName());
            lineContent.push_back(string(""));
            lineContent.push_back(string(""));
            overviewTable->addLine(lineContent, itoa(counter));
            counter++;
            if(tvmanager->getCurrentChannel().getName() == i->getChannelName()){
              overviewTable->selectLine(counter);
            }
          }

          overviewDetails->deleteAllLines();

          app->getOsdManager()->setGreenColorBarText("Switch");
          app->getOsdManager()->setYellowColorBarText("Close");
          
          overviewTable->show();
          overviewTable->update();
          
          overviewDetails->show();
          overviewDetails->update();
          
          app->getOsdManager()->update();
          
          epg->updateNow();
          
        }
        break;
      }
    case NEXT:
      {
        if(overview_mode == NEXT){
          overview_mode = BACK;
          if(osd_enabled){
            win->setVisible(true);
            win->update();
          }
          overviewTable->hide();
          overviewTable->update();
          
          overviewDetails->hide();
          overviewDetails->update();
          
        }else{
          overview_mode = NEXT;
          win->setVisible(false);
          win->update();
          
          epgWin->hide();
          epgWin->update();
          
          detailWin->hide();
          detailWin->update();
          
          //fill Overview with content
          overviewTable->deSelectLine();
          overviewTable->deleteAllLines();
          TVChannelList *cc = tvmanager->getChannelConfig();
          int counter = 0;
          list<TVChannel> l = cc->getChannels();
          for(list<TVChannel>::iterator i = l.begin(); i != l.end(); i++){
            list<string> lineContent;
            lineContent.push_back(i->getChannelName());
            lineContent.push_back(string(""));
            lineContent.push_back(string(""));
            overviewTable->addLine(lineContent, itoa(counter));
            counter++;
            if(tvmanager->getCurrentChannel().getName() == i->getChannelName()){
              overviewTable->selectLine(counter);
            } 
          }

          overviewDetails->deleteAllLines();
          
          app->getOsdManager()->setGreenColorBarText("Switch");
          app->getOsdManager()->setBlueColorBarText("Close");
          
          overviewTable->show();
          overviewTable->update();
          
          overviewDetails->show();
          overviewDetails->update();
          
          epg->updateNext();

          app->getOsdManager()->update();
        }
        break;
      }
    default:
      overview_mode = BACK;
      if(osd_enabled){
        win->setVisible(true);
        win->update();
      }

      overviewTable->hide();
      overviewTable->update();
      
      overviewDetails->hide();
      overviewDetails->update();

      break;
    }    

    updateOSD();

  }


  void DvbState::restartGraph(){
    try{
      //if(switchingCards){
      if( !(app->getAudioNode()->isStarted() &&  app->getDisplayNode()->isStarted())){

        cerr << " restart graph" << endl;

        connectGlobalSinks();
        synchronize(); 
      
        cerr << "synced" << endl;

        graph.activateGraph();
        currentChannel = tvmanager->getCurrentChannelNumber();
                
      
        INode *ts_node = graph.getINode(*ts_nd);
        if(ts_node){
          itimeshift = ts_node->getParentObject()->getInterface<ITimeshifting>();
        }else{
          itimeshift = NULL;
        }

        itimeshift->setModeLive(true);

        graph.startGraph();

        app->getOsdNode()->reachStarted();
        app->getDisplayNode()->reachStarted();
        app->getAudioNode()->reachStarted();

        if(tvmanager->getInterface<ITVManager>()->getTVCard())
          tvmanager->getInterface<ITVManager>()->getTVCard()->getParentObject()->registerEventListener(ITVCard::setChannel_event, &setChannel_listener);    

      }else{
        //don't know why I have to restart the graph here, because it is not stopped. seems to be a Bug in SessionSharing
        itimeshift->setModeLive(true);

        graph.startGraph();
      }
 
      }catch(const Exception &e){
        cerr << "Restaring failed: " << e << endl;
      }


    if(switchingCards){
      app->getOsdManager()->deleteMessage();
      app->getOsdManager()->update();
      switchingCards = false;
    }
  }

  Result DvbState::connectGlobalSinks(){

    Result res = FAILURE;
    //connect the Audio-Sink
    try{
      //but only, if not already connected
      if(!app->getAudioNode()->isInputConnected()){
        INode *mpa_inode = graph.getINode(mpa_dec_nd);
        if(mpa_inode){
          //bring the node in correct states for connect
          mpa_inode->reachOutputInitialized();
          app->getAudioNode()->reachInitialized();
          //and connect them
          connect(mpa_inode, app->getAudioNode());
          
          mpa_inode->reachActivated();
          res = SUCCESS;
        }
      }
    }catch(const Exception &e){
      cerr << "Erorr connecting Audio: " << e << endl;
    }

  //connect the Video-Sink
    try{
      //but only, if not already connected
      if(!app->getOsdNode()->isInputConnected("stream1")){ 
        INode *last_video_inode = graph.getINode(last_video_nd);
        if(last_video_inode){
          last_video_inode->reachOutputInitialized();
          app->getOsdNode()->reachInitialized();
          connect(last_video_inode, app->getOsdNode(),"stream1");
          last_video_inode->reachActivated();
          res = SUCCESS;
        }
      }
    }catch(const Exception &e){
      cerr << "Error connection Video: " << e << endl;
      res = FAILURE;
    }

    return res;
  }
  /***************************************/
  /*  DvbStateEPG-Class implementation   */
  /***************************************/
  DvbStateEPG::DvbStateEPG(MMBoxApplication *application, DvbState* dvbState)
    : app(application),
      cName(""),
      state(dvbState),
      currentMode(DvbStateEPG::UNKNOWN),
      newShow_listener(this, &DvbStateEPG::newShow),
      iepg(NULL),
      mutex()
  {
    time_t now = time(NULL); localtime(&now); //make sure, the timezone-vaiable is initialized
    char tmp[6];
    double d = -1*timezone/60/60*100;
    sprintf(tmp, "%0+5.0f", d);
    tz = string(tmp);

    //Connect to TVGuide and register EventListeners
    NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
    if(o){
      iepg = o->getInterface<ITVGuide>();
      iepg->getChannelsList(); //refresh our channels-List
    }else{
      GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
    }
    iepg->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newShow_event, &newShow_listener);    

  }

  DvbStateEPG::~DvbStateEPG(){
    if(iepg){
      iepg->getParentObject()->removeEventListener(&newShow_listener);    
      delete iepg;
      iepg = NULL;
    }
    if(state){
      state = NULL;
    }
  }

  Result DvbStateEPG::updateShort(string channelName){

    MutexGuard mg(mutex); 

    currentMode = SHORT;
    cName = channelName;
    activateThread();
    return SUCCESS;
  }

  Result DvbStateEPG::updateNow(){
    MutexGuard mg(mutex); 

    currentMode = NOW;
    activateThread();
    return SUCCESS;
  }

  Result DvbStateEPG::updateNext(){

    MutexGuard mg(mutex); 

    currentMode = NEXT;
    activateThread();
    return SUCCESS;

  }

  string DvbStateEPG::formatTimeString(string s){
    string ret("");
    try{
      ret += s.substr(6,2) + ".";
      ret += s.substr(4,2) + ".";
      ret += s.substr(0,4) + " ";
      
      ret += s.substr(8,2) + ":";
      ret += s.substr(10,2);
      return ret;
    }catch(const Exception &e){
      return string("");
    }catch(...){
      return string("");
    }
    
  }  

  Result DvbStateEPG::updateOverviewDescription(unsigned int index){

    if(!pdList.empty() && index <= pdList.size() && index > 0){
      list<ProgramDescription>::iterator i = pdList.begin();
      advance(i, index-1);
      state->overviewDetails->deleteAllLines();
      if(i != pdList.end()){
        if(!i->empty()){
          
          string time_line(formatTimeString(i->getStartTime()));
          if(!i->getEndTime().empty()){
            time_line += " - " + formatTimeString(i->getEndTime());
          }
          state->overviewDetails->addLine(time_line);
          state->overviewDetails->addLine(i->getChannelDescription().getChannelNames().front());
          state->overviewDetails->addLine("");
          state->overviewDetails->addLine(i->getTitles().front());
          state->overviewDetails->addLine("");
          state->overviewDetails->addMultiLine(i->getDescription());
        }
      }
      state->overviewDetails->update();
    }
    return SUCCESS;
  }


  ProgramDescription DvbStateEPG::getShow(unsigned int index){
    ProgramDescription res;
    if(!pdList.empty() && index <= pdList.size() && index > 0){
      list<ProgramDescription>::iterator i = pdList.begin();
      advance(i, index-1);
      res = *i;
    }
    return res;
  }

  Result DvbStateEPG::updateOverview(ProgramDescription &pd){
    
    if(state->overviewTable == NULL) return FAILURE;
    list<ProgramDescription>::iterator j = pdList.begin();
    for(int i = 1; i <= state->overviewTable->getNumLines(); i++){
      string channel = state->overviewTable->getLineContent(0,i);
      string channelID = iepg->getChannelByName(channel);

      if(channelID == pd.getChannelID()){
        if(state->overviewTable->getLineContent(1,i).empty()){
          pd.localizeTimes();
          state->overviewTable->setLineContent(1, i, pd.getStartTime().substr(8,4).insert(2,":"));
          state->overviewTable->setLineContent(2, i, pd.getTitles().front());
          pdList.insert(j, pd);
          pdList.erase(j); 

          state->overviewTable->update();
          app->getOsdManager()->update();
          return SUCCESS;
        }
      }
      j++;
    }
    
    return SUCCESS;
  }

  void DvbStateEPG::run() {

    switch(currentMode){
    case SHORT:
      updateShortWindows();
      break;
    case NOW:
      updateNowWindows();
      break;
    case NEXT:
      updateNextWindows();
      break;
    default:
      break;
    }


    //cleanup
    deactivateThread();
  }
  
  Result DvbStateEPG::updateShortWindows(){   

    MutexGuard mg(mutex); 

    if(state->epgWin && state->detailWin && iepg){

      state->epgWin->deleteAllLines();
      state->detailWin->deleteAllLines();
    
      string c = iepg->getChannelByName(cName);      
      currentShow = iepg->getCurrentEventOnChannel(c);
      
      if(currentShow.empty()){
        //fill with an empty line
        list<string> s;
        s.push_back(""); s.push_back("");
        state->epgWin->addLine(s);
      } else {
        currentShow.localizeTimes();
        list<string> lineContent;
        try{
          lineContent.push_back(currentShow.getStartTime().substr(8,4).insert(2,":"));
          lineContent.push_back(currentShow.getTitles().front());
        }catch(const Exception &e){
          return FAILURE;
        }
        
        state->epgWin->addLine(lineContent);
        
        list<string> titleList = currentShow.getTitles();
        list<string>::iterator titleIter = titleList.begin();
        string displayTime = currentShow.getStartTime().substr(8,4).insert(2,":") ;
        if(!currentShow.getEndTime().empty()){
          displayTime += " - " 
            + currentShow.getEndTime().substr(8,4).insert(2,":");
        }
        
        state->detailWin->addLine(displayTime);
        
        while(titleIter != titleList.end()){
          state->detailWin->addLine(*titleIter);
          titleIter++;
        }
        state->detailWin->addLine(string()); //add empty line
        
        string description = currentShow.getDescription();
        
        state->detailWin->addMultiLine(description); 
        state->detailWin->hide();
        state->detailWin->update();
        
        //register a timer at end-time of show that the display will be updated
        //Event* e = new Event("DVBEPGUPDATE", 1);
        //app->getTimerProducer()->registerTimer(stringToTime(currentShow.getEndTime())+10, e);
        //cerr << "register timer " <<  currentShow.getEndTime() << " " << stringToTime(currentShow.getEndTime()) << endl;
      }

      ProgramDescription followingShow = iepg->getFollowingEventOnChannel(c);
      if(followingShow.getTitles().empty()){
        state->epgWin->addLine(" ");
      }else{
        followingShow.localizeTimes();
        list<string> lineContent;
        try{
          lineContent.push_back(followingShow.getStartTime().substr(8,4).insert(2,":"));
          lineContent.push_back(followingShow.getTitles().front());
        }catch(const Exception &e){
          }
        
        state->epgWin->addLine(lineContent);
        
      }
 
      app->getOsdManager()->displayColorBar();
     
      if(currentShow.empty() && followingShow.empty()){
        state->epgWin->hide();
        state->epgWin->update();
        app->getOsdManager()->setColorBarText("Record","","Now","Next");
      }else{
        state->epgWin->show();
        state->epgWin->update();
        if(currentShow.empty()){
          app->getOsdManager()->setColorBarText("Record","","Now","Next");
        }else{
          app->getOsdManager()->setColorBarText("Record","Details","Now","Next");
        }
      }
    
      return SUCCESS;
    }    

    state->epgWin->hide();
    state->epgWin->update();

    return FAILURE;
  }


  Result DvbStateEPG::updateShort(ProgramDescription pd){

    MutexGuard mg(mutex); 

    if(state->epgWin && state->detailWin && !pd.empty() && iepg){

      state->epgWin->deleteAllLines();
      state->epgWin->hide();
      state->detailWin->deleteAllLines();
    

      pd.localizeTimes();
      list<string> lineContent;
      try{
        lineContent.push_back(pd.getStartTime().substr(8,4).insert(2,":"));
        lineContent.push_back(pd.getTitles().front());
      }catch(const Exception &e){
        return FAILURE;
      }
        
      state->epgWin->addLine(lineContent);
        
      list<string> titleList = pd.getTitles();
      list<string>::iterator titleIter = titleList.begin();
      string displayTime = pd.getStartTime().substr(8,4).insert(2,":") ;
      if(!pd.getEndTime().empty()){
        displayTime += " - " 
          + pd.getEndTime().substr(8,4).insert(2,":");
      }
        
      state->detailWin->addLine(displayTime);
        
      while(titleIter != titleList.end()){
        state->detailWin->addLine(*titleIter);
        titleIter++;
      }
      state->detailWin->addLine(string()); //add empty line
        
      string description = pd.getDescription();
        
      state->detailWin->addMultiLine(description); 
      state->detailWin->hide();
      state->detailWin->update();

      return SUCCESS;
    }    
    return FAILURE;

    
  }
  Result DvbStateEPG::updateNowWindows(){

    MutexGuard mg(mutex); 
    
    //fill Overview with content
    if(state->overviewTable && state->overviewDetails && iepg){

      int lines = state->overviewTable->getNumLines();
      pdList.clear();
      int sel = state->overviewTable->getSelectedLine();
      for(int i = 1; i <= lines; i++){
        if(iepg){
          string channelId = iepg->getChannelByName(state->overviewTable->getLineContent(0,i));
          ProgramDescription pd = iepg->getCurrentEventOnChannel(channelId);
          pdList.push_back(pd);
          if(!pd.empty()){
            pd.localizeTimes();
            state->overviewTable->setLineContent(1, i, pd.getStartTime().substr(8,4).insert(2,":"));
            state->overviewTable->setLineContent(2, i, pd.getTitles().front());
            state->overviewTable->update();
            app->getOsdManager()->update();
          }          
        }
      }
      state->overviewTable->selectLine(sel);
      state->overviewTable->update();
      updateOverviewDescription(sel);
      app->getOsdManager()->update();
      return SUCCESS;
    }
    return FAILURE;
  }
  
  Result DvbStateEPG::updateNextWindows(){

    MutexGuard mg(mutex); 

    //fill Overview with content
    if(state->overviewTable && state->overviewDetails && iepg){

      int lines = state->overviewTable->getNumLines();      
      pdList.clear();
      int sel = state->overviewTable->getSelectedLine();
      state->overviewTable->deSelectLine();
      for(int i = 1; i <= lines; i++){
        if(iepg){
          string channelId = iepg->getChannelByName(state->overviewTable->getLineContent(0,i));
          ProgramDescription pd = iepg->getFollowingEventOnChannel(channelId);
          pdList.push_back(pd);
          if(!pd.empty()){
            pd.localizeTimes();
            state->overviewTable->setLineContent(1, i, pd.getStartTime().substr(8,4).insert(2,":"));
            state->overviewTable->setLineContent(2, i, pd.getTitles().front());
            state->overviewTable->update();
            app->getOsdManager()->update();
          }
        }
      }
      state->overviewTable->selectLine(sel);
      state->overviewTable->update();
      updateOverviewDescription(sel);
      app->getOsdManager()->update();

      return SUCCESS;
    }
    return FAILURE;
  }

  Result DvbStateEPG::newShow(string& channel){

    if(!state->hasFocus()) return FAILURE;
    if(iepg == NULL) return FAILURE;
    ProgramDescription pd;

    switch(currentMode){
    case SHORT:
      cerr << "update short windows" << endl;
      updateShortWindows();
      return SUCCESS;
    case NOW:
      cerr << "update NOW windows" << endl;
      pd = iepg->getCurrentEventOnChannel(channel);
      break;
    case NEXT:
      cerr << "update NEXT windows" << endl;
      pd = iepg->getFollowingEventOnChannel(channel);
      break;
    default:
      return SUCCESS;
      break;
    }
    updateOverview(pd);
    return SUCCESS;
  }

}

