/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include <sstream>
#include <sys/stat.h>
#include "nmm/interfaces/net/IObjectAcceptor.hpp"
#include <time.h>
#include <unistd.h>

#include "epg.hpp"

#include "apps/mmbox/apps/webserver/webserver.hpp"
#include "apps/mmbox/apps/webserver/handler/EPGHandler.hpp"


#include "nmm/misc/dvb/channel.hpp"

#include "nmm/utils/xml/xml++.hpp"
#include "nmm/utils/utils.hpp"     // for string + int
#include "nmm/utils/DirReader.hpp"

#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"
#include "nmm/base/ServiceProxy.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"

#include "nmm/base/sync/LocalTime.hpp"
#include "nmm/base/utils/GraphURL.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

//EPG-Update-Plugins. Remove them, when Singletion-Bug is fixed
#include "nmm/interfaces/misc/tvguide/IXMLTVPlugin.hpp"
#include "nmm/interfaces/misc/tvguide/IDVBEPGPlugin.hpp"
#include "nmm/interfaces/misc/tvguide/INxtvepgPlugin.hpp"

//TVTimer-stuff
#include "nmm/misc/tvtimer/TVTimerEntry.hpp"

#include "../dvb/dvb.hpp"

// we want to be in:
namespace NMM {

  EPG::EPG(string id, MMBoxApplication* _app) 
    : MMBoxState(id, _app),
      rootpath("/"),
      now(""),
      today(""),
      currentDay(0),
      currentChannel(0),
      epgsource(""),
      currentEventIndex(1),
      currenttime(0),
      itvtimer(NULL),
      manager(NULL),
      win_focus(DAYS),
      newChannel_listener(this, &EPG::newChannel),
      newShow_listener(this, &EPG::newShow)
  {

    for( int i=0; i<4; i++ ) {
      epgWins[i] = NULL;
    }
    
    MMBoxConfig* config = app->getMMBoxConfig();
    if (!(config->getValue("rootpath", rootpath))) {
      rootpath="/";
    }    

    manager = app->getOsdManager();

    //get the interface
    string epg_location("localhost");
    if (!(config->getValue("epglocation", epg_location))) {
      GLOBAL_DEBUG_STREAM("TVGuide on localhost");
    }      


    string epgDirectory(rootpath + "/epg/");
    if(epg_location == "localhost"){
      //Test, if epg-subdir exists. if not, create it
      DirReader dr;
      dr.setRootDir(rootpath);
      if(!dr.changeDir("/epg")){
        dr.makeDir(rootpath+"/epg");
      }
      
      //Initialize TVGuide-Object and register it as Service
      itvgserver = STVGuide::getInstance().getInterface<ITVGuide>();
      if(itvgserver){
        itvgserver->setDirectory(epgDirectory);
        
        string epgPeer;
        int peerHostNumber = 0;
        if (config->getValue("epgpeers_number" , peerHostNumber)) {
          for(int i = 0; i < peerHostNumber; i++){
            char tmp[3];
            sprintf(tmp, "%i", i);
            if (config->getValue("epgpeer_" + string(tmp), epgPeer)) {
              try{
                unsigned int pos = epgPeer.find_first_of(":");
                string host(epgPeer.substr(0,pos));
                int port = 22802;
                if(pos != std::string::npos){
                  port = atoi(epgPeer.substr(++pos).c_str());
                }
                itvgserver->addPeerHost(host, port);
              }catch(const Exception &e){
                GLOBAL_ERROR_STREAM("host-String in wrong format" << endl);
              }
              
            }else{
              break;
            }
          }
        }
      }
    }else{
      
      ServiceProxy service_proxy;
      TCPAddress tcp_address = parseHostString(epg_location);

      try{
        service_proxy.connectTo(tcp_address, "TVGuide");
        itvgserver = service_proxy.getInterface<ITVGuide>();
      } catch (const BindingException& e) {
        GLOBAL_ERROR_STREAM("Could not connect to session location: " << epg_location << " because of " << e << endl);
        itvgserver = NULL;
      }
    }
    
    /*----------------------------
     * Initialize the plugins
     -----------------------------*/
    if(itvgserver){
      //register TVGuide as service internally so other states can access the same object
      SReferenceManager::getInstance().registerService("TVGuide", itvgserver->getParentObject());
      
      ITVGuidePlugin_var plugin(NULL);

      int epgSourceNumber = 0;
      if (config->getValue("epgsources_number" , epgSourceNumber)) {
        for(int i = 0; i < epgSourceNumber; i++){
          char tmp[3];
          sprintf(tmp, "%i", i);
          if (config->getValue("epgsource_" + string(tmp), epgsource)) {
              
          }else{
            break;
          }
        

          if(epgsource == "xmltv"){
            plugin.reset(itvgserver->getGrabberInterface("IXMLTVPlugin"));
            if(plugin.get()){
              IXMLTVPlugin* ixmltv= plugin->getParentObject()->getInterface<IXMLTVPlugin>();
            
              if(ixmltv){
                ixmltv->setConfigFile(epgDirectory + "/XMLTVConf.xml");
                ixmltv->setFillHolesInDB(false);
                
                string xmltv_locale("de");
                if (!(config->getValue("xmltv_locale", xmltv_locale))) {
                  GLOBAL_WARNING_STREAM("XMLTV for germany (default)" << endl);
                }
                ixmltv->setLocale(xmltv_locale);
                
                string xmltv_period;
                if (!(config->getValue("xmltv_period", xmltv_period))) {
                  GLOBAL_WARNING_STREAM("XMLTV: using default-period: d1" << endl);
                }else{
                  ixmltv->setPeriod(xmltv_period);
                }
              }
            }
          }else if(epgsource == "nxtvepg"){
            plugin.reset(itvgserver->getGrabberInterface("INxtvepgPlugin"));
            if(plugin.get()){
              INxtvepgPlugin* inxtvepg= plugin->getParentObject()->getInterface<INxtvepgPlugin>();
              
              if(inxtvepg){
                inxtvepg->setConfigFile(epgDirectory + "/NxtvepgConf.xml");
                string nxtvepgDBDir = "/usr/tmp/nxtvdb"; //Nxtvepg-default
                if (!(config->getValue("nxtvepg_dbdir", nxtvepgDBDir))) {
                  GLOBAL_DEBUG_STREAM("NXTVEPG-DatabaseDir " << nxtvepgDBDir << endl);
                }
                inxtvepg->setDBDir(nxtvepgDBDir);
                
                string nxtvepgProvider = "0d92"; //hex for RTL2
                if (!(config->getValue("nxtvepg_provider", nxtvepgProvider))) {
                  GLOBAL_DEBUG_STREAM("adding NXTVEPG-Provider " << nxtvepgProvider << "(default)" << endl);
                }
                inxtvepg->addProvider(nxtvepgProvider);
                
                string nxtvepg_period = "d1";
                if (!(config->getValue("nxtvepg_period", nxtvepg_period))) {
                  GLOBAL_WARNING_STREAM("Nxtvepg: using default-period: d1" << endl);
                }
                inxtvepg->setPeriod(nxtvepg_period); 
              }
            }
          }else if(epgsource == "dvb"){
            string dvbChannelConfig;
            if (!(config->getValue("dvbepgChannelConfig", dvbChannelConfig))) {
              dvbChannelConfig="";
            }      
            
            plugin.reset(itvgserver->getGrabberInterface("IDVBEPGPlugin"));

            if(plugin.get()){
              IDVBEPGPlugin* idvb= plugin->getParentObject()->getInterface<IDVBEPGPlugin>();
              if(idvb){
                TVManager *tvmanager = NULL;
                if(dvbChannelConfig.empty()){
                  tvmanager = new TVManager(IUsage::DVBEPG);
                }else{
                  tvmanager = new TVManager(IUsage::DVBEPG, dvbChannelConfig);
                }
              
                if(tvmanager){
                  GraphDescription *gd = new GraphDescription();
                  NodeDescription *null_nd = new NodeDescription("DevNullNode");
                
                  tvmanager->setApplication(app->getNMMApplication());
                  tvmanager->setGraphDescription(gd, null_nd);
                  
                  idvb->setTVManager(tvmanager->getInterface<ITVManager>());
                  cerr << "DVBSetup done" << endl;
                }
              }
            }
          }else{
          }
        }
      }

      //start updates only when TVGuide runs on localhost.
      if(epg_location == "localhost"){
        GLOBAL_DEBUG_STREAM("Starting TVGuide-Server locally" << endl);
        itvgserver->init();
        itvgserver->start();
        //call this here for distributed TVGuide that it starts searching for available channels
        itvgserver->getChannelsList();
      }else{
        GLOBAL_DEBUG_STREAM("Connected to TVGuide on " << epg_location << endl);
      }
    }

    //register TVGuide to WebServer-Handler
    SMMBoxWebServer::getInstance().registerHandler(new EPGHandler(itvgserver));
    
  }//end Constructor

  EPG::~EPG(){
    if(itvgserver){
      itvgserver->stop();
      delete itvgserver;
      itvgserver = NULL;
    }
  }

  void EPG::setWinProperties( int winnr, win_properties_t* prop ) {
    memcpy( &winprops[winnr], prop, sizeof( win_properties_t ) );
  }

  
  /******************** MMBoxState interface ********************/
  Result EPG::doCreate(string) {
    GLOBAL_DEBUG_STREAM("EPG: create: " << stateid << endl);

    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }
    
    //reconnect to tvguide in case the server crashed. only when location != localhost
    //get the interface
    MMBoxConfig* config = app->getMMBoxConfig();
    string epg_location("localhost");
    if (!(config->getValue("epglocation", epg_location))) {
      GLOBAL_DEBUG_STREAM("TVGuide on localhost");
    }      

    if(epg_location != "localhost"){
      ServiceProxy service_proxy;
      TCPAddress tcp_address = parseHostString(epg_location);

      try{
        SReferenceManager::getInstance().removeService("TVGuide");
        service_proxy.connectTo(tcp_address, "TVGuide");
        itvgserver = service_proxy.getInterface<ITVGuide>();
        SReferenceManager::getInstance().registerService("TVGuide", itvgserver->getParentObject());
      } catch (const BindingException& e) {
        GLOBAL_ERROR_STREAM("Could not connect to session location: " << epg_location << " because of " << e << endl);
      }
    }

    //Connect to TVTimer 
    NMMObject* o = SReferenceManager::getInstance().getService("TVTimer");
    if(o){
      itvtimer.reset(o->getInterface<ITVTimer>());
    }else{
      GLOBAL_ERROR_STREAM("Service TVTimer not available" << endl);
    }

    //get time of now :)
    currenttime = time(NULL);
    struct tm *t = localtime(&currenttime);

    //set timezone-information
    tzset();

    char tmp[25];
    strftime(tmp, sizeof(tmp), "%H%M", localtime(&currenttime));     
    now = string(tmp);
    strftime(tmp, sizeof(tmp), "%Y%m%d%H%M", localtime(&currenttime));  
    
    double timeDifference = -1*timezone/60/60 * 100; //in hours
    
    if(daylight != 0) timeDifference += t->tm_isdst*100;
    sprintf(tmp, "%s %0+5.0f", tmp, timeDifference);
    today=string(tmp);
    strftime(tmp, sizeof(tmp), "%u", localtime(&currenttime));     
    currentDay = atoi(tmp);
    

    currentChannel = 0;

    createWindow(DAYS    , today.substr(6,2) + "." + today.substr(4,2) + "." + today.substr(0,4));
    createWindow(CHANNELS, string("Channel"));
    createWindow(SCHEDULE, string("Schedule"));
    createWindow(EXTINFO,  string("Extended Information"));

    createContent();

    selectWindow(win_focus);
    showOSD();
    updateContent();
    updateOSD();

    if(itvgserver){
      itvgserver->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newChannel_event, &newChannel_listener);
      itvgserver->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newShow_event, &newShow_listener);
    }   
    return SUCCESS;
  }

  Result EPG::doKill() {

    if(itvgserver){
      itvgserver->getParentObject()->removeEventListener(&newChannel_listener);
      itvgserver->getParentObject()->removeEventListener(&newShow_listener);
    }
    hideOSD();
    deleteWindows();

    if (isBackgroundFree()) {
      app->disconnectVideoGraph();
    }
    win_focus = DAYS;
    shows.clear();
    return SUCCESS;
  }

  Result EPG::doInitialize(string) {
    GLOBAL_DEBUG_STREAM("EPG: initialize: " << stateid << endl);

    if(itvgserver){
      itvgserver->getParentObject()->removeEventListener(&newShow_listener);
      itvgserver->getParentObject()->removeEventListener(&newChannel_listener);
    }
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }
   
    showOSD();
    updateContent();
    updateOSD();

    if(itvgserver){
      itvgserver->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newChannel_event, &newChannel_listener);
      itvgserver->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newShow_event, &newShow_listener);
    }
    return SUCCESS;
  }

  Result EPG::doUninitialize() {
    GLOBAL_DEBUG_STREAM("EPG: initialize: " << stateid << endl);
 
    hideOSD();

    return SUCCESS;
  }

  Result EPG::Down() {
    GLOBAL_DEBUG_STREAM("EPG: Down" << endl);

    int line = epgWins[win_focus]->getSelectedLine();

    if(win_focus == DAYS && line >= 7){
      //go to the days of next week
      string selectedDate(epgWins[DAYS]->getLineURL(line));
      today = getDate(selectedDate, 1);
      currentDay = 1;
      createContent();
    }else if(win_focus == SCHEDULE){
      epgWins[win_focus]->selectLine(line+1);
      currentChannel = epgWins[CHANNELS]->getSelectedLine();
      int max = shows.size();
      if(currentEventIndex < max) currentEventIndex++;

      list<ProgramDescription>::iterator i = shows.begin();
      advance(i, currentEventIndex-1);
      if(i != shows.end()){i->localizeTimes(); now = (*i).getStartTime().substr(8,4);}
    }else{
      epgWins[win_focus]->selectLine(line+1);
      currentChannel = epgWins[CHANNELS]->getSelectedLine();
    }

    updateContent();
    updateOSD();
    return SUCCESS;
  }

  Result EPG::Up() {
    GLOBAL_DEBUG_STREAM("EPG: Up" << endl);

    int line = epgWins[win_focus]->getSelectedLine();

    if(win_focus == DAYS && line <= 1){
      string selectedDate(epgWins[DAYS]->getLineURL(line));
      today = getDate(selectedDate, -1);
      currentDay = 7;
      createContent();
    }else if(win_focus == SCHEDULE){
      epgWins[win_focus]->selectLine(line-1);
      currentChannel = epgWins[CHANNELS]->getSelectedLine();
      if(currentEventIndex > 1){
        currentEventIndex--;
      }else{
      }

      list<ProgramDescription>::iterator i = shows.begin();
      advance(i, currentEventIndex-1);
      if(i != shows.end()){i->localizeTimes(); now = (*i).getStartTime().substr(8,4);}
    }else{
      epgWins[win_focus]->selectLine(line-1);
      currentChannel = epgWins[CHANNELS]->getSelectedLine();
    }

    updateContent();
    updateOSD();

    return SUCCESS;
  }

  Result EPG::Left() {
    GLOBAL_DEBUG_STREAM("EPG: Left" << endl);

    int temp = win_focus;
    win_focus = (wintype_e)--temp;
    if(win_focus < DAYS)
      win_focus = EXTINFO;

    selectWindow(win_focus);
    updateOSD();
    return SUCCESS;
  }

  Result EPG::Right() {
    GLOBAL_DEBUG_STREAM("EPG: Right" << endl);

    int temp = win_focus;
    win_focus = (wintype_e)++temp;
    if(win_focus > EXTINFO)
      win_focus = DAYS;

    selectWindow(win_focus);
    updateOSD();
    return SUCCESS;
  }

  Result EPG::Back() {
    GLOBAL_DEBUG_STREAM("EPG: Back" << endl);
    hideOSD();

    return goToLastState();
  }


  Result EPG::specialKey01() {
    GLOBAL_DEBUG_STREAM("EPG: specialKey01" << endl);

    return SUCCESS;
  }

  Result EPG::specialKey02() {
    GLOBAL_DEBUG_STREAM("EPG: specialKey02" << endl);

    return SUCCESS;
  }

  Result EPG::colorKeyRed(){
    GLOBAL_DEBUG_STREAM("EPG: colorKeyRed" << endl);

    if(!epgWins[SCHEDULE]->getLineURL(epgWins[SCHEDULE]->getSelectedLine()).empty()){
      return FAILURE;
    }

    TvTimer::TvTimer* timer = dynamic_cast<TvTimer*>(app->getState("TV-Timer"));
    int line =epgWins[SCHEDULE]->getSelectedLine();
    
    if(line != 0){
      try{
        list<ProgramDescription>::iterator it = shows.begin();
        if(line > 0){
          advance(it, line-1);
        }
        if(it != shows.end()){
          ProgramDescription *pd = new ProgramDescription(*it);
          pd->localizeTimes();
          if(pd->getEndTime().empty()){
            it++;
            pd->setEndTime(it->getStartTime());
          }

          int res = timer->programTimer(*pd);
          delete pd;
          if(res >= 0){
            
            char tmp[5];
            sprintf(tmp, "%i", res);
            hideOSD();
            changeState("TV-Timer",string(tmp));
            
          }else{
            manager->displayMessage("Channel Not Configured" , "Timer Not programmed!", "OK" );
            manager->update();
            changeState("OSDInputState");
          }
        }
        return SUCCESS;
      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("Something went wrong during timer-programming: " << e << endl);
        manager->displayMessage("ERROR" , "Timer Not programmed!", "OK" );
        manager->update();
        changeState("OSDInputState");
        return FAILURE;
      } 
    }else{
      return FAILURE;
    }
  }
  Result EPG::colorKeyBlue(){
    int line = epgWins[CHANNELS]->getSelectedLine();

    if(line != 0){
      try{
        string channel = itvgserver->getChannelById(epgWins[CHANNELS]->getLineURL(line));
        DvbState* dvbstate = dynamic_cast<DvbState*>(app->getState("DVB-TV"));
        int c = dvbstate->getChannelConfig()->getChannelByName(channel).getChannelNumber();
        if(c >= 0){
          if(dvbstate) dvbstate->setTvChannel(c);
          changeState("DVB-TV"); 
        }else{
          manager->displayMessage("Error" , "Channel not available", "OK" );
          manager->update();
          changeState("OSDInputState");
        }
      }catch(const Exception &e){
        cerr << "args" << endl;
      }

      return SUCCESS;
    }else{
      return FAILURE;
    }
  }
  Result EPG::colorKeyYellow(){
    return SUCCESS;
  }
  Result EPG::colorKeyGreen(){
    return SUCCESS;
  }
  
  Result EPG::Return() {
    return SUCCESS;
  }

  Result EPG::Record(){
    GLOBAL_DEBUG_STREAM("EPG: Record" << endl);

    if(win_focus == SCHEDULE || win_focus == EXTINFO){
      TvTimer::TvTimer* timer = dynamic_cast<TvTimer*>(app->getState("TV-Timer"));

      try{
        list<ProgramDescription>::iterator it = shows.begin();
        advance(it, currentEventIndex-1);
        if(it != shows.end()){
          it->localizeTimes();
          int res = timer->programTimer(*it);
          if(res >= 0){

            char tmp[5];
            sprintf(tmp, "%i", res);
	    hideOSD();
            changeState("TV-Timer",string(tmp));

          }else{
            manager->displayMessage("Channel Not Configured" , "Timer Not programmed!", "OK" );
            manager->update();
	    changeState("OSDInputState");
          }
        }
        return SUCCESS;
      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("Something went wrong during timer-programming: " << e << endl);
        manager->displayMessage("ERROR" , "Timer Not programmed!", "OK" );
        manager->update();
	changeState("OSDInputState");
        return FAILURE;
      } 
    }
    return FAILURE;
  }
                        
  Result EPG::newChannel(){
    if(hasFocus()){
      if(epgWins[CHANNELS]->getSelectedLine() == 0){
        currentChannel = 1;
        updateChannels();
        updateShows();
        updateExtInfo();
      }else{
        updateChannels();
      }
      updateOSD();
    }
    return SUCCESS;
  }

  Result EPG::newShow(string& channelID){
    if(hasFocus()){
      int sel = epgWins[CHANNELS]->getSelectedLine();
      if(epgWins[CHANNELS]->getLineURL(sel) == channelID){
        //cerr << epgWins[CHANNELS]->getLineURL(sel) << " == " <<  channelID << endl;
        updateShows();
        updateExtInfo();
        updateOSD();
      }else{
        //cerr << epgWins[CHANNELS]->getLineURL(sel) << " != " <<  channelID << endl;
      }
    }
    return SUCCESS;
  }

  /******************** MMBoxState interface ********************/


  /******************** private helpers ********************/
  void EPG::deleteWindows() {
    for( int i=0; i<4; i++ ) {
      if ( epgWins[i] ) {
        delete epgWins[i];
        epgWins[i] = NULL;
      }
    }
  }

  void EPG::createWindow( wintype_e w, string h ) {
    if ( epgWins[w] ) {
      delete epgWins[w];
      epgWins[w] = NULL;
    }
    
    IWidgetFactory& factory = app->getOSDWidgetFactory();
    epgWins[w] = new TableUnit( manager, &factory, "medium");
    epgWins[w]->setHeader( h );
    
    if(w == DAYS){
      epgWins[w]->setScrollbarEnabled(false);
    }

    if(w == SCHEDULE){
      epgWins[w]->addColumn(17);
      epgWins[w]->addColumn(70);
    }

    /* set colors according to properties */
    epgWins[w]->setColors( winprops[w].win_active_color,
                                         winprops[w].win_inactive_color,
                                         winprops[w].sel_active_color,
                                         winprops[w].sel_active_color2,
                                         winprops[w].sel_inactive_color );
    
    /* set width and height of window */
    epgWins[w]->setDimension( winprops[w].width, winprops[w].height );
    
    /* set position of the window */
    epgWins[w]->setPosition( winprops[w].xpos, winprops[w].ypos );
    
    epgWins[w]->deSelectWindow();

    epgWins[w]->update();
    epgWins[w]->show();
    
    /* add window to OSDManagerNode */
    //manager->update();
  }

  void EPG::updateOSD(){    
    for(int i=0; i<4; i++)
      epgWins[i]->update();


    manager->setColorBarText("", "", "", "");


    int line = epgWins[CHANNELS]->getSelectedLine();
    string channel = itvgserver->getChannelById(epgWins[CHANNELS]->getLineURL(line));      
    if(itvtimer.get()){
      try{
        itvtimer->getChannelConfig().getChannelByName(channel);
        if(!shows.empty()){
          manager->setRedColorBarText("Record");
        }
      }catch(const Exception &e){
        manager->setRedColorBarText("");
      }
    }

    try{
      DvbState* dvbstate = dynamic_cast<DvbState*>(app->getState("DVB-TV"));
      if(dvbstate){
        dvbstate->getChannelConfig()->getChannelByName(channel).getChannelNumber();
        manager->setBlueColorBarText("Switch");
      }
    }catch(const Exception &e){
      manager->setBlueColorBarText("");
    }

    manager->displayColorBar();

    manager->update();

  }

  Result EPG::showOSD(){
    for(int i=0; i<4; i++)
      epgWins[i]->show();

    manager->displayColorBar();

    
    return SUCCESS;
  }

  Result EPG::hideOSD(){
    manager->hideColorBar();

    for(int i=0; i<4; i++)
      epgWins[i]->hide();

    return SUCCESS;
  }

  void EPG::selectWindow(int id){
    win_focus = (wintype_e)id;
    // activate/deactivate windows according to focus 
    switch(win_focus) {
    case DAYS:
      epgWins[DAYS]->selectWindow();
      epgWins[CHANNELS]->deSelectWindow();
      epgWins[SCHEDULE]->deSelectWindow();
      epgWins[EXTINFO]->deSelectWindow();
      break;
    case CHANNELS:
      epgWins[DAYS]->deSelectWindow();
      epgWins[CHANNELS]->selectWindow();
      epgWins[SCHEDULE]->deSelectWindow();
      epgWins[EXTINFO]->deSelectWindow();
      break;
    case SCHEDULE:
      epgWins[DAYS]->deSelectWindow();
      epgWins[CHANNELS]->deSelectWindow();
      epgWins[SCHEDULE]->selectWindow();
      epgWins[EXTINFO]->deSelectWindow();
      break;
    case EXTINFO:
      epgWins[DAYS]->deSelectWindow();
      epgWins[CHANNELS]->deSelectWindow();
      epgWins[SCHEDULE]->deSelectWindow();
      epgWins[EXTINFO]->selectWindow();
      break;
    default:
      break;
    }

    //updateContent();
    updateOSD();
  }

  void EPG::createContent(){
    GLOBAL_DEBUG_STREAM("EpgState: creating content ..." << endl);

    for( unsigned int i=0; i<4; i++ ) {
      epgWins[i]->deleteAllLines();
    }

    //write days to window
    for(int i=1; i < 8; i++){
      switch(i){
      case MO:       
        epgWins[DAYS]->addLine("Monday", getDate(today, i-currentDay));
        break;
      case DI:
        epgWins[DAYS]->addLine("Tuesday",getDate(today, i-currentDay));
        break;
      case MI:
        epgWins[DAYS]->addLine("Wednesday",getDate(today, i-currentDay));
        break;
      case DO:
        epgWins[DAYS]->addLine("Thursday",getDate(today, i-currentDay));
        break;
      case FR:
        epgWins[DAYS]->addLine("Friday",getDate(today, i-currentDay));
        break;
      case SA:
        epgWins[DAYS]->addLine("Saturday",getDate(today, i-currentDay));
        break;
      case SO:
        epgWins[DAYS]->addLine("Sunday",getDate(today, i-currentDay));
        break;
      default:
        break;
      }
    }

    if(itvgserver){
      list<ChannelDescription> lChanDesc = itvgserver->getChannelsList();

      for(list<ChannelDescription>::iterator it = lChanDesc.begin(); it != lChanDesc.end(); it++){
        epgWins[CHANNELS]->addLine((it->getChannelNames()).front(), it->getID());
      }
      
    }else{
      GLOBAL_ERROR_STREAM("Interface ITVGuide not available!!" <<endl);
    }
    GLOBAL_DEBUG_STREAM("EpgState: content created!" << endl);

    epgWins[DAYS]->selectLine(currentDay);
    epgWins[CHANNELS]->selectLine(currentChannel);
  }

  string EPG::convertTime(string time, bool longtime){
    //convert the time in a more readable format
    string year, month, day, hour, minute;
    
    try{
      year = time.substr(0,4);
      month = time.substr(4,2);
      day = time.substr(6,2);
      hour = time.substr(8,2);
      minute = time.substr(10,2);
      
      if(longtime){
        return string(day + "." + month + "." + year + " " + hour + ":" + minute); 
      }else{
        return string(hour + ":" + minute);
      }
    }catch(const Exception &e){
      return string();
    }catch(...){
      return string();
    }

  }
  
  void EPG::updateContent(){
    GLOBAL_DEBUG_STREAM("EpgState: updateing content ..." << endl);

    if(win_focus == DAYS){
      updateChannels();
      updateShows();
      updateExtInfo();
    }else if(win_focus == CHANNELS){
      updateShows();
      updateExtInfo();
    }else if(win_focus == SCHEDULE){ 
      updateExtInfo();
    }else if(win_focus == EXTINFO){
      
    }//end else if
    GLOBAL_DEBUG_STREAM("EpgState: content updated!" << endl);

  }

  void EPG::updateChannels(){
      currentEventIndex = epgWins[SCHEDULE]->getSelectedLine();
      
      MutexGuard mg(mutex);

      int line = epgWins[DAYS]->getSelectedLine();

      string filename = epgWins[DAYS]->getLineURL(line) + ".xml";
      
      epgWins[DAYS]->setHeader(filename.substr(6,2) + "." + filename.substr(4,2) + "." + filename.substr(0,4));

      if(itvgserver){
        list<ChannelDescription> lChanDesc = itvgserver->getChannelsList();        
        list<ChannelDescription>::iterator it = lChanDesc.begin();
        epgWins[CHANNELS]->deleteAllLines();

        while (it != lChanDesc.end()){
          ChannelDescription cur = *it;
          it++;
          epgWins[CHANNELS]->addLine((cur.getChannelNames()).front(), cur.getID()); 
        }       
      }else{
        GLOBAL_ERROR_STREAM("Interface ITVGuide not available!!" <<endl);
      }
      epgWins[CHANNELS]->selectLine(currentChannel);
  }

  void EPG::updateShows(){
    epgWins[SCHEDULE]->deleteAllLines();
    
    currentEventIndex = 1;

    int line = epgWins[CHANNELS]->getSelectedLine();
    string curChannel(epgWins[CHANNELS]->getLineURL(line));

    line = epgWins[DAYS]->getSelectedLine();
    string curDay(epgWins[DAYS]->getLineURL(line));
    curDay = curDay.substr(0,8);

    int linecount = 1;


    time_t n = time(NULL); 
    struct tm *t = localtime(&n);
    double d = -1*timezone/60/60*100;
    
    if(daylight != 0) d += t->tm_isdst*100;

    char tmp[6];
    sprintf(tmp, " %0+5.0f", d);
    string tz(tmp);

    if(itvgserver){
      shows = itvgserver->getEventsOnDayOnChannel(curDay + tz, curChannel);
    }else{
      GLOBAL_ERROR_STREAM("Interface ITVGuide not available!!" << endl);
    }

    TvTimer::TvTimer* timer = dynamic_cast<TvTimer*>(app->getState("TV-Timer"));
    list<TVTimerEntry> l = timer->getTimers(curDay, curChannel);

    for(list<ProgramDescription>::iterator iter = shows.begin(); iter != shows.end(); iter++){
      iter->localizeTimes();
      list<string> sl;

      string showInfo("");
      sl.push_back(convertTime(iter->getStartTime()));

      try{
        //select current show
        if( stringToTime(curDay + now) >= stringToTime(iter->getStartTime()))
          {
            string endTime(iter->getEndTime());
            if(endTime.empty()){
              list<ProgramDescription>::iterator tmpiter = iter;
              tmpiter++;
              if(tmpiter != shows.end()) {
                endTime = tmpiter->getStartTime();
              }else{
                currentEventIndex = l.size()-1;
              }
            }
            if(!endTime.empty()){
              if(stringToTime(curDay + now) < stringToTime(endTime))
                currentEventIndex=linecount;
            }
          }
      }catch(const Exception &e){
      }catch(...){
      }

      //get first title of show and display it. all titles will be displayed in EXTINFO
      sl.push_back(iter->getTitles().front());

      //check, if timer for show is programmed
      string recorded;
      string periodInformation;
      for(list<TVTimerEntry>::iterator i = l.begin(); i != l.end(); i++){
        if(i->getStartTime() <= stringToTime(iter->getStartTime()) && i->getEndTime() >= stringToTime(iter->getEndTime())
           || i->getStartTime() > stringToTime(iter->getStartTime()) && i->getEndTime() < stringToTime(iter->getEndTime())){
          recorded = "R";
          switch(i->getPeriod()){
          case TVTimerEntry::ONCE:
            periodInformation = "Once";
            break;
          case TVTimerEntry::DAILY:
            periodInformation = "Daily";
            break;
          case TVTimerEntry::WEEKLY:
            periodInformation = "Weekly";
            break;
          case TVTimerEntry::MONTHLY:
            periodInformation = "Monthly";
            break;
          case TVTimerEntry::WEEKDAYS:
            periodInformation = "Weekdays";
            break;
          case TVTimerEntry::WEEKENDS:
            periodInformation = "Weekends";
            break;
          case TVTimerEntry::RANDOM:
            periodInformation = "RANDOM";
            if(!i->getNextTimers().empty())
              periodInformation += ": " + convertTime(i->getNextTimers().front().getStartTime(), true);
            break;
          default:
            break;
          }
        }
      }//end for(TVTimerEntry)
      
      sl.push_back(recorded);
      epgWins[SCHEDULE]->addLine(sl, periodInformation);
      linecount ++;


    }//end while shows

    epgWins[SCHEDULE]->selectLine(currentEventIndex);
    
  }
  
  void EPG::updateExtInfo(){
    
    epgWins[EXTINFO]->deleteAllLines();
    epgWins[EXTINFO]->deSelectLine();   
   
    list<ProgramDescription>::iterator iter= shows.begin();
    if(currentEventIndex > 0)
      advance(iter, currentEventIndex-1);
    if(iter == shows.end()) return;
    
    //get titles of show and display them
    list<string> titleList = iter->getTitles();
    for(list<string>::iterator titleIter = titleList.begin();  titleIter != titleList.end(); titleIter++){
      epgWins[EXTINFO]->addLine(*titleIter);    
    }//end while titleList
    
    unsigned int winWidth; 
    unsigned int winHeight;
    epgWins[EXTINFO]->getWindow()->getDimension(winWidth, winHeight);

    stringstream s(iter->getDescription());
    unsigned int linelength = 0;
    string tmp;
    string line;
       
    //format the string for display
    while(s>>tmp)
      {
        linelength += tmp.length() *20; //length of word in pixel (approx)
        
        if(tmp == "/"){
          epgWins[EXTINFO]->addLine(line);
          line="";
          linelength = 0;
        }else if(tmp == ":"){
          line += tmp + " ";
          epgWins[EXTINFO]->addLine(line);
          line="";
          linelength = 0; 
        }else if(tmp == "..."){
          line += tmp + " ";
          epgWins[EXTINFO]->addLine(line);
          epgWins[EXTINFO]->addLine(string());
          line="";
          linelength = 0;
        }else{
          if(linelength < winWidth)
            {    
              line += tmp + " ";
            }else{
              line += tmp + " ";
              epgWins[EXTINFO]->addLine(line);
              line="";
              linelength = 0;
            }
        }
      }

    epgWins[EXTINFO]->addLine(line);    

    string periodInfo = epgWins[SCHEDULE]->getLineURL(epgWins[SCHEDULE]->getSelectedLine());
    if(periodInfo != "" ){
      epgWins[EXTINFO]->addLine("");
      epgWins[EXTINFO]->addLine("Period:");
      epgWins[EXTINFO]->addLine(epgWins[SCHEDULE]->getLineURL(epgWins[SCHEDULE]->getSelectedLine()));
    }

  }
  
  TCPAddress EPG::parseHostString(const string s){
    string host("localhost");
    int port = ClientRegistry::default_port;

      
    try{
      unsigned int pos = s.find_first_of(":");
      host = s.substr(0,pos);
      if(pos != std::string::npos){
        port = atoi(s.substr(++pos).c_str());
      }
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("host-String in wrong format" << endl);
    }

    return TCPAddress(host, port);

  }

  string EPG::getDate(string date, int div){

    int year, month, day, h, min;
    h = min = 0;
    try{
      if(date.length() >= 4){
        year = atoi(date.substr(0,4).c_str());
      }else{
        year = 0;
      }
      if(date.length() >= 6){
        month = atoi(date.substr(4,2).c_str());
      }else{
        month = 0;
      }
      if(date.length() >= 8){
        day = atoi(date.substr(6,2).c_str());
      }else{
        day = 0;
      }

      if(date.length() >= 10){
        h    = atoi(date.substr(8,2).c_str() );
      }else{
        h = 0;
      }
      if(date.length() >=12){
        min  = atoi(date.substr(10,2).c_str() );    
      }else{
        min = 0;
      }

      time_t inDate = time(NULL);
      struct tm *id = localtime(&inDate);
      id->tm_sec = 0;
      id->tm_min = min;
      id->tm_hour = h;
      id->tm_mday = day + div;
      id->tm_mon = month-1;
      id->tm_year = year-1900;
      inDate = mktime(id);
      
      char outDate[12];
      strftime(outDate, sizeof(outDate), "%Y%m%d%H%M", localtime(&inDate));     
      if(date.length() >= 13){
        return string(outDate) + date.substr(13);
      }else{
        return string(outDate);
      }
    }catch(const Exception &e){
      return string();
    }
  }

  /******************** private helpers ********************/


} // end namespace NMM  


   
