/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include <sstream>
#include <sys/stat.h>
#include <time.h>
#include <fstream>

#include "search.hpp"

#include "nmm/base/registry/ReferenceManager.hpp"
#include "nmm/base/sync/LocalTime.hpp"

#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/xml/xml++.hpp"
#include "nmm/utils/utils.hpp"     // for string + int
#include "nmm/utils/DirReader.hpp"

#include "search_result.hpp"



// we want to be in:
namespace NMM {

  EPGSearch::EPGSearch(string id, MMBoxApplication* _app) 
    : MMBoxState(id, _app),
      mutex(),
      itvgserver(NULL),
      manager(NULL),
      bookmarks(NULL),
      starttime(NULL),
      endtime(NULL),
      phrase(NULL),
      channel(NULL),
      categoryInput(NULL),
      category(NULL),
      bookmarkName(NULL),
      win_focus(BOOKMARKS),
      newChannel_listener(this, &EPGSearch::newChannel)
  {

  }//end Constructor

  EPGSearch::~EPGSearch(){

  }

  void EPGSearch::setWinProperties( int winnr, win_properties_t* prop ) {
    memcpy( &winprops[winnr], prop, sizeof( win_properties_t ) );
  }

  
  /******************** MMBoxState interface ********************/
  Result EPGSearch::doCreate(string) {
    GLOBAL_DEBUG_STREAM("EPGSearch: create: " << stateid << endl);

    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }

    manager = app->getOsdManager();

    createWindow(BOOKMARKS, "Bookmarks");
    createWindow(STARTTIME, "Start Time");
    createWindow(ENDTIME, "End Time");
    createWindow(PHRASE, "Search Phrase");
    createWindow(CHANNEL, "Channel");
    createWindow(CATINPUT, "Filter Categories");
    createWindow(CATEGORY, "Category");
    createWindow(BOOKMARKNAME,"Enter Bookmark Name");

    //itvgserver = STVGuide::getInstance().getInterface<ITVGuide>();
    NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
    if(o){
      itvgserver = o->getInterface<ITVGuide>();
    }else{
      GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
    }

    readBookmarks();
    initializeWindows();
    win_focus = BOOKMARKS;
    updateOSD();

    itvgserver->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newChannel_event, &newChannel_listener);

    return SUCCESS;
  }

  Result EPGSearch::doKill() {
    GLOBAL_DEBUG_STREAM("EPGSearch: doKill()" << endl);
    
    itvgserver->getParentObject()->removeEventListener(&newChannel_listener);

    manager->deleteMessage();
    manager->update();

    bookmarkName->hide();
    bookmarkName->update();
   
    saveBookmarks();
    hideFilterOSD();
    deleteWindows();
    return SUCCESS;
  }

  Result EPGSearch::doInitialize(string) {
    GLOBAL_DEBUG_STREAM("EPGSearch: initialize: " << stateid << endl);

    
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }
    
    showFilterOSD();
    fillFromBookmarks(bookmarks->getSelectedLine()-1);
    updateOSD();

    itvgserver->getParentObject()->registerEventListener(ITVGuideUpdateFinished::newChannel_event, &newChannel_listener);

    return SUCCESS;
  }

  Result EPGSearch::doUninitialize() {
    GLOBAL_DEBUG_STREAM("EPGSearch: uninitialize: " << stateid << endl);
    itvgserver->getParentObject()->removeEventListener(&newChannel_listener);

    bookmarkName->hide();
    bookmarkName->update();
    hideFilterOSD();
    return SUCCESS;
  }

  Result EPGSearch::Down() {
    GLOBAL_DEBUG_STREAM("EPGSearch: Down" << endl);
    int line = 0;
    switch(win_focus){
    case BOOKMARKS: 
      line = bookmarks->getSelectedLine();
      bookmarks->selectLine(line+1);
      fillFromBookmarks(line);
      break;
    case STARTTIME:
      //starttime->selectLine(1);
      break;
    case ENDTIME:
      //endtime->selectLine(1);
      break;
    case PHRASE:
      //phrase->selectLine(1);
      break;
    case CHANNEL:
      line = channel->getSelectedLine();
      channel->selectLine(line+1);
      break;
    case CATEGORY:
      line = category->getSelectedLine();
      category->selectLine(line+1);
      break;
    default:
      break;
    }

    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::Up() {
    GLOBAL_DEBUG_STREAM("EPGSearch: Up" << endl);
    int line = 0;
    switch(win_focus){
    case BOOKMARKS: 
      line = bookmarks->getSelectedLine();
      bookmarks->selectLine(line-1);
      fillFromBookmarks(line-2);
      break;
    case STARTTIME:
      //starttime->selectLine(1);
      break;
    case ENDTIME:
      //endtime->selectLine(1);
      break;
    case PHRASE:
      //phrase->selectLine(1);
      break;
    case CHANNEL:
      line = channel->getSelectedLine();
      channel->selectLine(line-1);
      break;
    case CATEGORY:
      line = category->getSelectedLine();
      category->selectLine(line-1);
      break;
    default:
      break;
    }

    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::Right() {
    GLOBAL_DEBUG_STREAM("EPGSearch: Right" << endl);
    int line = 0;
    switch(win_focus){
    case BOOKMARKS:
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->selectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = CHANNEL;
      break;
    case CHANNEL:
      line = bookmarks->getSelectedLine();
      if(line == 2 || line == 3){
        bookmarks->deSelectWindow();
        starttime->deSelectWindow();
        endtime->deSelectWindow();
        phrase->selectWindow();
        channel->deSelectWindow();
        categoryInput->deSelectWindow();
        category->deSelectWindow();
        win_focus = PHRASE;
      }else{
        line = bookmarks->getSelectedLine();
        bookmarks->deSelectWindow();
        starttime->selectWindow();
        endtime->deSelectWindow();
        phrase->deSelectWindow();
        channel->deSelectWindow();
        categoryInput->deSelectWindow();
        category->deSelectWindow();
        win_focus = STARTTIME;
      }
      break;
    case STARTTIME:
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->selectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = ENDTIME;
      break;
    case ENDTIME:
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->selectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = PHRASE;
      break;
    case PHRASE:
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->selectWindow();
      category->deSelectWindow();
      win_focus = CATINPUT;
      break;
    case CATINPUT:
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->selectWindow();     
      win_focus = CATEGORY;
      break;
   case CATEGORY:
      bookmarks->selectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = BOOKMARKS;
      break;
    default:
      break;
    }

    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::Left() {
    GLOBAL_DEBUG_STREAM("EPGSearch: Left" << endl);
    int line = 0;
    switch(win_focus){
    case BOOKMARKS: 
      line = bookmarks->getSelectedLine();
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->selectWindow();
      win_focus = CATEGORY;
      break;
    case CHANNEL:
      bookmarks->selectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = BOOKMARKS;  
      break;
    case STARTTIME:
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->selectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = CHANNEL;
      break;
    case ENDTIME:
      bookmarks->deSelectWindow();
      starttime->selectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = STARTTIME;
      break;
    case PHRASE:
      line = bookmarks->getSelectedLine();
      if(line == 2 || line == 3){
        bookmarks->deSelectWindow();
        starttime->deSelectWindow();
        endtime->deSelectWindow();
        phrase->deSelectWindow();
        channel->selectWindow();
        categoryInput->deSelectWindow();
        category->deSelectWindow();
        win_focus = CHANNEL;    
      }else{
        bookmarks->deSelectWindow();
        starttime->deSelectWindow();
        endtime->selectWindow();
        phrase->deSelectWindow();
        channel->deSelectWindow();
        categoryInput->deSelectWindow();
        category->deSelectWindow();
        win_focus = ENDTIME;    
      }
      break;
    case CATINPUT:      
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->selectWindow();
      channel->deSelectWindow();
      categoryInput->deSelectWindow();
      category->deSelectWindow();
      win_focus = PHRASE;
      break;
    case CATEGORY:      
      bookmarks->deSelectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      category->deSelectWindow();
      categoryInput->selectWindow();
      win_focus = CATINPUT;
      break;
    default:
      break;
    }

    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::Back() {
    GLOBAL_DEBUG_STREAM("EPGSearch: Back" << endl);

    bookmarkName->hide();
    bookmarkName->update();

    hideFilterOSD();
    
    return goToLastState();
  }

  Result EPGSearch::Return() {
    GLOBAL_DEBUG_STREAM("EPGSearch: Return" << endl);
    manager->deleteMessage();

    if(win_focus == BOOKMARKNAME){
      
      
      return SUCCESS;
    }else{
    
    }  
    
    return SUCCESS;
    
  }

  Result EPGSearch::Record(){
    return SUCCESS;
  }

  Result EPGSearch::specialKey01(){
    return SUCCESS;
  }

  Result EPGSearch::specialKey02(){
    return SUCCESS;
  }

  Result EPGSearch::numKey0(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "0");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(0);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(0);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(0);
    }
    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::numKey1(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "1");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(1);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(1);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(1);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey2(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "2");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(2);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(2);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(2);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey3(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "3");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(3);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(3);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(3);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey4(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "4");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(4);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(4);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(4);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey5(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "5");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(5);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(5);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(5);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey6(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "6");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(6);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(6);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(6);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey7(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "7");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(7);
      //}else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(7);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(7);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey8(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "8");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(8);
      //   }else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(8);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(8);
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::numKey9(){
    if(win_focus == STARTTIME || win_focus == ENDTIME){
      dateInput(win_focus, "9");
    }else if(win_focus == PHRASE){
      phrase->keyPressed(9);
      //    }else if(win_focus == CATEGORY){
    }else if(win_focus == CATINPUT){
      categoryInput->keyPressed(9);
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->keyPressed(9);
    }
    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::Rewind(){
    if(win_focus == STARTTIME){
      int stringLength = 0;
      stringLength = starttime->getContent().size();
      switch(stringLength){
      case 4:
        starttime->erase();
        break;
      case 7:
        starttime->erase();
        break;
      case 11:
        starttime->erase();
        break;
      case 14:
        starttime->erase();
        break;
      }
      starttime->erase();
      
    }else if(win_focus == ENDTIME){
      int stringLength = 0;
      stringLength = endtime->getContent().size();
      switch(stringLength){
      case 4:
        endtime->erase();
        break;
      case 7:
        endtime->erase();
        break;
      case 11:
        endtime->erase();
        break;
      case 14:
        endtime->erase();
        break;
      }
      endtime->erase();

    }else if(win_focus == PHRASE){
      phrase->erase();
    }else if(win_focus == CATINPUT){
      categoryInput->erase();
      int numLines = category->getTextWidget()->getNumLines();
      int strSize = categoryInput->getContent().size();
      
      for(int i = 2; i <= numLines; i++){
        string s1 = category->getLineContent(i).substr(0, strSize) ;
        string s2 = categoryInput->getContent();
        transform(s1); transform(s2);
        if(s1 == s2){
          category->selectLine(i);
          break;
        }
      }

    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->erase();
    }

    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::colorKeyRed(){
    if(!itvgserver) return FAILURE;

    if(win_focus == BOOKMARKS){
      //delete Bookmarks
      string name = bookmarks->getLineContent(bookmarks->getSelectedLine());
      int line = bookmarks->getSelectedLine();
      bookmarkList.erase(name);

      /** Fill Bookmark-Window */
      bookmarks->deleteAllLines();
      bookmarks->addLine("Custom", "");
      bookmarks->addLine("Now", "");
      bookmarks->addLine("Next", "");

      for(map<string, ProgramFilter>::iterator i = bookmarkList.begin(); i != bookmarkList.end(); i++){
        bookmarks->addLine(i->first);
      }

      bookmarks->selectLine(line-1);
      fillFromBookmarks(line-2);

      updateOSD();
      saveBookmarks();
    }else if(win_focus == STARTTIME){
      starttime->erase();
      starttime->setLineURL("1");
    }else if(win_focus == ENDTIME){
      endtime->erase();
    }else if(win_focus == PHRASE){
      phrase->erase();
    }else if(win_focus == CHANNEL){
    }else if(win_focus == CATINPUT){
      categoryInput->erase();
    }else if(win_focus == CATEGORY){
      categoryInput->erase();
    }else if(win_focus == BOOKMARKNAME){
      bookmarkName->erase();
    }else{
    }    
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearch::colorKeyGreen(){
    if(!itvgserver) return FAILURE;

    /* Marking and unmarking lines */
    unsigned int sel = 0;
    list<unsigned int> lineList;
    switch(win_focus){
    case BOOKMARKS:
      break;
    case BOOKMARKNAME:
      bookmarkName->clear();
      break;
    case STARTTIME:
      starttime->clear();
      starttime->setLineURL("1");
      break;
    case ENDTIME:
      endtime->clear();
      break;
    case PHRASE:
      phrase->clear();
      break;
    case CATINPUT:
      categoryInput->clear();
      break;
    case CHANNEL:
      sel = channel->getSelectedLine();
      lineList = channel->getMarkedLines();
      if(sel == 1){
        if(lineMarked(lineList, 1)){
          for(list<unsigned int>::iterator i = lineList.begin(); i != lineList.end(); i++){
            channel->unmarkLine(*i);
          }
        }else{
          for(unsigned int i = 0; i <= channel->getTextWidget()->getNumLines() ; i++){
            channel->markLine(i);
          }
        }
      }else{
        for(list<unsigned int>::iterator i = lineList.begin(); i != lineList.end(); i++){
          if (*i == sel){
            channel->unmarkLine(sel);
            channel->unmarkLine(1); //cosmetic: if not all lines are selected, 'All' is deselected
            updateOSD();
            return SUCCESS;
          }
        }
        channel->getTextWidget()->markLine(sel);
      }
      break;
    case CATEGORY:
      sel = category->getSelectedLine();
      lineList = category->getMarkedLines();
      if(sel == 1) { 
        if(lineMarked(lineList, 1)){
          for(list<unsigned int>::iterator i = lineList.begin(); i != lineList.end(); i++){
            category->unmarkLine(*i);
          }
        }else{
          for(unsigned int i = 0; i <= category->getTextWidget()->getNumLines() ; i++){
            category->markLine(i);
          }
        }
      }else{
        for(list<unsigned int>::iterator i = lineList.begin(); i != lineList.end(); i++){
          if (*i == sel){
            category->unmarkLine(sel);
            category->unmarkLine(1);//cosmetic: if not all lines are selected, 'All' is deselected
            updateOSD();
            return SUCCESS;
          }
        }
        category->getTextWidget()->markLine(sel);
      }
      break;
    default:
      break;
    }
    updateOSD();
    return SUCCESS;
  }

  Result EPGSearch::colorKeyYellow(){
    if(!itvgserver) return FAILURE;

    if(win_focus == BOOKMARKNAME){
      string bName = bookmarkName->getContent();

      addBookmark(bName);
      saveBookmarks();
      
      /** Fill Bookmark-Window */
      bookmarks->deleteAllLines();
      bookmarks->addLine("Custom", "");
      bookmarks->addLine("Now", "");
      bookmarks->addLine("Next", "");
      for(map<string, ProgramFilter>::iterator i = bookmarkList.begin(); i != bookmarkList.end(); i++){
        bookmarks->addLine(i->first);
      }

      bookmarks->selectLine(1);
      fillFromBookmarks(-1);
      bookmarks->selectWindow();
      starttime->deSelectWindow();
      endtime->deSelectWindow();
      phrase->deSelectWindow();
      channel->deSelectWindow();
      category->deSelectWindow();
      bookmarkName->deSelectWindow();
      win_focus = BOOKMARKS;
      bookmarkName->clear();
      bookmarkName->hide();
      bookmarkName->update();

      showFilterOSD();
      updateOSD();
    }else{
      hideFilterOSD();
      bookmarkName->selectWindow();
      
      //Set Default-Name
      int line = bookmarks->getTextWidget()->getNumLines();
      char tmp[3];
      sprintf(tmp, "%i", line-3);
      bookmarkName->addString("Bookmark" + string(tmp));
      
      bookmarkName->show();
      win_focus = BOOKMARKNAME;
      updateOSD();
    }


    return SUCCESS;
  }
  Result EPGSearch::colorKeyBlue(){
    if(!itvgserver) return FAILURE;

    if(win_focus == BOOKMARKNAME) return SUCCESS;

    EPGSearchResult::EPGSearchResult* result = dynamic_cast<EPGSearchResult*>(app->getState("EPGSearchResult"));
    changeState("EPGSearchResult");

    //determine the local timezone
    tzset();
    time_t now = time(NULL);
    struct tm *t = localtime(&now);
    char tmp[6];
    double d = -1*timezone/60/60*100;
    if(daylight != 0) d += t->tm_isdst*100;
    sprintf(tmp, " %0+5.0f", d);
    string tz(tmp);

    ProgramFilter pf;
    if(starttime->getContent() != ""){
      if(starttime->getContent() == "Now" || starttime->getContent() == "Next"){
        pf.setStartTime(starttime->getContent() + tz );
      }else{
        pf.setStartTime(formatTimeSearchString(starttime->getContent()) + tz);
      }
    }
 
    if(endtime->getContent() != ""){
      if(endtime->getContent() == "Now" || endtime->getContent() == "Next"){
        pf.setEndTime(endtime->getContent() + tz);
      }else{
        pf.setEndTime(formatTimeSearchString(endtime->getContent()) + tz);
      }
    }

    if(phrase->getContent() != ""){
      pf.setSearchPhrase(phrase->getContent());
    }

    list<unsigned int> lc = channel->getMarkedLines();
    if(channel->getSelectedLine() > 1)
      lc.push_back(channel->getSelectedLine());
    for(list<unsigned int>::iterator i = lc.begin(); i != lc. end(); i++){
      if (channel->getLineURL(*i) != "")
        pf.addChannel(channel->getLineURL(*i));
    }

    list<unsigned int> l = category->getMarkedLines();
    bool allSelected = false;
    for(list<unsigned int>::iterator i = l.begin(); i != l. end(); i++){
      if (*i == 0){
        allSelected = true;
        break;
      }
    }

    if(!allSelected){
      if(category->getSelectedLine() > 1)
        l.push_back(category->getSelectedLine());
      for(list<unsigned int>::iterator i = l.begin(); i != l. end(); i++){
        pf.addCategory(category->getLineContent(*i));
      }
    }

    if(result) result->startSearch(pf);

    return SUCCESS;
  }
  
  Result EPGSearch::newChannel(){

    MutexGuard mg(mutex);

    list<ChannelDescription> cdList = itvgserver->getChannelsList();

    for(list<ChannelDescription>::iterator i = cdList.begin(); i != cdList.end(); i++){
      bool insert = true;
      for(unsigned int j = 1; j <= channel->getTextWidget()->getNumLines(); j++){
        if(channel->getLineURL(j) == i->getID()){
          insert = false;
        }
      }

      if(insert){
        channel->addLine(i->getChannelNames().front(), i->getID());
      }
    }
    updateOSD();
    return SUCCESS;
  }

  /******************** MMBoxState interface ********************/


  /******************** private helpers ********************/

  void EPGSearch::deleteWindows() {
    if(bookmarks){
      delete bookmarks;
      bookmarks = NULL;
    }
    if(starttime){
      delete starttime;
      starttime = NULL;
    }
    if(endtime){
    delete endtime;
    endtime = NULL;
    }
    if(phrase){
      delete phrase;
      phrase = NULL;
    }
    if(channel){
      delete channel;
      channel = NULL;
    }
    if(categoryInput){
      delete categoryInput;
      categoryInput = NULL;
    }
    if(category){
      delete category; 
      category = NULL;
    }    
  }

  void EPGSearch::createWindow( wintype_e w, string h ) {

    IWidgetFactory& factory = app->getOSDWidgetFactory();

    if ( w == BOOKMARKS ) {
      delete bookmarks;
      bookmarks = NULL;
      
      bookmarks = new ListUnit( manager, &factory, "medium" );
      bookmarks->setHeader( h );
      /* set colors according to properties */
      bookmarks->setColors( winprops[w].win_active_color,
			     winprops[w].win_inactive_color,
			     winprops[w].sel_active_color,
			     winprops[w].sel_active_color2,
			     winprops[w].sel_inactive_color );
      /* set width and height of window */
      bookmarks->setDimension( winprops[w].width, winprops[w].height );
      /* set position of the window */
      bookmarks->setPosition( winprops[w].xpos, winprops[w].ypos );
      bookmarks->selectWindow();
      bookmarks->show();
    }else if ( w == STARTTIME ) {
      delete starttime;
      starttime = NULL;
      
      starttime = new TextInput( manager, &factory, "medium" );
      starttime->setHeader( h );
      /* set colors according to properties */
      starttime->setColors( winprops[w].win_active_color,
			     winprops[w].win_inactive_color,
			     winprops[w].sel_active_color,
			     winprops[w].sel_active_color2,
			     winprops[w].sel_inactive_color );
      /* set width and height of window */
      starttime->setDimension( winprops[w].width, winprops[w].height );
      /* set position of the window */
      starttime->setPosition( winprops[w].xpos, winprops[w].ypos );
      starttime->deSelectWindow();
      starttime->show();
    }else if ( w == ENDTIME ) {
      delete endtime;
      endtime = NULL;
      
      endtime = new TextInput( manager, &factory, "medium" );
      endtime->setHeader( h );
      /* set colors according to properties */
      endtime->setColors( winprops[w].win_active_color,
			     winprops[w].win_inactive_color,
			     winprops[w].sel_active_color,
			     winprops[w].sel_active_color2,
			     winprops[w].sel_inactive_color );
      /* set width and height of window */
      endtime->setDimension( winprops[w].width, winprops[w].height );
      /* set position of the window */
      endtime->setPosition( winprops[w].xpos, winprops[w].ypos );
      endtime->deSelectWindow();
      endtime->show();
    }else if ( w == PHRASE ) {
      delete phrase;
      phrase = NULL;
      
      phrase = new TextInput( manager, &factory, "medium" );
      phrase->setHeader( h );
      /* set colors according to properties */
      phrase->setColors( winprops[w].win_active_color,
			     winprops[w].win_inactive_color,
			     winprops[w].sel_active_color,
			     winprops[w].sel_active_color2,
			     winprops[w].sel_inactive_color );
      /* set width and height of window */
      phrase->setDimension( winprops[w].width, winprops[w].height );
      /* set position of the window */
      phrase->setPosition( winprops[w].xpos, winprops[w].ypos );
      phrase->deSelectWindow();
      phrase->show();
    }else if ( w == CHANNEL ) {
      delete channel;
      channel = NULL;
      
      channel = new ListUnit( manager, &factory, "medium" );
      channel->setHeader( h );
      /* set colors according to properties */
      channel->setColors( winprops[w].win_active_color,
			     winprops[w].win_inactive_color,
			     winprops[w].sel_active_color,
			     winprops[w].sel_active_color2,
			     winprops[w].sel_inactive_color );
      /* set width and height of window */
      channel->setDimension( winprops[w].width, winprops[w].height );
      /* set position of the window */
      channel->setPosition( winprops[w].xpos, winprops[w].ypos );
      channel->deSelectWindow();
      channel->show();
    }else if ( w == CATINPUT ) {
      delete categoryInput;
      categoryInput = NULL;
      
      categoryInput = new TextInput( manager, &factory, "medium" );
      categoryInput->setHeader( h );
      /* set colors according to properties */
      categoryInput->setColors( winprops[w].win_active_color,
			     winprops[w].win_inactive_color,
			     winprops[w].sel_active_color,
			     winprops[w].sel_active_color2,
			     winprops[w].sel_inactive_color );
      /* set width and height of window */
      categoryInput->setDimension( winprops[w].width, winprops[w].height );
      /* set position of the window */
      categoryInput->setPosition( winprops[w].xpos, winprops[w].ypos );
      categoryInput->deSelectWindow();
      categoryInput->show();
    }else if ( w == CATEGORY ) {
      delete category;
      category = NULL;
      
      category = new ListUnit( manager, &factory, "medium" );
      category->setHeader( h );
      /* set colors according to properties */
      category->setColors( winprops[w].win_active_color,
			   winprops[w].win_inactive_color,
			   winprops[w].sel_active_color,
			   winprops[w].sel_active_color2,
			   winprops[w].sel_inactive_color );
      /* set width and height of window */
      category->setDimension( winprops[w].width, winprops[w].height );
      category->setPosition( winprops[w].xpos, winprops[w].ypos );
      category->deSelectWindow();
      category->show();
    }else if(w == BOOKMARKNAME){
      delete bookmarkName;
      bookmarkName = NULL;

      bookmarkName = new TextInput(manager, &factory, "medium");
      bookmarkName->setHeader( h );
      bookmarkName->setColors( winprops[PHRASE].win_active_color,
                               winprops[PHRASE].win_inactive_color,
                               winprops[PHRASE].sel_active_color,
                               winprops[PHRASE].sel_active_color2,
                               winprops[PHRASE].sel_inactive_color );

      bookmarkName->setDimension(500,50);
      bookmarkName->setPosition(50,50);
      bookmarkName->deSelectWindow();
      bookmarkName->hide();

    }
  }

  void EPGSearch::updateOSD(){
    bookmarks->update();
    starttime->update();
    endtime->update();
    phrase->update();
    channel->update();
    categoryInput->update();
    category->update();
    
    bookmarkName->update();

    //Handle Display of ColorBars
    if(itvgserver){
      if(win_focus == BOOKMARKS){
        if(bookmarks->getSelectedLine() <= 3){
          if(bookmarks->getSelectedLine() > 1){
            manager->setColorBarText("", "", "", "Search");
          }else{
            manager->setColorBarText("", "", "Save", "Search");
          }
        }else{
          manager->setColorBarText("Delete", "", "Save", "Search");
        }
      }else if(win_focus == STARTTIME){
        manager->setColorBarText("Delete", "Clear", "Save", "Search");
      }else if(win_focus == ENDTIME){
        manager->setColorBarText("Delete", "Clear", "Save", "Search");
      }else if(win_focus == PHRASE){
        manager->setColorBarText("Delete", "Clear", "Save", "Search");
      }else if(win_focus == CHANNEL){
        int line = channel->getSelectedLine();
        list<unsigned int> lineList = channel->getMarkedLines();
        if(lineMarked(lineList, line)){
          manager->setColorBarText("", "Deselect", "Save", "Search");
        }else{
          manager->setColorBarText("", "Select", "Save", "Search");
        }
        //manager->setColorBarText("", "Select", "", "Search");
      }else if(win_focus == CATINPUT){
        manager->setColorBarText("Delete", "Clear", "Save", "Search");
      }else if(win_focus == CATEGORY){
        int line = category->getSelectedLine();
        list<unsigned int> lineList = category->getMarkedLines();
        if(lineMarked(lineList, line)){
          manager->setColorBarText("", "Deselect", "Save", "Search");
        }else{
          manager->setColorBarText("", "Select", "Save", "Search");
        }
        
        //manager->setColorBarText("", "Select", "", "Search");
      }else if(win_focus == BOOKMARKNAME){
        manager->setColorBarText("Delete", "Clear", "Save", "");
      }else{
        manager->setColorBarText("", "", "Save", "Search");
      }
    }else{
      manager->setColorBarText("","","","");
    }

    manager->displayColorBar();
    manager->update();
  }

  Result EPGSearch::showFilterOSD(){
    //manager->displayColorBar();
    bookmarks->show();
    starttime->show();
    endtime->show();
    phrase->show();
    channel->show();
    categoryInput->show();
    category->show();

    manager->update();
    return SUCCESS;
  }

  Result EPGSearch::hideFilterOSD(){
    bookmarks->hide();
    starttime->hide();
    endtime->hide();
    phrase->hide();
    channel->hide();
    categoryInput->hide();
    category->hide();

    manager->setColorBarText("", "", "", "");
    manager->hideColorBar();

    manager->update();
    return SUCCESS;
  }


  void EPGSearch::initializeWindows(){
    GLOBAL_DEBUG_STREAM("EPGSearch: initilizeWindows()" << endl);

    /** Fill Bookmark-Window */
    bookmarks->deleteAllLines();
    bookmarks->addLine("Custom", "");
    bookmarks->addLine("Now", "");
    bookmarks->addLine("Next", "");
    for(map<string, ProgramFilter>::iterator i = bookmarkList.begin(); i != bookmarkList.end(); i++){
      bookmarks->addLine(i->first);
    }
    bookmarks->selectLine(1);

    Clock clock;
    LocalTime t = clock.getTime();
    starttime->clear();
    starttime->addString(t.toString("%d.%m.%y %H:%M")); //initialize with current date

    if(!itvgserver) return;

    /** Fill Channel-Window */
    channel->deleteAllLines();
    channel->addLine("All");
    list<ChannelDescription> cdList = itvgserver->getChannelsList();
    for(list<ChannelDescription>::iterator i = cdList.begin(); i != cdList.end(); i++){
      channel->addLine(i->getChannelNames().front(), i->getID());
    }
    channel->selectLine(1);

    /** Fill Categories-Window */
    category->deleteAllLines();
    category->addLine("All");

    list<string> catList = itvgserver->getCategories();
    for(list<string>::iterator i = catList.begin(); i != catList.end(); i++){
      category->getTextWidget()->addLine(*i);
      category->update();
    }
    category->selectLine(1);
    
    updateOSD();
  }  

  void EPGSearch::readBookmarks(){
    string rootpath;
    if(!(app->getMMBoxConfig()->getValue("rootpath", rootpath))){
      rootpath = "/";
    }
    string filename = rootpath + "/epg/bookmarks.xml";
    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }
    string name;

    XMLTree* xml = new XMLTree(filename);
    XMLNode* root = NULL;
    if(xml) {
      root = xml->root();
      if(root){
        bookmarkList.clear();
        XMLNodeList bmList = root->children("bookmark");
        for(XMLNodeIterator bmIt = bmList.begin(); bmIt != bmList.end(); bmIt++){
          XMLNode* curNode = *bmIt;
          XMLProperty* nameProp = curNode->property("name");
          if(nameProp)name = nameProp->value();
          
          ProgramFilter pf;
          XMLNodeList starttimeList = curNode->children("start");
          if(!starttimeList.empty())pf.setStartTime(starttimeList.front()->getData());
        
          XMLProperty* editProp = starttimeList.front()->property("edited");
          if(editProp){
            if(editProp->value() == "1"){
              pf.setStarttimeEdited(true);
            }else{
              pf.setStarttimeEdited(false);
            }
          }
  
          XMLNodeList endtimeList = curNode->children("end");
          if(!endtimeList.empty()) pf.setEndTime(endtimeList.front()->getData());
        
          XMLNodeList phraseList = curNode->children("phrase");
          if(!phraseList.empty()) pf.setSearchPhrase(phraseList.front()->getData());
          
          XMLNodeList channelList = curNode->children("channel");
          for(XMLNodeIterator j = channelList.begin(); j != channelList.end(); j++){
            XMLNode* c = *j;
            if(c) pf.addChannel(c->getData());
          }
          
          XMLNodeList categoryList = curNode->children("category");
          for(XMLNodeIterator j = categoryList.begin(); j != categoryList.end(); j++){
            XMLNode* c = *j;
            if(c) pf.addCategory(c->getData());
          }
          
          bookmarkList[name] = pf;
          
        }//end for(bmIt)
      }
    }
  }

  void EPGSearch::saveBookmarks(){
    string rootpath;
    if(!(app->getMMBoxConfig()->getValue("rootpath", rootpath))){
      rootpath = "/";
    }
    string filename = rootpath + "/epg/bookmarks.xml";

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    /**
     * write filters to file
     */
    ofstream fout(filename.c_str(), ios::out);
    if(fout){

      fout << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?> \n" << endl;
      fout << "<!DOCTYPE tv SYSTEM \"xmltv.dtd\"> \n" << endl;
      fout << "<bookmarks generator-info-name=\"NMM MMBox EPGSearch\">\n" << endl;; 

      for(map<string, ProgramFilter>::iterator i = bookmarkList.begin(); i != bookmarkList.end(); i++){
	fout << "<bookmark name=\"" << i->first << "\">" << endl;
	fout << "<start edited=\"" << i->second.getStarttimeEdited() << "\">" << i->second.getStartTime() << "</start>" << endl;
	fout << "<end>" << i->second.getEndTime() << "</end>" << endl;
	fout << "<phrase>" << i->second.getSearchPhrase() << "</phrase>" << endl;

	list<string> catList = i->second.getCategories();
	for(list<string>::iterator j = catList.begin(); j != catList.end(); j++){
	  fout << "<category>" << *j << "</category>" << endl;
	}

	list<string> chanList = i->second.getChannels();
	for(list<string>::iterator j = chanList.begin(); j != chanList.end(); j++){
	  fout << "<channel>" << *j << "</channel>" << endl;
	}

	fout << "</bookmark>" << endl;
      }//end for
      fout << "</bookmarks>" << endl;
    }//end if(out)

    updateOSD();
  }

  void EPGSearch::addBookmark(string name){
    /**
     * add current filter to map
     */
    ProgramFilter pf;
    
    if(starttime->getContent() != "")
      pf.setStartTime(formatTimeSearchString(starttime->getContent()));

    if(starttime->getLineURL() == "1"){
      pf.setStarttimeEdited(true);
    }else{
      pf.setStarttimeEdited(false);
    }

    if(endtime->getContent() != "")
      pf.setEndTime(formatTimeSearchString(endtime->getContent()));
    
    if(phrase->getContent() != "")
      pf.setSearchPhrase(phrase->getContent());
    
    list<unsigned int> lc = channel->getMarkedLines();
    if(channel->getSelectedLine() > 1)
      lc.push_back(channel->getSelectedLine());
    lc.sort(); lc.unique();
    for(list<unsigned int>::iterator i = lc.begin(); i != lc. end(); i++){
      pf.addChannel(channel->getLineURL(*i));
      }
    
    list<unsigned int> l = category->getMarkedLines();
    if(category->getSelectedLine() > 1)
      l.push_back(category->getSelectedLine());
    l.sort(); l.unique();
    for(list<unsigned int>::iterator i = l.begin(); i != l. end(); i++){
      pf.addCategory(category->getLineContent(*i));
    }

    bookmarkList[name] = pf;

  }

  string EPGSearch::formatTimeString(string s){

    try{
      string ret = s.substr(6,2) + ".";
      ret += s.substr(4,2) + ".";
      ret += s.substr(2,2) + " ";
      ret += s.substr(8,2) + ":";
      ret += s.substr(10,2);
      return ret;
    }catch(const Exception &e){
      return string("");
    }
  }

  string EPGSearch::formatTimeSearchString(string s){

    try{
      string ret;
      ret = "20" + s.substr(6,2) + s.substr(3,2) + s.substr(0,2) + s.substr(9,2) + s.substr(12,2);
      return ret;
    }catch(const Exception &e){
      return string();
    }
  }
  void EPGSearch::fillFromBookmarks(int i){
    GLOBAL_DEBUG_STREAM("EPGSearch: fillFromBookmarks " << endl);
    Clock clock;
    
    if( i <= 0){
      //if first line is selceted
      starttime->clear();
      LocalTime t = clock.getTime();
      starttime->clear();
      starttime->addString(t.toString("%d.%m.%y %H:%M")); //initialize with current date
      endtime->clear();
      phrase->clear();
      categoryInput->clear();
      //Display the windows again, when needed
      starttime->show();
      endtime->show();
      phrase->show();
      categoryInput->show();
      category->show();
      channel->selectLine(0);
      list<unsigned int> l = channel->getMarkedLines();
      for(list<unsigned int>::iterator j = l.begin(); j != l.end(); j++)
        channel->unmarkLine(*j);
      category->selectLine(0);
      l = category->getMarkedLines();
      for(list<unsigned int>::iterator j = l.begin(); j != l.end(); j++)
        category->unmarkLine(*j);
      
      return;
    }

    if(i == 1){
      //Bookmark NOW
      starttime->hide();
      starttime->clear();
      starttime->addString("Now");
      endtime->hide();
      phrase->clear();
      categoryInput->clear();
      for(unsigned int i = 0; i <= channel->getTextWidget()->getNumLines(); i++)
        channel->markLine(i);
      category->selectLine(0);
      for(unsigned int i = 0; i <= category->getTextWidget()->getNumLines(); i++)
        category->markLine(i);

      return;
    }
    if(i == 2){
      //Bookmark NEXT
      starttime->hide();
      starttime->clear();
      starttime->addString("Next");
      endtime->hide();
      phrase->clear();
      categoryInput->clear();
      for(unsigned int i = 0; i <= channel->getTextWidget()->getNumLines(); i++)
        channel->markLine(i);
      category->selectLine(0);
      for(unsigned int i = 0; i <= category->getTextWidget()->getNumLines(); i++)
        category->markLine(i);

      return;
    }
    

    //at end of list, show last result
    if(static_cast<unsigned int>(i) > bookmarkList.size() + 2) return;

    //Display the windows again, when needed
    starttime->show();
    endtime->show();
    phrase->show();
    categoryInput->show();
    category->show();

    i -= 3; //to get the correct bookmark from the list, because the first 3 ones are defaults
    //otherwise, display the actual searchfilter
    starttime->clear();
    endtime->clear();
    phrase->clear();
    categoryInput->clear();
    channel->selectLine(0);
    list<unsigned int> l = channel->getMarkedLines();
    for(list<unsigned int>::iterator j = l.begin(); j != l.end(); j++)
      channel->unmarkLine(*j);
    category->selectLine(0);
    l = category->getMarkedLines();
    for(list<unsigned int>::iterator j = l.begin(); j != l.end(); j++)
      category->unmarkLine(*j);


    map<string, ProgramFilter>::iterator it = bookmarkList.begin();
    advance(it, i);
    if(it == bookmarkList.end()) it--;

    if(it->second.getStarttimeEdited()){
      if(it->second.getStartTime()    != "") starttime->addString(formatTimeString(it->second.getStartTime()));
    }else{
      LocalTime t = clock.getTime();
      starttime->addString(t.toString("%d.%m.%y %H:%M")); //initialize with current date
      it->second.setStartTime(t.toString("%Y%m%d%H%M"));
    }

    if(it->second.getEndTime()      != "") endtime->addString(formatTimeString(it->second.getEndTime()));
    if(it->second.getSearchPhrase() != "") phrase->addString(it->second.getSearchPhrase());
    
    list<string> chanList = it->second.getChannels();
    for(list<string>::iterator si = chanList.begin(); si != chanList.end(); si++){ 
      for(unsigned int j = 0; j < channel->getTextWidget()->getNumLines(); j++){
        if(*si == channel->getLineURL(j)){
          channel->markLine(j);
        }
      }
    }

    list<string> catList = it->second.getCategories();
    for(list<string>::iterator si = catList.begin(); si != catList.end(); si++){
      for(unsigned int j = 0; j <= category->getTextWidget()->getNumLines(); j++){
        if(*si == category->getLineContent(j)){
          category->markLine(j);
        }
      }
    }
    
  }

  string EPGSearch::dateInput(int w, string s){

    int stringLength = 0;
    TextInput* tmp;
    if(w == STARTTIME){
      stringLength = starttime->getContent().size();
      tmp = starttime;
    }else if(w == ENDTIME){
      stringLength = endtime->getContent().size();
      tmp = endtime;
    }
    
    switch(stringLength){
    case 1:
      tmp->addString(s + ".");
      break;
    case 2:
      tmp->addString("." + s);
      break;
    case 4:
      tmp->addString(s + ".");
      break;
    case 5:
      tmp->addString("." + s);
      break;
    case 7:
      tmp->addString(s + " ");
      break;
    case 8:
      tmp->addString(" " + s);
      break;
    case 10:
      tmp->addString(s + ":");
      break;
    case 11:
      tmp->addString(":" + s);
      break;
    case 14:
      break;
    default:
      tmp->addString(s);
      break;
    }

    return s;
  }

  bool EPGSearch::lineMarked(list<unsigned int> l, unsigned int i){
    for(list<unsigned int>::iterator j = l.begin(); j != l.end(); j++){
      if (*j == i){
        return true;
      }
    }
    return false;
  }

} // end namespace NMM  


   
