/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002, 2003
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass (dama@graphics.cs.uni-sb.de)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>

#include "transcode.hpp"

#include "apps/mmbox/apps/webserver/webserver.hpp"
#include "apps/mmbox/apps/webserver/handler/TranscodeHandler.hpp"

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"

#include "nmm/utils/gdparse/GDParse.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/DirReader.hpp"

namespace NMM {
  
  Transcode::Transcode(string id, MMBoxApplication* _app)
    : MMBoxState(id, _app),
      group_list(NULL),
      info_window(NULL),
      rootpath("/"),
      _mode(JOBS),
      manager(NULL),
      transcode_format(""),
      _spymode(NONE),
      _focus(GROUPLIST)
  {
    MMBoxConfig* config = app->getMMBoxConfig();
    if (!(config->getValue("rootpath", rootpath))) {
      rootpath="/";
    }

    if (!(config->getValue("resourcepath", gdpath))) {
      gdpath = rootpath;
    }
    

    if (!(config->getValue("transcodingdir", transpath))) {
      transpath = rootpath;
    } else transpath = rootpath + "/" + transpath;
    
    if (!(config->getValue("cdrompath",cdrompath))) {
      cdrompath="/cdrom";
    }

    if (!(config->getValue("distribute",distribute))) {
      distribute = true;
    }

    if (!(config->getValue("parallel",parallel))) {
      parallel = 5;
    }

    if (!(config->getValue("priorize",priori))) {
      priori = true;
    }

    if (!(config->getValue("use_localhost",use_localhost))) {
      use_localhost = false;
    }

    jobm = new JobManager(this);

    /** get Transcodinghosts from config-file **/
    int  transcodinghosts_number = 0;
    if(config->getValue("transcodinghosts_number", transcodinghosts_number)) {
      if(transcodinghosts_number > 0) {
	char tmp[128];
	for(int i = 0; i < transcodinghosts_number; ++i) {
	  sprintf(tmp, "transcodinghost_%u", i); 
	  string current_host_string(tmp);
	  string host_and_port = "";
	  string host = "";
	  int port = 0;
	  if(config->getValue(current_host_string, host_and_port)) {
	    string::size_type colon_pos = host_and_port.find_first_of(':');
	    if (colon_pos != string::npos) {
	      host = host_and_port.substr(0, colon_pos);
	      port = atoi(host_and_port.substr(colon_pos+1, string::npos).c_str());
	    }
	    else {
	      host = host_and_port;
	      port = ClientRegistry::default_port;
	    }
	    host_and_port = host + ":" + port;
          }
          GLOBAL_DEBUG_STREAM("adding transcodinghost "<<host<<":"<<port<<endl);
          reg_list.push_back(pair<string,int>(host,port));
        }
      }
    }

    manager = app->getOsdManager();

    playlist_gui = new WindowSelectorTranscode(app->getOsdManager(),  &app->getOSDWidgetFactory(), this);
    playlist_gui->hide();
    config_gui = new TranscodeConfig(app->getOsdManager(), &app->getOSDWidgetFactory(),this);
    config_gui->hide();
    config_gui->setApplication(app);
    playlist_gui->setRootpath(rootpath);
    playlist_gui->setApplication(app);
    playlist_gui->setCDRomPath(cdrompath);
    playlist_gui->addprefix(true);

    //Register Webhandler
    SMMBoxWebServer::getInstance().registerHandler(new TranscodeHandler(this));
    }
  
  Transcode::~Transcode(){
    delete playlist_gui;
    delete config_gui;
    delete jobm;
  }
  
   void Transcode::setWinProperties( int winnr, OSDElementComposit::win_properties_t* prop ) {
     if (winnr > 2)
       config_gui->setWinProperties(winnr-3,prop);
     else
       playlist_gui->setWinProperties(winnr,prop);
   }


  Result Transcode::doCreate( string ) {
    GLOBAL_DEBUG_STREAM("Transcode: create: " << stateid << endl);
    _spymode = NONE;
    _focus = GROUPLIST;
    _mode = JOBS;

    if (!DirReader::makeDir(transpath)) {
      GLOBAL_ERROR_STREAM("Failed to mkdir outputdirectory "<<transpath<<endl);
      return FAILURE;
    }  

    if ( isBackgroundFree() ) {
      if (FAILURE == app->createDefaultGraph( background )) {
	return FAILURE;
      }
      
      if (FAILURE == app->startDefaultGraph()) {
	return FAILURE;
      }
    }

    playlist_gui->createWindows();
    config_gui->createWindows();
    playlist_gui->hide();
    config_gui->hide();
    config_gui->setConfig(app->getMMBoxConfig(),shown_options);
    createStatusWindow();
    showGroups();
    updateColorBar();
    manager->update();
    return( SUCCESS );
  }
  
  Result Transcode::doKill() {
    GLOBAL_DEBUG_STREAM("Transcode: doKill begin: "<< stateid << endl);
    manager->hideProgressBar();
    bool killall = true;
    MMBoxConfig* config = app->getMMBoxConfig();
    config->getValue("kill_jobs_on_exit",killall);
    if (killall) {
      //kill all running/waiting/finished transcoding jobs
      for (uint i=0;i<_transcoding_groups.size();++i) {
	if (_transcoding_groups[i]) {
          delete _transcoding_groups[i];
	  _transcoding_groups[i]=NULL;
	}
      }
      _transcoding_groups.clear();
    }

    if (isBackgroundFree()) {
      /* hide all on-screen items if state is left */
      app->disconnectDefaultGraph();
    }
    
    /* delete all windows */
    playlist_gui->deleteWindows();
    config_gui->deleteWindows();
    if (group_list) {
      delete group_list;
      group_list = NULL;
    }
    if (job_list) {
      delete job_list;
      job_list = NULL;
    }
    if (info_window) {
      delete info_window;
      info_window = NULL;
    }
    manager->hideColorBar();
    GLOBAL_DEBUG_STREAM("Transcode: doKill end: "<< stateid << endl);
    return( SUCCESS );
  }

  Result Transcode::doInitialize( string ) {
    GLOBAL_DEBUG_STREAM("Transcode: initialize begin: " << stateid << endl);

    if ( isBackgroundFree() ) {    
      if (FAILURE == app->createDefaultGraph( background )) {
        return FAILURE;
      }
      
      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }
    
    switch (_mode) {
    case JOBS:
      group_list->show();
      group_list->update();
      job_list->show();
      job_list->update();
      info_window->show();
      info_window->update();
      updateInfo();
      manager->update();
      break;
    case FILES:
      manager->displayColorBar();
      playlist_gui->show();
      playlist_gui->update();
      playlist_gui->updateColorBar();
      break;
    case CONFIG:
      config_gui->show();
      config_gui->updateColorBar();
      config_gui->update();
      break;
    }
    manager->update();
    
    GLOBAL_DEBUG_STREAM("Transcode: initialize begin: " << stateid << endl);
    return( SUCCESS );
  }

  Result Transcode::doUninitialize() {
    GLOBAL_DEBUG_STREAM("Transcode: uninitialize begin: " << stateid << endl);

    /* hide all on-screen items */
    playlist_gui->hide();
    config_gui->hide();
    group_list->hide();
    job_list->hide();
    info_window->hide();
    manager->hideColorBar();
    manager->hideProgressBar();

    GLOBAL_DEBUG_STREAM("Transcode: uninitialize end: " << stateid << endl);
    return( SUCCESS );
  }

  Result Transcode::receiveMessage( string message ) {
    manager->deleteMessage();
    manager->update();
    
    /* the only message we can get is "DEL:filename" that indicates to delete
       the given file */
    unsigned int pos = message.find_first_of(':');
    if ( pos != string::npos ) {
      string temp;
      temp = message.substr( 0, pos );
      if ( temp == "DEL" ) {
	temp = "rm -rf '" + (message.substr( pos+1, message.length())+"'");
	system( temp.c_str() );
	playlist_gui->refreshFiles();
      }
    }
    return SUCCESS;
  }
  
  Result Transcode::showOption( string id, string shortform ) {
    shown_options[shortform] = id;
    return SUCCESS;
  }

  Result Transcode::Down() {
    if (_spymode != NONE)
      return SUCCESS;
    switch (_mode) {
    case JOBS:
      manager->hideProgressBar();
      switch(_focus) {
      case GROUPLIST: {
        group_list->getTextWidget()->selectDown();
        group_list->update();
        int selection = group_list->getSelectedLine();
        if (selection) {
          job_list->deSelectLine();
          showJobs(atoi(group_list->getLineURL(selection).c_str()));
          showGroupInfo(atoi(group_list->getLineURL(selection).c_str()));
        }
        else {
          info_window->deleteAllLines();
          info_window->update();
        }
        updateColorBar();
        manager->update();
        break;
      }
      case JOBLIST: {
        updatemutex.lock();
        job_list->selectDown();
        job_list->update();
        //update info window
        int selection = group_list->getSelectedLine();
        int selection2 = job_list->getSelectedLine();
        if (selection && selection2) {
          int value = atoi(job_list->getLineContent(0,selection2).c_str());
          showJobInfo(atoi(group_list->getLineURL(selection).c_str()),value);
        }
        updateColorBar();
        manager->update();
        updatemutex.unlock();
        break;
      }
      }
      break;
    case FILES:
      playlist_gui->selectDown();
      break;
    case CONFIG:
      config_gui->selectDown();
      break;
    }
    return SUCCESS;
  }

  Result Transcode::Up() {
    if (_spymode != NONE)
      return SUCCESS;
    switch (_mode) {
    case JOBS:
      manager->hideProgressBar();
      switch(_focus) {
      case GROUPLIST: {
        group_list->getTextWidget()->selectUp();
        group_list->update();
        int selection = group_list->getSelectedLine();
        if (selection) {
          job_list->deSelectLine();
          showJobs(atoi(group_list->getLineURL(selection).c_str()));
          showGroupInfo(atoi(group_list->getLineURL(selection).c_str()));
        }
        else {
          info_window->deleteAllLines();
          info_window->update();
        }        
        updateColorBar();
        manager->update();
        break;
      }
      case JOBLIST: {
        updatemutex.lock();
        job_list->selectUp();
        job_list->update();
        //update info window
        int selection = group_list->getSelectedLine();
        int selection2 = job_list->getSelectedLine();
        if (selection && selection2) {
          int value = atoi(job_list->getLineContent(0,selection2).c_str());
          showJobInfo(atoi(group_list->getLineURL(selection).c_str()),value);
        }

        updateColorBar();
        manager->update();
        updatemutex.unlock();
        break;
      }
      }
      break;
    case FILES:
      playlist_gui->selectUp();
      break;
    case CONFIG:
      config_gui->selectUp();
      break;
    }
    return SUCCESS;
  }
  
  Result Transcode::Left() {
    if (_spymode != NONE)
      return SUCCESS;
    switch(_mode) {
    case JOBS:
      manager->hideProgressBar();
      switch(_focus) {
      case GROUPLIST: {
        _focus = JOBLIST;
        info_window->setHeader("Job information");
        info_window->update();
        group_list->deSelectWindow();
        job_list->selectWindow();
        job_list->update();
        group_list->update();
        updateColorBar();
        updateInfo();
        manager->update();
        break;
      }
      case JOBLIST: {
        _focus = GROUPLIST;
        info_window->setHeader("Group information");
        info_window->update();
        job_list->deSelectWindow();
        group_list->selectWindow();
        job_list->update();
        group_list->update();
        updateInfo();
        updateColorBar();
        manager->update();
        break;
      }
      }
      break;
    case FILES:
      playlist_gui->selectLeft();
      break;
    case CONFIG:
      config_gui->selectLeft();
      break;
    }
    return SUCCESS;
  }
  
  Result Transcode::Right() {
    if (_spymode != NONE)
      return SUCCESS;
    switch(_mode) {
    case JOBS:
      manager->hideProgressBar();
      switch(_focus) {
      case GROUPLIST: {
        _focus = JOBLIST;
        info_window->setHeader("Job information");
        info_window->update();
        group_list->deSelectWindow();
        job_list->selectWindow();
        job_list->update();
        group_list->update();
        updateColorBar();
        updateInfo();
        manager->update();
        break;
      }
      case JOBLIST: {
        _focus = GROUPLIST;
        info_window->setHeader("Group information");
        info_window->update();
        job_list->deSelectWindow();
        group_list->selectWindow();
        job_list->update();
        group_list->update();
        updateColorBar();
        updateInfo();
        manager->update();
        break;
      }
      }
      break;
    case FILES:
      playlist_gui->selectRight();
      break;
    case CONFIG:
      config_gui->selectRight();
      break;
    }
    return SUCCESS;
  }
  
  Result Transcode::Back() {
    GLOBAL_DEBUG_STREAM( "Transcode: Back" << endl );
    
    return( goToLastState() );
  }
  
  Result Transcode::colorKeyRed() {
    switch (_mode) {
    case FILES:
      return playlist_gui->colorKeyRed();
      break;
    case CONFIG:
      return config_gui->colorKeyRed();
      break;
    case JOBS:  {
      manager->hideProgressBar();
      if (_spymode != NONE) {
	int selection = group_list->getSelectedLine();
	int selection2 = job_list->getSelectedLine();
	//	if (selection && selection2) {
	string url = group_list->getTextWidget()->getLineURL(selection);
	int value = atoi(job_list->getLineContent(0,selection2).c_str());
	int pos = atoi(url.c_str());
	_transcoding_groups[pos]->endSpymode(value);
        _spymode=NONE;
        group_list->show();
        job_list->show();
        info_window->show();
        app->disconnectDefaultGraph();
        app->createDefaultGraph(background);
        app->startDefaultGraph();
        updateColorBar();
        return SUCCESS;
      }

      switch(_focus) {
      case GROUPLIST:{
        int selection = group_list->getSelectedLine();

        if (selection) {
          string url = group_list->getTextWidget()->getLineURL(selection); 
          int pos = atoi(url.c_str());
          delete _transcoding_groups[pos];
          _transcoding_groups[pos] = NULL;

          group_list->getTextWidget()->deleteLine(selection);
	  selection = group_list->getSelectedLine();

          if (selection) {
            showJobs(atoi(group_list->getLineURL(selection).c_str()));
            showGroupInfo(atoi(group_list->getLineURL(selection).c_str()));
          }
          else {
	    job_list->deleteAllLines();
	    job_list->update();
            info_window->deleteAllLines();
            info_window->update();
          }
          group_list->update();
          updateColorBar();
          manager->update();
          break;
        }
      }
      case JOBLIST:{
        int selection = group_list->getSelectedLine();
        int selection2 = job_list->getSelectedLine();
        if (selection && selection2) {
          string url = group_list->getTextWidget()->getLineURL(selection); 
          int pos = atoi(url.c_str());
          int value = atoi(job_list->getLineContent(0,selection2).c_str());
          _transcoding_groups[pos]->deleteJob(value);
          showJobs(pos);
          updateInfo();
          info_window->update();
          updateColorBar();
          manager->update();
        }
        break;
      }
      }
      break;
    }
    }
    return SUCCESS;
  }

  Result Transcode::colorKeyGreen() {
    switch (_mode) {
    case JOBS: {
      manager->hideProgressBar();
      if (_spymode==NONE) {
        _mode = FILES;
        group_list->hide();
        job_list->hide();
        info_window->hide();
        playlist_gui->show();
        playlist_gui->update();
        manager->update();
        playlist_gui->updateColorBar();
        manager->update();
      }
      break;
    }
    case FILES:
      playlist_gui->colorKeyGreen();
      break;
    case CONFIG:
      return config_gui->colorKeyGreen();
      break;
    }
    return SUCCESS;
  }

  Result Transcode::colorKeyYellow() {
    switch (_mode) {
    case JOBS: 
      {
        if (_focus == JOBLIST) {
          int selection = group_list->getSelectedLine();
          int selection2 = job_list->getSelectedLine();
          if (selection && selection2) {
            string url = group_list->getTextWidget()->getLineURL(selection);
            int pos = atoi(url.c_str());
            int value = atoi(job_list->getLineContent(0,selection2).c_str());
            //if we are in spymode switch between raw/transcoded video view
            if (_spymode!=NONE) {
              app->disconnectDefaultGraph();
              app->getOsdNode()->reachInitialized();
              if (_spymode==TRANS) {
                if (_transcoding_groups[pos]->spymodeRaw(value,app->getOsdNode(),"stream1")) {
                  app->setSynchronized(false);
                  manager->setPause(false);
                  app->getOsdNode()->reachStarted();
                  _spymode=RAW;
                } else {
                  app->createDefaultGraph(background);
                  app->startDefaultGraph();
                }
              } else {
                if (_transcoding_groups[pos]->spymodeTrans(value,app->getOsdNode(),"stream1")) {
                  app->setSynchronized(false);
                  manager->setPause(false);
                  app->getOsdNode()->reachStarted();
                  _spymode=TRANS;
                } else {
                  app->createDefaultGraph(background);
                  app->startDefaultGraph();
                }
              }
              updateColorBar();
              return SUCCESS;
            }
            
            if (_transcoding_groups[pos])
              _transcoding_groups[pos]->pauseJob(value);
	    showJobInfo(atoi(group_list->getLineURL(selection).c_str()),value);
            updateColorBar();
	    manager->update();
          }
        }
        return SUCCESS;
      }
    case FILES:
      return playlist_gui->colorKeyYellow();
    default:
      return SUCCESS;
    }
  }

  Result Transcode::colorKeyBlue() {
    switch(_mode) {
    case JOBS: {
      manager->hideProgressBar();
      if (_focus == JOBLIST) {
        int selection = group_list->getSelectedLine();
        int selection2 = job_list->getSelectedLine();
        if (selection && selection2) {
          string url = group_list->getTextWidget()->getLineURL(selection);
          int pos = atoi(url.c_str());
          int value = atoi(job_list->getLineContent(0,selection2).c_str());
          //if we have a running job which allows spymode switch to spymode raw
          // XXX: fw: spymode is disabled
//           if (_transcoding_groups[pos]->spymode(value)) {
//             group_list->hide();
//             job_list->hide();
//             info_window->hide();
//             app->disconnectDefaultGraph();
//             app->getOsdNode()->reachInitialized();
//             if (_transcoding_groups[pos]->spymodeRaw(value,app->getOsdNode(),"stream1")) {
//               app->setSynchronized(false);
//               manager->setPause(false);
//               app->getOsdNode()->reachStarted();
//               _spymode=RAW;
//             } else {
//               app->createDefaultGraph(background);
//               app->startDefaultGraph();
//             }
//             updateColorBar();
//           }
        }
      }
    return SUCCESS;
    } 
    case FILES:
      return playlist_gui->colorKeyBlue();
    case CONFIG:
      return config_gui->colorKeyBlue();
    default:
      return SUCCESS;
    }
  }
  
  Result Transcode::Return() {
    switch (_mode) {
    case JOBS:
      manager->update();
      return SUCCESS;
      break;
    case FILES:
      return playlist_gui->select();
      break;
    case CONFIG:
      return config_gui->select();
      break;
    }
    return FAILURE;
  }

  Result Transcode::doSelectGDWindow(GraphDescription* gd, list<NodeConfiguration>* nodeConfig, const string& des) {
    /** Time to create a new TranscodingGroup **/
    TranscodingGroup* trans_group = new TranscodingGroup(gd, nodeConfig, transcode_format, app->getNMMApplication()->getRegistry() , app->getNMMApplication()->getPluginManager(),jobm,this);
    trans_group->setDescription(des);
    trans_group->setDistribute(distribute);
    trans_group->setPriorize(priori);
    trans_group->useLocalhost(use_localhost);
    trans_group->setWorkingDirectory(transpath);
    trans_group->setParallel(parallel);
    trans_group->useRegistrys(reg_list);
    trans_group->useScript(config_gui->getScript());
    
    if (_url != "") {
      /** add one file to job **/
      trans_group->addInputFile(GraphURL(_url).getURI());
      if (transcode_format == "audiocd")
        trans_group->addInputFileName(playlist_gui->getActualFile());
    } else {
      /** add multiple files to job **/
      vector<GraphURL> fileurls;
      playlist_gui->getAllPlaylistFiles(fileurls);
      trans_group->addInputFiles(fileurls);

      if (transcode_format == "audiocd") {
        vector<string> filenames;
        playlist_gui->getAllPlaylistNames(filenames);
        trans_group->addInputFileNames(filenames);
      }
    }
    
    _transcoding_groups.push_back(trans_group);
    trans_group->setup();
    
    _mode=JOBS;
    config_gui->hide();
    playlist_gui->hide();
    showGroups();
    group_list->selectLine(1);
    group_list->update();
    showGroups();
    updateColorBar();
    group_list->show();
    job_list->update();
    job_list->show();
    info_window->update();
    info_window->show();

    manager->update();
    
    return SUCCESS;
  }

  Result Transcode::doSelectPlaylistWindow(string) {
    /** get all Files and check if they all have the same filetype **/
    playlist_gui->hide();
    manager->displayMessage( "Please Wait...", "Analysing Filetypes" , "", "", "");
    manager->setColorBarText("","","","");
    manager->update();
    vector<GraphURL> fillitup;
    playlist_gui->getAllPlaylistFiles(fillitup);
    string protocol = fillitup[0].getProtocolName();

    if (protocol != "file") {
      transcode_format = protocol;
      for (uint i=1;i<fillitup.size();i++) {
        if (fillitup[i].getProtocolName() != transcode_format) {
          manager->deleteMessage();
          playlist_gui->show();
          manager->displayMessage("Error", "not all files have the same type - unable to Continue", "OK","","",0);
          changeState("OSDInputState");
          manager->update();
          return FAILURE;
        }
      }
    } else {
      transcode_format = fillitup[0].getFormat()->getFormatType();
      for (uint i=1;i<fillitup.size();i++) {
        if (fillitup[i].getFormat()->getFormatType() != transcode_format) {
          cerr<<"not matching: "<<fillitup[i].getFormat()->getFormatType()<<endl;
          manager->deleteMessage();
          playlist_gui->show();
          manager->displayMessage("Error", "not all files have the same type - unable to Continue", "OK","","",0);
          changeState("OSDInputState");
          manager->update();
          return FAILURE;
        }
      }
    }

    //do some remapping of Formats to directorynames
    string oldf(transcode_format);
    transcode_format = mapFormat(transcode_format);
    manager->deleteMessage();

    if (transcode_format=="") {
      playlist_gui->show();
      manager->displayMessage("Error", "unknown filetype: "+oldf,"OK","");
      changeState("OSDInputState");
      manager->update();
      return FAILURE;
    }
    
    config_gui->setRootpath(gdpath+"/transcode_gd/"+transcode_format);
    config_gui->reset();
    config_gui->analyzeDir();
    manager->deleteMessage();
    config_gui->show();
    _url="";
    _mode=CONFIG;
    return SUCCESS;
  }
  
  Result Transcode::doSelectTrackWindow(string url) {
    GraphURL gurl = GraphURL(url);
    string protocol = gurl.getProtocolName();

    /** check file type and show configgui **/
    if (protocol != "file") {
      transcode_format = protocol;
    } else {
      transcode_format=gurl.getFormat()->getFormatType();
    }
    string oldf(transcode_format);
    transcode_format = mapFormat(transcode_format);

    if (transcode_format=="") {
      manager->displayMessage("Error", "unknown filetype: "+oldf,"OK","");
      changeState("OSDInputState");
      manager->update();
      return FAILURE;
    }

    config_gui->setRootpath(gdpath+"/transcode_gd/"+transcode_format);
    config_gui->reset();
    config_gui->analyzeDir();
    manager->deleteMessage();
    playlist_gui->hide();
    config_gui->show();
    config_gui->updateColorBar();
    _url = url;
    _mode=CONFIG;
    return SUCCESS;
  }

  void Transcode::doChangeState(string state) {
    changeState(state);
  }

  string Transcode::mapFormat(const string& trans_format) {
    MMBoxConfig* config = app->getMMBoxConfig();
    string retval;
    if (!config->getValue(trans_format,retval)) {
      cerr<<"Unknown format "<<trans_format<<endl;
      return "";
    } else {
      return retval;
    }
  }

  void Transcode::createStatusWindow() {
    group_list = new ListUnit( manager,  &app->getOSDWidgetFactory(), "medium" );
    group_list->setHeader( "Transcoding groups" );
    group_list->setDimension(300, 250);
    group_list->setPosition(50, 50);
    group_list->setColors(0x80ff0000,
                          0x80808080,
                          0xff00ff80,
                          0xff0080ff,
                          0xff808080);
    group_list->getTextWidget()->deleteAllLines();
    group_list->show();
    group_list->selectWindow();
    group_list->update();
    
    job_list = new TableUnit( manager,  &app->getOSDWidgetFactory(), "medium" );
    job_list->setHeader( "Transcoding jobs" );
    job_list->setDimension(300, 250);
    job_list->addColumn(10);
    job_list->addColumn(90);
    job_list->setPosition(370, 50);
    job_list->setColors(0x80ff0000,
                        0x80808080,
                        0xff00ff80,
                        0xff0080ff,
                        0xff808080);
    job_list->deleteAllLines();
    job_list->show();
    job_list->deSelectWindow();
    job_list->update();

    info_window = new ListUnit(manager, &app->getOSDWidgetFactory(),"medium");
    info_window->setHeader("Group information");
    info_window->setDimension(620,132);
    info_window->setPosition(50,310);
    info_window->setColors(0x80ff0000,
                           0x80808080,
                           0x800000ff,
                           0x800000ff,
                           0x800000ff);
    info_window->getTextWidget()->deleteAllLines();
    info_window->show();
    info_window->selectWindow();
    info_window->update();
  }

  void Transcode::showGroups() {
    if (app->isStateActive(this)) {
      int selection = group_list->getSelectedLine();
      group_list->getTextWidget()->deleteAllLines();
      for (uint i=0;i<_transcoding_groups.size();++i) {
        if (_transcoding_groups[i]) {
          group_list->addLine(_transcoding_groups[i]->getDescription(),itoa(i));
        }
      }
      if (selection) {
        group_list->selectLine(selection);
        showJobs(atoi(group_list->getLineURL(selection).c_str()));
        showGroupInfo(atoi(group_list->getLineURL(selection).c_str()));
      } else {
        info_window->deleteAllLines();
        info_window->update();
      }
      group_list->update();
      manager->update();
    }
  }

  void Transcode::updateInfo() {
    //do not do anything if we're in background mode
    if (!hasFocus()) {
      return;
    }
    /** is triggered by JobManager if something has changed **/
    updatemutex.lock();
    if (group_list && job_list) {
      switch (_mode) {
      case JOBS:
	switch (_focus) {
	case GROUPLIST: {
	  int selection = group_list->getSelectedLine();
	  if (selection) {
	    showGroupInfo(atoi(group_list->getLineURL(selection).c_str()));
	  }
	  break;
	}
	case JOBLIST: {
	  int selection = group_list->getSelectedLine();
	  int selection2 = job_list->getSelectedLine();
	  if (selection && selection2 ) {
            /*if (job_list->scrolling()) {
              string line = job_list->getLineContent(selection2);
              int important_pos = line.find_first_of(":")+2;
              string newline = line.substr(0,important_pos);
              newline+=line.substr(important_pos+1);
              newline+=line[important_pos];
              job_list->getTextWidget()->changeLineContent(selection2,newline);
              job_list->update();
            }
            */
            job_list->update();
	    int value = atoi(job_list->getLineContent(0,selection2).c_str());
	    showJobInfo(atoi(group_list->getLineURL(selection).c_str()),value);
	  } else {
            info_window->deleteAllLines();
            info_window->update();
          }
	  break;
	}
	}
	manager->update();
	break;
      default:
	break;
      }
    }
    updatemutex.unlock();
  }

  void Transcode::showJobs(const int& i) {
    int selection = job_list->getSelectedLine();
    job_list->deleteAllLines();
    if (_transcoding_groups[i]) {
      list<list<string> > slist = _transcoding_groups[i]->getStatus();
      for (list<list<string> >::iterator iter=slist.begin();iter!=slist.end();iter++) {
        job_list->addLine(*iter);
      }
    }
    job_list->selectLine(selection);
    job_list->update();
  }

  void Transcode::showJobInfo(const int& i,const int& j) {
    info_window->deleteAllLines();
    if (_transcoding_groups[i]) {
      list<string> slist = _transcoding_groups[i]->getJobStatus(j);
      for (list<string>::iterator iter=slist.begin();iter!=slist.end();iter++) {
        info_window->addLine(*iter);
      }
    }
    int val,max;
    _transcoding_groups[i]->getJobProgress(j,val,max);
    if (val!=0 && max !=0) {
      manager->displayProgressBar( "", 50, 420, 620);
      manager->setMaxProgress(max);
      manager->setProgress(val);
      manager->updateProgress();
    }
    info_window->update();
  }
  
  void Transcode::showGroupInfo(const int& i) {
    info_window->deleteAllLines();
    if (_transcoding_groups[i]) {
      list<string> slist = _transcoding_groups[i]->getGroupStatus();
      for (list<string>::iterator iter=slist.begin();iter!=slist.end();iter++) {
        info_window->addLine(*iter);
      }
    }
    info_window->update();
  }

  void Transcode::setConfiguration(const string& first, const string& second) {
    if (first=="trans_distri")
      second=="yes" ? distribute=true : distribute=false;
    else if (first=="trans_prio")
      second=="yes" ? priori=true : priori=false;
    else if (first=="trans_parallel") {
      parallel=atoi(second.c_str());
    }
    else if (first=="trans_local") {
      use_localhost=atoi(second.c_str());
    }
  }

  void Transcode::updateColorBar() {
    manager->displayColorBar();
    switch(_spymode) {
    case RAW:
      manager->setColorBarText("Leave spy","","Show trans","");
      return;
    case TRANS:
      manager->setColorBarText("Leave spy","","Show raw","");
      return;
    default:
      break;
    }

    int selection1 = group_list->getSelectedLine();
    int selection2 = job_list->getSelectedLine();
    switch (_focus) {
    case GROUPLIST: {
      if (selection1) {
        manager->setColorBarText("Del. group","New group","","");
      } else 
        manager->setColorBarText("","New group","","");
      return;
    }
    case JOBLIST: {
      if (selection1 && selection2) {
        int value = atoi(job_list->getLineContent(0,selection2).c_str());
          if (_transcoding_groups[atoi(group_list->getLineURL(selection1).c_str())]->spymode(value))
            // XXX: fw: spymode is disabled
            manager->setColorBarText("Delete job","","Pause","");
//             manager->setColorBarText("Delete job","","Pause","Spymode");
          else 
            manager->setColorBarText("Delete job","","Pause","");
        return;
      }
    }
    }
    manager->setColorBarText("","","","");
    return;
  }

  list<pair<int,string> > Transcode::getTranscodingGroups()
  {
    list<pair<int,string> > group_list;
    for (uint i=0;i<_transcoding_groups.size();i++) {
      if (_transcoding_groups[i])
        group_list.push_back(pair<int,string>(i,_transcoding_groups[i]->getDescription()));
    }
    return group_list;
  }

  list<string> Transcode::getTranscodingJobs(const int& group_number)
  {
    list<list<string> > slist = _transcoding_groups[group_number]->getStatus();
    list<string> slist2;
    string s;
    for (list<list<string> >::iterator iter=slist.begin();iter!=slist.end();iter++) {
      for (list<string>::iterator iter2=(*iter).begin();iter2!=(*iter).end();iter2++)
        s+=(*iter2)+" ";
      slist2.push_back(s);
      s="";
    }
    return slist2;
  }

  list<string> Transcode::getJobInformation(const int& group_number, const int& job_number)
  {
    return _transcoding_groups[group_number]->getJobStatus(job_number);
  }

  list<string> Transcode::getGroupInformation(const int& group_number) {
    return _transcoding_groups[group_number]->getGroupStatus();
  }
}
