/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include "application.hpp"
#include "mmboxstate.hpp"
#include "DistMMBoxState.hpp"
#include "../apps/taskmgr/DistTaskState.hpp"
#include "nmm/base/connect.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/ProxyObject.hpp"

#include "nmm/interfaces/base/IInstream.hpp"

#include "nmm/interfaces/base/sync/ISyncReset.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/file/IMultipleSourceFileHandler.hpp"
#include "nmm/interfaces/video/image/ICreateImage.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"

#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

#include "nmm/misc/tvguide/TVGuide.hpp"
#include "nmm/utils/fork/Fork.hpp"

namespace NMM {

MMBoxConfig::MMBoxConfig() 
{
  string filename = string(getenv("HOME")) + "/.mmboxrc";
  parseFile(filename);
}

MMBoxConfig::MMBoxConfig(const string filename) 
{
  parseFile(filename);
}

/* Begin> MMBoxEventDispatcher */
MMBoxEventDispatcher::MMBoxEventDispatcher() :
  EventDispatcher("MMBoxEventDispatcher"),
  IEventReceiver(),
  ThreadedObject(),
  event_queue(),
  event_queue_mutex(),
  event_queue_cv(event_queue_mutex),
  current_sender(NULL),
  dispatching(false),
  is_running(false) {

  // we activate the thread in ctor, no need for start() and stop()
  is_running = true;
  activateThread();
}

MMBoxEventDispatcher::~MMBoxEventDispatcher() {
  if (dispatching) {
    cerr << "Trying to stop " << getName() << " while dispatching" << endl;
  }

  // we activate the thread in ctor, no need for start() and stop()
  is_running = false;
  {
    MutexGuard mg(event_queue_mutex);
    event_queue_cv.notifyAll();
  }
  deactivateThread();

  while (!event_queue.empty()) {
    Event* e = event_queue.front().first;
    if (e) {
      cerr << "Deleting remaining Event " << e->getKey() << " in " << getName() << endl;
      delete e;
    }
    event_queue.pop();
  }
}

Result MMBoxEventDispatcher::receiveEvent(const CEvent& ce, const IEventSender* sender /*= 0*/ ) {
  Result result = SUCCESS;
  for (unsigned int i = 0; i< ce.size(); ++i) {
    Result r = receiveEvent(ce[i], sender);
    if (SUCCESS != r) {
      result = r;
    }
  }

  return result;
}

Result MMBoxEventDispatcher::receiveEvent(const Event& e, const IEventSender* sender /*= 0*/ ) {
  /* Begin> Hack for CeBIT 2004. Prehandling of Event KEY_EXIT! If we receive the event KEY_EXIT,
   * we call abort()
   */
  string event_value;
  if (e.getValue(0)) {
    const TValue<string>* tmp_value = dynamic_cast<const TValue<string>* > (e.getValue(0) );
    if (tmp_value) {
      event_value = tmp_value->getRValue();
    }
  }
  if (event_value == "KEY_EXIT" || e.getKey() == "KEY_EXIT") {      
    cerr << "Leaving MMBox ... " << endl;
    ITVGuide* itvgserver = STVGuide::getInstance().getInterface<ITVGuide>();
    if(itvgserver){
      itvgserver->stop();
    }
    
    SFork::getInstance().killall();
    abort();
  }
  /* End< */

  MutexGuard mg(event_queue_mutex);
  event_queue.push(make_pair(new Event(e), sender));
  event_queue_cv.notifyAll();

  return SUCCESS;
}

void MMBoxEventDispatcher::run() {
  while (true) {
    if (!is_running) { // we have to stop
      cerr << "Not running any more: break" << endl;
      break;           // so break
    }
    
    event_queue_mutex.lock();
    if (event_queue.empty()) {
      event_queue_cv.wait();      // this also unlockes the event_queue_mutex
    }
    event_queue_mutex.unlock();

    if (!is_running) { // we have to stop
      cerr << "Not running any more: break" << endl;
      break;           // so break
    }
    
    Event* e = NULL;
    const IEventSender* sender = NULL;
    {
      MutexGuard mg(event_queue_mutex); // lock the mutex and get the to be dispatched event
      e = event_queue.front().first;
      sender = event_queue.front().second;
      event_queue.pop();
    }

    if (!e) { // no event
      continue;
    }

    string event_key = e->getKey();
    dispatching = true;
    DEBUG_STREAM("Dispatching Event " << event_key << endl);
    Result r = SUCCESS;
    try {
      r = internal_dispatch(*e, sender);
    } catch (const Exception& e) {
      cerr << NamedObject::color_array[NamedObject::LBLUE_COLOR] 
           << getName() << ": while dispatch " << event_key << ":" << endl
           << e << endl;
      if (e.getParent()) {
        cerr << "Parent of Exception is " << e.getParent() << endl;
      }
      cerr << "Exiting!" << endl;
      exit(1);
    }
    dispatching = false;
    delete e;

    if (SUCCESS != r) {
      WARNING_STREAM("Dispatching Event " << event_key << " returned " << r << endl);
    }
  }
}

Result MMBoxEventDispatcher::internal_dispatch(const Event& e, const IEventSender* sender /*= 0*/) {
  Event send_event;
  if (e.getKey() == IDisplayEvents::keyReleased_event) {
    const TValue<string>* tvalue = dynamic_cast<const TValue<string>*>(e.getValue(0));
    if (!tvalue) {
      cerr << "could not dynamic cast TValue<string>" << endl;
      send_event = e;
    } else {
      const string& key = tvalue->getRValue();
      send_event = Event(key);
    }
  } else {
    send_event = e;
  }
  return EventDispatcher::dispatch(send_event, sender);
}
/* End< MMBoxEventDispatcher */

Result MMBoxApplication::Up() {
  return( _state->Up() );
}

Result MMBoxApplication::Down() {
  return( _state->Down() );
}

Result MMBoxApplication::VCR_Up() {
  if (audio_node) {
    IAudioDevice_var audio(audio_node->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio->increaseVolume();
  }
  return SUCCESS;
}

Result MMBoxApplication::VCR_Down() {
  if (audio_node) {
    IAudioDevice_var audio(audio_node->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio->decreaseVolume();
  }

  return SUCCESS;
}

Result MMBoxApplication::Left() {
  return( _state->Left() );
}

Result MMBoxApplication::Right() {
  return( _state->Right() );
}

Result MMBoxApplication::Return() {
  return( _state->Return() );
}

Result MMBoxApplication::Back() {
  return( _state->Back() );
}

Result MMBoxApplication::Pause() {
  return( _state->Pause() );
}

Result MMBoxApplication::Rewind() {
  return( _state->Rewind() );
}

Result MMBoxApplication::Forward() {
  return( _state->Forward() );
}

Result MMBoxApplication::Record() {
  return( _state->Record() );
}

Result MMBoxApplication::Stop() {
  return( _state->Stop() );
}

Result MMBoxApplication::OSDKey() {
  return( _state->OSDKey() );
}

Result MMBoxApplication::specialKey01() {
  return( _state->specialKey01() );
}

Result MMBoxApplication::specialKey02() {
  return( _state->specialKey02() );
}

Result MMBoxApplication::colorKeyRed() {
  return( _state->colorKeyRed() );
}

Result MMBoxApplication::colorKeyGreen() {
  return( _state->colorKeyGreen() );
}

Result MMBoxApplication::colorKeyYellow() {
  return( _state->colorKeyYellow() );
}

Result MMBoxApplication::colorKeyBlue() {
  return( _state->colorKeyBlue() );
}

Result MMBoxApplication::numKey0() {
  return( _state->numKey0() );
}

Result MMBoxApplication::numKey1() {
  return( _state->numKey1() );
}

Result MMBoxApplication::numKey2() {
  return( _state->numKey2() );
}

Result MMBoxApplication::numKey3() {
  return( _state->numKey3() );
}

Result MMBoxApplication::numKey4() {
  return( _state->numKey4() );
}

Result MMBoxApplication::numKey5() {
  return( _state->numKey5() );
}

Result MMBoxApplication::numKey6() {
  return( _state->numKey6() );
}

Result MMBoxApplication::numKey7() {
  return( _state->numKey7() );
}

Result MMBoxApplication::numKey8() {
  return( _state->numKey8() );
}

Result MMBoxApplication::numKey9() {
  return( _state->numKey9() );
}

Result MMBoxApplication::REVEAL() {
  return( _state->REVEAL() );
}

Result MMBoxApplication::reInitActualState() {
  changeState( getActualStateID() );

  return( SUCCESS );
}

Result MMBoxApplication::receiveMessage( string message ) {
  return( _state->receiveMessage( message ) );
}

void MMBoxApplication::addToActiveStates( MMBoxState* s ) {
  active_states.push_back( s );
}

void MMBoxApplication::delFromActiveStates( MMBoxState* std ) {
  list<MMBoxState*>::iterator iter, end;

  iter = active_states.begin();
  end  = active_states.end();
  bool del = false;

  MMBoxState* s;
  while( (del == false) && (iter != end) ) {
    s = *iter;
    if ( s == std ) {
      active_states.erase( iter );
      del = true;
    } else {
      iter++;
    }
  }

  if ( del ) {
    /* if deactivated state used audio node, free audio node */
    if ( s == getAudioState() )  registerAudioState( NULL );

    /* if deactivated state used background, free background */
    if ( s == getVideoState() ) registerVideoState( NULL );
  }
}

bool MMBoxApplication::isStateActive( MMBoxState* std ) {
  list<MMBoxState*>::iterator iter, end;

  iter = active_states.begin();
  end  = active_states.end();
  bool found = false;

  while( (found == false) && (iter != end) ) {
    MMBoxState* s = *iter;
    if ( s == std ) found = true;
    iter++;
  }

  return( found );
}

bool MMBoxApplication::isStateActive( string stateid ) {
  MMBoxState* state = stateinstance[stateid];

  if ( state ) {
    return( isStateActive( state ) );
  };

  return( false );
}

const MMBoxApplication::mmbox_state_list_t* MMBoxApplication::getActiveStates() {
  return( &active_states );
}

MMBoxState* MMBoxApplication::getState( string stateid ) {
  return( stateinstance[stateid] );
}

NMMApplication* MMBoxApplication::getNMMApplication() {
  return nmmapp;
}

bool MMBoxApplication::hasProxyApplication() const {
  return (dynamic_cast<ProxyApplication*>(nmmapp) ? true : false);
}


void MMBoxApplication::setNMMApplication(NMMApplication* app) {
  nmmapp = app;
}

MMBoxState* MMBoxApplication::getActualState() {
  return( _state );
}

string MMBoxApplication::getActualStateID() {
  if ( _state ) return( _state->getID() );

  return("");
}

MMBoxApplication::MMBoxApplication()
  : nmmapp(NULL),
    pngread_node(NULL),
    converter_node(NULL),
    display_node(NULL),
    audio_node(NULL),
    osd_node(NULL),
    sync(NULL),
    isync(NULL),
    audio_sync(NULL),
    video_sync(NULL),
    manager(NULL),
    factory(NULL),
    config(NULL),
    timerproducer(NULL),
    dispatcher(NULL),
    fullscreen(false),
    use_mga(false),
    __use_external_synchronizer(false),
    width(0),
    height(0),
    framerate(0),
    symbol_mutex(),
    symbol_cond(symbol_mutex),
    current_symbol(""),
    _state(NULL),
    audio_state(NULL),
    video_state(NULL)
{
  /* Format specific stuff */
  width     = 720;
  height    = 576;
  framerate = 15.0;
}

MMBoxApplication::~MMBoxApplication(){
  deleteStateInstances();
  if (!config) {
    delete config;
  }

  deinit();

  if (nmmapp) {
    delete nmmapp;
  }
}

Result MMBoxApplication::init() {
  /* Default Graph Node creation */

  // NMM-operations may throw exception:
  // catch them, output them, clean up and return FAILURE
  try {
    /*Begin> Create the request for the PNGReadNode */
    NodeDescription pngread_req;
    pngread_req.setNodeName("PNGReadNode");
    /*End< */
    
    /*Begin> Create the request for the RGBtoYV12ConverterNode  */
    NodeDescription converter_req;
    converter_req.setNodeName("RGBtoYV12ConverterNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = getNMMApplication()->getRegistry();
    registry.setOptimization(false);
    {
      RegistryLock lock(registry);
      
      list<Response> pngread_response = registry.initRequest(pngread_req);
      list<Response> converter_response = registry.initRequest(converter_req);
      
      if (pngread_response.empty() ||
          converter_response.empty()) {
        GLOBAL_ERROR_STREAM("Size of pngread_response " << pngread_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of converter_response " << converter_response.size() << endl);
        throw Exception(string("A required node is not available\n") +
                        string("Size of pngread_response ") + 
                        pngread_response.size() + 
                        string("\n") +
                        string("Size of converter_response ") + converter_response.size());
      }
      
      pngread_node = registry.requestNode(pngread_response.front());
      converter_node = registry.requestNode(converter_response.front());
    }
    /*End< */

    /* audio, video synchronizer */
    sync = new MultiAudioVideoSynchronizer();
    isync = sync->getCheckedInterface<IMultiAudioVideoSynchronizer>();

  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("MMBoxApplication::init():" << endl);
    GLOBAL_ERROR_STREAM(e << endl);

    // cleanup allocated stuff
    if (pngread_node) {
      getNMMApplication()->getRegistry().releaseNode(*pngread_node);
      delete pngread_node;
      pngread_node = NULL;
    }

    if (converter_node) {
      getNMMApplication()->getRegistry().releaseNode(*converter_node);
      delete converter_node;
      converter_node = NULL;
    }

    if (isync) {
      delete isync;
      isync = NULL;
    }

    if (sync) {
      delete sync;
      sync = NULL;
    }

    return FAILURE;
  } catch (...) {
    cerr << "Something strang went wrong in MMBoxApplication::init()" << endl;

    // no idea, how to clean up, if this happens!
    return FAILURE;
  }
  
  return SUCCESS;
}

IMultiAudioVideoSynchronizer* MMBoxApplication::getSynchronizer() {
  return( isync );
}

Result MMBoxApplication::deinit() {
  if (pngread_node) {
    pngread_node->reachConstructed();
    getNMMApplication()->getRegistry().releaseNode(*pngread_node);
    delete pngread_node;
    pngread_node = NULL;
  }
  
  if (converter_node) {
    converter_node->reachConstructed();
    getNMMApplication()->getRegistry().releaseNode(*converter_node);
    delete converter_node;
    converter_node = NULL;
  }

  if (video_sync) {
    delete video_sync;
    video_sync = NULL;
  }

  if (display_node) {
    display_node->reachConstructed();
    getNMMApplication()->getRegistry().releaseNode(*display_node);
    delete display_node;
    display_node = NULL;
  }

  if (factory) {
    delete factory;
    factory = NULL;
  }

  if (manager) {
    delete manager;
    manager = NULL;
  }

  if (osd_node) {
    osd_node->reachConstructed();
    getNMMApplication()->getRegistry().releaseNode(*osd_node);
    delete osd_node;
    osd_node = NULL;
  }
  
  if (audio_node) {
    audio_node->reachConstructed();
    getNMMApplication()->getRegistry().releaseNode(*audio_node);
    delete audio_node;
    audio_node = NULL;
  }
  
  if ( sync != NULL ) {
    delete isync;
    isync = NULL;
    
    delete sync;
    sync = NULL;
  }
  
  return SUCCESS;

}

void MMBoxApplication::setConfigFile( const string _filename) {
  config_filename = _filename;
  if (!config) {
    delete config;
  }
  config = new MMBoxConfig(config_filename);
  if (!(config->getValue("fullscreen",fullscreen))) {
    fullscreen = false;
  }
  if (!(config->getValue("matrox", use_mga))) {
    use_mga = false;
  }
}

void MMBoxApplication::setDisplayNode( INode* _display ) {
  display_node = _display;
}

void MMBoxApplication::setOsdNode( INode* _osd ) {
  osd_node = _osd;
}

void MMBoxApplication::setAudioNode( INode* _audio ) {
  audio_node = _audio;
}

INode* MMBoxApplication::getDisplayNode() {
  return( display_node );
}

INode* MMBoxApplication::getOsdNode() {
  return( osd_node );
}

IOSDManager* MMBoxApplication::getOsdManager() {
  return( manager );
}

IWidgetFactory& MMBoxApplication::getOSDWidgetFactory() {
  if (!factory) {
    factory = getOsdManager()->getParentObject()->getCheckedInterface<IWidgetFactory>();
  }

  return *factory;
}

INode* MMBoxApplication::getAudioNode() {
  return( audio_node );
}

MMBoxConfig* MMBoxApplication::getMMBoxConfig() {
  return config;
}

void MMBoxApplication::setFont( string id, string xmlfilename, string fontname  ) {
  manager->loadFont(id, xmlfilename, fontname);
}

void MMBoxApplication::setTimerProducer( TimerProducer* tp ) {
  timerproducer = tp;
}

TimerProducer* MMBoxApplication::getTimerProducer() {
  return( timerproducer );
}

void MMBoxApplication::setDispatcher(  MMBoxEventDispatcher* d ) {
  dispatcher = d;
}

MMBoxEventDispatcher* MMBoxApplication::getDispatcher() {
  return( dispatcher );
}

Result MMBoxApplication::initStates( string id ) {
  _state = stateinstance[id];
  last_stateid = id;

  if ( _state ) {
    _state->create();
    return( SUCCESS );
  }
  return( FAILURE );
}

Result MMBoxApplication::changeState( string stateid, string message ) {
  MMBoxState* state = stateinstance[stateid];
  MMBoxState* old_state = _state;
  MMBoxState* audio_state;
  MMBoxState* video_state;
  
  /* if new state needs audio and no audio node is available, fail */
  if ( !audio_node && state->isNeedAudio() ) {
    return FAILURE;
  }
  
  if ( state ) {
    /* if new state needs audio and no audio node is available, fail */
    if ( !audio_node && state->isNeedAudio() ) {
      return FAILURE;
    }

    if ( isStateActive( _state ) ) {
      GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "uninitialize() " << _state->getID() << endl);
      // Call uninitialize method from prior state
      _state->uninitialize();
      GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "uninitialize() finished " << _state->getID() << endl);
    } else {
      GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "kill() ID:" << _state->getID() << endl);
      _state->kill();
      GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "killed!" << endl);
    }

    _state = state;
    
    audio_state = getAudioState();
    if ( _state->isNeedAudio() ) {
      /* if new state needs audio and audio node is already used, terminate this state */
      if ( audio_state ) {
        if ( audio_state != _state ) {
          GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "audio_state->kill();" << endl);
	  audio_state->kill();
        } else {
          GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "reenter state (a)" << endl);
	}
      }

      /* set new audio state */
      registerAudioState( _state );
      if ( audio_state == video_state ) {
	// Do not register NULL as video state, because some state might use video only, another state audio only
	// registerVideoState( NULL );
      }
      _state->setAudioFree( true );
    }
    else {
      /* if no audio is needed and no audio state is registered, tell state that it can use audio node */
      if ( audio_state ) {
        _state->setAudioFree( false );
      }
      else {
        _state->setAudioFree( true );
      }
    }

    video_state = getVideoState();
    /* check if new state needs background */
    if ( _state->isLazyBackground() == false ) {
      /* if new state needs background and background is already used, terminate this state */
      if ( video_state ) {
        if ( video_state != _state ) {
          GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "video_state->kill();" << endl);
	  video_state->kill();
	} else {
          GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << "reenter state (v)" << endl);
	}
      }
      registerVideoState( _state );
      _state->setBackgroundFree( true );
    } else {
      /* if no background is needed and no video state is registered, tell state that it can use background */
      if ( video_state ) {
        _state->setBackgroundFree( false );
      }
      else {
        _state->setBackgroundFree( true );
      }
    }

    if ( isStateActive( _state ) ) {
      GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << _state->getID() << " state active -> init" << endl);
      // Call initialize method from new state
      if ( _state->initialize( message ) == FAILURE ) {
        GLOBAL_DEBUG_STREAM("state initialization failed, switching to prior state..." << endl);
        // if initialize failed, call init from prior state
        _state = old_state;
        _state->initialize( message );
        return( FAILURE );
      } else {
        GLOBAL_DEBUG_STREAM("state initialization ok" << endl);

        return(SUCCESS);
      }
    } else {
      GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): " << _state->getID() << " state inactive -> create" << endl);
      // Call create method from new state
      if ( _state->create( message ) == FAILURE ) {
        GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): state creation failed, switching to prior state..." << endl);
        // if create failed, call create from prior state
        _state = old_state;
        _state->create( message );
        return( FAILURE );
      } else {
        GLOBAL_DEBUG_STREAM("MMBoxApplication::changeState(): state initialization ok" << endl);

        return(SUCCESS);
      }
    }

  }
  else return(FAILURE);
}

/****************************************************************************/

void MMBoxApplication::registerState( MMBoxState* state ) {
  stateinstance[ state->getID() ] = state;
}

void MMBoxApplication::killState(const string& state_id)
{
  MMBoxState* state = stateinstance[state_id];
  if (state) {
    state->deactivate();
    state->goToLastState();
  }
}

void MMBoxApplication::deleteStateInstances() {
  list<MMBoxState*> current_active_states = active_states;
  for (list<MMBoxState*>::iterator state = current_active_states.begin();
       state != current_active_states.end();
       state++) {
    if (*state) {
      (*state)->kill();
    }
  }

  for(string_to_state_t::iterator i = stateinstance.begin();
      i != stateinstance.end();
      i++) {
    if (i->second) {
      delete i->second;
    }
  }

  /* delete all map entries */
  stateinstance.clear();

  active_states.clear();
}

void MMBoxApplication::registerAudioState( MMBoxState* state ) {
  audio_state = state;
}

void MMBoxApplication::registerVideoState( MMBoxState* state ) {
  video_state = state;
}

MMBoxState* MMBoxApplication::getAudioState() {
  return( audio_state );
}

MMBoxState* MMBoxApplication::getVideoState() {
  return( video_state );
}

Result MMBoxApplication::createGlobalNodes() {

  // NMM-operations may throw exception:
  // catch them, output them, clean up and return FAILURE
  try {
    /*Begin> Create the request for the *DisplayNode */
    NodeDescription display_req;
    display_req.setNodeName((use_mga ? "MatroxDisplayNode" : "XDisplayNode"));
    /*End< */
    
    /*Begin> Create the request for the OSDManagerNode  */
    NodeDescription osd_req;
    osd_req.setNodeName("OSDManagerNode");
    /*End< */
    
    /*Begin> Create the request for the PlaybackNode  */
    NodeDescription audio_req;
    audio_req.setNodeName("PlaybackNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = getNMMApplication()->getRegistry();
    registry.setOptimization(false);

    {
      RegistryLock lock(registry);
      
      list<Response> display_response = registry.initRequest(display_req);
      list<Response> osd_response = registry.initRequest(osd_req);
      list<Response> audio_response = registry.initRequest(audio_req);
      
      if (display_response.empty() ||
          osd_response.empty() ||
          audio_response.empty()) {
        GLOBAL_ERROR_STREAM("Size of display_response " << display_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of osd_response " << osd_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of audio_response " << audio_response.size() << endl);
        if (display_response.empty() ||
            osd_response.empty()) {
          throw Exception(string("A required node is not available\n") +
                          string("Size of display_response ") + display_response.size() + string("\n") +
                          string("Size of osd_response ") + osd_response.size() + string("\n") +
                          string("Size of audio_response ") + audio_response.size());
        } 
        if (audio_response.empty()) {
          GLOBAL_WARNING_STREAM("no soundcard available");
          //HACK FOR CEBIT2004 ONLY
          //when arts-deamon is running, the playbackNode cannot be initialized
          cerr << "/dev/dsp already in use ... exiting with errorcode -15" << endl;
          exit(-15);
        }
      }
      
      display_node = registry.requestNode(display_response.front());
      osd_node = registry.requestNode(osd_response.front());
      
      if (!audio_response.empty()) {
        audio_node = registry.requestNode(audio_response.front());
      }
    }
    /*End< */

  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("MMBoxApplication::createGlobalNodes(): requesting failed!" << endl);
    GLOBAL_ERROR_STREAM(e << endl);

    // cleanup allocated stuff
    if (display_node) {
      getNMMApplication()->getRegistry().releaseNode(*display_node);
      delete display_node;
      display_node = NULL;
    }

    if (osd_node) {
      getNMMApplication()->getRegistry().releaseNode(*osd_node);
      delete osd_node;
      osd_node = NULL;
    }

    if (audio_node) {
      getNMMApplication()->getRegistry().releaseNode(*audio_node);
      delete audio_node;
      audio_node = NULL;
    }

    return FAILURE;
  } catch (...) {
    cerr << "Something strang went wrong while requesting in MMBoxApplication::createGlobalNodes()" << endl;

    // no idea, how to clean up, if this happens!
    return FAILURE;
  }

  // step 2: configure & initialize the nodes
  try {
    IVideoDisplay_var display(display_node->getParentObject()->getCheckedInterface<IVideoDisplay>());
    if (display->hasFullscreen() && !display->isFullscreenOnly()) {
      display->setFullscreen(fullscreen);
    }
    video_sync = display_node->getParentObject()->getCheckedInterface<ISynchronizedSink>();
    IBiDirQueuedNode_var display_queue(display_node->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    display_queue->setDownstreamMaxSize(1);
    display_node->init();

    manager = osd_node->getParentObject()->getCheckedInterface<IOSDManager>();
    // IBiDirQueuedNode_var osd_queue(osd_node->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    // osd_queue->setDownstreamMaxSize(100);
    osd_node->init();

    if (audio_node) {
      
      IAudioDevice_var audio_device(audio_node->getParentObject()->getCheckedInterface<IAudioDevice>());
      audio_device->setDevice("/dev/dsp");
      
      audio_sync = audio_node->getParentObject()->getCheckedInterface<ISynchronizedSink>();
      // IBiDirQueuedNode_var audio_queue(audio_node->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      // audio_queue->setDownstreamMaxSize(100);
      audio_node->init();
    }
  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("MMBoxApplication::creatGlobalNodes(): configuring and initializing failed!" << endl);
    GLOBAL_ERROR_STREAM(e << endl);

    if (display_node->isInitialized()) {
      display_node->deinit();
    }

    if (video_sync) {
      delete video_sync;
      video_sync = NULL;
    }

    if (manager) {
      delete manager;
      manager = NULL;
    }

    if (audio_node && audio_node->isInitialized()) {
      audio_node->deinit();
    }

    if (audio_sync) {
      delete audio_sync;
      audio_sync = NULL;
    }

    return FAILURE;
  } catch (...) {
    cerr << "Something strang went wrong while configuring & initializing in MMBoxApplication::createGlobalNodes()" << endl;

    // no idea, how to clean up, if this happens!
    return FAILURE;
  }
  /*End< */
  
  /* register symbolChanged handler */
  display_node->getParentObject()->registerEventListener(ISymbolChange::symbolChanged_event, 
                                                         new TEDObject1<MMBoxApplication,string,Result(MMBoxApplication::*)(const string&)>
                                                         (this, &MMBoxApplication::symbolChanged));
  
  return SUCCESS;
}

Result MMBoxApplication::startDefaultGraph() {
  manager->setPause(true);  
  /* start the nodes */
  if ( display_node->isActivated() ) {
    display_node->start();
  }
  if ( osd_node->isActivated() ) {
    osd_node->start();
  }
  if ( converter_node->isActivated() ) {
    converter_node->start();
  }
  if ( pngread_node->isActivated() ) {
    pngread_node->start();
  }
  
  // Stop the readnode, because the osd will cache the image,
  // and we don't need to get a Buffer from pngread 25 times/second.
  // errror 20030418: no need to stop pngread any more as it used the
  // producing flag to send the image only once
  //pngread_node->stop();
  manager->update();
  
  // don't forget the audio node later for menu sound events
  
  return SUCCESS;
}

Result MMBoxApplication::createDefaultGraph( string _background ) {
  /* PNGRead Node */
  GLOBAL_DEBUG_STREAM("initing PNG" << endl);
  pngread_node->reachInitialized();

  /* if no background is specified, create black image */
  if ( _background == "" ) {
    ICreateImage_var image_source(pngread_node->getParentObject()->getCheckedInterface<ICreateImage>());
    image_source->createImage(0x00000000, width, height);
  } else {
    IMultipleSourceFileHandler_var file_source(pngread_node->getParentObject()->
					       getCheckedInterface<IMultipleSourceFileHandler>());
    file_source->addFilename(_background);
  }
  GLOBAL_DEBUG_STREAM("inited PNG" << endl);
  
  try {
    
    /* stop and disconnect first. won't hurt (I hope!!) */
    disconnectVideoGraph();
    
    /* Converter RGB to YUV */
    GLOBAL_DEBUG_STREAM("initing Converter" << endl);
    if (converter_node->isConstructed()) {
      try {
        converter_node->init();
      } catch(...) {
        GLOBAL_ERROR_STREAM("Converter init failed" << endl);
        return FAILURE;
      }
    }
    GLOBAL_DEBUG_STREAM("inited Converter" << endl);
  
    /* connect the nodes */
    GLOBAL_DEBUG_STREAM("connecting PNG" << endl);
    if (pngread_node->isInitialized()) {
      pngread_node->initOutput();
    }
    if (pngread_node->isStarted()) {
      pngread_node->stop();
    }
    if (pngread_node->isActivated()) {
      pngread_node->deactivate();
    }
    GLOBAL_DEBUG_STREAM("prepared PNG" << endl);
    {
      IOutputJack_var pngout_jack(pngread_node->getOutputStream("default"));
      if (!pngout_jack.get() || !(pngout_jack->isConnected())) {
        connect(pngread_node, converter_node);
      }
    }
    GLOBAL_DEBUG_STREAM("connected PNG" << endl);
    
    GLOBAL_DEBUG_STREAM("connecting to OSD" << endl);
    if (pngread_node->isOutputInitialized()) {
      pngread_node->activate();
    }
    if (converter_node->isInitialized()) {
      converter_node->initOutput();
    }
    if (converter_node->isStarted()) {
      converter_node->stop();
    }
    if (converter_node->isActivated()) {
      converter_node->deactivate();
    }
    {
      IOutputJack_var convout_jack(converter_node->getOutputStream("default"));
      if (!convout_jack.get() || !(convout_jack->isConnected())) {
        connect(converter_node, osd_node, "stream1");
      }
    }
    GLOBAL_DEBUG_STREAM("connected to OSD" << endl);
    
    GLOBAL_DEBUG_STREAM("connecting to display" << endl);
    converter_node->activate();
    if (osd_node->isInitialized()) {
      osd_node->initOutput();
    }
    if (osd_node->isStarted()) {
      osd_node->stop();
    }
    if (osd_node->isActivated()) {
      osd_node->deactivate();
    }
    {
      IOutputJack_var osdout_jack(osd_node->getOutputStream("default"));
      if (!osdout_jack.get() || !(osdout_jack->isConnected())) {
        connect(osd_node, display_node);
      }
    }
    GLOBAL_DEBUG_STREAM("connected to display" << endl);
  
    GLOBAL_DEBUG_STREAM("activating OSD & display" << endl);
    if (osd_node->isOutputInitialized()) {
      osd_node->activate();
    }
    if (display_node->isInitialized()) {
      display_node->initOutput();
    }
    if (display_node->isOutputInitialized()) {
      display_node->activate();
    }
    GLOBAL_DEBUG_STREAM("activated OSD & display" << endl);

  } catch (const Exception& e) {
    cerr << e << endl;
    if (e.getParent()) {
      cerr << "Parent is " << e.getParent()->getName() << endl;
    }
    
    return FAILURE;
  }

  // errror 20030225: I don't think that this is neccessary, as the Formats should be exact enough  
//   /*
//    * Here we must send the initial resolution.
//    * since some sub applications use other resolution (DVD uses PAL or NTSC)
//    *
//    * fw: is this really necessary? we already specify the resolution AND aspect ratio in the format...
//    */
// #warning "initial resolution"
// //   CEvent* event = new CEvent();
// //   Event* event_ptr;
// //   event_ptr = new Event( "resolution_changed", 2 );
// //   event_ptr->setValue( 0, new TValue<int>(width) );
// //   event_ptr->setValue( 1, new TValue<int>(height) );
// //   event->insertEvent( 0, *event_ptr );

// //   event_ptr = new Event( "aspect_ratio", 2 );
// //   event_ptr->setValue( 0, new TValue<int>( 4 ) );
// //   event_ptr->setValue( 1, new TValue<int>( 3 ) );
// //   event->insertEvent( 1, *event_ptr );

// //   CEvent* event2 = new CEvent();
// //   *event2 = *event;

// // #warning using SetEventDispatcher
// // //   display_node->sendSetEvent( event );
// // //   osd_node->sendSetEvent( event2 );
  

  video_sync->setSynchronized(false);  
  ISinkController_var video_sync_controller(video_sync->getController());
  isync->addVideoSinkController(video_sync_controller.get() );
  // video_sync_controller->permitDiscardBuffers( false );
  
  if (audio_sync) {
    setAudioSynchronized(false);
    
    if (!getUseExternalSynchronizer()) {
      ISinkController_var audio_sync_controller(audio_sync->getController());
      isync->addAudioSinkController( audio_sync_controller.get() );
    }
    
    // audio_sync_controller->setMaxDelay( Interval(5,0) );
    // audio_sync_controller->permitDiscardBuffers( false );
  }
  
  return SUCCESS;
}

void MMBoxApplication::setSynchronized( bool s ) {
  if (audio_sync && !getUseExternalSynchronizer()) {
    audio_sync->setSynchronized(s);
  }
  
  if (video_sync) {
    video_sync->setSynchronized(s);
  }
}

void MMBoxApplication::setAudioSynchronized( bool s ) {
  if (audio_sync && !getUseExternalSynchronizer()) {
    audio_sync->setSynchronized(s);
  }
}
  
void MMBoxApplication::setVideoSynchronized( bool s ) {
  if (video_sync) {
    video_sync->setSynchronized(s);
  }
}

void MMBoxApplication::syncReset() {

  cerr << "reset Synchronizer()" << endl;
  if (audio_sync && !getUseExternalSynchronizer()) {
    audio_sync->syncReset();
  }
  
  if (video_sync) {
    video_sync->syncReset();
  }
}

Result MMBoxApplication::disconnectDefaultGraph() {
  GLOBAL_DEBUG_STREAM("Begin> MMBoxApplication::disconnectDefaultGraph" << endl);
  
  GLOBAL_DEBUG_STREAM("MMBoxApplication::disconnectDefaultGraph: disconnecting video graph" << endl);
  disconnectVideoGraph();
  GLOBAL_DEBUG_STREAM("MMBoxApplication::disconnectDefaultGraph: disconnected video graph" << endl);
  
  GLOBAL_DEBUG_STREAM("End< MMBoxApplication::disconnectDefaultGraph" << endl);
  return SUCCESS;
}

Result MMBoxApplication::disconnectVideoGraph() {
  /*****************************************************************************************/
  /* disconnect Video Graph */
  
//   //cerr << "disconnectVideoGraph: start" << endl;
  
//   //if ( display_node->getState() == INode::STARTED ) display_node->stop();
//   GLOBAL_DEBUG_STREAM("Stopping OSD" << endl);
//   if ( osd_node->getState() == INode::STARTED ) {
//     osd_node->stop();
//     osd_node->flush();
//   }
//   GLOBAL_DEBUG_STREAM("Stopped OSD" << endl);

//   GLOBAL_DEBUG_STREAM("Stopping Converter" << endl);
//   if ( converter_node->getState() == INode::STARTED ) {
//     converter_node->stop();
//     //converter_node->flush();
//   }
//   GLOBAL_DEBUG_STREAM("Stopped Converter" << endl);

//   GLOBAL_DEBUG_STREAM("Stopping PNG" << endl);
//   if ( pngread_node->getState() == INode::STARTED )   pngread_node->stop();
//   GLOBAL_DEBUG_STREAM("Stopped PNG" << endl);
  
//   //if ( display_node->getState() == INode::ACTIVATED ) display_node->deactivate();
//   //if ( display_node->getState() == INode::OUTPUT_INITIALIZED ) display_node->deinitOutput();
//   GLOBAL_DEBUG_STREAM("Deactivating OSD" << endl);
//   if ( osd_node->getState() == INode::ACTIVATED ) osd_node->deactivate();  
//   GLOBAL_DEBUG_STREAM("Deactivated OSD" << endl);
  
//   //display_node->disconnectInput();
//   GLOBAL_DEBUG_STREAM("Deiniting Output OSD" << endl);
//   if ( osd_node->getState() == INode::OUTPUT_INITIALIZED ) osd_node->deinitOutput();
//   GLOBAL_DEBUG_STREAM("Deinited Output OSD" << endl);

//   GLOBAL_DEBUG_STREAM("Deactivating Converter" << endl);
//   if ( converter_node->getState() == INode::ACTIVATED ) converter_node->deactivate();
//   GLOBAL_DEBUG_STREAM("Deactivated Converter" << endl);
  
//   GLOBAL_DEBUG_STREAM("Disconnecting input of OSD" << endl);
//   {
//     IInputJack_var stream1_jack(osd_node->getInputStream("stream1"));
//     if (stream1_jack.get() && stream1_jack->isConnected()) {
//       disconnect_input(osd_node, string("stream1"));
//     }
//     IInputJack_var stream2_jack(osd_node->getInputStream("stream2"));
//     if (stream2_jack.get() && stream2_jack->isConnected()) {
//       disconnect_input(osd_node, string("stream2"));
//     }
//   }
//   GLOBAL_DEBUG_STREAM("Disconnected input of OSD" << endl);

//   GLOBAL_DEBUG_STREAM("Deiniting Output Converter" << endl);
//   if ( converter_node->getState() == INode::OUTPUT_INITIALIZED ) converter_node->deinitOutput();
//   GLOBAL_DEBUG_STREAM("Deinited Output Converter" << endl);

//   GLOBAL_DEBUG_STREAM("Deactivating PNG" << endl);
//   if ( pngread_node->getState() == INode::ACTIVATED ) pngread_node->deactivate();  
//   GLOBAL_DEBUG_STREAM("Deactivated PNG" << endl);
  
//   if (pngread_node->isInitialized()) {
//     pngread_node->initOutput();
//   }
//   GLOBAL_DEBUG_STREAM("Disconnecting output of PNG which is in state " << pngread_node->getState() << endl);
//   if (pngread_node->isOutputInitialized()) {
//     IOutputJack_var jack(pngread_node->getOutputStream("default"));
//     if (jack.get() && jack->isConnected()) {
//       disconnect_output(pngread_node);
//     }
//   } else {
//     GLOBAL_ERROR_STREAM("PNG is in state " << pngread_node->getState() << ", ignoring disconnect call!" << endl);
//   }
//   GLOBAL_DEBUG_STREAM("Disconnected output of PNG" << endl);

//   GLOBAL_DEBUG_STREAM("Deiniting Output PNG" << endl);
//   if ( pngread_node->getState() == INode::OUTPUT_INITIALIZED ) pngread_node->deinitOutput();
//   GLOBAL_DEBUG_STREAM("Deinited Output PNG" << endl);
  
  setSynchronized( false );
  
  //cerr << "disconnectVideoGraph: end" << endl;
  
  try {

    /* BEGIN> stop and flush video graph */
//     if (display_node->isStarted()) {
//       display_node->stop();
//     }
    if (osd_node->isStarted()) {
      osd_node->stop();
    }
    if (converter_node->isStarted()) {
      converter_node->stop();    
    }
    if (pngread_node->isStarted()) {
      pngread_node->stop();
    }
    
//     if (display_node->isActivated()) {
//       display_node->flush();
//     }
    if (osd_node->isActivated()) {
      osd_node->flush();
    }
    if (converter_node->isActivated()) {
      converter_node->flush();
    }
    if (pngread_node->isActivated()) {
      pngread_node->flush();
    }
    /* END< stop and flush video graph */
    
    /* BEGIN> disconnect video graph bottom-up */
//     if (display_node->isActivated()) {
//       display_node->deactivate();
//     }
//     if (display_node->isOutputInitialized()) {
//       display_node->deinitOutput();
//     }
//     IInputJack_var displayin_jack(display_node->getInputStream("default"));
//     if (displayin_jack.get() && displayin_jack->isConnected()) {
//       if (osd_node->isActivated()) {
//         osd_node->deactivate();
//       }
//       disconnect_input(display_node, string("default"));
//     }
    
    if (osd_node->isActivated()) {
      osd_node->deactivate();
    }
    if (osd_node->isOutputInitialized()) {
      osd_node->deinitOutput();
    }
    IInputJack_var osdin_jack(osd_node->getInputStream("stream1"));
    if (osdin_jack.get() && osdin_jack->isConnected()) {
      if (converter_node->isActivated()) {
        converter_node->deactivate();
      }
      disconnect_input(osd_node, string("stream1"));
      /* do NOT disconnect stream2! the state that connected stream2 is responsible for disconnecting it */
      /* it is NOT necessary to disconnect stream2 to have corect state transitions */
    }
    
    if (converter_node->isActivated()) {
      converter_node->deactivate();
    }
    if (converter_node->isOutputInitialized()) {
      converter_node->deinitOutput();
    }
    if (converter_node->isInputConnected()) {
//       if (pngread_node->isActivated()) {
//         pngread_node->deactivate();
//       }
      pngread_node->reachOutputInitialized();
      converter_node->reachInitialized();
      disconnect_input(converter_node, string("default"));
    }
    
    if (pngread_node->isActivated()) {
      pngread_node->deactivate();
    }
    if (pngread_node->isOutputInitialized()) {
      pngread_node->deinitOutput();
    }
    /* END< disconnect video graph bottom-up */
    
  } catch(const Exception& e) {
    
    GLOBAL_ERROR_STREAM("exception while disconnecting video graph" << endl);
    GLOBAL_ERROR_STREAM(e << endl);
    return FAILURE;
    
  } catch(...) {
    
    GLOBAL_ERROR_STREAM("something strange went wrong while disconnecting video graph" << endl);
    return FAILURE;
    
  }
  
  return SUCCESS;
}

IMMBoxState* MMBoxApplication::createDistState(const string& id)
{
  DistMMBoxState* wrapper_state = 0;
  if(id == "DistMMBoxState") {
    wrapper_state = new DistMMBoxState(id, this);
  } else if(id == "DistTaskState") {
    wrapper_state = new DistTaskState(id, this);
  } else {
    return 0;
  }
  IMMBoxState* mmbox_state = wrapper_state->getInterface<IMMBoxState>();
  registerState(wrapper_state);
  changeState(id);
  return mmbox_state;
}

void MMBoxApplication::deleteDistState(const string& id) 
{
  MMBoxState* state = stateinstance[id];
  if(state) {
    // if state to kill is *not* the actual state, call kill, and go back to actual state
    // (code taken from taskmgr)
    if(getActualStateID() != id) {
      GLOBAL_DEBUG_STREAM("kill *other* state than actual state" << endl);
      state->kill();
      reInitActualState();
    }
    // if state to kill is the actual state, call deactivate, then go to last state
    // (goToLastState will call changeState where the actual state is killed)
    else {
      GLOBAL_DEBUG_STREAM("kill *actual* state" << endl);
      state->deactivate();
      state->goToLastState();
    }
    stateinstance.erase(id);
    delete state;
  }
}

Result MMBoxApplication::symbolChanged(const string& in) {
  MutexGuard mg(symbol_mutex);
  current_symbol = in;
  symbol_cond.notify();
  
  return SUCCESS;
}

void MMBoxApplication::waitForSymbolChange(const string& symbol) {
  MutexGuard mg(symbol_mutex);
  while (current_symbol != symbol) {
    symbol_cond.wait();
  }
  current_symbol = "";
}


string MMBoxApplication::getStateSessionLocation(const string &id){
  //list<MMBoxState*>         active_states

  for(list<MMBoxState*>::iterator i = active_states.begin(); i != active_states.end(); i++){
    if((*i)->getID() == id){
      return (*i)->getSessionLocation();
    }
  }
  
  return NULL;
}

}
