/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "RegistryView.hpp"
#include "nmm/base/registry/NodeDescription.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/Event.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/interfaces/base/registry/IServerRegistry.hpp"

#include <qlayout.h>
#include <qlistview.h>
#include <algorithm>
#include <iostream>


RegistryView::RegistryView(NMM::NMMApplication& na,QWidget *parent, const char *name ) : View(parent,name) {
  __registry = &na.getRegistry();

  allocate();
  initConnect();
  updateView();
}

RegistryView::~RegistryView() {
}

void RegistryView::updateView() {
  __registries->clear();
  __internal_map.clear();


  for (map<NMM::RegistryName, NMM::ClientRegistry::RegistryStruct>::iterator it = __registry->getRegistryList().begin(); it != __registry->getRegistryList().end(); ++it)
    createEntry(it->second.server_registry , 0, it->second.server_registry->getRegistryID());
}

void RegistryView::createEntry(NMM::IRegistry* registry, QListViewItem* item, const NMM::RegistryID& reg_id) {
  try {
    QListViewItem* tmp_item = 0;
    if ( item )
      tmp_item = new QListViewItem(item, reg_id.getRegistryName().c_str());
    else
      tmp_item = new QListViewItem(__registries, (reg_id.getRegistryName() + "@"+reg_id.getUserInfo().getHostname()).c_str());

    __internal_map.insert(pair<QListViewItem*, NMM::RegistryID>(tmp_item, reg_id));

    /* Begin> Get the NodeInfo list */
    list<NMM::NodeInfo> node_list;
    try  {
      list<NMM::NodeInfo> nodel = registry->getNodeInfoList(reg_id);
      node_list = nodel;
    } catch (NMM::Exception& e) {
      cerr << e << endl;
    }
    /* End< */

    QString current_count;
    QString max_count;
    for (list<NMM::NodeInfo>::iterator it = node_list.begin(); it != node_list.end(); ++it)
    {
      if (it->getMaxInstanceCount() == -1)
        max_count = "unlimited";
      else
        max_count.setNum(it->getMaxInstanceCount());

      current_count.setNum(it->getInstanceCount());

      /* Begin> Convert the ObjectID in a string */
      strstream stream;
      stream << hex << it->getObjectID();
      string s;
      stream >> s;
      QString string_id(s.c_str());
      /* End< */

      /* Begin> set the NodeType" */
      QString node_type;
      switch (it->getNodeType()) {
        case NMM::INode::UNKNOWN: node_type = "Unknwown"; break;
        case NMM::INode::SOURCE : node_type = "Source"  ; break;
        case NMM::INode::SINK   : node_type = "Sink" ;break;
        case NMM::INode::CONVERTER : node_type = "Converter"; break;
        case NMM::INode::FILTER :  node_type = "Filter"; break;
        case NMM::INode::MULTIPLEXER: node_type = "Multiplexer"; break;
        case NMM::INode::DEMULTIPLEXER: node_type = "Demultiplexer"; break;
        case NMM::INode::COMPOSITE : node_type = "Composite"; break;
        case NMM::INode::MUX_DEMUX : node_type = "Multiplexer/Demultiplexer"; break;
      }
      /* End< */

      /* add the item */
      new QListViewItem(tmp_item, it->getNodeName().c_str(), node_type, it->getLibraryName().c_str(), current_count, max_count, string_id);
    }

    /* Begin> create the entry for the other registries */
    list<NMM::RegistryID> child_list = registry->getChildList(reg_id);
    for (list<NMM::RegistryID>::iterator it = child_list.begin(); it != child_list.end(); ++it)
    {
      createEntry(registry, tmp_item, *it);
    }
    /* End< */
  } catch (const NMM::RegistryException& re) {
    cerr << re << endl;
  }
}

void RegistryView::initConnect() {
  connect(__registries, SIGNAL(clicked(QListViewItem*)), this, SLOT(slotShowDevices(QListViewItem*)));
}

void RegistryView::allocate() {
  __top_level = new QGridLayout(this, 1, 1);
  __registries = new QListView(this);
  __registries->setRootIsDecorated(true);
  __registries->addColumn("Registry");
  __registries->addColumn("Node type");
  __registries->addColumn("Plugin");
  __registries->addColumn("Used");
  __registries->addColumn("Max Count");
  __registries->addColumn("ObjectID");
  __top_level->addWidget(__registries, 0,0);
}


void RegistryView::slotShowDevices(QListViewItem*) {

}
