/* 
 * Copyright (C) 2005 Andrs Otn Urbano
 *
 * Maintainer:        Andrs Otn <andres.oton@gmail.com>
 * Author:            Andrs Otn <andres.oton@gmail.com>
 * From:  PlaybackNode.hpp by NMM work group to use Alsa                    
 *                     
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_ALSAPLAYBACKNODE_HPP_
#define _NMM_ALSAPLAYBACKNODE_HPP_

#include "nmm/base/registry/DevicePlugin.hpp"
#include "nmm/base/sync/GenericBSyncSinkNode.hpp"
#include "nmm/base/sync/Clock.hpp"

#include "nmm/interfaces/device/audio/IAudioDeviceImpl.hpp"
#include "nmm/interfaces/file/ITrackImpl.hpp"
#include "nmm/interfaces/audio/IMuteImpl.hpp"
#include <alsa/asoundlib.h>

namespace NMM {

  //! This class plays an audio stream to a given output device.
  /*!
   *  The ALSAPlaybackNode uses the ALSA audio driver.
   *  The Node itself supports the following formats:
   *
   *  - Logarithmic mu-law encoding
   *  - Logarithmic A-law encoding
   *  - Linear encoding:
   *    -  8-bit, (un)signed
   *    - 16-bit, (un)signed, [big|little] endian
   *  
   *
   *  Possible sample rates: 8000, 11025, 22050, 44100, 48000, 96000 Hz
   *
   *  Possible channel numbers: 1, 2, 4, 5, 6
   *
   *  The actually supported formats may be only a subset of these depending on your
   *  sound hardware.
   *
   *  In this new version, much functionality is implemented in
   *  GenericSinkNode.
   */
	class ALSAPlaybackNode :
		public GenericBSyncSinkNode,
		protected IAudioDeviceImpl,
		protected ITrackImpl,
		protected IMuteImpl
	{
	private:

		/* Handle for the PCM device */ 
	 	snd_pcm_t *pcm_handle;

		/* Playback stream */
		snd_pcm_stream_t stream;

		/* This structure contains information about    */
		/* the hardware and can be used to specify the  */
		/* configuration to be used for the PCM stream. */
		snd_pcm_hw_params_t *hwparams;

		unsigned int buffer_time;
		unsigned int period_time;
		snd_pcm_uframes_t period_frames;
		snd_pcm_uframes_t buffer_frames;
		
		snd_pcm_status_t *pcm_status;

		// Used to set the pcm access to INTERLEAVED
		snd_pcm_access_t pcm_access;

		//! Formats to check in the doInit method.
		static snd_pcm_format_t test_formats[];
		
		//! Channels to check in the doInit method.
		static const int test_channels[];
		
		//! Sample rates to check in the doInit method.
		static const int test_sample_rates[];
		
		//! Converts an ALSA_Format into an NMM::Format.
		Format& getNMMFormat(snd_pcm_format_t alsa_format, Format& nmm_format);
		
		//! Converts an NMM::Format into an ALSA_Format.
		snd_pcm_format_t getALSAFormat(const Format& nmm_format);
		
		unsigned int sample_rate;                //!< The sample_rate (depends on input_format)
		int channels;                   //!< The number of channels (depends on input_format)
		int bit_per_sample;             //!< The bits per sample (depends on input format)
		int byte_per_sample;            //!< The bytes per sample (depends on input format)
		int byte_per_second;            //!< The bytes per second (depends on input format)
		
		int internal_buffer_size;
		Time internal_buffer_time;
		
		Clock clock;
		
		void setCLOEXEC(int fd);

	protected:
		string dev;              //!< The name of the sound device.
		string __mixer_device;
		
		Result xrun(void);
		
		Result suspend(void);
		
		Result doSetFormat(const Format& format);
		
		//! The doFlush method from Node.
		/*!
		*  Flushes the audio device and checks the supported formats.
		*  \return - SUCCESS If the device could be flushed
		*          - FAILURE otherwise
		*/
		virtual Result doFlush();
		
		//! The doInit method from Node.
		/*!
		*  Simply calls doInitInput
		*/
		virtual Result doInit();
		
		//! The doInitOutput method from Node.
		/*!
		* In this method, the sound device is configured to the values determined in the formats.
		* \return SUCCESS if the device has been configured correctly, FAILURE otherwise.
		*/
		virtual Result doInitOutput();
		
		//! The doDeinit method from Node.
		/*!
		*  Simply calls doDeinitInput
		*/
		virtual Result doDeinit();
		
		//! Sets the output device.
		/** \warning obsolete method
		*/
		Result oldSetDevice(string& );
		
		//! Gets the output device.
		/** \warning obsolete method
		*/
		Result oldGetDevice(string& );
		
		//! The getOutputDelay - method from GenericSyncSinkNode.
		Interval getOutputDelay();
		
		//! The doStart method from Node.
		virtual Result doStart();
		
		//! The doStop method from Node.
		/*! 
		*  In this method the data in the internal buffer of the sound device will be played.
		*  The method returns when the internal buffer is empty. Depending on the fragment size,
		*  that means this method can take up to one second.
		*/
		virtual Result doStop();
		
		//! The presentBuffer method from GenericSyncSinkNode.
		virtual Message* presentBuffer(Buffer *in_buffer);
		
		//! Event handler for debug 
		Result EventDebug(int&);
		
		//! startTrack from ITrack interface
		virtual Result startTrack(const string&);
		
		//! endTrack from ITrack interface
		virtual Result endTrack();
		
		//! finished from ITrack interface
		virtual Result finished();
		
		//! new seek method
		Result eventSeek(Interval&, Time&, long&);
		
		//! speed up or delay playback
		Result eventSpeedUp(int& speedup);
		
		/**
		* Set audio format.
		* This method configures the audio device to a given audio format.
		* It is called on activation and on format changes. It has been renamed to keep interface
		* and implementation seperate.
		* @param format NMM audio format
		* @return SUCCESS if the device has been configured correctly, FAILURE otherwise.
		*/
		Result setFormatInternal(Format& input_format);

	        /** Begin> Added dummy implementation of two new interface methods to compile this node */
		void setVolume(const int& ) {}
		int getVolume() { return 0;}
		/** End< */ 
	public:
		//! The main constructor.
		/*!
		*  Can take a BufferManager to use instead of creating its own Allocator.
		*  \param dev the device to open.
		*  \param manager the buffer manager to use.
		*/
		ALSAPlaybackNode(const char* name = "ALSAPlaybackNode",
				const StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND,
				const int size = StreamQueue::DEFAULT_SIZE);
		
		//! The main destructor.
		/*!
		*  Closes the audio device if it is still open.
		*/
		~ALSAPlaybackNode();
		
		void setLineInVolume( int vol );
		int getLineInVolume();
		
		//! Sets the output device.
		void setDevice(const string& ) throw (const DeviceNotFoundException&);
		
		//! Sets the mixer devce
		void setMixerDevice(const string&) throw (const DeviceNotFoundException&);
		
		//! Sets the output device.
		string getDevice() const;
		
		//! Gets the output device.
		string getMixerDevice() const;

	protected:

		virtual void increaseVolume();
		
		virtual void decreaseVolume();
		
		/**
		* Mutes an audio source or sink
		*/
		virtual void mute();
		
		/**
		* Un-mutes an audio source or sink
		*/
		virtual void unmute();
	};

	class ALSAPlaybackNodePlugin
		: public TDevicePlugin<ALSAPlaybackNode>
	{
	public:
		ALSAPlaybackNodePlugin(const string& node_name, int max_instance = -1);

	protected:
		virtual void checkForStandardAuxDevice(IDevice& device) throw();
		virtual void checkForAuxDevice(IDevice& device, unsigned device_num) throw();
	};
}

#endif // _NMM_ALSAPLAYBACKNODE_HPP_


