/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <getopt.h>
#include <unistd.h>
#include <iostream>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/video/tv/ITV.hpp"
#include "nmm/interfaces/video/tv/IIVTV.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/file/IFileModeHandler.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/NodeDescription.hpp"
#include "nmm/base/registry/GraphDescription.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

// import namespace
using namespace NMM;

void usage( char *name ) {
  cerr << endl 
       << "Usage: " << name << " [options] <channel> <filename>" << endl
       << endl
       << "  -r --remote  : specify the remote source" << endl
       << "  -s --source  : specify either IVTV or DVB-Source. default is DVB" << endl
       << "  -t --timeout : records specified minutes and exits then" << endl
       << "  -q --quiet   : be realy quiet, only report abording errors" << endl
       << "  -v --verbose : increase verbosity (only without -q)" << endl
       << "  -h --help    : show this help" << endl;
};

int main(int argc, char* argv[]) {
  bool quiet = false;
  int verbose = 0;
  string filename("");
  int channel;
  unsigned int timeout = 0;
  string source("dvb");
  string remoteHost("localhost");
  int remotePort = 22801;

  /* command line parameter - see getopt(3) */
  const char *opt_string = "t:qvhs:r:";
  struct option options [] = {
        {"timeout",1,NULL,'t'},
        {"quiet",0,NULL,'q'},
        {"verbose",0,NULL,'v'},
        {"help",0,NULL,'h'},
        {"source",1,NULL,'s'},
        {"remote",1,NULL,'r'},
        {NULL,0,NULL,0}
  };

  char c;
  int long_option_index;
  while ((c=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF) {
    // which option did we get?
    switch (c) {
    case 't':
      timeout = atoi(optarg);
      break;
    case 'q':
      quiet = true;
      break;
    case 'v':
      verbose++;
      break;
    case 's':
      source = string(optarg);
      break;
    case 'r':
      if(optarg) {
        string remote_registry = string(optarg);
        string::size_type colon_pos = remote_registry.find_first_of(':');
        if (colon_pos == string::npos) {
          remoteHost = remote_registry;
          remotePort = ClientRegistry::default_port;
        } else {
          remoteHost = remote_registry.substr(0, colon_pos);
          remotePort = atoi(remote_registry.substr(colon_pos+1, string::npos).c_str());
        }
      }
      break;
    default:
    case 'h':
      usage(argv[0]);
      return 0;
      break;
    }
  }
  
  /* handle *Stream output */
  if (quiet) {
    verbose = 0;
  }
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  if (argc - optind < 2) {
    usage( argv[0] );
    return 0;
  } else {
    channel = atoi(argv[optind++]);
    filename = string(argv[optind]);
  }

  /* the NMM application for the example, intialized below */
  NMMApplication* app = 0;
  GraphDescription gd;
  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    app = ProxyApplication::getApplication(argc, argv);

    /*Begin> Create the 1. request for the MP3ReadNode*/
    NodeDescription dvb;
    if(source == "ivtv"){
      cerr << "using IVTV from " << remoteHost << endl;
      dvb.setNodeName("IVTVReadNode");
      //dvb.addNodeEvent(Event(IIVTV::setBitrate_event, new TInValue<int>(800000)), INode::INITIALIZED);
      //dvb.addNodeEvent(Event(IIVTV::setPeakBitrate_event, new TInValue<int>(1000000)), INode::INITIALIZED);

    }else{
      cerr << "using DVB from " << remoteHost << endl;
      dvb.setNodeName("DVBReadNode");
    }
    /*End< */

    /*Begin> Create the 2. request for the MPEGAudioDecodeNode*/
    NodeDescription writer;
    writer.setNodeName("RawWriteNode");
    /*End< */

    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();

    if (remoteHost != "localhost") { 
      registry.addRegistry(TCPAddress(remoteHost, remotePort));
      dvb.setLocation(remoteHost);
    }

    gd.addEdge(dvb, writer);

    registry.requestGraph(gd);

    ISinkFileHandler_var file_sink(gd.getINode(writer)->getParentObject()->getCheckedInterface<ISinkFileHandler>());
    file_sink->setFilename(filename);
    file_sink->setWorkingDirectory(string("."));
    file_sink->setFileExtension(string("mpeg"));

    IFileModeHandler_var file_mode(gd.getINode(writer)->getParentObject()->getCheckedInterface<IFileModeHandler>());
    file_mode->setFormat(IFileModeHandler::SIZE_TIMESTAMP_DATA);
  
    gd.realizeGraph();
    gd.startGraph();

    ITV_var tv(gd.getINode(dvb)->getParentObject()->getInterface<ITV>());
    if(tv.get()){
      if (0 == channel) {
        tv->setCurrentInput(0);
        tv->setStandardByName("PAL");
      }else{
        tv->setCurrentInput(4);
        tv->setStandardByName("PAL");
        tv->setFrequencyTable("pal-europe");
      }
    }

    ITVCard_var tuner(gd.getINode(dvb)->getParentObject()->getInterface<ITVCard>());
    if(tuner.get()) tuner->setChannel(channel);

    if (timeout) { // if timeout was given
      sleep(60 * timeout); // sleep timeout minutes and exit then
    } else {
      char key;
      do {
        cin >> key;
      } while (key != 'q'); // else wait for a 'q' on stdin
    }

    // stop nodes, reset and destroy them
    registry.releaseGraph(gd);
  }
  // catch NMM-exception and print it
  catch(Exception e) {
    cerr << e << endl;    
  }

  return 0;
}
