/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <getopt.h>
#include <iostream>
#include <sstream>
#include <unistd.h>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

// import namespace
using namespace NMM;


void usage( char *name ) {
  cerr << endl 
       << "Usage: " << name << " [options] <channel> <filename>" << endl
       << endl
       << "  -t --timeout   : records specified minutes and exits then" << endl
       << "  -T --timestamp : prepend date timestamp to filename" << endl
       << "  -q --quiet     : be realy quiet, only report abording errors" << endl
       << "  -v --verbose   : increase verbosity (only without -q)" << endl
       << "  -h --help      : show this help" << endl;
};

int main(int argc, char* argv[]) {
  bool quiet = false;
  int verbose = 0;
  string filename("");
  int channel;
  unsigned int timeout = 0;
  bool timestamp(false);

  /* command line parameter - see getopt(3) */
  const char *opt_string = "t:Tqvh";
  struct option options [] = {
        {"timeout",1,NULL,'t'},
        {"timestamp",0,NULL,'T'},
        {"quiet",0,NULL,'q'},
        {"verbose",0,NULL,'v'},
        {"help",0,NULL,'h'},
        {NULL,0,NULL,0}
  };

  char c;
  int long_option_index;
  while ((c=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF) {
    // which option did we get?
    switch (c) {
    case 't':
      timeout = atoi(optarg);
      break;
    case 'q':
      quiet = true;
      break;
    case 'T':
      timestamp = true;
      break;
    case 'v':
      verbose++;
      break;
    default:
    case 'h':
      usage(argv[0]);
      return 0;
      break;
    }
  }
  
  /* handle *Stream output */
  if (quiet) {
    verbose = 0;
  }
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  if (argc - optind < 2) {
    usage( argv[0] );
    return 0;
  } else {
    channel = atoi(argv[optind++]);
    filename = string(argv[optind]);
  }

  if (0 == channel) {
    cerr << "Wrong channel number" << endl;
    usage(argv[0]);
    return 0;
  }

  if (timestamp) {
    // compute current time
    time_t now = time(0);
    struct tm* t = localtime(&now);

    // get relevat fields
    int day  = t->tm_mday;
    int mon  = t->tm_mon+1;
    int year = t->tm_year+1900;

    // compute timestamp
    ostringstream o;
    o << year;
    o.width(2);
    o.fill('0');
    o << mon;
    o.width(2);
    o.fill('0');
    o << day;
    o << "-";
    string timestamp(o.str());

    // place timestamp into file
    string::size_type slash_pos = filename.find_last_of('/');
    if (slash_pos != string::npos) {
      //         /home/errror/                     "20030627-" test.mpeg
      filename = filename.substr(0, slash_pos+1) + timestamp + filename.substr(slash_pos+1, string::npos);
    } else {
      filename = o.str() + filename;
    }
  }

  /* the NMM application for the example, intialized below */
  NMMApplication* app = 0;

  /* Allocate _var variables for the INode's*/
  INode_var dvb;
  INode_var writer;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    app = ProxyApplication::getApplication(argc, argv);

    /*Begin> Create the 1. request for the MP3ReadNode*/
    NodeDescription dvb_req;
    dvb_req.setNodeName("DVBReadNode");
    /*End< */

    /*Begin> Create the 2. request for the MPEGAudioDecodeNode*/
    NodeDescription writer_req;
    writer_req.setNodeName("GenericWriteNode");
    /*End< */

    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    registry.initTransaction();

    list<Response> dvb_resp = registry.initRequest(dvb_req);
    list<Response> writer_resp = registry.initRequest(writer_req);


    if (dvb_resp.empty() || writer_resp.empty() ) {
      GLOBAL_ERROR_STREAM("Size of DVBReadNode response " << dvb_resp.size() << endl);
      GLOBAL_ERROR_STREAM("Size of RawWriteNode response " << writer_resp.size() << endl);
      throw Exception("A required node is not available");
    }

    /* Begin> request the nodes */
    dvb.reset(registry.requestNode(dvb_resp.front()));
    writer.reset(registry.requestNode(writer_resp.front()));
    registry.releaseTransaction();
    /* End< */

    dvb->init();

    ISinkFileHandler_var file_sink(writer->getParentObject()->getCheckedInterface<ISinkFileHandler>());
    file_sink->setFilename(filename);
    file_sink->setWorkingDirectory(string("/home/errror/tmp/nmm2/examples/dvb_ts"));
    file_sink->setFileExtension(string("mpeg"));
    writer->init();

    dvb->initOutput();
    connect(dvb, writer);
    dvb->activate();

    writer->initOutput();
    writer->activate();

    dvb->start();
    writer->start();

    ITVCard_var tuner(dvb->getParentObject()->getInterface<ITVCard>());
    tuner->setChannel(channel);

    if (timeout) { // if timeout was given
      sleep(60 * timeout); // sleep timeout minutes and exit then
    } else {
      char key;
      do {
        cin >> key;
      } while (key != 'q'); // else wait for a 'q' on stdin
    }

    // stop nodes, reset and destroy them
    GLOBAL_DEBUG_STREAM("Stopping all Nodes" << endl);
    dvb->stop();
    writer->stop();

    GLOBAL_DEBUG_STREAM("Flushing all Nodes" << endl);
    dvb->flush();
    writer->flush();

    GLOBAL_DEBUG_STREAM("Deactivating all Nodes" << endl);
    dvb->deactivate();
    writer->deactivate();

    GLOBAL_DEBUG_STREAM("Deiniting-Output all Nodes" << endl);
    dvb->deinitOutput();
    writer->deinitOutput();

    GLOBAL_DEBUG_STREAM("Deinitializing all Nodes" << endl);
    dvb->deinit();
    writer->deinit();

    /*Release the reserved nodes*/
    GLOBAL_DEBUG_STREAM("Releasing all Nodes" << endl);
    registry.releaseNode(*dvb);
    registry.releaseNode(*writer);
  }
  // catch NMM-exception and print it
  catch(Exception e) {
    cerr << e << endl;

    /* release the tranaction */
    app->getRegistry().releaseTransaction();

    /*Release the reserved nodes*/
    if (dvb.get())
      app->getRegistry().releaseNode(*dvb);
    if (writer.get())
      app->getRegistry().releaseNode(*writer);

    return 1;
  }
  // any other exception
  catch(...) {
    cerr << "Something went wrong..." << endl;

    /* release the tranaction */
    app->getRegistry().releaseTransaction();

    if (dvb.get())
      app->getRegistry().releaseNode(*dvb);
    if (writer.get())
      app->getRegistry().releaseNode(*writer);

    return 1;
  }

  return 0;
}
