/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <getopt.h>   // for getopt()
#include <unistd.h>   // for open/close/dup/dup2
#include <sys/file.h> // for open/close/dup/dup2
#include <iostream>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"
#include "nmm/interfaces/base/sync/IMultiAudioVideoSynchronizer.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/base/sync/ISinkController.hpp"
#include "nmm/interfaces/device/dvd/IDVD.hpp"
#include "nmm/interfaces/device/dvd/IDVDNav.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/av/mpeg/IMPEGDemux.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/general/progress/IProgress.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

// include the synchronizer
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

using namespace NMM;

/* Global options */
int usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options]" << endl
       << endl
       << "    -d --device          Set the DVD device" << endl
       << "    -a --audiostream     Set the audio stream number" << endl
       << "    -t --title           Select title" << endl
       << "    -c --chapter         Select chapter" << endl
       << "    -f --fullscreen      Starts in fullscreen mode" << endl
       << "    -p --ac3passthrough  Send ac3 data directly to soundcard" << endl
       << "    -b --deinterlace     Deinterlace video frames" << endl
       << "    -v --verbose         increase verbosity (may be given multiple times)" << endl
       << "    -S --no-subtitles    Disable subtitle display" << endl
       << "    -r --remote-source   Use specified remote registry as DVD-Source" << endl
       << "                         registry must be specified with: reghost[:regport]" << endl
       << "    -i --information     Shows number of title, chapters and stream informations" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  return 0;
}


int fderr = -1;
int fd    = -1;

void setStderr( bool on ) {
  if ( !on ) {
    fderr = dup( 2 );
    close( 2 );
    fd = open( "/dev/null", O_WRONLY );
  } else {
    dup2( fderr, fd );
  }
}

bool fullscreen       = false;

/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool finished;
  bool paused;
  string ss_prefix;
  unsigned int ss_counter;

  ThreadMutex finished_mutex;
  ThreadCondition finished_cv;
  
  IVideoDisplay* ivideo_display;
  IMultiAudioVideoSynchronizer* isync;
  IDVDNav* dvdnav_interface;
  IDVD* dvd_interface;
  
public:
  EventCatcher() : finished(false),
                   paused(false),
                   ss_prefix("dvdtest"),
                   ss_counter(0),
                   finished_cv(finished_mutex),
                   ivideo_display(NULL),
                   isync(NULL),
                   dvdnav_interface(NULL),
                   dvd_interface(NULL) {
  }
  
  void setVideoDisplay(IVideoDisplay* value) {
    ivideo_display = value;
  }
  
  void setSynchronizer(IMultiAudioVideoSynchronizer* value) {
    isync = value;
  }
  
  void setDVDNavInterface(IDVDNav* value) {
    dvdnav_interface = value;
  }
  
  void setDVDInterface(IDVD* value) {
    dvd_interface = value;
  }
  
  Result endTrack() {
    signalExit();
    return SUCCESS;
  }
  
  Result setProgress(u_int64_t& actual, u_int64_t& total) {
    int progress = static_cast<int>(actual * 100.0 / static_cast<float>(total));
    // print out progress like " 04% extracted, 2000/50000
    cout << "Progress: ";
    cout << progress << "%";
    cout << "\r"; // no endl, we want to overwrite this again
    cout.flush(); // flush the output to make sure, its printed out.
    return SUCCESS;
  }

  Result keyReleased(string& key) {
    if (key == "KEY_Left") {
      dvdnav_interface->buttonLeft();
    } else if (key == "KEY_Right") {
      dvdnav_interface->buttonRight();
    } else if (key == "KEY_Up") {
      dvdnav_interface->buttonUp();
    } else if (key == "KEY_Down") {
      dvdnav_interface->buttonDown();
    } else if (key == "KEY_Return") {
      dvdnav_interface->activateSelectedButton();
    } else if (key == "KEY_ESC") {
      //dvdnav_interface->goUp();
      dvd_interface->reset();
    } else if (key == "KEY_q") {
      signalExit();
    } else if (key == "KEY_f") {
      fullscreen = (ivideo_display->isFullscreenOnly() || !fullscreen) && ivideo_display->hasFullscreen();
      ivideo_display->setFullscreen(fullscreen);
    } else if (key == "KEY_p") {
      nextChapter();
    } else if (key == "KEY_n") {
      prevChapter();
    } else if (key == "KEY_s") {
      makeScreenshot();
    } else if (key == "KEY_space") {
      paused = !paused;
      if (paused) {
        isync->pause();
      } else {
        isync->wakeup();
      }
    } else {
      cerr << "Unsupported key pressed: " << key << endl;
    }
    
    return SUCCESS;
  }
  
  void prevChapter() {
    int title, chapter, angel;
    angel = 1;
    title = dvd_interface->getCurrentTitle();
    chapter = dvd_interface->getCurrentChapter();
    chapter++;
    dvd_interface->sendFlush();
    dvd_interface->select( title, chapter, angel );
  }
  
  void nextChapter() {
    int title, chapter, angel;
    angel = 1;
    title = dvd_interface->getCurrentTitle();
    chapter = dvd_interface->getCurrentChapter();
    chapter--;
    dvd_interface->sendFlush();
    dvd_interface->select( title, chapter, angel );
  }
  
  void makeScreenshot() {
    if (!ivideo_display) {
      cout << "No VideoDisplay set!" << endl;
      return;
    }

    string filename = ss_prefix + string ("-") + ss_counter + string(".png");
    if (!ivideo_display->makeScreenshot(filename)) {
      cout << "Unable to make Screenshot!" << endl;
      return;
    }
    ss_counter++;
  }
  
  bool isFinished() const { return finished; }
  
  void signalExit() {
    MutexGuard guard(finished_mutex);
    finished = true;
    finished_cv.notifyAll();
  }
  
  void handleEvents() {
    finished_mutex.lock();
    while (!finished) {
      finished_cv.wait();
    }
    finished_mutex.unlock();
  }
};


int main(int argc, char ** argv) {
  bool show_information = false;
  bool ac3_passthrough  = false;
  bool deinterlace      = false;
  bool use_spu          = false;
  bool no_subtitles     = false;
  int audio_track       = 0;
  string device         = "/dev/cdrom";
  int chapter           = 1;
  int title             = 1;
  unsigned int verbose(0);
  bool remote_source      = false;
  string remote_source_host = "localhost";
  int remote_source_port = 0;

  const char *optstring = "d:a:t:c:fpvbSr:ih";
  struct option longopts[] = {
    {"device",         1, NULL, 'd'},
    {"audiostream",    1, NULL, 'a'},
    {"title",          1, NULL, 't'},
    {"chapter",        1, NULL, 'c'},
    {"fullscreen",     0, NULL, 'f'},
    {"ac3passthrough", 0, NULL, 'p'},
    {"deinterlace",    0, NULL, 'b'},
    {"no-subtitles",   0, NULL, 'S'},
    {"remote-source",  1, NULL, 'r'},
    {"verbose",        0, NULL, 'v'},
    {"information",    0, NULL, 'i'},
    {"help",           0, NULL, 'h'},
    {NULL         ,    0, NULL, 0  } /* last row must be zero */
  };

  int ch, optionindex;

  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'd':
      if (optarg) {
	device = optarg;
      }
      break;
    case 'a':
      if (optarg) {
	audio_track = atoi(optarg);
      }
      break;
    case 'p':
      ac3_passthrough = true;
      break;
    case 'r':
      remote_source = true;
      if(optarg) {
        string remote_registry = string(optarg);
        string::size_type colon_pos = remote_registry.find_first_of(':');
        if (colon_pos == string::npos) {
          remote_source_host = remote_registry;
          remote_source_port = ClientRegistry::default_port;
        } else {
          remote_source_host = remote_registry.substr(0, colon_pos);
          remote_source_port = atoi(remote_registry.substr(colon_pos+1, string::npos).c_str());
        }
        if (!remote_source_port) {
          return usage(argv[0]);
        }
      }
      break;
      break;
    case 't':
      if (optarg) {
	title = atoi(optarg);
      }
      break;
    case 'c':
      if (optarg) {
	chapter = atoi(optarg);
      }
      break;
    case 'S':
      no_subtitles = true;
      break;
    case 'f':
      fullscreen = true;
      break;
    case 'b':
      deinterlace = true;
      break;
    case 'i':
      show_information = true;
      break;
    case 'v':
      verbose++;
      break;
    default: //includes 'h'
      return usage(argv[0]);
    }
  }
  
  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
  
  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var dvd;
  INode_var mpg2yuv;
  INode_var deinterlacer;
  INode_var ac3_dec;
  INode_var audio_play;
  INode_var display;
  INode_var demux;
  INode_var overlay;
  INode_var spudec;
  /*End< */
  
  /* Allocate an event listener */
  EventCatcher ec;
  
  try {
    app = ProxyApplication::getApplication(argc, argv);

    /*Begin> Init requests for nodes */
    NodeDescription dvd_request;
    dvd_request.setNodeName("DVDNavReadNode");

    NodeDescription mpg2yuv_request;
    mpg2yuv_request.setNodeName("MPEGVideoDecodeNode");
    
    NodeDescription deinterlacer_request;
    deinterlacer_request.setNodeName("YUVDeInterlaceNode");
    
    NodeDescription ac3_dec_request;
    ac3_dec_request.setNodeName("AC3DecodeNode");
    
    NodeDescription audio_play_request;
    audio_play_request.setNodeName("PlaybackNode");
    
    NodeDescription display_request;
    display_request.setNodeName("XDisplayNode");
    
    NodeDescription demux_request;
    demux_request.setNodeName("MPEGDemuxNode");
    
    NodeDescription overlay_request;
    overlay_request.setNodeName("OverlayNode");
    
    NodeDescription spudec_request;
    spudec_request.setNodeName("SPUDecodeNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    cerr << "Remote-Source: " << remote_source << endl;
    if (remote_source) { 
      cerr << "Adding registry " << remote_source_host << endl;
      registry.addRegistry(TCPAddress(remote_source_host, remote_source_port));
    }

    {
      RegistryLock lock(registry);
      
      list<Response> dvd_response;
      if(remote_source) {
	dvd_request.setLocation(remote_source_host);      
	dvd_response = registry.initRequest(dvd_request);
      }
      else {
	dvd_response = registry.initRequest(dvd_request);
      }
      list<Response> mpg2yuv_response = registry.initRequest(mpg2yuv_request);
      list<Response> deinterlacer_response = registry.initRequest(deinterlacer_request);
      list<Response> ac3_dec_response = registry.initRequest(ac3_dec_request);
      list<Response> audio_play_response = registry.initRequest(audio_play_request);
      list<Response> display_response = registry.initRequest(display_request);
      list<Response> demux_response = registry.initRequest(demux_request);
      list<Response> overlay_response = registry.initRequest(overlay_request);
      list<Response> spudec_response = registry.initRequest(spudec_request);
      
      if (dvd_response.empty() ||
          mpg2yuv_response.empty() ||
          deinterlacer_response.empty() ||
          ac3_dec_response.empty() ||
          audio_play_response.empty() ||
          display_response.empty() ||
          demux_response.empty() ||
          overlay_response.empty() ||
          spudec_response.empty()) {
        GLOBAL_DEBUG_STREAM("Size of dvd_response " << dvd_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of mpg2yuv_response " << mpg2yuv_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of deinterlacer_response " << deinterlacer_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of ac3_dec_response " << ac3_dec_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of audio_play_response " << audio_play_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of display_response " << display_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of demux_response " << demux_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of overlay_response " << overlay_response.size() << endl);
        GLOBAL_DEBUG_STREAM("Size of spudec_response " << spudec_response.size() << endl);
        throw Exception("A required node is not available");
      }
      
      dvd.reset(registry.requestNode(dvd_response.front()));
      mpg2yuv.reset(registry.requestNode(mpg2yuv_response.front()));
      deinterlacer.reset(registry.requestNode(deinterlacer_response.front()));
      ac3_dec.reset(registry.requestNode(ac3_dec_response.front()));
      audio_play.reset(registry.requestNode(audio_play_response.front()));
      display.reset(registry.requestNode(display_response.front()));
      demux.reset(registry.requestNode(demux_response.front()));
      overlay.reset(registry.requestNode(overlay_response.front()));
      spudec.reset(registry.requestNode(spudec_response.front()));
    }
    /*End< */
    
    /*Begin> Initialize DVD */
    IDVD_var dvd_interface(dvd->getParentObject()->getCheckedInterface<IDVD>());
    dvd_interface->setDevice(device);
    ec.setDVDInterface(dvd_interface.get());
    IDVDNav_var dvdnav_interface(dvd->getParentObject()->getInterface<IDVDNav>());
    if (dvdnav_interface.get()) {
      ec.setDVDNavInterface(dvdnav_interface.get());
    }
    
    setStderr( false );
    
    // enable progress informations
    IProgress_var progress(dvd->getParentObject()->getInterface<IProgress>());
    if (progress.get()) {
      progress->sendProgressInformation(true);
    }

    GLOBAL_MESSAGE_STREAM("Open DVD (" << device << "), this can take a while if the DVD is encrypted" << endl);
    dvd->init();
    dvd->initOutput();
    
    setStderr( true );
    /*End< */
    
    /*Begin> Get DVD information */
    int chapters[100];
    int titles = dvd_interface->getTitles();
    if (titles > 100) throw Exception("DVD has too many titles");
    for(int i=1; i<=titles; i++) {
      dvd_interface->select(i,1,1);
      chapters[i] = dvd_interface->getChapters(i);
    }
    if ( title > titles ) {
      GLOBAL_ERROR_STREAM("You selected title " << title << ", but there are only " << titles << " titles on this DVD" << endl);
      throw Exception("Invalid DVD title/chapter");
    }
    if ( chapter > chapters[title] ) {
      GLOBAL_ERROR_STREAM("You selected chapter " << chapter << ", but there are only " <<
                          chapters[title] << " chapters in title " << title << endl);
      throw Exception("Invalid DVD title/chapter");
    }
    /*End< */
    
    /*Begin> Show DVD information */
    if ( show_information ) {
      cout << "DVD-Contents:" << endl;
      cout << "-------------" << endl;
      cout << "There are " << titles << " titles on this DVD." << endl;
      
      for(int i=1; i<=titles; i++) {
        cout << "Title " << i << " has " << chapters[i] << " chapters" << endl;
      }
      
      cout << endl;      
      return 0;
    }
    
    //dvd_interface->select(title, chapter, 1);
    dvd_interface->reset();
    
    /*Begin> Initialize playback node */
    IAudioDevice_var iaudio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
    iaudio_device->setDevice("/dev/dsp");
    audio_play->init();
    /*End< */
    
    /* Initialize display */
    IVideoDisplay_var ivideo_display(display->getParentObject()->getCheckedInterface<IVideoDisplay>());
    ec.setVideoDisplay(ivideo_display.get());
    if (fullscreen && ivideo_display->hasFullscreen()) {
      ivideo_display->setFullscreen(fullscreen);
    }
    display->init();
    /*End< */

    /*Begin> Initialize queue sizes */
    IBiDirQueuedNode_var mpg2yuv_q(mpg2yuv->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var spudec_q(spudec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var ac3_dec_q(ac3_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    
    mpg2yuv_q->setDownstreamMaxSize(100);
    spudec_q->setDownstreamMaxSize(100);
    ac3_dec_q->setDownstreamMaxSize(100);
    /*End< */
    
    /*Begin> Initialize all other nodes */
    mpg2yuv->init();
    demux->init();
    overlay->init();
    spudec->init();
    if (deinterlace) { deinterlacer->init(); }
    if (!ac3_passthrough) { ac3_dec->init(); }
    /*End< */
    
    /*Begin> Init synchronization */
    MultiAudioVideoSynchronizer av_sync;
    IMultiAudioVideoSynchronizer_var sync(av_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());
    ec.setSynchronizer(sync.get());
    
    ISynchronizedSink_var audio_sync(audio_play->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    ISynchronizedSink_var video_sync(display->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    audio_sync->setSynchronized(true);
    video_sync->setSynchronized(true);
    ISinkController_var audio_controller(audio_sync->getController());
    ISinkController_var video_controller(video_sync->getController());
    sync->addAudioSinkController(audio_controller.get());
    sync->addVideoSinkController(video_controller.get());
    /*End< */
    
    /*Begin> Connect nodes */
    string spu_jack_string = string("spu") + 0;
    string audio_jack_string = string("ac3_audio") + audio_track;
    string video_jack_string = string("mpeg_video") + 0;    
    
    connect(dvd, demux);
    
    dvd->activate();
    
    IMPEGDemux_var mpeg_demux(demux->getParentObject()->getCheckedInterface<IMPEGDemux>());
    mpeg_demux->setDVDMode(true);
    
    demux->initOutput();
    
    /*Begin> Check MPEG format */
    Property_var demux_wout_property(demux->getWorkingOutputProperty());
    
    const Format* mpeg_video_format = demux_wout_property->getDefaultFormat(video_jack_string);
    if ( mpeg_video_format != NULL && mpeg_video_format->getFormatType() == "video/mpeg" ) {
      cout << "Found MPEG video stream version " << mpeg_video_format->getIntValue("version") << endl;
    } else {
      throw Exception("No video stream found");
    }
    
    const Format* ac3_audio_format = demux_wout_property->getDefaultFormat(audio_jack_string);
    if ( ac3_audio_format != NULL && ac3_audio_format->getFormatType() == "audio/ac3" ) {
      cout << "Found AC3 audio stream" << endl;
    } else {
      throw Exception("No audio stream found");
    }
    
    if (!no_subtitles) {
      const Format* spu_format = demux_wout_property->getDefaultFormat(spu_jack_string);
      if ( spu_format != NULL && spu_format->getFormatType() == "video/spu" ) {
        cout << "Found subtitle picture unit" << endl;
        use_spu = true;
      } else {
        throw Exception("No subtitle stream found");
      }
    }
    /*End< Check MPEG format*/
    
    connect(demux, video_jack_string, mpg2yuv);
    if (use_spu) {
      connect(demux, spu_jack_string, spudec); 
    }
    if (!ac3_passthrough) {
      connect(demux, audio_jack_string, ac3_dec);
    } else {
      connect(demux, audio_jack_string, audio_play);
    }
    
    demux->activate();
    if (!ac3_passthrough) {
      ac3_dec->initOutput();
      connect(ac3_dec, audio_play);
      
      ac3_dec->activate();
    }
    
    audio_play->initOutput();
    audio_play->activate();
    
    mpg2yuv->initOutput();
    
    /*Begin> Check video format */
    Property_var mpg2yuv_wout_property(mpg2yuv->getWorkingOutputProperty());
    
    const Format* raw_video_format = mpg2yuv_wout_property->getDefaultFormat();
    if ( raw_video_format != NULL ) {
      cout << "video format: " << endl;
      int x_resolution = raw_video_format->getIntValue(Format::x_resolution_param);
      int y_resolution = raw_video_format->getIntValue(Format::y_resolution_param);
      int ratio_x = raw_video_format->getIntValue(Format::ratio_x_param);
      int ratio_y = raw_video_format->getIntValue(Format::ratio_y_param);
      float framerate = raw_video_format->getFloatValue(Format::framerate_param);
      cout << "resolution: " << x_resolution << "x" << y_resolution << endl;
      cout << "Ratio     : " << ratio_x << ":" << ratio_y << endl;
      cout << "Framerate : " << framerate << " fps" << endl;
    }
    /*End< Check video format */
    
    /* connect SPU graph */
    if (use_spu) {
      spudec->initOutput();
      connect(spudec, overlay, string("foreground"));
      
      spudec->activate();
    }
    
    /* connect video graph */
    INode* video_buffer_source = mpg2yuv.get();
    INode* video_source = mpg2yuv.get();
    if (use_spu) {
      connect(video_source, overlay.get(), string("background"));
      video_source->activate();
      overlay->initOutput();
      video_source = overlay.get();
    }
    if (deinterlace) {
      connect(video_source, deinterlacer.get());
      video_source->activate();
      deinterlacer->initOutput();
      video_source = deinterlacer.get();
    }
    connect(video_source, display.get());
    video_source->activate();
    display->initOutput();
    display->activate();
    /*End< Connect nodes */
    
    /* Use shared buffer manager */
    {
      IBufferManager_var bm1(display->getParentObject()->getCheckedInterface<IBufferManager>());
      IBufferManager_var bm2(video_buffer_source->getParentObject()->getCheckedInterface<IBufferManager>());
      bm2->useBufferManager(*bm1.get());
    }
    
    /*Begin> Start nodes */
    dvd->start();
    demux->start();
    mpg2yuv->start();
    if (use_spu) {
      spudec->start();
      overlay->start();
    }
    if (deinterlace) { deinterlacer->start(); }
    if (!ac3_passthrough) { ac3_dec->start(); }
    audio_play->start();
    display->start();
    /*End< */
    
    /* Begin> Register event handlers */
    // EDObject for progress receiving
    TEDObject2<EventCatcher, u_int64_t, u_int64_t> progress_listener(&ec, &EventCatcher::setProgress);
    display->getParentObject()->registerEventListener(IProgressListener::setProgress_event,
                                                      &progress_listener);
    /* End< Register event handlers */

    /*Begin> Event loop */
    IDisplayEventSender_var display_event_sender(display->getParentObject()->getInterface<IDisplayEventSender>());
    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->enableDisplayEvents();
      display->getParentObject()->
        registerEventListener(IDisplayEvents::keyReleased_event,
                              new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    }
    ec.handleEvents();
    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->disableDisplayEvents();
    }
    /*End< */
    
    /* Begin> Deregister event handlers */
    display->getParentObject()->removeEventListener(&progress_listener);
    /* End< Deregister event handlers */

    /*Begin> Stop nodes */
    if (dvd->isStarted()) {
      dvd->stop();
    }
    if (demux->isStarted()) {
      demux->stop();
    }
    if (mpg2yuv->isStarted()) {
      mpg2yuv->stop();
    }
    if (spudec->isStarted()) {
      spudec->stop();
    }
    if (overlay->isStarted()) {
      overlay->stop();
    }
    if (deinterlacer->isStarted()) {
      deinterlacer->stop();
    }
    if (ac3_dec->isStarted()) {
      ac3_dec->stop();
    }
    if (audio_play->isStarted()) {
      audio_play->stop();
    }
    if (display->isStarted()) {
      display->stop();
    }
    /*End< */
    
    /*Begin> Flush nodes */
    if (dvd->isActivated()) {
      dvd->flush();
    }
    if (demux->isActivated()) {
      demux->flush();
    }
    if (mpg2yuv->isActivated()) {
      mpg2yuv->flush();
    }
    if (spudec->isActivated()) {
      spudec->flush();
    }
    if (overlay->isActivated()) {
      overlay->flush();
    }
    if (deinterlacer->isActivated()) {
      deinterlacer->flush();
    }
    if (ac3_dec->isActivated()) {
      ac3_dec->flush();
    }
    if (audio_play->isActivated()) {
      audio_play->flush();
    }
    if (display->isActivated()) {
      display->flush();
    }
    /*End< */
    
    /*Begin> Deactivate nodes */
    if (dvd->isActivated()) {
      dvd->deactivate();
    }
    if (demux->isActivated()) {
      demux->deactivate();
    }
    if (mpg2yuv->isActivated()) {
      mpg2yuv->deactivate();
    }
    if (spudec->isActivated()) {
      spudec->deactivate();
    }
    if (overlay->isActivated()) {
      overlay->deactivate();
    }
    if (deinterlacer->isActivated()) {
      deinterlacer->deactivate();
    }
    if (ac3_dec->isActivated()) {
      ac3_dec->deactivate();
    }
    if (audio_play->isActivated()) {
      audio_play->deactivate();
    }
    if (display->isActivated()) {
      display->deactivate();
    }
    /*End< */
    
    /*Begin> DeinitOutput nodes */
    if (dvd->isOutputInitialized()) {
      dvd->deinitOutput();
    }
    if (demux->isOutputInitialized()) {
      demux->deinitOutput();
    }
    if (mpg2yuv->isOutputInitialized()) {
      mpg2yuv->deinitOutput();
    }
    if (spudec->isOutputInitialized()) {
      spudec->deinitOutput();
    }
    if (overlay->isOutputInitialized()) {
      overlay->deinitOutput();
    }
    if (deinterlacer->isOutputInitialized()) {
      deinterlacer->deinitOutput();
    }
    if (ac3_dec->isOutputInitialized()) {
      ac3_dec->deinitOutput();
    }
    if (audio_play->isOutputInitialized()) {
      audio_play->deinitOutput();
    }
    if (display->isOutputInitialized()) {
      display->deinitOutput();
    }
    /*End< */
    
    /*Begin> Deinit nodes */
    if (dvd->isInitialized()) {
      dvd->deinit();
    }
    if (demux->isInitialized()) {
      demux->deinit();
    }
    if (mpg2yuv->isInitialized()) {
      mpg2yuv->deinit();
    }
    if (spudec->isInitialized()) {
      spudec->deinit();
    }
    if (overlay->isInitialized()) {
      overlay->deinit();
    }
    if (deinterlacer->isInitialized()) {
      deinterlacer->deinit();
    }
    if (ac3_dec->isInitialized()) {
      ac3_dec->deinit();
    }
    if (audio_play->isInitialized()) {
      audio_play->deinit();
    }
    if (display->isInitialized()) {
      display->deinit();
    }
    /*End< */
    
    /*Begin> Release nodes */
    registry.releaseNode(*dvd);
    registry.releaseNode(*demux);
    registry.releaseNode(*mpg2yuv);
    registry.releaseNode(*spudec);
    registry.releaseNode(*overlay);
    registry.releaseNode(*deinterlacer);
    registry.releaseNode(*ac3_dec);
    registry.releaseNode(*audio_play);
    registry.releaseNode(*display);
    /*End< */
  } catch(Exception e) {
    /*Begin> Release nodes */
    if (app != NULL) {
      if (dvd.get()) {
        app->getRegistry().releaseNode(*dvd);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (mpg2yuv.get()) {
        app->getRegistry().releaseNode(*mpg2yuv);
      }
      if (spudec.get()) {
        app->getRegistry().releaseNode(*spudec);
      }
      if (overlay.get()) {
        app->getRegistry().releaseNode(*overlay);
      }
      if (deinterlacer.get()) {
        app->getRegistry().releaseNode(*deinterlacer);
      }
      if (ac3_dec.get()) {
        app->getRegistry().releaseNode(*ac3_dec);
      }
      if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
      }
      if (display.get()) {
        app->getRegistry().releaseNode(*display);
      }
      /*End< */
    }
    
    cerr << e << endl;
    return 1;
  } catch(...) {
    /*Begin> Release nodes */
    if (app != NULL) {
      if (dvd.get()) {
        app->getRegistry().releaseNode(*dvd);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (mpg2yuv.get()) {
        app->getRegistry().releaseNode(*mpg2yuv);
      }
      if (spudec.get()) {
        app->getRegistry().releaseNode(*spudec);
      }
      if (overlay.get()) {
        app->getRegistry().releaseNode(*overlay);
      }
      if (deinterlacer.get()) {
        app->getRegistry().releaseNode(*deinterlacer);
      }
      if (ac3_dec.get()) {
        app->getRegistry().releaseNode(*ac3_dec);
      }
      if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
      }
      if (display.get()) {
        app->getRegistry().releaseNode(*display);
      }
    }
    /*End< */
    
    cerr << "Something went wrong..." << endl;
    return 1;
  }
  
  return 0;
}
