/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"
#include "nmm/base/graph/GraphBuilder2.hpp"
#include <iostream>

using namespace NMM;

int main(int argc, char * argv[]) 
{
  if(argc < 4) {
    cerr << "Usage:" << endl
	 << "  " << argv[0] << " <url> <location of audio sink> <location of video sink>" << endl
	 << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see http://graphics.cs.uni-sb.de/NMM/current/Docs/messages/index.html
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the NMM application for the example
  NMMApplication* app = 0;
  // the composite node that holds the flow graph created by the graph builder
  CompositeNode* composite = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create node descriptions for possible sink nodes for audio and video
    NodeDescription playback_nd("PlaybackNode");
    NodeDescription display_nd("XDisplayNode");
    
    // set location for audio sink to argv[2] and for video sink to argv[3]
    playback_nd.setLocation(argv[2]);
    display_nd.setLocation(argv[3]);

    // create synchronizer for audio/video
    MultiAudioVideoSynchronizer av_sync;
    IMultiAudioVideoSynchronizer_var sync(av_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());

    // setup graph builder
    GraphBuilder2 gb;
    if(!gb.setURL(argv[1])) {
      throw Exception("Invalid URL given");
    }
    gb.setMultiAudioVideoSynchronizer(sync.get());
    gb.setAudioSink(playback_nd);
    gb.setVideoSink(display_nd);

    // create flow graph for given URL
    composite = gb.createGraph(*app);

    // start all internal nodes of composite node
    composite->reachStarted();

    // wait ...
    cerr << "Playing ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
  } 
  catch (const Exception& e) {
    cerr << e << endl;
  }
  catch(...) {
    cerr << "something went wrong..." << endl;
  }
  
  // release composite node
  if(composite) {
    composite->reachActivated();
    composite->flush();
    composite->reachConstructed();
    delete composite;
  }

  return 0;
}

