/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

// simple mp3 player using GraphDescription with error handling and access to 'serverregistry'
// same code as helloworld11, but added listener for endTrack event

#include <iostream>

// include NMM 
#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"

// include used interfaces
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"

// import namespace
using namespace NMM;

// a simple listener class
class Listener2 {
public:
  Result endTrack() 
  {
    cerr << "endTrack" << endl;
    return SUCCESS;
  }
};

int main(int argc, char ** argv)
{
  if(argc != 2) {
    cerr << "Usage:" << endl
	 << "  " << argv[0] << " <mp3file>" << endl
	 << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see http://graphics.cs.uni-sb.de/NMM/current/Docs/messages/index.html
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create the descriptions for the GenericReadNode, MPEGAudioDecodeNode, and PlaybackNode
    NodeDescription readfile_descr("GenericReadNode");
    NodeDescription decoder_descr("MPEGAudioDecodeNode");
    NodeDescription audioplay_descr("PlaybackNode");

    // create a flow graph: readfile -> decoder -> audioplay
    graph.addEdges(&readfile_descr, &decoder_descr, &audioplay_descr);

    // request complete graph
    cerr << "request GraphDescription from registry" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.requestGraph(graph);

    // set the filename by requesting appropriate interface
    cerr << "set filename to " << argv[1] << endl;
    INode* readfile = graph.getINode(readfile_descr);
    IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler->setFilename(argv[1]);

    // register listener for endTrack event of ITrack interface at audio sink node
    // (see nmm/interfaces/file/ITrack.idl for more information)
    // -> the method Listener2::endTrack will get triggered when the stream ended at audio sink
    INode* audioplay = graph.getINode(audioplay_descr);
    Listener2 end_track_listener;
    audioplay->getParentObject()->registerEventListener(ITrack::endTrack_event, 
							new TEDObject0<Listener2>(&end_track_listener, &Listener2::endTrack));

    // realize and start graph
    cerr << "realizeGraph()" << endl;
    graph.realizeGraph();
    cerr << "startGraph()" << endl;
    graph.startGraph();
    
    // wait ...
    cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    
    // stop and release graph
    cerr << "stopGraph()" << endl;
    graph.stopGraph();

    cerr << "releaseGraph()" << endl;
    registry.releaseGraph(graph);
    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    return 1;
  }
}
