/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyApplication.hpp"

#include "nmm/plugins/audio/PlaybackNode.hpp"
#include "nmm/plugins/audio/oggvorbis/OggVorbisDecodeNode.hpp"
#include "nmm/plugins/file/GenericReadNode.hpp"

#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/audio/oggvorbis/IOggVorbis.hpp"

#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyObject.hpp"

#include <getopt.h>
#include <iostream>

using namespace NMM;

void usage(void) {
  cerr << "USAGE:" << endl;
  cerr << "  oggvorbisplay [options] <filename>" << endl;
  cerr << endl;
  cerr << "OPTIONS:" << endl;
  cerr << " -d --demux    : ude ogg demux." << endl;
  cerr << " -v --verbose  : enable debug output." << endl;
  cerr << " -h --help     : show this help and exits." << endl;
  exit(1);
}

int main(int argc, char* argv[]) {
  /* command line parameter - see getopt(3) */
  const char *opt_string = "dvh";
  struct option options [] = {
        {"demux",0,NULL,'d'},
        {"verbose",0,NULL,'v'},
        {"help",0,NULL,'h'},
        {NULL,0,NULL,0}
  };
  int ch,long_option_index;

  string filename;
  bool verbose = false;
  bool use_demux = false;

  while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
    // which option did we get?
    switch (ch) {
    case 'd':
      use_demux = true;
      break;
    case 'v':
      verbose = true;
      break;
    case 'h':
      usage();
      break;
    default:
      usage();
      break;
    }
  }

  if ((argc - optind) != 1) {
    usage();
  } else {
    filename = string(argv[optind]);
  }

  if (! verbose) {
    NamedObject::getGlobalInstance().setErrorStream(NULL, NamedObject::ALL_LEVELS);
    NamedObject::getGlobalInstance().setDebugStream(NULL, NamedObject::ALL_LEVELS);
    NamedObject::getGlobalInstance().setWarningStream(NULL, NamedObject::ALL_LEVELS);
    NamedObject::getGlobalInstance().setMessageStream(NULL, NamedObject::ALL_LEVELS);
  }

  NMMApplication& app = *ProxyApplication::getApplication(argc, argv);
  list<INode*> node_list;

  INode_var read;
  INode_var demux;
  INode_var decode;
  INode_var play;
  
  try {
    
    NodeDescription read_request;
    read_request.setNodeName("GenericReadNode");
    
    NodeDescription demux_request;
    demux_request.setNodeName("OGMDemuxNode");
    
    NodeDescription decode_request;
    decode_request.setNodeName("OggVorbisDecodeNode");
    
    NodeDescription play_request;
    play_request.setNodeName("PlaybackNode");
    
    ClientRegistry& registry = app.getRegistry();
    
    list<Response> read_response = registry.initRequest(read_request);
    list<Response> demux_response = registry.initRequest(demux_request);
    list<Response> decode_response = registry.initRequest(decode_request);
    list<Response> play_response = registry.initRequest(play_request);
    
    if (read_response.empty() || (use_demux && demux_response.empty()) || decode_response.empty() || play_response.empty() ) {
      GLOBAL_ERROR_STREAM("Size of read_response " << read_response.size() << endl);
      GLOBAL_ERROR_STREAM("Size of demux_response " << demux_response.size() << endl);
      GLOBAL_ERROR_STREAM("Size of decode_response " << decode_response.size() << endl);
      GLOBAL_ERROR_STREAM("Size of play_response " << play_response.size() << endl);
      return 0;
    }
    
    read.reset(registry.requestNode(read_response.front()));
    node_list.push_back(read.get());
    
    if (use_demux) {
      demux.reset(registry.requestNode(demux_response.front()));
      node_list.push_back(demux.get());
    }
    
    decode.reset(registry.requestNode(decode_response.front()));
    node_list.push_back(decode.get());
    
    play.reset(registry.requestNode(play_response.front()));
    node_list.push_back(play.get());
    
    IFileHandler_var iread (read->getParentObject()->getCheckedInterface<IFileHandler>());
    IOggVorbis_var idecode (decode->getParentObject()->getCheckedInterface<IOggVorbis>());
    IAudioDevice_var iplay(play->getParentObject()->getCheckedInterface<IAudioDevice>());
  
    iread->setFilename(filename);
    //iplay->setDevice("dev/dsp");
  
    read->init();
    if (use_demux) {
      demux->init();
    }
    decode->init();
    play->init();
    
    read->initOutput();
    if (use_demux) {
      connect(read,demux);
      read->activate();
      
      demux->initOutput();
      connect(demux, string("audio0"), decode);
      demux->activate();
    } else {
      connect(read,decode);
      read->activate();
    }
    
    decode->initOutput();
    connect(decode,play);
    decode->activate();
  
    play->initOutput();
    play->activate();
  
  
    /* -- Start --- */
    read->start();
    GLOBAL_DEBUG_STREAM( "Playback Node started" << endl);
    if (use_demux) {
      demux->start();
      GLOBAL_DEBUG_STREAM( "Demux Node started" << endl);
    }
    decode->start();
    GLOBAL_DEBUG_STREAM( "OggVorbis Node started" << endl);
    play->start();
    GLOBAL_DEBUG_STREAM( "Read Node started" << endl);


    /* --- dance --- */
    bool finished = false;
    char buffer = 0;
    while (!finished) {
      cin >> buffer;
	
      switch (buffer) {
      case 'i':
        {cout<<"UserComments: "<<endl<<idecode-> getUserComments()<<endl;
        cout<<"Vendor: "<<endl<<idecode->getVendor()<<endl;
        }
        break;
      case 'q':
        finished = true;
        break;
      default:
        break;
      }
    }
    GLOBAL_DEBUG_STREAM("Stopping nodes"<<endl);
    read->stop();
    if (use_demux) {
      demux->stop();
    }
    decode->stop();
    play->stop();
    GLOBAL_DEBUG_STREAM("nodes stopped"<<endl);  

    GLOBAL_DEBUG_STREAM("Flushing nodes"<<endl);
    read->flush();
    if (use_demux) {
      demux->flush();
    }
    decode->flush();
    play->flush();
    GLOBAL_DEBUG_STREAM("nodes flushing done"<<endl);
 
    GLOBAL_DEBUG_STREAM("deactivating nodes"<<endl);
    read->deactivate();
    if (use_demux) {
      demux->deactivate();
    }
    decode->deactivate();
    play->deactivate();
    GLOBAL_DEBUG_STREAM("nodes deactivated"<<endl);

    GLOBAL_DEBUG_STREAM("Deinitializing Output"<<endl);
    read->deinitOutput();
    if (use_demux) {
      demux->deinitOutput();
    }
    decode->deinitOutput();
    play->deinitOutput();
    GLOBAL_DEBUG_STREAM("Output Deinitialized"<<endl);

    GLOBAL_DEBUG_STREAM("Deinitializing nodes"<<endl);
    read->deinit();
    if (use_demux) {
      demux->deinit();
    }
    decode->deinit();
    play->deinit();
    GLOBAL_DEBUG_STREAM("nodes deinitialized"<<endl);

    for (list<INode*>::iterator it = node_list.begin(); it != node_list.end(); ++it)
      registry.releaseNode(**it);
    return 0;
    
  } catch( const Exception& e ) {
    
    cerr << e << endl;
    return 1;
    
  } catch( ... ) {
    
    cerr << "..." << endl;
    return 1;
    
  }
}






