/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
//#include <cstdio>
//#include <time.h>
#include <unistd.h>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"

#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"


using namespace NMM;

/* Listener class, catch events from the nodes */
class Listener2 {
private:
  bool finished;
public:
  Listener2() { finished = false; }
  virtual Result receiveEvent(const Event&) { return SUCCESS; }  
  Result setfinishedflag() {finished = true; return SUCCESS; }
  bool getfinishedflag() { return finished;}
};


int main(int argc, char *argv[]) {
  Listener2* l = new Listener2();
  char *source;
  float quality;
  NamedObject::getGlobalInstance().setErrorStream(NULL, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(NULL, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(NULL, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(NULL, NamedObject::ALL_LEVELS);

  NMMApplication* app = ProxyApplication::getApplication(argc, argv);

  if (argc < 2) {
    GLOBAL_ERROR_STREAM("usage:  <file.wav>  [quality]"<< endl);
    GLOBAL_ERROR_STREAM("quality from -1 (low quality) to 1 (high quality"<<endl);
    return 1;
  } else {
    source = argv[1];
  }
  if (argv[2] != NULL) {
    quality = atof(argv[2]);
    if (quality<-1) quality=-1;
    else if (quality > 1) quality=1;
  } else quality=0.4;
  
  INode_var read,encode,write;
  try{
    ClientRegistry& registry = app->getRegistry();
    
    NodeDescription request1;
    request1.setNodeName("WavReadNode");
    NodeDescription request2;
    request2.setNodeName("VorbisEncodeNode");
    NodeDescription request3;
    request3.setNodeName("GenericWriteNode");
    
    list<Response> response1 = registry.initRequest(request1);
    list<Response> response2 = registry.initRequest(request2);
    list<Response> response3 = registry.initRequest(request3);
    
    if (response1.empty() || response2.empty() || response3.empty() ) {
      GLOBAL_ERROR_STREAM("Size of 1. response " << response1.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 2. response " << response2.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 3. response " << response3.size() << endl);
      throw Exception ("A required node is not available");
    }
   
    read.reset(registry.requestNode(response1.front()));
    encode.reset(registry.requestNode(response2.front()));
    write.reset(registry.requestNode(response3.front()));
    
    IFileHandler_var iread(read->getParentObject()->getInterface<IFileHandler>());
    ISinkFileHandler_var iwrite(write->getParentObject()->getInterface<ISinkFileHandler>());
    
    /* -- setting Parameters in the Nodes using the Interfaces -- */
    iread->setFilename(source);
    iwrite->setFileExtension("ogg");

    /* -- registering ITrack::finished_event -- */
    write->getParentObject()->registerEventListener(ITrack::finished_event, new TEDObject0<Listener2>(l, &Listener2::setfinishedflag));

    /* -- try to init and activate the nodes -- */
    try {
      read->init();
      encode->init();
      write->init();
      
      read->initOutput();
      connect(read,encode);
      read->activate();
      
      encode->initOutput();

      const Format* enc_format = encode->getWorkingOutputProperty()->getDefaultFormat();
      Format* encode_format = new Format(Format::audio_x_ogg_type);
      encode_format->addIntParamValue(Format::audio_codec_param, enc_format->getIntValue(Format::audio_codec_param));
      encode_format->addIntParamValue(Format::sample_rate_param, enc_format->getIntValue(Format::sample_rate_param));
      encode_format->addIntParamValue(Format::channels_param, enc_format->getIntValue(Format::channels_param));
      encode_format->addFloatParamValue(Format::quality_param, quality);
      
      connect(encode,write,encode_format);   
      encode->activate();
      
      write->initOutput();
      write->activate();
    }
    catch(...) {
      GLOBAL_ERROR_STREAM("FAILED to init/activate the needed notes - I'm sorry"<<endl);
      return 1;
    }
    
    /* -- Starting Nodes -- */
    try {
      write->start();
      encode->start();
      read->start();
    } 
    catch (...) {
      GLOBAL_ERROR_STREAM("FAILED to start the nodes" << endl);
      return 1;
    }
    
    /* -- doint something useful -- */
    while (!(l->getfinishedflag())) {sleep(1);}
    
    GLOBAL_DEBUG_STREAM("Stopping nodes"<<endl);
    
    /* -- Stopping Nodes -- */
    try {
      read->stop();
      encode->stop();
      write->stop();
    } 
    catch (...) {
      GLOBAL_ERROR_STREAM("FAILED to stop the nodes" << endl);
      return 1;
    }
    
    /* -- Deactivate/Deinit Nodes -- */
    try {
      
      read->deactivate();
      encode->deactivate();
      write->deactivate();
      read->deinitOutput();
      encode->deinitOutput();
      write->deinitOutput();
      
      read->deinit();
      encode->deinit();
      write->deinit();
    } 
    catch (...) {
      GLOBAL_ERROR_STREAM("FAILED to deactivate/deinit the nodes" << endl);
      return 1;
    }
    registry.releaseNode(*read);
    registry.releaseNode(*encode);
    registry.releaseNode(*write);
    GLOBAL_DEBUG_STREAM("\033[0;31m" << "We did it! ... Enjoy your OGG-File" << "\033[0m" << endl); 
   
  }catch(...) {
    if (read.get())
      app->getRegistry().releaseNode(*read);
    if (encode.get())
      app->getRegistry().releaseNode(*encode);
    if (write.get())
      app->getRegistry().releaseNode(*write);
    cerr<<"So sorry - but something went wront..."<<endl;
    return 1; }
  return 0;
}









