/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef __NMM_COMCHANNEL_HPP
#define __NMM_COMCHANNEL_HPP

#include "nmm/base/serialize/ComplexValue.hpp"
#include "TSAddress.hpp"
#include "Interface.hpp"

#include <algorithm>
using std::find;

#include <memory>
using std::auto_ptr;

namespace NMM {

  /** This class is base class for all CommunicationChannels. A CommunicationChannel
   * represents a connection between two objects and is a first class object in the RM-ODP
   * manner. The application can configure the TransportStrategy that should be used and
   * access their interfaces.
   */
  class CommunicationChannel : public SerializedValue {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    CommunicationChannel(const TypeInfo& ti = getStaticTypeInfo()) : SerializedValue(ti) {}

    /** Destructor
     */
    virtual ~CommunicationChannel() {}

    /** virtual copy Constructor
     */
    virtual CommunicationChannel* copy()                  const { return 0;}
    //@}

    /** @name Configure the Communication Channel
     */
    //@{
    /** This method sets the TransportStrategy that should be used
     */
    virtual      void  setTransportStrategy(const string&) {};

    /** This method sets the ValueStream in the TransportStrategy that should be used
     */
    virtual      void  setValueStream(const IValueStream&, const OValueStream& ) {};

    /** This method returns a list of all TransportStrategies availble for the connection
     */
    virtual const list<string>& getStrategyList()          const { return __strategy_list; }

    /** This method returns a list of all TransportStrategies availble for the connection
     */
    virtual       list<string>& getStrategyList()                { return __strategy_list; }
    //@}

    /** @name Access the TransportStrategy Interfaces
     */
    //@{
    /** This template method returns a pointer to the Interface of type T or NULL if this interface
     * is not supported by this object. The interface is from the ProxyObjects or the InputJack.
     * The use of this method is:
     * \code
     * CommunicationChannel channel;
     * const ITCPControl* i_tcp = channe.getInterface<ITCPControl>();
     * if (i_tcp)
     *   cout << "ITCPControl Interface supported" << endl;
     * else
     *   cout << "ITCPControl Interface not supported" << endl;
     * \endcode
     */
    template<class T>
    T* getInterface();

    /** This template method returns a pointer to the Interface of type T or NULL if this interface
     * is not supported by this object. The interface is from the ProxyObjects or the InputJack.
     * The use of this method is:
     * \code
     * CommunicationChannel channel;
     * const ITCPControl* i_tcp = channe.getInterface<ITCPControl>();
     * if (i_tcp)
     *   cout << "ITCPControl Interface supported" << endl;
     * else
     *   cout << "ITCPControl Interface not supported" << endl;
     * \endcode
     */
    template<class T>
    const T* getInterface()          const;

    /** This template method returns a pointer to the Interface of type T or NULL if this interface
     * is not supported by this object. The interface is from the NMMObjects or the OutputJack.
     * The use of this method is:
     * \code
     * CommunicationChannel channel;
     * const ITCPControl* i_tcp = channe.getInterface<ITCPControl>();
     * if (i_tcp)
     *   cout << "ITCPControl Interface supported" << endl;
     * else
     *   cout << "ITCPControl Interface not supported" << endl;
     * \endcode
     */
    template<class T>
    T* getPartnerInterface();

    /** This template method returns a pointer to the Interface of type T or NULL if this interface
     * is not supported by this object. The interface is from the NMMObjects or the OutputJack.
     * The use of this method is:
     * \code
     * CommunicationChannel channel;
     * const ITCPControl* i_tcp = channe.getInterface<ITCPControl>();
     * if (i_tcp)
     *   cout << "ITCPControl Interface supported" << endl;
     * else
     *   cout << "ITCPControl Interface not supported" << endl;
     * \endcode
     */
    template<class T>
    const T* getPartnerInterface()   const;

    /** This method returns a list of all Interfaces from the TransportStrategy
     */
    virtual list<Interface*>& getInterfaceList()                { return __interface_list; }

    /** This method returns a list of all Interfaces from the TransportStrategy of the Partner Strategy.
     */
    virtual list<Interface*>& getPartnerInterfaceList()         { return __partner_list; }
    //@}

    /** @name Obsolete method
     */
    //@{
    virtual void deinit() {};
    //@}
  protected:
    /** @name Access the TransportStrategy Interfaces
     */
    //@{
    /** This methods adds a new TransportStrategy
     */
    inline void  addStrategyType(const string& type)            { if (find (__strategy_list.begin(), __strategy_list.end(), type) == __strategy_list.end()) __strategy_list.push_back(type); }

    /** This methods sets new TransportStrategies
     */
    inline void  setStrategyList(const list<string>& type)      { __strategy_list = type; }

    /** This method adds an interface to the interface list
     */
    inline void  addInterface(Interface* interface)             { __interface_list.push_back(interface); }

    /** This method adds an interface to the partner interface list
     */
    inline void  addPartnerInterface(Interface* interface)      { __partner_list.push_back(interface); }

    /** This method must be called after the staregy list has been created
     * @param binding_type The binding type for this channel
     */
    void finishStrategyList(TSAddress::BindingType binding_type);
    //@}
  private:
    list<string>        __strategy_list;                  ///< stores the TransportStrategeis available for the connection
    list<Interface*>    __interface_list;                 ///< stores all interfaces of the TransportStrategy
    list<Interface*>    __partner_list;                   ///< stores all interfaces of the partner TransportStrategy

  public:
    /** This method returns the TypeInfo for this class.
     */
    static TypeInfo getStaticTypeInfo();
  };




  template <typename T>
  T* CommunicationChannel::getInterface() {
    T* interface = 0;
    for (list<Interface*>::iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      {
        interface = dynamic_cast<T*>(*it);
        if (interface != 0)
          return dynamic_cast<T* >(interface->copy());
      }
    return 0;
  }


  template <typename T>
  const T* CommunicationChannel::getInterface() const {
    const T* interface = 0;
    for (list<Interface*>::const_iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      {
        interface = dynamic_cast<const T*>(*it);
        if (interface != 0)
          return dynamic_cast<T* >(interface->copy());
      }
    return 0;
  }


  template <typename T>
  T* CommunicationChannel::getPartnerInterface() {
    T* interface = 0;
    for (list<Interface*>::iterator it = getPartnerInterfaceList().begin(); it != getPartnerInterfaceList().end(); ++it)
      {
        interface = dynamic_cast<T*>(*it);
        if (interface != 0)
          return dynamic_cast<T* >(interface->copy());
      }
    return 0;
  }


  template <typename T>
  const T* CommunicationChannel::getPartnerInterface() const {
    const T* interface = 0;
    for (list<Interface*>::const_iterator it = getPartnerInterfaceList().begin(); it != getPartnerInterfaceList().end(); ++it)
      {
        interface = dynamic_cast<const T*>(*it);
        if (interface != 0)
          return dynamic_cast<T* >(interface->copy());
      }
    return 0;
  }

  typedef auto_ptr<CommunicationChannel> CommunicationChannel_var;

}

#endif


