/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_CONNECTOR_ADDRESS_HPP_
#define _NMM_CONNECTOR_ADDRESS_HPP_

#include "TSAddress.hpp"

namespace NMM {
  
  class IConnectionManager;
  
  /**
   *  @brief Connector address
   *
   *  This address is the default address of a %ConnectorAcceptor. It refers to a remote %IConnectionManager interface
   *  which is used for sending connection requests. The address can be used with any type of transport strategy.
   *  There is no default transport strategy type.
   *
   *  A %ConnectorAddress also contains the NMM reference of the remote transport strategy. This field is used during
   *  address exchange when a connection is initiated through the %ConnectionManager. It is needed for sending reliable
   *  configuration events between the partner transport strategies.
   *
   *  This class should be the base class of all %TSAddress classes of transport strategies that support connection
   *  setup through the %ConnectionManager.
   *
   *  @see ConnectorAcceptor
   *  @see IConnectionManager
   */
  class ConnectorAddress
    : public TSAddress
  {
  public:
    /**
     *  @name Constructor & Destructor
     */
    //@{
    /**
     *  @brief Construct a %ConnectorAddress with default values
     *
     *  This default constructor is only needed for deserialization and should not be used.
     */
    ConnectorAddress();
    
    /**
     *  @brief Construct a %ConnectorAddress
     *  @param strategy_type Type of transport strategy
     *  @param ref NMM Reference of addressed object
     */
    ConnectorAddress(const string& strategy_type, const NMMReference& ref = 0);
    
    /**
     *  @brief Copy constructor
     */
    ConnectorAddress(const ConnectorAddress& other);
    
    /**
     *  @brief Virtual destructor
     */
    ~ConnectorAddress();
    
    /**
     *  @brief Virtual copy constructor
     */
    virtual ConnectorAddress* copy() const;
    
    /**
     *  @brief Assignment operator
     */
    ConnectorAddress& operator=(const ConnectorAddress& other);
    //@}
    
    
    /**
     *  @name Accessors
     */
    //@{
    /**
     *  @brief Set connection manager interface
     *  @param icm Connection manager interface
     */
    void setConnectionManager(const IConnectionManager& icm);
    
    /**
     *  @brief Get connection manager interface.
     *  @return The connection manager interface
     */
    IConnectionManager* getConnectionManager() const;
    
    /**
     *  @brief Get NMM reference of remote transport strategy
     *  @return The NMM reference of remote transport strategy
     */
    const NMMReference& getTSReference() const;
    
    /**
     *  @brief Set NMM reference of remote transport strategy
     *  @param ts_reference The NMM reference of remote transport strategy
     */
    void setTSReference(const NMMReference& ts_reference);
    //@}

  protected:
    /**
     *  @brief Construct a %ConnectorAddress
     *  @param strategy_type Type of transport strategy
     *  @param ref NMM Reference of addressed object
     *  @param info Type info
     */
    ConnectorAddress(const TypeInfo& info, unsigned int binding_type, const string& strategy_type, const NMMReference& ref);
    
  private:
    TValue<IConnectionManager*> __icm;       ///< Connection manager interface
    TValue<NMMReference> __ts_reference;     ///< NMM reference of transport strategy
  };
  
  
  inline IConnectionManager* ConnectorAddress::getConnectionManager() const {
    return __icm.getRValue();
  }
  
  inline const NMMReference& ConnectorAddress::getTSReference() const {
    return __ts_reference.getRValue();
  }
  
  inline void ConnectorAddress::setTSReference(const NMMReference& ts_reference) {
    __ts_reference.setValue(ts_reference);
  }
  
}

#endif
