/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "NMMApplication.hpp"
#include "TransportStrategy.hpp"
#include "ConnectionManager.hpp"

#include "nmm/base/registry/LocalRegistry.hpp"
#include "nmm/base/registry/PluginManager.hpp"
#include "nmm/base/registry/ServerRegistry.hpp"
#include "nmm/base/ObjectAcceptor.hpp"
#include "nmm/base/proxy/tcp/TCPAcceptor.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"
#include "nmm/base/serialize/MagicNumberValueStream.hpp"
#include "nmm/base/serialize/StdStream.hpp"
#include "nmm/utils/utils.hpp"

#include <fstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <iostream>

namespace NMM {
jmp_buf                     plugin_loop_position;        ///<stack position to the loop which loads all plugins
string                      current_plugin;              ///<current plugin
bool                        seg_fault_occured = false;   ///<flag which is set to true if a segmentaion fault occurs during plugin loading


/** This method catch segmentation faults during plugin loading */
void segfaultHandler( int ) {
  seg_fault_occured = true;
  GLOBAL_ERROR_STREAM(" ********************************************** " << endl);
  GLOBAL_ERROR_STREAM(" Plugin " << current_plugin << " causes a segfault " << endl);
  GLOBAL_ERROR_STREAM(" ********************************************** " << endl << endl << endl);

  longjmp(plugin_loop_position, 0);
}

NMMApplication::NMMApplication(int argc_, char** argv_, bool is_local, bool enable_port_listening) {
  argc = argc_;
  argv = argv_;
  if (is_local) {
    __client_registry.reset(new ClientRegistry());
    __server_registry.reset(new ServerRegistry());
    __plugin_manager.reset(SPluginManager::getInstance().getInterface<IPluginManager>());
    __client_registry->addRegistry(LocalAddress(__server_registry->getReference()), true);
    try {
      readRegistryConfig();
    } catch (const Exception& e) {
      GLOBAL_WARNING_STREAM( e << endl);
    }
  }
  if (enable_port_listening)
    __default_acceptor.reset(new TCPAcceptor(5000, 8000));
}

NMMApplication::~NMMApplication() {

}

ClientRegistry& NMMApplication::getRegistry() {
  return *__client_registry.get();
}
const ClientRegistry& NMMApplication::getRegistry() const {
  return *__client_registry.get();
}

IPluginManager& NMMApplication::getPluginManager(){
  return *__plugin_manager.get();
}
const IPluginManager& NMMApplication::getPluginManager() const{
  return *__plugin_manager.get();
}

IServerRegistryImpl* NMMApplication::getServerRegistry() {
  return __server_registry.get();
}

const IServerRegistryImpl* NMMApplication::getServerRegistry() const {
  return __server_registry.get();
}

void NMMApplication::enableServerRegistry(int port) {
  if ( getServerRegistry() ) {
    /* The TCPAcceptor accpets incoming tcp connections on port 28001
     * This port is used by clients to connect to objcects running in this address space */
    __registry_acceptor.reset( new TCPAcceptor(port) );

    SReferenceManager::getInstance().registerService("Registry", getServerRegistry());
    SReferenceManager::getInstance().registerService("PluginManager", &SPluginManager::getInstance());
    /* End< */
  } else {
    GLOBAL_WARNING_STREAM ("No local ServerRegistry available that can be enabled incoming requests. Maybe a global ServerRegistry is running!" << endl);
  }
}

void NMMApplication::setServerRegistry(IServerRegistryImpl* registry) {
  __server_registry.reset(registry);
}

void NMMApplication::readRegistryConfig(const string& file) {
  string plugin_file;
  if (file.empty()) {
    string::size_type pos = 0;
    string path = getNMMDir();
    path = path.substr(1, path.size()-2);
    pos  = path.find("/",0);
    while (pos != string::npos) {
      path[pos] = '_';
      pos = path.find("/", pos);
    }

    plugin_file =
      SProcessUserInfo::getInstance().getHomeDirectory() +
      string("/.nmm/plugins.") +
      string(VERSION) +
      string(".") +
      SProcessUserInfo::getInstance().getHostname() +
      string(".") +
      path;
  } else {
    plugin_file = SProcessUserInfo::getInstance().getHomeDirectory() + string("/.nmm/") + file;
  }

  ifstream stream;
  stream.open(plugin_file.c_str());
  /* Begin> If no plugin information available, create it*/
  if (!stream) {
    cerr << endl << "No plugin information available! If you start NMM for the first time this information is created automatically." << endl;
    cerr << "Note: If you ever change your hardware configuration or update the NMM version you must delete the file " << plugin_file << endl << endl;

    writeRegistryConfig(plugin_file);
    stream.open(plugin_file.c_str());
    if (!stream)
      throw Exception(string("Could not load plugin information from file ")+ plugin_file);
  }
  /* End< */

  /* Begin> Write the information to the stream */
  StdIStream s_stream(&stream);
  MagicNumberIValueStream m_stream;
  m_stream.setIPartnerStream(&s_stream);
  m_stream.readData();
  ServerRegistry* reg = dynamic_cast<ServerRegistry*>(getServerRegistry());
  if (reg) {
    reg->readFromStream(m_stream);
  } else {
    GLOBAL_WARNING_STREAM("NMMApplication uses not a ServerRegistry" << file << endl);
  }
  /* End< */
}

void NMMApplication::writeRegistryConfig(const string& file) {
  /* Begin> Check if directoy exist ~/.nmm exist, if not create it */
  string directory = SProcessUserInfo::getInstance().getHomeDirectory() + string("/.nmm") ;
  if (opendir(directory.c_str()) == NULL) {
    if (errno == ENOENT) {
      GLOBAL_DEBUG_STREAM2("Creating directory "<< directory << endl);
      if ( mkdir(directory.c_str(),0777) != 0) {
        GLOBAL_DEBUG_STREAM2("Unable to create directory "<< directory << endl);
	throw Exception(string("Could not create directory ") + directory);
      }
    }
  }
  /* End< */

  string filename;
  if (file.empty()) {
    filename =
      directory +
      string("/plugins.") +
      string(VERSION) +
      string(".") +
      SProcessUserInfo::getInstance().getHostname();
  } else {
    filename = file;
  }

  ofstream stream(filename.c_str());
  if (!stream) {
    throw Exception(string( "Could not open file " ) + filename );
    return;
  }
  cerr << "Create config file with plugin information ..." << endl;
  /* Load Plugins */
  loadPlugins();

  /* Create the stream */
  StdOStream s_stream(&stream);
  MagicNumberOValueStream m_stream;
  m_stream.setOPartnerStream(&s_stream);

  /* write the plugin configuration */
  dynamic_cast<ServerRegistry*>(getServerRegistry())->writeToStream(m_stream);
  m_stream.writeData();
  cerr << endl << "Config file successfully written. " << endl;
}

void NMMApplication::loadPlugins () {
  /*Begin> Load all plugins available*/
  IPluginManager_var pm(SPluginManager::getInstance().getInterface<IPluginManager>());
  if (!pm.get()) {
    GLOBAL_ERROR_STREAM("Could not get PluginManager interface" << endl);
    return;
  }

  typedef void (*sigmethod_t)(int);
  sigmethod_t old_sig_handler = ::signal(SIGSEGV, NMM::segfaultHandler); //register segfault handler

  list<string> plugin_list = pm->getPlugins();
  if (plugin_list.empty()) {
    throw Exception(string("No NMM plugins found. Please set the NMM_DEV_DIR environment variable to your NMM installation"));
  }
  
  cerr << "Loading plugins... " << endl;
  for (list<string>::const_iterator it = plugin_list.begin(); it != plugin_list.end(); ++it)
  {
    setjmp(plugin_loop_position); //stores the current position

    if (seg_fault_occured) {
      seg_fault_occured = false;
      continue;
    }

    size_t max_size = 48;
    current_plugin = *it;
    string plugin_name =  it->substr(6,it->size()-9);

    int plugin_size = max_size - plugin_name.size();

    cerr << plugin_name;
    while (plugin_size > 0 ) {
      cerr << "\t";
      plugin_size -= 8;
    }

    try {
      pm->loadLibrary(*it);
      if ( SPluginManager::getInstance().isFactoryPartActive(*it) ) {
        cerr << " available " << endl;
      } else {
        cerr << " not available " << endl;
      }
    } catch (const Exception& e) {
      cerr << " not available " << endl;
      GLOBAL_ERROR_STREAM(e <<  endl);
      GLOBAL_ERROR_STREAM("Error loading library " << *it << endl << endl << endl);
    }
  }

  /* If the old signal handler was vaild, use it. Otherwise set the SIG_DFL handler*/
  if (old_sig_handler != SIG_ERR)
    ::signal(SIGSEGV, old_sig_handler);
  else
    ::signal(SIGSEGV, SIG_DFL);
}

int            NMMApplication::argc = 0;
char**         NMMApplication::argv = 0;
}//end namespace

