/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_NMMOBJECT_HPP_
#define _NMM_NMMOBJECT_HPP_

#include "IReceiver.hpp"
#include "Interface.hpp"
#include "EventDispatcher.hpp"

namespace NMM {
  class Event;
  class CEvent;
  class EDObject;
  class TSAddress;
  class CommunicationChannel;
  class OutOfBandChannel;
  class ITransportStrategy;
  class IServerRegistry;

/** A NMMObject implements the IEventReceiver interface. For event dispatching a standard eventdispatcher is
  * used that can be accessed by base class to register their event specific methods. In the
  * NMM-Framework all NMMObjects are reachable trough the network.
  */
  class NMMObject : public IEventReceiver {
    public:
   /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor
      */
      NMMObject();

    /** Destructor
      */
      virtual ~NMMObject();

    /** Copy constructor
      */
      NMMObject(const NMMObject&);
    //@}

   /** @name Event Processing
     */
    //@{
    /** The method is used to send events to this NMMObject. 
      */
      virtual Result receiveEvent(const Event&  event, const IEventSender* = 0);

    /** The method is used to send an composite event to this NMMObject.
      */
      virtual Result receiveEvent(const CEvent& event, const IEventSender* = 0);

    /** The method is used to send an event to this NMMObject.
      */
      virtual Result receiveEvent(const Event&  event, const IEventSender* = 0) const;

    /** The method is used to send an composite event to this NMMObject.
      */
      virtual Result receiveEvent(const CEvent& event, const IEventSender* = 0) const;

    /** Returns an prototype event with an given event key. The caller of this method
      * is responsible to delete this event.
      */
      virtual Event* getEventPrototype(const string& event_key)  const;

    /** Returns an prototype event with a substring of an event key. The caller of this method
      * is responsible to delete this event.
      */
      virtual Event* matchEventPrototype(const string& event_key)  const;

    /** Returns the reference of NMMObjects eventdispatcher
      */
      inline        EventDispatcher& getEventDispatcher()                { return __event_dispatcher; }

    /** returns the list with all supported event keys
      */
      virtual list<string> getEventList() const;
    //@}

   /** @name Event Listener
     */
    /** This method registers an event listener. All registered IEventSender receives the event. 
      */
      virtual void registerEventListener(IEventReceiver*);

    /** Adds an TED-Object as EventListener. Incoming events with corresponding
      * event-keys are forwarded to the TED-Object
      */
      virtual void registerEventListener(const string&, EDObject* );

    /** This method registers an event listener for a specific key. 
      */
      virtual void registerEventListener(const string&, IEventReceiver*);

    /** This method removes an event listener.
      */
      virtual void removeEventListener(IEventReceiver*);

    /** This method removes an event listener.
      */
      virtual void removeEventListener(const string&, IEventReceiver*);

    /** removes a certain TEDObject
      */
      virtual void removeEventListener(EDObject*);
    //@}

   /** @name Access interfaces
     */
    //@{
    /** This template method returns a pointer to the Interface of type T or NULL if this interface
      * is not supported by this object. The use of this method is:
      * \code
      * NMMObject my_object;
      * INode* i_node = my_object.getInterface<INode>();
      * if (i_node)
      *   cout << "INode Interface supported" << endl;
      * else
      *   cout << "INode Interface not supported" << endl;
      * \endcode
      */
      template <typename T>
      T* getInterface() { T* tmp_interface = getOwnInterface<T>(); if (tmp_interface) return dynamic_cast<T*>(tmp_interface->copy()); return 0; }

    /** This template method returns a pointer to the Interface of type T or NULL if this interface
      * is not supported by this object. The use of this method is:
      * \code
      * const NMMObject my_object;
      * const INode* i_node = my_object.getInterface<INode>();
      * if (i_node)
      *   cout << "INode Interface supported" << endl;
      * else
      *   cout << "INode Interface not supported" << endl;
      * \endcode
      */
      template <typename T>
      const T* getInterface() const { const T* tmp_interface = getOwnInterface<T>(); if (tmp_interface) return dynamic_cast<T*>(tmp_interface->copy()); return 0; }

    /** This template method returns a pointer to the Interface of type T or throws an exception if this interface
      * is not supported by this object.
      */
      template <typename T>
      const T* getCheckedInterface() const
        throw (const InterfaceNotFoundException&) { const T* tmp_interface = getOwnInterface<T>(); if (tmp_interface) return dynamic_cast<T*>(tmp_interface->copy()); throw InterfaceNotFoundException(); }

    /** This template method returns a pointer to the Interface of type T or throws an exception if this interface
      * is not supported by this object.
      */
      template <typename T>
      T* getCheckedInterface()
        throw (const InterfaceNotFoundException&) { T* tmp_interface = getOwnInterface<T>(); if (tmp_interface) return dynamic_cast<T*>(tmp_interface->copy()); throw InterfaceNotFoundException(); }


    /** The method returns all Interfaces supported by this node.
      */
      virtual const list<Interface*>& getInterfaceList()  const     {return __interface_list;}

    /** The method returns all Interfaces supported by this node.
      */
      virtual       list<Interface*>& getInterfaceList()            {return __interface_list;}

    /** adds an interface supported by this object
      */
      inline  void   addInterface(Interface* interface)                  { if (interface) __interface_list.push_back(interface); }
    //@}

    /** @name Methods to specify the position of an interface
      */
    //@{
    /** returns true if the object of this interface and the object of the given interface runs on the same system
      * and in the same process
      */
      virtual bool isLocalTo(const Interface& );

    /** returns true if the object of this interface and the object of the given interface runs on the same system
      * but in a different process
      */
      virtual bool isSystemTo(const Interface&);

    /** returns true if the object of this interface and the object of the given interface runs on different systems.
      */
      virtual bool isNetworkTo(const Interface&);
    //@}

    /** @name Additional information
      */
    //@{
    /** stores the default address which can be used to connect this node with strategy type strategy_type
      */
      virtual TSAddress* getDefaultAddress(const string& strategy_type) const;

    /** Returns the address of this object as NMMReference
      */
      virtual const NMMReference& getReference()                  const  { getReference(__reference); return __reference; }

    /** returns the interface of the registry that manages the object
      */
      virtual IServerRegistry* getRegistry();

    /** sets the manager registry.
      */
      virtual void setRegistry(const IServerRegistry& );
    //@}
    protected:

    /** Objects which inherit NMMObject can get a non const eventdispatcher in const methods
      */
      inline        EventDispatcher& getEventDispatcher()        const  { return __event_dispatcher; }

    /** This method returns the current EventSender
      */
      inline  const IEventSender* getCurrentSender()             const  { return __current_sender;}

    /** This method returns the current EventSender
      */
      inline        IEventSender* getCurrentSender()                    { return __current_sender;}

    /** This method sets the current EventSender wich is done by the receiveEvent method. If you
      * overload the receiveEvent method you MUST call this method and set the current EventSender.
      * If the receiveEvent method has processed the event the current_sender should be set to NULL.
      * \sa receiveEvent
      */
      inline void setCurrentSender(IEventSender* current_sender)        { __current_sender = current_sender; }

    /** This method sets the current EventSender wich is done by the receiveEvent method. If you
      * overload the receiveEvent method you MUST call this method and set the current EventSender.
      * If the receiveEvent method has processed the event the current_sender should be set to NULL.
      * \sa receiveEvent
      */
      inline void setCurrentSender(IEventSender* current_sender) const  { __current_sender = current_sender; }

    /** This method removes the event key method an returns the corresponding DispatcherOject.
      * The caller of this method is responsible to delte the object. If the same object is
      * registered again, the NMMObjects takes care of the memorymanagent for this object and must
      * not be deleted again.
      * \sa registerEvent
      */
      inline EDObject* removeEvent(const string& key)                   { return __event_dispatcher.removeEvent(key); }

    /** This method registers an EventDispatcherObject under the given key.
      *\sa removeEvent
      */
      inline void registerEvent(const string& key, EDObject* edo)       { __event_dispatcher.registerEvent(key, edo); }

      template <typename T>
      T* getOwnInterface();

      template <typename T>
      const T* getOwnInterface() const;

    protected: //event methods
      /** This method stores this NMMRefernce in the given NMMReference
        */
      virtual Result getReference(NMMReference& reference) const;

      /** This method requests the supproted interfaces. This method is obsolete
        */
      virtual Result getInterfaceList(list<string>& interface_list) const;

      /** This method requests the UserInfo
        */
      virtual Result getUserInfo(UserInfo& user_info) const;

      /** This method requests the TransportStratgeies
        */
      virtual Result getStrategyList(list<string>& strategy_list) const;

      /** This method returns a list of all Interfaces
        */
      SequenceValue<Interface> getInternalInterfaceList() const;

      /** This method returns the interface count.
        */
      int getInterfaceSize();

      /** This method requests a specific interface.
        */
      Interface* getInternalInterface( int i );
    /**
      */
      virtual void connectWithOutOfBandChannel();

    /** Disconnect from an object
      */
      virtual void disconnectFromOutOfBandChannel(const bool& delete_strategy);

   /** @name Access OutOfBandChannels
     * Each connection from a ProxyObject to this NMMObject is represented by an OutOfBandChannel.
     * The OutOfBandChannel stores a pointer to a TranportStrategy which can be used to send
     * Events to the ProxyObject.
     */
    //@{
    /** This method locks the internal list of OutOfBandChannels. If this mehtod is called a 
      * connecetion can be closed or established to this object until the unlockChannelList
      * is called.
      * \sa unlockChannelList
      */
    inline void lockChannelList()                             const   { __internal_mutex.lock(); }

    /** This method unlocks the internal list of OutOfBandChannels. 
      * \sa lockChannelList
      */
    inline void unlockChannelList()                           const   { __internal_mutex.unlock();}

    /** This method returns the current OutOfBandChannels which represent a connection from a ProxyObject
      * to this NMMObject
      */
    inline const list<OutOfBandChannel*>& getChannelList()       const   { return __channel_list; }
    //@}
    private:
      virtual void registerEventListener()                            { registerEventListener(dynamic_cast<IEventReceiver*>(getCurrentSender())); }
      virtual void removeEventListener()                              { removeEventListener(dynamic_cast<IEventReceiver*>(getCurrentSender())); }
      virtual void registerSpecificEventListener(const string& key)   { registerEventListener(key, dynamic_cast<IEventReceiver*>(getCurrentSender())); }
      virtual void removeSpecificEventListener(const string& key)     { removeEventListener(key, dynamic_cast<IEventReceiver*>(getCurrentSender()));   }
      virtual ITransportStrategy* getPartnerStrategy();

    /** returns a new OutOfBandChannel that can be used to set TransportStrategies.
      */
      virtual OutOfBandChannel* requestCommunicationChannel();

    /** releases a requested communication channel;
      */
      virtual void releaseCommunicationChannel(CommunicationChannel*);

    private:
      void registerEvents();
    private:
      list<Interface*>              __interface_list;             ///< stores a list with all interfaces
      list<OutOfBandChannel*>          __channel_list;               ///< stores a list of connected ProxyObjects
      mutable EventDispatcher       __event_dispatcher;           ///< The EventDispatcher to process incoming events
      mutable IEventSender*         __current_sender;             ///< The current event sender
      mutable NMMReference          __reference;                  ///< The NMMReference
      mutable ThreadMutex           __internal_mutex;             ///< Internal Mutex for thread safe access
      IServerRegistry*              __i_registry;                 ///< registry which manages the node
      int                           __ref_count;                  ///< internal ref count.
  };




template <typename T>
T* NMMObject::getOwnInterface() {
  T* interface = 0;
  for (list<Interface*>::iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
  {
    interface = dynamic_cast<T*>(*it);
    if (interface != 0)
      return interface;
  }
  return 0;
}


template <typename T>
const T* NMMObject::getOwnInterface() const {
  const T* interface = 0;
  for (list<Interface*>::const_iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
  {
    interface = dynamic_cast<const T*>(*it);
    if (interface != 0)
      return interface;
  }
  return 0;
}

}

#endif // _NMMOBJECT_HPP_
