/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include "ProxyObject.hpp"
#include "nmm/base/InterfaceFactory.hpp"
#include "nmm/base/TSFactory.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "nmm/interfaces/net/IObjectAcceptor.hpp"
#include "nmm/base/registry/PluginManager.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"
#include "NMMApplication.hpp"

namespace NMM {
ProxyObject::ProxyObject(bool auto_delete)
  : NamedObject("Proxy"), __object_channel(*this), __parent_reference(0), __is_connected(false), __interface_counter(0),
                          __auto_delete(auto_delete)
{
  /* Register the Events */
  registerEvents();

  /* Register methods to be informed if the connection to an object changes */
  NMMObject::registerEventListener(IObjectAcceptor::connectToObject_event, new TEDObject1<ProxyObject, NMMReference, void (ProxyObject::*)(const NMMReference&), void>(this, &ProxyObject::connectToObjectListener));
  NMMObject::registerEventListener(IObjectAcceptor::connectToService_event, new TEDObject1<ProxyObject, string, void (ProxyObject::*)(const string&), void>(this, &ProxyObject::connectToServiceListener));
}

ProxyObject::ProxyObject(const Interface& interface, bool auto_delete)
  : NamedObject("Proxy"), __object_channel(*this), __parent_reference(0), __is_connected(false), __interface_counter(0),
                          __auto_delete(auto_delete)
{
  /* Register the Events */
  registerEvents();

  /* Register methods to be informed if the connection to an object changes */
  NMMObject::registerEventListener(IObjectAcceptor::connectToObject_event, new TEDObject1<ProxyObject, NMMReference, void (ProxyObject::*)(const NMMReference&), void>(this, &ProxyObject::connectToObjectListener));
  NMMObject::registerEventListener(IObjectAcceptor::connectToService_event, new TEDObject1<ProxyObject, string, void (ProxyObject::*)(const string&), void>(this, &ProxyObject::connectToServiceListener));

  __object_channel.updateStrategyList(interface);
}


ProxyObject::~ProxyObject( ) {
  /* Begin> Unregister the ProxyObejct by the ReferenceManager */
  if (NMMObject::getInterfaceList().empty())
    SReferenceManager::getInstance().unregisterProxyObject( *this, UserInfo() );
  else
    SReferenceManager::getInstance().unregisterProxyObject( *this, NMMObject::getInterfaceList().front()->getParentInfo() );
  /* End< */

  /* If the ProxyObject is still connected disconnect it */
  if (__is_connected) {
    disconnectFromObject();
  }

  /* Destroy the TransportStrategy if availabel */
  if (getOutOfBandChannel().getTransportStrategy()) {
    getOutOfBandChannel().getTransportStrategy()->scheduleDestruction();
    getOutOfBandChannel().setTransportStrategy(0);
  }

  /* Reset the parante pointer in our interfaces, so they dont dereference the internal interface counter */
  for (list<Interface*>::iterator it = NMMObject::getInterfaceList().begin(); it != NMMObject::getInterfaceList().end(); ++it)
    (*it)->setParentObject(0);
}

void ProxyObject::registerEvents() {
  getEventDispatcher().registerEvent(updateInterfaces_event, new TEDObject0<ProxyObject, void (ProxyObject::*)( ) , void >(this, &ProxyObject::updateInterfaces));
}


const list<Interface*>& ProxyObject::getInterfaceList()   const {
  /* If the ProxyObject is not connected, connect it */
  if (!__is_connected)
    const_cast<ProxyObject*>(this)->sendConnectToObject();

  /* return the interface list of the base class */
  return NMMObject::getInterfaceList();
}

list<Interface*>& ProxyObject::getInterfaceList() {
  /* If the ProxyObject is not connected, connect it */
  if (!__is_connected)
    sendConnectToObject();

  /* return the interface list of the base class */
  return NMMObject::getInterfaceList();
}       

Result ProxyObject::receiveEvent(const CEvent& event, const IEventSender* sender)  {
  /* If not connected, connect with NMMObject */
  if (!__is_connected)
    sendConnectToObject();

  /* If the TransportStrategy is the sender we dispatch the Event only*/
  if (sender == __object_channel.getTransportStrategy()) 
    return NMMObject::receiveEvent(event, sender);

  /* Otherwise we send the Event over the TransportStrategy */
  if (__object_channel.getTransportStrategy() ) {
    Result result = __object_channel.getTransportStrategy()->receiveEvent(event, sender);
    for (unsigned int i = 0; i< event.size(); ++i) {
      const_cast<EventDispatcher&>(getEventDispatcher()).forwardEvent(const_cast<CEvent&>(event).getEvent(i), sender);
      const_cast<EventDispatcher&>(getEventDispatcher()).externalDispatch(const_cast<CEvent&>(event).getEvent(i), sender);
    }
    return result;
  }

  return FAILURE;
}

Result ProxyObject::receiveEvent(const Event&  event, const IEventSender* sender)  {
  /* If not connected, connect with NMMObject */
  if (!__is_connected)
    sendConnectToObject();

  /* If the TransportStrategy is the sender we dispatch the Event only*/
  if (sender == __object_channel.getTransportStrategy()) 
    return NMMObject::receiveEvent(event, sender);

  /* Otherwise we send the Event over the TransportStrategy */
  if (__object_channel.getTransportStrategy() ) {
    Result result = __object_channel.getTransportStrategy()->receiveEvent(event, sender);
    /* forward the event to all listener */
    getEventDispatcher().forwardEvent(event, sender);
    /* forward the event to all listener */
    getEventDispatcher().externalDispatch(event, sender);
    return result;
  }

  return FAILURE;
}

Result ProxyObject::receiveEvent(const CEvent& event, const IEventSender* sender) const {
  /* If not connected, connect with NMMObject */
  if (!__is_connected)
    const_cast<ProxyObject*>(this)->sendConnectToObject();

  /* If the TransportStrategy is the sender we dispatch the Event only*/
  if (sender == __object_channel.getTransportStrategy()) 
    return NMMObject::receiveEvent(event, sender);

  /* Otherwise we send the Event over the TransportStrategy */
  if (__object_channel.getTransportStrategy() ) {
    Result result = __object_channel.getTransportStrategy()->receiveEvent(event, sender);
    for (unsigned int i = 0; i< event.size(); ++i) {
      const_cast<EventDispatcher&>(getEventDispatcher()).forwardEvent(const_cast<CEvent&>(event).getEvent(i), sender);
      const_cast<EventDispatcher&>(getEventDispatcher()).externalDispatch(const_cast<CEvent&>(event).getEvent(i), sender);
    }
    return result;
  }

  return FAILURE;
}

Result ProxyObject::receiveEvent(const Event&  event, const IEventSender* sender) const {
  /* If not connected, connect with NMMObject */
  if (!__is_connected)
    const_cast<ProxyObject*>(this)->sendConnectToObject();

  /* If the TransportStrategy is the sender we dispatch the Event only*/
  if (sender == __object_channel.getTransportStrategy()) 
    return NMMObject::receiveEvent(event, sender);

  /* Otherwise we send the Event over the TransportStrategy */
  if (__object_channel.getTransportStrategy() ) {
    Result result = __object_channel.getTransportStrategy()->receiveEvent(event, sender);
    getEventDispatcher().forwardEvent(event, sender);
    getEventDispatcher().externalDispatch(event, sender);
    return result;
  }

  return FAILURE;
}

Result ProxyObject::receiveReliableEvent(Event&  event, const IEventSender* sender) const {
  /* If not connected, connect with NMMObject */
  if (!__is_connected)
    const_cast<ProxyObject*>(this)->sendConnectToObject();

  /* If the TransportStrategy is the sender we dispatch the Event only*/
  if (sender == __object_channel.getTransportStrategy()) 
    return NMMObject::receiveEvent(event, sender);

  /* Otherwise we send the Event over the TransportStrategy */
  if (__object_channel.getTransportStrategy() ) {
    Result result = __object_channel.getTransportStrategy()->receiveReliableEvent(event, sender);
    getEventDispatcher().forwardEvent(event, sender);
    getEventDispatcher().externalDispatch(event, sender);
    return result;
  }

  return FAILURE;
}


void ProxyObject::connectToObject(const TSAddress& address) 
    throw (const BindingException&)
{
  /* copy the address to connect to */
  getOutOfBandChannel().setDefaultAddress(address);

  /* store the parent reference */
  __parent_reference  = address.getNMMReference();

  /* set the internal connection flag to false */
  __is_connected      = false;
}

void ProxyObject::disconnectFromObject() {
  if (!__is_connected)
    return;

  /* Unregister the ProxyObject by the ReferenceManager */
  if (NMMObject::getInterfaceList().empty())
    SReferenceManager::getInstance().unregisterProxyObject( *this, UserInfo() );
  else
    SReferenceManager::getInstance().unregisterProxyObject( *this, NMMObject::getInterfaceList().front()->getParentInfo() );

  /* set the parent reference to  0 */
  __parent_reference  = 0;

  /* we are not connected */
  __is_connected      = false;

  /* delete the old TransportStrategy */
  if (getOutOfBandChannel().getTransportStrategy())
    getOutOfBandChannel().getTransportStrategy()->scheduleDestruction();

  /* Currently no TransportStrategy is used */
  getOutOfBandChannel().setTransportStrategy(0);
}


void ProxyObject::establishBinding()
    throw (const BindingException&)
{
  if (!__is_connected)
    sendConnectToObject();
}

void ProxyObject::sendConnectToObject()
    throw (const BindingException&)
{
  try {
    if ( !getOutOfBandChannel().getDefaultAddress() ) {
      throw BindingException("Invalid address (0-ptr)");
    }

    /* destroy an old TransportStrategy */
    if ( getOutOfBandChannel().getTransportStrategy() )
      getOutOfBandChannel().getTransportStrategy()->scheduleDestruction();

    /* Set the Transport Strategy */
    getOutOfBandChannel().setTransportStrategy(getOutOfBandChannel().getDefaultAddress()->getStrategyType());

    if (!getOutOfBandChannel().getTransportStrategy() )
      throw BindingException("Can not connect to object because no TransportStrategy is set");

    /* connect to address */
    getOutOfBandChannel().getTransportStrategy()->setAddress( *getOutOfBandChannel().getDefaultAddress() );

    __is_connected = true;

    /* Establish a binding between the NMMObject and this object */
    if (getOutOfBandChannel().getDefaultAddress()->getNMMReference() != 0) {
      Event e("NMMObject::connectWithOutOfBandChannel");
      if (receiveReliableEvent(e) != SUCCESS) {
        WARNING_STREAM("Could not establish binding" << endl);
      }
    }

    /* update the interface list */
    updateInterfaces();
  } catch (const Exception& e) {
    throw BindingException(e.getComment() );
  } catch ( ... ) {
    throw BindingException("Unknown exception in ProxyObject::sendConnectToObject");
  }
}

void ProxyObject::updateInterfaces()  {
  /* Begin> Delete the old interfaces and clear the list */
  for (list<Interface*>::iterator it = NMMObject::getInterfaceList().begin(); it != NMMObject::getInterfaceList().end(); ++it) {
    (*it)->setParentObject(0);
    delete *it;
  }

  NMMObject::getInterfaceList().clear();
  /* End< */


  /* Create the return value for method getInterfaceList */
  TOutValue<SequenceValue<Interface> > interface_list;
  interface_list.getRValue().useTValue(false);

  Event e("NMMObject::getInterfaceList");
  e.setReturnValue(&interface_list);
  if(receiveEvent(e) != SUCCESS) {
    ERROR_STREAM("Could not get InterfaceList ");
  } else {
    /* */
    if (!interface_list.getRValue().empty()) {
      for(unsigned int i = 0; i < interface_list.getRValue().size(); ++i) {
        /* If an interface pointer is NULL corresponding interface is unknown to the application */
        if ( interface_list.getRValue()[i] ) {
          interface_list.getRValue()[i]->setParentObject(this);
          addInterface(interface_list.getRValue()[i]);
          interface_list.getRValue()[i] = NULL;
        } else {
          WARNING_STREAM("Could not create Interface " << endl);
        }
      }
    } else {
      WARNING_STREAM("Distributed object without interfaces, so this object can only be accessed through its Proxy. " << endl);
    }
  }
  /*End< */
}

Result ProxyObject::getInterfaceList(SequenceValue<Interface>& interface_list) const {
  TOutValue<SequenceValue<Interface> > tmp_value(interface_list);
  Event e("NMMObject::getInterfaceList");
  e.setReturnValue(&tmp_value);
  if (receiveEvent(e) == FAILURE) {
    WARNING_STREAM("Could not get interfaceList"<< endl);
    return FAILURE;
  }
  interface_list = tmp_value.getRValue();
  return SUCCESS;
}

Result ProxyObject::getUserInfo(UserInfo& user_info) const {
  TOutValue<UserInfo > value;
  if (receiveEvent(Event("NMMObject::getUserInfo", &value)) == FAILURE) {
    return FAILURE;
  }

  user_info = value.getRValue();
  return SUCCESS;
}

Result ProxyObject::getStrategyList(list<string>& strategy_list) const {
  TOutValue<list<string> > value;
  if (receiveEvent(Event("NMMObject::getStrategyList", &value)) == FAILURE) {
    return FAILURE;
  }

  strategy_list = value.getRValue();
  return SUCCESS;
}

TSAddress* ProxyObject::getDefaultAddress(const string& strategy) const {
  TOutValue<TSAddress*> address;
  Event e("NMMObject::getDefaultAddress", new TInValue<string>(strategy));
  e.setReturnValue(&address);
  if (receiveEvent(e) != SUCCESS) {
    if (e.getException())
      throw *e.getException();
  }
  return address.getRValue();
}

list<string> ProxyObject::getEventList() const {
  TOutValue<list<string> > event_list;
  Event e("NMMObject::getEventList");
  e.setReturnValue(&event_list);
  if (receiveEvent(e) != SUCCESS) {
    if (e.getException())
      throw *e.getException();
  }

  return event_list.getRValue();
}

Event* ProxyObject::getEventPrototype(const string& event_key)  const  {
  TOutValue<Event* > event(0);
  Event e("NMMObject::getEventPrototype", new TInValue<string>(event_key));
  e.setReturnValue(&event);
  if (receiveEvent(e) != SUCCESS) {
    if (e.getException())
      throw *e.getException();
  }

  return event.getRValue();
}

Event* ProxyObject::matchEventPrototype(const string& event_key)  const  {
  TOutValue<Event* > event(0);
  Event e("NMMObject::matchEventPrototype", new TInValue<string>(event_key));
  e.setReturnValue(&event);
  if (receiveEvent(e) != SUCCESS) {
    if (e.getException())
      throw *e.getException();
  }

  return event.getRValue();
}


void ProxyObject::registerEventListener(IEventReceiver* receiver) {
  bool register_listener = !getEventDispatcher().hasListener();
  NMMObject::registerEventListener(receiver);

  if (register_listener) {
    /* If we must register a geneirc listener, we test if specfifc listener are registered */
    if (getEventDispatcher().hasSpecificListener()) {
      /* If specific listener are registered we send an "remove.." event to unregister Proxy for all specifc listener */
      if (receiveEvent(Event("NMMObject::removeEventListener")) == FAILURE) {
        WARNING_STREAM("Could not unregister EventListener"<<endl);
      }
    }
    /* Then we register an eventhandler for all events*/
    if (receiveEvent(Event("NMMObject::registerEventListener")) == FAILURE) {
      ERROR_STREAM("Could not register EventListener");
      return;
    }
  }
}

void ProxyObject::registerEventListener(const string& key, EDObject* ed) {
  /* Check if another EDObject or IEventReceiver is listening to the given event*/
  bool forward_registration = !(getEventDispatcher().hasSpecificListener(key));

  NMMObject::registerEventListener(key, ed);
  if (forward_registration && !getEventDispatcher().hasListener()) {
    if (receiveEvent(Event("NMMObject::registerSpecificEventListener", new TInValue<string>(key))) == FAILURE) {
      ERROR_STREAM("Could not register EventListener");
    }
  }
}

void ProxyObject::registerEventListener(const string& key, IEventReceiver* handler) {
  /* Check if another EDObject or IEventReceiver is listening to the given event*/
  bool forward_registration = !(getEventDispatcher().hasSpecificListener(key));

  NMMObject::registerEventListener(key, handler);
  if (forward_registration &&  !getEventDispatcher().hasListener()) {
    if (receiveEvent(Event("NMMObject::registerSpecificEventListener", new TInValue<string>(key))) == FAILURE) {
      ERROR_STREAM("Could not register EventListener");
    }
  }
}

void ProxyObject::removeEventListener(IEventReceiver* handler) {
  /* This bool is an additional test. If we have no listener and given listener is not in our map, we must not send the event*/
  bool had_listener_before_remove = getEventDispatcher().hasListener();
  NMMObject::removeEventListener(handler);

  if (had_listener_before_remove && !getEventDispatcher().hasListener()) {
    if (receiveEvent(Event("NMMObject::removeEventListener")) == FAILURE) {
      WARNING_STREAM("Could not unregister EventListener"<< endl);
    }
    /* If we have specific listener the proxy object must send an registration event for these events*/
    if (getEventDispatcher().hasSpecificListener()) {
      list<string> specific_listener = getEventDispatcher().getSpecificListener();
      for (list<string>::iterator it = specific_listener.begin(); it != specific_listener.end(); ++it) {
        if (receiveEvent(Event("NMMObject::registerSpecificEventListener", new TInValue<string>(*it))) == FAILURE) { 
          ERROR_STREAM("Could not register EventListener");
        }
      }
    }
  }
}

void ProxyObject::removeEventListener(const string& key, IEventReceiver* handler) {
  NMMObject::removeEventListener(key, handler);

  /* We send "removeEvent..." if no other listener is available and no generic listener is available*/
  if (!getEventDispatcher().hasSpecificListener(key) && !getEventDispatcher().hasListener()) {
    if (receiveEvent(Event("NMMObject::removeSpecificEventListener", new TInValue<string>(key))) == FAILURE) {
      WARNING_STREAM("Could not unregister EventListener"<<endl);
    }
  }
}

void ProxyObject::removeEventListener(EDObject* edo) {
  list<string> unregistered_events = getEventDispatcher().removeEventListener(edo);

  if (!getEventDispatcher().hasListener()) {
    for (list<string>::iterator it = unregistered_events.begin(); it != unregistered_events.end(); ++it) {
      if (!getEventDispatcher().hasSpecificListener(*it)) {
        if (receiveEvent(Event("NMMObject::removeSpecificEventListener", new TInValue<string>(*it))) == FAILURE) {
          WARNING_STREAM("Could not unregister EventListener"<<endl);
        }
      }
    }
  }
}

IServerRegistry* ProxyObject::getRegistry() {
  TOutValue<IServerRegistry*> r_value;
  Event e("NMMObject::getRegistry");
  e.setReturnValue(&r_value);
  if (receiveEvent(e) != SUCCESS) {
    if (e.getException())
      throw *e.getException();
  }
  return r_value.getRValue();
}

void ProxyObject::setRegistry(const IServerRegistry& registry) {
  Event e("NMMObject::setRegistry", new TInValue<IServerRegistry>(registry));
  if (receiveEvent(e) != SUCCESS) {
    if (e.getException())
      throw *e.getException();
  }
}


void ProxyObject::connectToObjectListener(const NMMReference& reference) {
  __parent_reference = reference;
  Event e("NMMObject::connectWithOutOfBandChannel");
  if (receiveReliableEvent(e) != SUCCESS) {
    WARNING_STREAM("Could not establish binding" << endl);
  }
  updateInterfaces();
}

void ProxyObject::connectToServiceListener(const string& ) {
  Event e("NMMObject::connectWithOutOfBandChannel");
  if (receiveReliableEvent(e) != SUCCESS) {
    WARNING_STREAM("Could not establish binding" << endl);
  }

  TOutValue<NMMReference> reference;
  if (receiveEvent(Event("NMMObject::getReference", &reference)) != SUCCESS) {
    WARNING_STREAM("Could not getNMMReference" << endl);
  } else {
    __parent_reference = reference.getRValue();
  }

  updateInterfaces();
}


const string ProxyObject::updateInterfaces_event = "ProxyObject::updateInterfaces";
}



