/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_PROXYOBJECT_HPP
#define NMM_PROXYOBJECT_HPP

#include "nmm/base/OutOfBandChannel.hpp"
#include "nmm/base/NMMObject.hpp"

namespace NMM {
  
  class ProxyObject  : public NMMObject, public NamedObject {
    friend class Interface;
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor
     */
    ProxyObject(bool auto_delete = false);

    /** Constructor, that uses the connection addresses inside the given interface
     */
    ProxyObject(const Interface&, bool auto_delete = false);

    /** Destructor
     */
    virtual ~ProxyObject();
    //@}

    /** @name Interface mehtods
     */
    //@{
    /** The method returns all Interfaces supported by this node.
     */
    virtual const list<Interface*>& getInterfaceList()   const;

    /** The method returns all Interfaces supported by this node.
     */
    virtual       list<Interface*>& getInterfaceList()        ;
    //@}

    /** @name Event Processing
     */
    //@{
    /** The method is used to send events to the NMMObject.
     */
    virtual Result receiveEvent(const CEvent& event, const IEventSender* = 0 );
    /** The method is used to send events to the NMMObject.
     */
    virtual Result receiveEvent(const Event&  event, const IEventSender* = 0 );

    /** The method is used to send events to the NMMObject.
     */
    virtual Result receiveEvent(const CEvent& event, const IEventSender* = 0 ) const;
    /** The method is used to send events to the NMMObject.
     */
    virtual Result receiveEvent(const Event&  event, const IEventSender* = 0 ) const;
    /** This method is used to send reliable configuration events to the NMMObject.
     */
    virtual Result receiveReliableEvent(Event&, const IEventSender* = 0 ) const;
    /** This mehtod returns a prototype for the given Event key
     */
    virtual Event* getEventPrototype(const string&)  const;

    /** Returns an prototype event with a substring of an event key. The caller of this method
     * is responsible to delete this event.
     */
    virtual Event* matchEventPrototype(const string& event_key)  const;

    /** returns the list with all supported event keys
     */
    virtual list<string> getEventList() const;
    //@}


    /** @name Connection methods
     */
    //@{
    /** This method returns the current OutOfBandChannel
     */
    OutOfBandChannel& getOutOfBandChannel()       { return __object_channel; }

    /** This method returns the current OutOfBandChannel
     */
    const OutOfBandChannel& getOutOfBandChannel() const { return __object_channel; }

    /** This method connects to the NMMObject
     */
    virtual void connectToObject(const TSAddress& address)
      throw (const BindingException&);

    /** This method enforces a connection between the proxy and a distributed object
     */
    virtual void establishBinding()
      throw (const BindingException&);

    /** This method closes a connection to the NMMObject
     */
    virtual void disconnectFromObject();
    //@}

    /** @name Event Listener mehtods
     */
    //@{
    /** This method registers an event listener. All registered IEventSender receives the event.
     */
    virtual void registerEventListener(IEventReceiver*);

    /** Adds an TED-Object as EventListener. Incoming events with corresponding
     * event-keys are forwarded to the TED-Object
     */
    virtual void registerEventListener(const string&, EDObject* );

    /** This method registers an event listener for a specific key.
     */
    virtual void registerEventListener(const string&, IEventReceiver*);

    /** This method removes an event listener.
     */
    virtual void removeEventListener(IEventReceiver*);

    /** This method removes an event listener.
     */
    virtual void removeEventListener(const string&, IEventReceiver*);

    /** removes a certain TEDObject
     */
    virtual void removeEventListener(EDObject*);
    //@}

    /** @name Additional information
     */
    //@{
    /** stores the default address which can be used to connect this node with strategy type strategy_type
     */
    virtual TSAddress* getDefaultAddress(const string& strategy_type) const;

    /** Returns the address of this object as NMMReference
     */
    virtual const NMMReference& getReference()                  const  { return __parent_reference; }

    /** Returns the address of this object as NMMReference
     */
    virtual const NMMReference& getParentReference()            const  { return __parent_reference; }
    //@}

    virtual IServerRegistry* getRegistry();

    virtual void setRegistry(const IServerRegistry& registry);


  protected:
    void    updateInterfaces();
    
    virtual Result getInterfaceList(SequenceValue<Interface>& interface_list) const;
    virtual Result getStrategyList(list<string>& strategy_list) const;
    virtual Result getUserInfo(UserInfo& user_info) const;

    void increaseInterfaceCounter()      {++__interface_counter;}

    void decreaseInterfaceCounter()      {--__interface_counter; if ( (__interface_counter <= 0) && __auto_delete) delete this;}

  private:
    void registerEvents();

    void sendConnectToObject()
      throw (const BindingException&);

    void connectToObjectListener(const NMMReference&);
    void connectToServiceListener(const string& );
  private:
    OutOfBandChannel      __object_channel;               ///< stores the OutOfBandChannel
    mutable NMMReference  __parent_reference;             ///< Stores the NMMReference of the parent
    mutable bool          __is_connected;                 ///< Is true if the ProxyObject is connected
    int                   __interface_counter;            ///< Counter to interfaces that use this ProxyObject
    bool                  __auto_delete;                  ///< If true, the ProxyObject deletes themself, if no Interfaces point to it
  public:
    static const string   updateInterfaces_event;         ///< event key for method updateInterfaces
  };


}
#endif



