/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_STREAMQUEUE_HPP_
#define _NMM_STREAMQUEUE_HPP_

#include "nmm/utils/NamedObject.hpp"
#include "ISignalReceiver.hpp"
#include "nmm/NMMTypes.hpp"
#include "sync/Timestamp.hpp"
#include "nmm/utils/thread/Thread.hpp"
#include <queue>

namespace NMM {
  class Message;
  class ThreadMutex;

   /**
   * This class represents an input queue.
   * A QueuedThreadedNode has a Queue at the input jack. 
   * The queue may be configured to behave in a specific way if a buffer
   * arrives when the queue is full: The node may chose to block its 
   * predecessor's thread, discard the newly incoming buffer in favour of 
   * the older ones in the queue, or discard the oldest waiting buffer in 
   * the queue in favour of the newly entered one.
   */
  class StreamQueue : public NamedObject{
  public:
    /** Specifies what to do when the queue is full and a new buffer
	arrives.                                                            */
    enum Mode { DEFAULT_MODE = 1,     /*!< Default mode, uses  */
		MODE_SUSPEND = 1,     /*!< Wait if the queue is full.   */
		MODE_KEEP_OLDEST, /*!< Keep oldest buffer if the
				    queue is full.                       */
		MODE_KEEP_NEWEST  /*!< Keep newest buffer if the
				    queue is full.                       */
    };

    /**
     * The main constructor.
     * \param mode : The QueueMode of the new queue.
     * \param size : The size of the new queue. Default is 16.
     */
    StreamQueue(const Mode mode = DEFAULT_MODE,
		const unsigned int size = DEFAULT_SIZE,
		const char* name = 0,
		const Signal free_queue_slot_sig = FREE_QUEUE_SLOT,
		const Signal queue_not_empty_sig = QUEUE_NOT_EMPTY);

    virtual ~StreamQueue();

    /**
     * This method is called by the node's receiveMessage method.
     * \returns BLOCKED if the queue is full now.
     *          SUCCESS if the buffer is handled according to the QueueMode.
     */
    Result receiveMessage(Message* buffer);

    /**
     * This function returns the next buffer or null if queue is empty. 
     */
    Message* dequeueMessage();

    /** 
     * Get the size (number of enqueued Buffers) of the queue.
     * \returns the size of the queue.
     */
    unsigned int getSize() const;
    unsigned int getMaxSize() const;
    Result setMaxSize(unsigned int size);
    bool empty() const; 
    bool full() const;

    /**
     * Sets the pre fetch queue size 
     */
    void setPrefetchQueueSize(unsigned int size);

    void waitPrefetchFinished();
  
    /**
     * Sets the internal max. queue length to 'infinite'
     */
    void unlimitSize();

    /**
     * Re-sets the internal max. queue length to the value specified by setMaxSize
     */
    void limitSize();

    Mode getMode() const;
    Result setMode(Mode mode);
    
    /** Set parent signal receiver.
     * Set @a parent as the signal receiver which is signaled when the queue is not empty.
     */
    Result setParentSignalReceiver(ISignalReceiver* parent);
    
    /** Set partner signal receiver.
     * Set @a partner as the signal receiver which is signaled when there is a free slot in the queue.
     */
    Result setPartnerSignalReceiver(ISignalReceiver* partner);
    
    /** Set parent signal receiver.
     * Set @a parent as the signal receiver which is signaled when the queue is not empty.
     * Also set @a queue_not_empty_sig as the signal which is sent.
     */
    Result setParentSignalReceiver(ISignalReceiver* parent, const Signal queue_not_empty_sig);

    /** Set parent signal receiver.
     * Set @a parent as the signal receiver which is signaled when the queue is not empty.
     * Also set @a free_queue_slot_sig as the signal which is sent.
     */
    Result setPartnerSignalReceiver(ISignalReceiver* partner, const Signal free_queue_slot_sig);
    
    Result signalFreeQueueSlot();
    
    /** Lock the internal mutex of the queue.
     */
    void lock() const;

    /** Unlock the internal mutex of the queue.
     */
    void unlock() const;
    
    Result flush(const Timestamp& flush_ts = Timestamp());
    
    static const unsigned int DEFAULT_SIZE = 16; /*!< The default size of a queue.   */

    /**
     * Try to removes at least drop_bytes in buffers from the queue
     * until the queue is empty.
     * (All events in the queue are not handled but kept in the queue.)
     *
     * \param drop_bytes number of bytes to drop
     * \return the actual number of bytes dropped 
     */
    virtual long dropBuffers(long drop_bytes);
  private:

    queue<Message*> message_queue;

    /* Behaviour on full queue (see enum QueueMode).                        */
    Mode queue_mode;

    ///< Store incoming buffers here.
    unsigned int queue_length;       ///< Length of the queue array.

    ThreadMutex* queue_mutex;        ///< Mutex for queue access.

    /**
     * This counter specifies how often we sent out an
     * AbortReceiveBufferException on a blocking enqueue. Only send this
     * many free slot events to avoid the premature end of a virtual stop.
     */
    int abort_receive_exception_count;


    /** Here, a ISignalReceiver-pointer to the parent of this stream queue's jack will be stored */
    ISignalReceiver* parent_of_jack;

    /** Here, a ISignalReceiver-pointer to the parent of this stream queue's partner jack will be stored */
    ISignalReceiver* parent_of_partner_jack;

    /** Signal to be sent for FREE_QUEUE_SLOT */
    Signal sig_free_queue_slot;

    /** Signal to be sent for QUEUE_NOT_EMPTY */
    Signal sig_queue_not_empty;

    /** Specifies whether queue size is limited or not, true per default */
    unsigned int limited_size;
 
    /** Specifies the size of messages that must queued if pre-fetch modus is enabled, 0 per default */
    unsigned int pre_fetch_queue_size;

    /** Specifies whether the pre-fetch modus is enabled or not, false per default */
    bool pre_fetch_flag;

    /** Mutex for pre_fetch_condition attribute */
    ThreadMutex pre_fetch_mutex;

    /** Condition variable used to wait until pre-fetching is finished */
    ThreadCondition pre_fetch_condition;
  };
  
}

#endif // _NMM_STREAMQUEUE_HPP_

