/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_TSADDRESS_HPP
#define NMM_TSADDRESS_HPP

#include "nmm/NMMTypes.hpp"
#include "registry/RegistryTypedef.hpp"

namespace NMM {
  
/** This abstract superclass specify an address wich can be used to connect a TransportStrategy
  * with a corresponding partner. It stores the NMMReference of the object that should be connected.
  * A concrete implementation of this class adds the TransportStrategy specific parameter, like
  * an IP-address and a port which is need for a tcp connection. It also stores the binding type
  * which specifiy if the corresponding TransportStrategy is a local, a system or a network binding.
  * A TransportStrategy can support all binding types.
  */
class TSAddress : public SerializedValue {
public:
  enum BindingType{ LOCAL = 1, SYSTEM = 2, NETWORK = 4};

/** @name Constructor & Destructor
  */
//@{
/** Default Constructor 
  */
  TSAddress(const NMMReference& reference = 0, const string& stratgey_type = "", const TypeInfo& t_info = TypeInfo("TSAddress", "NMM"), int binding_type = 0)
    : SerializedValue(t_info), __reference(reference), __strategy_type(stratgey_type), __binding_type(binding_type), __port(0)  { registerAttributes(); }

/** Copy Constructor
  */
  TSAddress(const TSAddress& address)
    : SerializedValue(address), __reference(address.getNMMReference())      { registerAttributes(); *this = address;}

/** virtual Copy Constructor
  */
  TSAddress* copy() const {return new TSAddress(*this); } //adjust the return value
//@}
/** @name Access information
  */
//@{
/** This method sets the NMMReference to reference
  */
  void          setNMMReference(const NMMReference& reference)       { __reference.setValue(reference); }

/** This method returns the current NMMReference
  */
  const NMMReference&  getNMMReference()                              const { return __reference.getRValue(); }

/** This method returns the type name of the TransportStrategy.
  */
  const string&        getStrategyType()                              const { return __strategy_type.getRValue(); }

/** This method returns if the given binding type is supported by the TransportStrategy
  */
  bool isBindingType(const BindingType& binding_type)                 const { if ( (__binding_type.getRValue() & binding_type) != 0) return true; return false;}
  
/** This method returns the binding types of this address
  */
  int  getBindingType()                                               const { return __binding_type.getRValue(); }

  virtual const string& getHostname()                                 const { return __host_name.getRValue(); }
  virtual void          setHostname(const string& hostname)                 { __host_name.setValue(hostname); }

  virtual int  getPort()                                              const { return __port.getRValue(); }
  virtual void setPort(int port)                                            { __port.setValue(port);}
//@}
protected:
  void setBindingType(int binding_type)                                     { __binding_type = binding_type; }
private:
  void registerAttributes() { insertArgument(&__reference); insertArgument(&__strategy_type); insertArgument(&__binding_type); insertArgument(&__host_name); insertArgument(&__port); }
private:
  TValue<NMMReference> __reference;               ///< stores the NMMReference
  TValue<string>       __strategy_type;           ///< stores the typename of the strategy
  TValue<int>          __binding_type;            ///< stores the supported binding types
  TValue<string>       __host_name;
  TValue<int>          __port;
};

}
#endif


