/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_TS_FACTORY_HPP
#define NMM_TS_FACTORY_HPP

#include "TransportStrategy.hpp"

namespace NMM {
  
  class ConnectorAcceptor;
  class Format;
  
  
  /** This abstract class is used to create TransportStrategies
   */
  class TSFactoryObject {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Destructor
     */
    virtual ~TSFactoryObject() {}

    /** Virtual copy constructor
     */
    virtual TSFactoryObject*       copy()           const = 0;
    //@}

    /** @name Factory method
     */
    //@{
    /** This method creates a new TransportStrategy
     */
    virtual TransportStrategy*     createInstance() const = 0;

    /** This method creates a connector, if necessary, and returns a pointer to it.
     */
    virtual ConnectorAcceptor*     requestConnectorAcceptor() const { return NULL; }
    //@}
  };

  /** This class is a Factory for TransportStrategies. To create a TransportStrategy a TSFactoryObject
   * is used. This Factory can be used to register these objects with the type name of the
   * TransportStrategy.
   *
   * This class offers so called "default addresses" for TransportStrategies. These addresses
   * can be used to establish a connection. So an NMMObject or other objects, must not allocate
   * ressources, they can use these addresses.
   *
   * This class ios designed as Singleton
   */
  class TSFactory : public NamedObject {
    friend class CreateUsingNew<TSFactory>;
  public:
    /** @name Factory & Registration methods
     */
    //@{
    /** Register transport strategy.
     * This method registers an TFSactoryObject that can create a TransportStrategy with its typename.
     * @param factory_obj Factory object which creates the instance
     * @param strategy_type Type name of strategy
     * @param format_list List of supported formats
     * @param no_events This transport strategy can not send events
     */
    void registerStrategy(const TSFactoryObject& factory_obj, const string& strategy_type,
                          const list<Format>& format_list, bool no_events);

    /** Unregister transport strategy.
     * This method unregisters the TSFactoryObject and default address of a TransportStrategy with the given typename.
     * @param strategy_type Type name of strategy
     */
    void unregisterStrategy(const string& strategy_type);

    /** Create transport strategy.
     * This method creates a TransportStrategy from the given type name.
     * @param strategy_type Type name of strategy
     */
    TransportStrategy* createStrategy(const string& interface_type);

    /** Set default address for transport strategy.
     * This method sets an default address for a specific strategy type.
     * @param strategy_type Type name of strategy
     * @param address Default address
     */
    void setDefaultAddress(const string& strategy_type, const TSAddress* address);
    //@}

    /** @name Access information
     */
    //@{
    /** Get default address for strategy.
     * This method returns the default address for specifc strategy type.
     * @param strategy_type Type name of strategy
     */
    TSAddress* getDefaultAddress(const string& strategy_type) const;

    /** Is transport strategy available.
     * This method returns true if the given strategy type is available, otherwise false.
     * @param strategy_type Type name of strategy
     */
    bool isStrategyAvailable(const string& strategy_type) const;

    /** Does transport strategy support events.
     * This method returns true if the given strategy type can send events, otherwise false.
     * @param strategy_type Type name of strategy
     */
    bool canStrategySendEvents(const string& strategy_type) const;

    /** Get transport strategy list.
     * This method returns a list of all TransportStrategies.
     */
    list<string> getStrategyList() const;
    //@}

    /** @name ConnectorAcceptors
     */
    //@{
    /** Request %ConnectorAcceptor.
     * This method requests a connector for the given strategy type.
     * @param strategy_type Type name of strategy
     */
    ConnectorAcceptor* requestConnectorAcceptor(const string& strategy_type) const;

    /** Add shared address for transport strategy.
     * This method registers a ConnectorAddress as the default address of the given strategy type.
     * @param strategy_type Type name of strategy
     */
    void addConnectorAddress(const string& strategy_type);
    //@}
    
  private:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor
     */
    TSFactory() ;

    /** Destructor
     */
    ~TSFactory();

    /** Disabled Copy constructor
     */
    TSFactory(const TSFactory&);
    //@}

  private:
    struct InternalStruct {
      InternalStruct()
        : factory_object(NULL),
          default_address(NULL),
          no_events(false)
      {}
      
      const TSFactoryObject* factory_object;
      TSAddress*             default_address;
      bool                   no_events;
    };
    
    typedef map<string, InternalStruct> InternalMap;
    typedef InternalMap::iterator InternalIt;
    typedef InternalMap::const_iterator CInternalIt;
    
    InternalMap __factory_map;          ///< stores the FactoryObjects and registered default addeesses
  };

  typedef Singleton<TSFactory, CreateUsingNew> STSFactory;


  /** This parametrized implementation of TSFactorObject uses the new operator to allocate a
   * TransportStrategy.
   */
  template <class T>
  class TTSFactoryObject : public TSFactoryObject {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Virtual copy constructor
     */
    virtual TTSFactoryObject<T>*  copy()           const { return new TTSFactoryObject<T>(*this); }
    //@} 
    /** @name Factory methods
     */
    //@{
    /** This method creates a new TransportStrategy using the new operator
     */

    /* GCC3.3                  */
    //virtual T*                  createInstance() const { return new T(); }
    virtual TransportStrategy*                  createInstance() const { return new T(); }
    //@}
  };


  /** This subclass of TTSFactoryObject automatically creates a %ConnectorAcceptor object.
   */
  template <class _TransportStrategy, class _ConnectorAcceptor = NMM::ConnectorAcceptor>
  class TConnectorTSFactoryObject : public TSFactoryObject {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor
     */
    TConnectorTSFactoryObject()
      : __connector_acceptor(NULL)
    {}
    
    /** Copy constructor
     */
    TConnectorTSFactoryObject(const TConnectorTSFactoryObject& other) 
      : TSFactoryObject(other),
        __connector_acceptor(NULL)
    {}
    
    /** Virtual copy constructor
     */
    virtual TConnectorTSFactoryObject* copy() const {
      return new TConnectorTSFactoryObject(*this);
    }
    
    /** Destructor
     */
    virtual ~TConnectorTSFactoryObject() {
      if (__connector_acceptor) {
        delete __connector_acceptor;
      }
    }
    //@}
  
    /** @name Factory methods
     */
    //@{
    /** This method creates a new TransportStrategy using the new operator
     */

    /* GCC3.3                  */
    //virtual T*                  createInstance() const { return new T(); }
    virtual TransportStrategy*                  createInstance() const { return new _TransportStrategy(); }
    //@}
  
    /** @name %ConnectorAcceptor Factory
     */
    //@{
    /** This method creates a connector, if necessary, and returns a pointerto it.
     */
    virtual _ConnectorAcceptor* requestConnectorAcceptor() const {
      if (!__connector_acceptor) {
        __connector_acceptor = new _ConnectorAcceptor(_TransportStrategy::getStaticStrategyType());
      }
      return __connector_acceptor;
    }
    //@}
  
  private:
    mutable _ConnectorAcceptor* __connector_acceptor;        ///< %ConnectorAcceptor
  };
  
  /** Format list for transport strategy.
   * Encapsulates a list of formats that should be registered with a transport strategy.
   */
  class TSFormatList {
  public:
    TSFormatList();
    
    inline const list<Format>& getFormatList() const { return __formats; }
    
  protected:
    inline void addFormat(const Format& format) { __formats.push_back(format); }
    
  private:
    list<Format> __formats;
  };
  
  /** Default format list for transport strategy.
   * Contains the top format. Use this list to register a transport strategy which supports all formats.
   */
  class DefaultTSFormatList : public TSFormatList {
  public:
    DefaultTSFormatList();
  };
  
  /** Transport strategy registration object.
   * This object is a wrapper for an instance of a subclass of TSFactoryObject. It is responsible
   * for construction, registration, deregistration and destruction of the TSFactoryObject.
   * Using this wrapper, we make sure that registered objects are always fully constructed.
   */
  template<class T>
  class TTSFactoryRegistrator {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor.
     * @param strategy_type Typename of strategy
     * @param reliable_connect This transport strategy has its own reliable connection setup
     * @param no_events This transport strategy can not send events
     * @param format_list List of supported formats
     */
    TTSFactoryRegistrator(const string& strategy_type,
                          bool reliable_connect = true, bool no_events = false,
                          const TSFormatList& format_list = DefaultTSFormatList())
      : __strategy_type(strategy_type),
        __factory_object()
    {
      /* Register the transport strategy */
      STSFactory::getInstance().registerStrategy(__factory_object, strategy_type, format_list.getFormatList(), no_events);
      
      if (strategy_type == "LocalStrategy") {
        /* Add default LocalAddress for LocalStrategy */
        STSFactory::getInstance().setDefaultAddress("LocalStrategy", new LocalAddress());
      }
      if (!reliable_connect) {
        /* Add default ConnectorAddress for strategy without reliable connection setup */
        STSFactory::getInstance().addConnectorAddress(strategy_type);
      }
    }
    
    /** Destructor
     */
    ~TTSFactoryRegistrator() {
      STSFactory::getInstance().unregisterStrategy(__strategy_type);
    }
    //@}
    
  private:
    string __strategy_type;       ///< Strategy typename
    T __factory_object;           ///< Wrapped factory object
  };

}

#endif
