/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "Usage.hpp"

#include <iostream>

namespace NMM{

  Usage::Usage(){}

  Usage::~Usage(){
    usages.clear();
    usage_mutex.unlock();
  }


  void Usage::setUsage(const IUsage::usage_t &mode, const IUsageNotify* user){
    
    MutexGuard mg(usage_mutex);

    //Test, if we already have someone with the given usage
    if(usages.find(mode) == usages.end()){
      //if not, we add the new usage to the map. Since a map is sorted, the usage is inserted at the correct position

      //But firstget the current Master to notify him, that he is no longer the master
      map<IUsage::usage_t, list<usermode_t> >::reverse_iterator r = usages.rbegin();
      if(r != usages.rend()){
        list<usermode_t>& userList = r->second;
        if(!userList.empty()){
          //the master is the first user in the list of the usage with the highest priority
          IUsageNotify* tmp = const_cast<IUsageNotify*>(userList.front().second);
          tmp->becomeSlave();
        }
      }
      
      list<usermode_t> l;
      l.push_back(usermode_t(0,user));
      usages[mode] = l;
    }else{
      //otherwise we have to insert the user 
      list<usermode_t> &l = usages[mode];

      
      for(list<usermode_t>::iterator i = l.begin(); i != l.end(); i++){
        if(i->second->getParentObject() == user->getParentObject()){
          //if the user already uses this node, we increase its usage-counter
          (i->first)++;
          return;
        }
      }
      //if the user sets its usage for the first time, we simply add it to the back of the users-list for this usage
      l.push_back(usermode_t(0, user));
    }
  }

  void Usage::unsetUsage(const IUsage::usage_t &mode, const IUsageNotify* user){
    MutexGuard mg(usage_mutex);

    list<usermode_t> &l = usages[mode];

    //get the user-list for the given mode
    for(list<usermode_t>::iterator i = l.begin(); i != l.end(); i++){
      if(i->second->getParentObject() == user->getParentObject()){
        //we found the user, wich releases this node
        if((i->first) == 0){
          //if the user had this node reqested one time only, we remove it from the list
          l.erase(i);
          
          //if it was the last user for the usage, we remove the usage from our map
          if(usages[mode].empty()){
            usages.erase(mode);
          }
          
          
          //If there are users left for the usage, we notify the new Master
          map<IUsage::usage_t, list<usermode_t> >::reverse_iterator r = usages.rbegin();
          if(r != usages.rend()){
            list<usermode_t>::iterator u = r->second.begin();
            if(u != r->second.end()){
              IUsageNotify* tmp = const_cast<IUsageNotify*>(u->second);
              if(tmp){
                tmp->becomeMaster();
              }
            }else{
            }
          }
          
          
        }else{
          //if the user has requested this node more than one time, we only decrease its usage-counter
          //the user will keep its status
          (i->first)--;
        }
        return;
      }
    }
  }


  IUsage::usage_t Usage::getUsage(){
    MutexGuard mg(usage_mutex);
    if(!usages.empty()){
      //the current Usage is the application with the highest priority. 
      //Since the map is sorted, the current usage is at the end of the map
      map<IUsage::usage_t, list<usermode_t> >::reverse_iterator i = usages.rbegin();
      return i->first;
    }
    return IUsage::UNKNOWN;
  }

  int Usage::getRank(const IUsage::usage_t &mode, const IUsageNotify* user){
    MutexGuard mg(usage_mutex);
    list<usermode_t> &l = usages[mode];
    int counter = 0;
    
    for(list<usermode_t>::iterator i = l.begin(); i != l.end(); i++){
      if(i->second->getParentObject() == user->getParentObject()){
        return counter;
      }
      counter++;
    }

    return counter;
  }

  Result Usage::isMaster(const IUsage::usage_t &mode, const IUsageNotify* user){
    //determine the current usage of the node
    IUsage::usage_t currentUsage = getUsage();
    
    //if the new usage will have a higher priority than the current one, the user will be master
    if(currentUsage <  mode) return SUCCESS;
    //if the new usage will have a lower priority than the current one, the user will be slave
    if(currentUsage >  mode) return FAILURE;

    map<IUsage::usage_t, list<usermode_t> >::reverse_iterator r = usages.rbegin();
    if(r != usages.rend()){
      list<usermode_t>& userList = r->second;
      if(!userList.empty()){
        //the master is the first user in the list of the usage with the highest priority
        if(userList.front().second->getParentObject() == user->getParentObject())
          return SUCCESS;
        
      }
    }
    return FAILURE;
      
    try{
      //determine the current usage of the node
      IUsage::usage_t currentUsage = getUsage();
      
      //if the new usage will have a higher priority than the current one, the user will be master
      if(currentUsage <  mode) return SUCCESS;
      //if the new usage will have a lower priority than the current one, the user will be slave
      if(currentUsage >  mode) return FAILURE;
      //if the new usage is equal to the current ...
      if(currentUsage == mode){
        //... we determine the rank of the new user ...
        int rank = getRank(mode, user);
        //... and if the rank will be 0, the new user will be master
        if(rank == 0) return SUCCESS;
      } 
    }catch(const Exception &e){
    }catch(...){
    }

    return FAILURE;
  }
}
