/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "FormatExceptions.hpp"

namespace NMM
{
  FormatException::FormatException(const char * message)
    : Exception( string("FormatException : ") + message)
  {
    m_message = message;
  }
  
  void FormatException::showException()
  {
    GLOBAL_DEBUG_STREAM2( "Message : " << m_message << endl);
  }

  ParameterException::ParameterException(const char * parameter_name)
  {
    m_parameter_name = parameter_name;
  }

  ParameterException::ParameterException(const char * parameter_name, Parameter * parameter)
  {
    m_parameter_name = parameter_name;
    m_parameter      = parameter;
  }

  ParameterException::ParameterException(const char * message, const char * parameter_name) : FormatException(message)
  {
    m_parameter_name = parameter_name;
  }

  ParameterException::ParameterException(const char * message, const char * parameter_name, Parameter *parameter) : FormatException(message)
  {
    m_parameter_name = parameter_name;
    m_parameter      = parameter;
  }

  void ParameterException::showException()
  {
    FormatException::showException();
    GLOBAL_DEBUG_STREAM2( "Parameter : " << m_parameter_name << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter : " << m_parameter << endl);
  }

  ParameterExceptionFalseType::ParameterExceptionFalseType(const char * parameter_name) : ParameterException(parameter_name)
  {
  }

  ParameterExceptionFalseType::ParameterExceptionFalseType(const char * parameter_name,
							   Parameter * parameter_false) : ParameterException(parameter_name)
  {
    m_parameter_false = parameter_false;
  }

  ParameterExceptionFalseType::ParameterExceptionFalseType(const char * parameter_name,
							   Parameter::PType first, 
							   Parameter::PType second) : ParameterException("ParameterExceptionFalseType",
													 parameter_name)
  {
    m_original = first;
    m_fail     = second;
  }

  void ParameterExceptionFalseType::showException()
  {
    ParameterException::showException();
    GLOBAL_DEBUG_STREAM2( "Parameter Types are not identical !" << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter Type  : " << m_original << endl);
    GLOBAL_DEBUG_STREAM2( "!=   Fail Type  : " << m_fail << endl);
    GLOBAL_DEBUG_STREAM2( "False Parameter : " << m_parameter_false << endl);
  }

  ParameterExceptionWeightRangeError::
  ParameterExceptionWeightRangeError(const char * parameter_name,
				     float weight) : ParameterException("ParameterExceptionWeightRangeError",
									parameter_name)
  {
    m_weight = weight;
  }
  
  void ParameterExceptionWeightRangeError::showException()
  {
    ParameterException::showException();
    GLOBAL_DEBUG_STREAM2( "The value of the weight - " << m_weight << " - is out of the range [0,1] " << endl);
  }

  ParameterExceptionNotExist::ParameterExceptionNotExist(const char * parameter_name) : ParameterException("ParameterExceptionNotExist",parameter_name)
  {
  }

  void ParameterExceptionNotExist::showException()
  {
    ParameterException::showException();
    GLOBAL_DEBUG_STREAM2( "Parameter " << m_parameter_name << " doesn't exist ! " << endl);
  }

  GraphManagerException::GraphManagerException(const char * message) : FormatException(message)
  {
  }

  void GraphManagerException::showException()
  {
    FormatException::showException();
  }

  GraphManagerExceptionDeviceNotExist::GraphManagerExceptionDeviceNotExist(const char * device_name) : GraphManagerException("Device is not in the graph")
  {
    m_device_name = device_name;
  }

  void GraphManagerExceptionDeviceNotExist::showException()
  {
    GraphManagerException::showException();
    GLOBAL_DEBUG_STREAM2( "The device " << m_device_name << " is not a valid device in the graph " << endl);
  }

  XmlErrorException::XmlErrorException(const char * file_name) : FormatException("Error in reading the xml_file")
  {
    m_file_name = file_name;
  }

  void XmlErrorException::showException()
  {
    FormatException::showException();
    GLOBAL_DEBUG_STREAM2( " [ " << m_file_name << " ] " << endl);
  }

  UsergraphException::UsergraphException(const char * message) : FormatException(message)
  {
  }

  void UsergraphException::showException()
  {
    FormatException::showException();
  }

//  UsergraphExceptionNoPath::UsergraphExceptionNoPath(user_node_list & list) : UsergraphException ("No Connection between ")
//  {
//    m_list  = list;
//  }

//  UsergraphExceptionNoPath::UsergraphExceptionNoPath(const char * begin, const char * end) : UsergraphException ("No Connection between ")
//  {
//    m_begin = begin;
//    m_end = end;
//  }

//  UsergraphExceptionNoPath::UsergraphExceptionNoPath(const char * destination) : UsergraphException ("No Connection to")
//  {
//    m_begin = "";
//    m_end = destination;
//  }
  
//  void UsergraphExceptionNoPath::showException()
//  {
//    UsergraphException::showException();
//    if (!m_list.empty()) {
//      for (user_node_list::const_iterator itr = m_list.begin() ; itr != m_list.end(); itr++) {
//	GLOBAL_DEBUG_STREAM2( (*itr)->getGuid() << endl);
//      }
//    }
//    else {
//      if (strlen(m_begin) > 0) {
//      GLOBAL_DEBUG_STREAM2( m_begin << " and " << m_end << endl);
//      }
//      else {
//	GLOBAL_DEBUG_STREAM2( m_end << endl);
//      }
//    }
//  }
}
