/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_PARAMETER_HPP
#define _NMM_PARAMETER_HPP

#include "definition.hpp"

namespace NMM 
{
  class IValueStream;
  class OValueStream;
  class Format;
  class R_INT;
  class R_FLOAT;

  /** 
    This is the abstract superclass for every kind of parameter.
  */

  class Parameter 
  {
  public:
  
    //! The enum PType represents the type of the parameter.
    enum PType{	INT32  = 0,    /*!< 32-Bit Integer. */
		FLOAT,    /*!< Float */
		STRING,   /*!< C++ String */
		FPOINTER, /*!< Format Pointer for recursive Formats */
		R_INT32,  /*!< Range of 32-Bit Integer */
		R_FLOAT,  /*!< Range of Floats */
		WILDCARD  /*!< Wildcart */
    };
    
  protected:
    
    //! The name of the Parameter.
    std::string m_parameter_name; 	
    
    //! The type of the parameter. See in enum PType for a description.
    PType m_parameter_type; 

    //! The weighting factor of the parameter.
    float m_parameter_weight;
    
  public:
    
    /**
       The default constructur.
       Note: You have to specify always a name for the parameter.
       \param string parameter_name: The parameter-name.
    **/
    Parameter(const std::string & parameter_name);
    
    /**
       The default destructor.
    **/
    virtual ~Parameter();
    
    /** 
	Returns the name of the parameter.
	\return string: name of the parameter.
    **/
    virtual const std::string & getParameterName() const;
    
    /**
       Sets the name of the parameter.
       \param string parameter_name: the name of the parameter.
    **/
    virtual void setParameterName(const std::string & parameter_name);

    /**
       Returns the type of the parameter.
       \return PType: type of the parameter.
    */
    virtual Parameter::PType getParameterType() const;
    
    /**
       Adds the value specified in pValue to the set or set of range 
       of this parameter. This is an abstract function an must be 
       implemented in every inherited class to cast the void-ptr in
       the proper type.
       \param void *: pValue
    **/
    virtual void addValue(void * /*pValue*/) {};
    
    /**
       Adds an int32_t value to the set or set of ranges.
       \param int32_t value: an integer value.
       \param float weight: the weight of the value.
    **/
    virtual void addValue(int32_t /*value*/, float /*weight = DEFAULT_WEIGHT*/){};
    
    /**
       Adds a float value to the set or set of ranges.
       \param float value: a float value.
       \param float weight: the weight of the value.
    **/
    virtual void addValue(float /*value*/, float /*weight = DEFAULT_WEIGHT*/){};

    /**
       Adds a string value to the set or set of ranges.
       \param string value: a string value.
       \param float weight: the weight of the value.
    **/
    virtual void addValue(const std::string & /*value*/, float /*weight = DEFAULT_WEIGHT*/){};
    
    /**
       Adds a Format-pointer to the set of values.
       \param Format * value: a pointer to an already existant format.
       \param float weight: the weight of the format.
    **/
    virtual void addValue(Format * /*value*/, float /*weight = DEFAULT_WEIGHT*/){};
    
    /**
       Adds the value specified in pValue with the appropriate weight.
       \param void * pValue: the value.
       \param float weight: the weight of this value.
    **/
    virtual void addValue(void * /*pValue*/, float /*weight*/) {};

    /**
       Adds an integer-range from lValue to hValue with the specified range of weights to the set of ranges.
       \param int32_t lValue: the lower bound of the range.
       \param int32_t hValue: the upper bound of the range.
       \param float lWeight: the weight of the lower bound.
       \param float hWeight: the weight of the upper bound.
    **/
    virtual void addValue(int32_t /*lValue*/,
			  int32_t /*hValue*/,
			  float /*lWeight = DEFAULT_WEIGHT*/,
			  float /*hWeight = DEFAULT_WEIGHT*/){};

    /**
       Adds an float-range from lValue to hValue with the specified range of weights to the set of ranges.
       \param float lValue: the lower bound of the range.
       \param float hValue: the upper bound of the range.
       \param float lWeight: the lower bound of the range.
       \param float hWeight: the upper bound of the range.
    **/
    virtual void addValue(float /*lValue*/,
			  float /*hValue*/,
			  float /*lWeight = DEFAULT_WEIGHT*/,
			  float /*hWeight = DEFAULT_WEIGHT*/){};

    /**
       Adds an already existing integer-range to the set of ranges.
       \param R_INT * value: the integer-range.
    **/
    virtual void addValue(R_INT * /*value*/){};

    /**
       Adds an already existing float-range to the set of ranges.
       \param R_FLOAT * value: the float-range.
    **/
    virtual void addValue(NMM::R_FLOAT * /*value*/){};

    /**	
	Debugging method, prints out the Parameter-Name.
    **/
    virtual void showParameterName() const;
    
    /** 
	Debugging method, prints out every Parameter-Value and is
	reimplemented in every inherited class (deprecated).
    **/
//     virtual void showParameter() const = 0;
    virtual void showParameter() const {};
    
    /** 
	Output method.
        prints out every Parameter-Value to the given ostream and is
	reimplemented in every inherited class.
    **/
    virtual std::ostream& showParameter(std::ostream& os) const { return os; }
    
    /** 
	Output method.
        prints out every Parameter-Value to the given ostream and is
	reimplemented in every inherited class in the old verbose format (deprecated).
    **/
    virtual std::ostream& showParameterVerbose(std::ostream& os) const { return os; }

    /**
       Checks if the parameter pValue is included in the set of values of this parameter.
       Abstract function which is implemented in all inherited classes.
       \param void *pValue: the value that will be checked.
       \return Returns true if the value pValue is included in the set, otherwise false.
    **/
    virtual bool isValueIncluded(void *pValue) const = 0;
//     virtual bool isValueIncluded(void *pValue) const {return true;};
    
    /**
       Checks if an integer value is included.
       \param int32_t value: integer value that will be checked.
       \return true if the value is included, otherwise false.
    **/
    virtual bool isValueIncluded(int32_t value) const;

    /**
       Checks if a float value is included.
       \param float value: float value that will be checked.
       \return true if the value is included, otherwise false.
    **/
    virtual bool isValueIncluded(float value) const;

    /**
       Checks if a string value is included.
       \param string value: string value that will be checked.
       \return true if the value is included, otherwise false.
    **/
    virtual bool isValueIncluded(const std::string & value) const;

    /**
       Checks if a format is included.
       \param Format * value: format pointer that will be checked.
       \return true if the format is included, otherwise false.
    **/
    virtual bool isValueIncluded(const Format * value) const;
    
    /**
       Checks if the range from lValue to hValue is included in the set of ranges of this parameter.
       Note: the functions returns true if there is also an intersection between the two ranges!
       \param void *lValue: the lower bound of the range
       \param void *hValue: the upper bound of the range
       \return returns true if the range is included in the set of ranges, otherwise false.
    **/
    virtual bool isValueIncluded(void *lValue,void *hValue) const;

    /**
       Checks if the range from lValue to hValue is included.
       \param float lValue: the lower bound of the range.
       \param float hValue: the upper bound of the range.
       \return true if the range is included otherwise false.
    **/
    virtual bool isValueIncluded(float lValue, float hValue) const;

    /**
       Checks if the range from lValue to hValue is included.
       \param int32_t lValue: the lower bound of the range.
       \param int32_t hValue: the uppper bound of the range.
       \retrun true if the range is included, otherwise false.
    **/
    virtual bool isValueIncluded(int32_t lValue, int32_t hValue) const;

    /**
       Checks if the range from lValue to hValue has an intersection with a range.
       \param float lValue: the lower bound of the range.
       \param float hValue: the upper bound of the range.
       \return true if the range has an intersetion, otherwise false.
    **/
    virtual bool isRangeMatching(float lValue, float hValue) const;

    /**
       Checks if the range form lValue to hValue has an intersection with a range.
       \param int32_t lValue: the lower bound of the range.
       \param int32_t hValue: the upper bound of the range.
       \retrun true if the range has an intersection, otherwise false.
    **/

    virtual bool isRangeMatching(int32_t lValue, int32_t hValue) const;

    /**
       This function iterates over the values and
       returns the next value or the next range of the set. 
       At the end of the list, the function returns NULL. Must be 
       reimplemented in every inherited class.
    **/
//     virtual void * nextValue() const = 0;
    virtual void * nextValue() const {return 0;}; /* Should work ?!?*/
    
    /**
       This function returns the first value stored in the
       appropriate set and increases the internal value counter.
       If you want to iterate over all values of a parameter, this
       function guaranties to start always with the first value. After a
       call of this function, just use the nextValue() function above.
    **/
//     virtual void * getFirstValue() const = 0;
    virtual void * getFirstValue() const {return 0;};/* Should work ?!?*/

    /**
       This functions makes a copy of the current parameter with all
       values and weights and returns a pointer to the new parameter.
       \return Parameter * : the new copy of the parameter.
    **/
    virtual Parameter *duplicate() const = 0;
//     virtual Parameter *duplicate() const ;
    
    /**
       Returns the weight of the argument pValue
       \param pValue: the desired value.
       \return float: the weight of the value.
    **/
//     virtual float getWeight(void *pValue) const=0;
    virtual float getWeight(void */*pValue*/) const{return 1.0;};/* Should work ?!?*/
    
    /**
       Checks if the given argument has the same co-domain.
       \param Parameter *: the parameter which will be compared with.
       \return true if the given parameter has the same co-domain,
       otherwise false.
    **/
//     virtual bool isParameterIdentical(const Parameter * reference) const = 0;
    virtual bool isParameterIdentical(const Parameter * /*reference*/) const {return false;};/* Should work ?!?*/

    /**
       Checks if the given argument has an intersection in the co-domain of the values.
       \param Parameter *reference: the parameter that will be matched.
       \return true if the parameters are matching, otherwise false.
     **/
    virtual bool isParameterMatching(const Parameter */*reference*/) const {return false;};/* Should work ?!?*/
//     virtual bool isParameterMatching(const Parameter *reference) const = 0;

    /**
       Returns the number of values stored in the data-structures.
       return uint: the number of values.
    **/
//     virtual uint getNumberOfValues() const  = 0 ;
    virtual uint getNumberOfValues() const {return 0;} ;/* Should work ?!?*/
    
    /**
     This function creates an intersection between the current parameter
     and given argument reference.
     \param Parameter *reference: a parameter which will be intersected with the current parameter.
     \return Parameter *: a pointer to the new created parameter.
    **/
    virtual Parameter * getIntersection(const Parameter *reference) const = 0 ;
//     virtual Parameter * getIntersection(const Parameter *reference) const {} ;

    /**
       This functions computes a standardized weight for the included
       co-domain, depending on the parameter-type.
       \return float: the standardized weight.
    **/
//     virtual float getStandardizedWeight() const = 0;
    virtual float getStandardizedWeight() const {
//       cout << m_parameter_name << " " << m_parameter_type << " " << this << endl;
      return DEFAULT_WEIGHT * m_parameter_weight;
      
    };
//       cout << typeid(this).name() << " haloo " << endl;
//       return DEFAULT_WEIGHT;};

    /**
       Sets the weight of the parameter.
       \param float weight: the weight of the parameter.
    **/
    virtual void setParameterWeight(float weight);

    /**
       Returns the weight of the parameter.
       \return float: the weight of the parameter.
    **/
    virtual float getParameterWeight() const;

    friend std::ostream & operator << (std::ostream & os, const Parameter & reference);

    virtual void fromStream(IValueStream& ) = 0;
    virtual void toStream(OValueStream&) const  = 0;
  };
}
#endif
