/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_R_INT_HPP
#define _NMM_R_INT_HPP

#include "definition.hpp"

namespace NMM {

/**
   This class implements an INT32 range.
*/

  class R_INT {

  private:	
    //! The lower bound of the range.
    int32_t m_l_value;
    //! The upper bound of the range.
    int32_t m_h_value;
    //! The lower weight of the range.
    float m_l_weight;
    //! The upper weight of the range.
    float m_h_weight;

  public:
    /**
       The default Constructor.
    */
    R_INT() {};
    
    /**
       Constructor with additional parameters wich set directly 
       the member-variables m_l_value and m_h_value.
       Additionally the weighting factor can be specified in the
       parameters l_weight and h_weight.
       \param int32_t lParam: the lower bound of the range.
       \param int32_t hParam: the upper bound of the range.
       \param float l_weight: the lower weight of the range.
       \param float h_weight: the upper weight of the range.
    */
    R_INT(int32_t lParam, int32_t hParam, float l_weight = DEFAULT_WEIGHT, float h_weight = DEFAULT_WEIGHT);
		
    //! The default Destructor
    ~R_INT();
    
    /**
       Returns the lower bound of the range.
       \return int32_t: the lower bound of the range.
    */
    int32_t getLowValue() const;
        
    /**
       Returns the upper bound of the range.
       \return int32_t: the upper bound of the range.
    */
    int32_t getHighValue() const;
    
    /**
       Returns the lower weight of the range.
       \return float: the lower weight of the range.
    **/
    float getLowWeight() const;

    /**
       Returns the upper weight of the range.
       \return float: the upper weight of the range.
    **/
    float getHighWeight() const;

    /**
       Sets the lower bound of the range.
       \param int32_t: the lower bound of the range.
    */
    void setLowValue(int32_t lParam);

    /**
       Sets the upper bound of the range.
       \param int32_t: the upper bound of the range.
    */
    void setHighValue(int32_t hParam);
    
    /**
       Sets the lower weight of the range.
       \param float: the lower weight of the range.
     **/
    void setLowWeight(float lWeight);

    /**
       Sets the upper weight of the range.
       \param float: the upper weight of the range.
    **/
    void setHighWeight(float hWeight);

    /**
       Debugging method. Prints out the values of the range.
    */
    void showValue();

    /**
       Returns the standardized weight for this range.
       \return float: the standardized weight.
    **/
    float getStandardizedWeight();
    
    /**
       Returns if the value is included in the range.
       \param int32_t value: the checked value.
       \return true if the value is included otherwise false.
    **/
    bool isValueIncluded(int32_t value);

    /**
       Returns if the range from lValue to hValue is included in the range.
       \param int32_t lValue: the lower bound of the range.
       \param int32_t hValue: the upper bound of the range.
       \return true if the range is included, otherwise false.
    **/
    bool isRangeIncluded(int32_t lValue, int32_t hValue);

    /**
       Returns if the range from lValue to hValue has an intersection with the range.
       \param int32_t lValue: the lower bound of the range.
       \param int32_t hValue: the upper bound of the range.
       \returns true if the range has an intersection otherwise false.
    **/
    bool isRangeMatching(int32_t lValue, int32_t hValue);

    /**
       Returns if the range from lValue to hValue is identical with the current instance.
       \param int32_t lValue: the lower bound of the range.
       \param int32_t hValue: the upper bound of the range.
       \returns true if the range is identical with the current instance, otherwise false.
    **/
    bool isRangeIdentical(int32_t lValue, int32_t hValue) const;
    
    friend bool operator == (const R_INT & lhs, const R_INT & rhs);

    friend ostream & operator << (ostream & os, const R_INT & reference);
  };
}

#endif
