/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_BUFFEREDSINK_HPP
#define NMM_BUFFEREDSINK_HPP NMM_BUFFEREDSINK_HPP

#include "nmm/utils/thread/Thread.hpp"
#include "nmm/utils/NamedObject.hpp"

#include "nmm/base/ISignalReceiver.hpp"

#include <live/MediaSink.hh>

namespace NMM {
  
  class IExternalBufferManager;
  class IMessageReceiver;
  class IMessageSender;
  class Buffer;
  
  /**
   *  @brief Buffered media sink
   *
   *  This class is a liveMedia sink which produces NMM buffers. It also reproduces the NMM timestamps
   *  from the RTP timestamps, the RTP sequence numbers and the RTP marker bit.
   *
   *  The buffered sink has no queue. It forwards messages directly to the message receiver, which is
   *  usually a Jack. This avoids an unnecessary context switch.
   */
  class BufferedSink
    : public MediaSink,
      public NamedObject,
      public ISignalReceiver
  { 
  public:
    /**
     *  @brief Create new BufferedSink
     *  @param env The liveMedia UsageEnvironment
     *  @param receiver Receiver to receive the NMM buffers
     *  @param sender Message sender (should be the RTPStrategy)
     *  @param buffer_manager External buffer manager to allocate buffers from
     *  @param buffer_size Size of buffers to allocate
     *  @param frame_size Size of a frame, must be less than or equal to buffer_size
     *
     *  The frame size is the minimum count of bytes that must be free in the buffer before a new one is
     *  allocated (the internal buffer may contain more than one frame)
     */
    static BufferedSink* createNew(UsageEnvironment& env,
                                   IMessageReceiver* receiver, IMessageSender* sender,
                                   IExternalBufferManager* buffer_manager,
                                   unsigned buffer_size, unsigned frame_size);
    
    /**
     *  @brief Flush the stream
     *
     *  This method sets a flag which tells liveMedia to stop getting frames. It
     *  must be called to unblock the thread before stopping the liveMedia flow graph.
     */
    void flush();
    
    /**
     *  @brief Signal handler
     *
     *  If @c FREE_QUEUE_SLOT is received from the message receiver, then this method notifies
     *  the livemedia thread to unblock and continue getting frames.
     */
    virtual void receiveSignal(Signal);
    
  protected:
    /**
     *  @brief Constructor
     *
     *  This constructor is called by createNew.
     */
    BufferedSink(UsageEnvironment&, IMessageReceiver*, IMessageSender*,
                 IExternalBufferManager*, unsigned, unsigned);
    
    /**
     *  @brief Destructor
     */
    virtual ~BufferedSink();
    
    /**
     *  @brief liveMedia after getting frame handler
     */
    static void afterGettingFrame(void*, unsigned, unsigned, struct timeval, unsigned);
    
  private:
    /**
     *  @name MediaSink functions
     */
    //@{
    virtual Boolean continuePlaying();
    //@}
    
    /**
     *  @brief Called by afterGettingFrame
     *  @param frameSize Frame size in bytes
     *  @param presentationTime Presentation time
     */
    virtual void handleFrame(unsigned frameSize, struct timeval presentationTime);
    
  private:
    /** Current buffer to write to
     */
    Buffer* __current_buffer;
    
    /** The buffer size.
     */
    size_t __buffer_size;
    
    /** The RTP frame size.
     */
    size_t __frame_size;
    
    /** Initial RTP timestamp value (to compute offset).
     */
    struct timeval __start_time;
    
    /** Receiver is currently blocked.
     */
    bool __blocked;
    
    /** Flush method was called.
     */
    bool __flush;
    
    /** Internal thread mutex.
     */
    ThreadMutex __internal_mutex;

    /** Internal thread condition.
     */
    ThreadCondition __internal_cond;
    
    /** Buffer manager.
     */
    IExternalBufferManager* __buffer_manager;
    
    /** Message receiver.
     */
    IMessageReceiver* __receiver;

    /** Message sender.
     */
    IMessageSender* __sender;
    
    /** No RTP frames received yet.
     */
    bool __first_rtp_frame;
    
    /** RTP sequence number of first RTP frame.
     */
    uint32_t __first_rtp_seqno;
    
    /** RTP sequence number of last RTP frame.
     */
    uint32_t __last_rtp_seqno;
  };
  
}

#endif
