/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_CLIENTREGISTRY_HPP
#define NMM_CLIENTREGISTRY_HPP

#include "nmm/utils/NamedObject.hpp"
#include "nmm/utils/thread/ThreadedObject.hpp"

#include "nmm/interfaces/base/registry/IServerRegistry.hpp"
#include "nmm/interfaces/base/registry/IPluginManager.hpp"


namespace NMM {
class IServerRegistry;
class ISourceSynchronizer;
class ISinkSynchronizer;
class ISession;

typedef pair<string, int> RegistryName;
typedef map<RegistryName, list<Response> > ResponseMap;

/** \class ClientRegistry
  * The ClientRegistry allows an application to access several ServerRegistries and reserves their
  * Nodes. It allows an local optimization too. If a node is requested, the ClientRegistry allocates the
  * node in the application address-space to avoid an additional interprocess communication between
  * the application and the node.
  *
  */
class ClientRegistry : public NamedObject, public ThreadedObject {
  class ReservationStruct;

public:
  class RegistryStruct;
  class ISessionStruct;


/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  ClientRegistry();

/** Destructor
  */
  virtual ~ClientRegistry();
//@} 

/** @name Request & Release methods
  */
//@{
/** 
  * These methods serach any Server-Registries for specific Nodes with specific requirements.
  * The requirements are specified in a NodeDescrtiption that is send as first argument.
  *
  */
//@{
/**  This method searches all Server-Registry for Nodes that match
  */
  virtual  list< Response >  initRequest(const NodeDescription& nd);

/**  This method searches all Server-Registry for Nodes that match and returns the fastest host in hostname
  */
  virtual  list< Response >  initDistributedRequest(const NodeDescription& nd, string& hostname);

/** This method searches specific Server-Registries  for Nodes with specific requirements.
  * The requirements are specified in a NodeDescrtiption that is send as first argument.
  * 
  * \param nd NodeDescription with the Node's requirements
  * \param registry_list list with the name of Server-Registries that should process the request
  */
  map<RegistryName, list<Response> > initRequest(const NodeDescription& nd, const list<RegistryName>& registry_list);
//@}

/** @name Request & Release methods
  */
//@{
/** This method tries to request one of the nodes in the Response object.
  * The response object contains a list with ObjectID's. This method starts to request the first Node in
  * this list. If this request fails this method tries to request another node in this list in the list's
  * order. If a request is successfully it returns an INode, otherwise a exception is thrown
  * \param resp The response object with list of ObjectID's.
  */
  virtual INode* requestNode(const Response& resp, NodeDescription* = 0)
          throw (const RegistryException& );

/** This method tries to request one of the nodes in the list of Response objects.
  * The response object contains a list with ObjectID's. This method starts to request the first Node in
  * this list. If this request fails this method tries to request another node in this list in the list's
  * order. If a request is successfully it returns an INode, otherwise a exception is thrown
  * \param resp The response object with list of ObjectID's.
  */
  virtual INode* requestNode(const list<Response>& resp, NodeDescription* = 0)
          throw (const RegistryException& );

/** This method tries to request a specific Node from a specific registry.
  * \param reg_id The RegistryID of the registry that manage the node
  * \param node_id The node's ObjectID.
  */
  virtual INode* requestNode(const RegistryID& reg_id, const ObjectID& node_id, NodeDescription* = 0)
          throw (const RegistryException& );

/** requests an NMMObject from given ObjectDescription.
  */
  virtual Interface* requestNMMObject(const ObjectDescription& );

/** releases a requested SourceSynchronizer
  */
  virtual void releaseNMMObject(const Interface&);

/** This method releases a requested node.
  * \param i_node The interface of the requested Node
  */
  virtual void releaseNode(INode& i_node);
//@}

/** @name access control methods
  */
//@{
/** This method enables the access-control to all Server-Registries that can be accessed through this
  * Client-Registry. If the access-control is enabled no other application can access these
  * Server-Registries. If this method fails, an exception is thrown.
  */
  virtual void initTransaction()
    throw (const RegistryException&);

/** This method disables the access-control.
  */
  virtual void releaseTransaction();
//@}

/** @name get and add Server-Registries
  */
//@{
/** This method retuns a map with all Server-Registries that can be accessed through this Client-Registry
  */
  const map<RegistryName, RegistryStruct>& getRegistryList()     const { return __named_registry; }

/** This method retuns a map with all Server-Registries that can be accessed through this Client-Registry
  */
  map<RegistryName, RegistryStruct>& getRegistryList()                 { return __named_registry; }

/** This method returns an interface to a server registry located on given host. The interface is
  * managed by the ClientRegistry. So the caller of this method must not delete it.
  */
  IServerRegistry* getServerRegistry(const RegistryName& registry_name);

/** This method returns an interface to a server registry located on given host. The interface is
  * managed by the ClientRegistry. So the caller of this method must not delete it.
  */
  const IServerRegistry* getServerRegistry(const RegistryName& registry_name) const ;

/** This method returns the RegistryName, which is the hostname and port of a ServerRegistry, that manages the object
  * of an given interface.  
  */
  const RegistryName& determineRegistryName(Interface& interface) const;
  
/** This method returns an interface to the plugin manager of a given host. The interface is
  * managed by the ClientRegistry. So the caller of this method must not delete it.
  */
  IPluginManager* getPluginManager(const RegistryName& registry_name);

/** This method return true, if plugins of a server registry which runs on the same system
  * may create the plugin in the address space of the ClientRegistry to avoid network communication.
  */
  bool isOptimizationAllowed(const IServerRegistry* ) const;

/** This method enables/disables the local optimization, which allows to create plugins of a server
  * registry running on the same system in the address space of the ClientRegistry to avoid network
  * communication. If the server registry runs on a different host, this flag is ignored.
  */
  void setOptimization(bool b, const RegistryName& registry_name = RegistryName("", ClientRegistry::default_port));

/** This method adds an additional Server-Registry that should be used by this Client-Registry.
  * If the Server-Registry can not be accessed an exception is thrown
  * \param address This argument specifies the address of the Server-Registry. The NMMReference can
  * be 0. In this case the Client-Registry tries to find the service "ServerRegistry"
  * \param registry_name This string is used to identify the new Server-Registry in the application.
  * \param is_optimization_allowed If this flag is true, the Client-Registry tries to allocate Nodes
  * in this address space if the Server-Registry is on the same host
  */
  void addRegistry(const TSAddress& address, bool is_optimization_allowed = false)
    throw (const BindingException&);
//@}

/** @name node specific methods
  */
//@{
/** This mehtod requests the NodeDescription from a specifc node.
  * \param reg_id The ID of the Registry that manages the node.
  * \param node_id The ObjectID of the specific Node
  */
  virtual  NodeDescription getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id);
//@}

/** @name reset methods
  */
//@{
/** This method releases ALL reserved nodes by all Server-Registry that can be accessed through this
  * Client-Registry and completly resets them.
  * This method is for debug-reason.
  */
  virtual void resetInternalReservation();

/** This method releases ALL reserved nodes from a specific user by all Server-Registry and completly
  * resets them. This method is for debug-reason.
  */
  virtual void resetSpecificInternalReservation(const UserInfo& user_info);
//@}

/** @name GraphDescription specific methods
  */
//@{
  void requestGraph(GraphDescription& gd, const RegistryName& registry_name = RegistryName("", ClientRegistry::default_port))
    throw (const RegistryException&);
        
  INode* requestCompositeNode(GraphDescription& gd)
        throw (const RegistryException&);

  void releaseGraph(GraphDescription& gd);

  void registerGraphDescription(GraphDescription& gd)
    throw (const RegistryException&);
//@}

  RegistryName getRegistryName( const IServerRegistry& reg_name ) const;
public:
  /* Stores the well known port to the ServerRegistry */
  static const int default_port;
protected:
/** This method refreshs reserved nodes and access-control.
  */
  virtual void run();

public: //methods

/* Begin> Some internal typedefs */
  typedef map<RegistryName, RegistryStruct>::iterator         RegistryIt;
  typedef map<RegistryName, RegistryStruct>::const_iterator   CRegistryIt;

  typedef map<NMMObject*, ReservationStruct>::iterator        ReservationIt;
  typedef map<NMMObject*, ReservationStruct>::const_iterator  CReservationIt;

  typedef map<ProxyObject*, ISessionStruct>::iterator         ISessionIt;
  typedef map<ProxyObject*, ISessionStruct>::const_iterator   CISessionIt;
/* End< */
private: //methods
/** inits a transaction on a specific server_registry.
  */
  void initTransaction( IServerRegistry* server_registry );
  
/** This method finds the Server-Registry that manages the Registry with given ID
  */
  RegistryIt findRegistryStruct(const RegistryID& );

/** This method creates a node from the ServerRegistry on this system in the same addresse space
  * as the Client-Registry runs
  */
  INode* createLocalNode(RegistryIt it, const RegistryID& reg_id, const ObjectID& node_id, NodeDescription* nd = 0)
    throw (const RegistryException&);

/** This method creates a distribued node in the address space of the Server-Registry
  */
  INode* createDistributedNode(RegistryIt it, const RegistryID& reg_id, const ObjectID& node_id, NodeDescription* nd = 0)
    throw (const RegistryException&);

/** Loads a plugin to instantiate a node in the local address space.
  */
  void loadLocalPlugin(IPluginManager*, const string& plugin)
    throw (const RegistryException& );

/** Releases a Node that is creates by the Client-Registry
  */
  void releaseLocalNode(ReservationIt& );

/** refresh the reservation of reserved nodes by all Server-registries
  */
  void refreshNodeReservation();

/** refresh an activated access control
  */
  void refreshTransactionLock();

  Interface* allocateNMMObject( IServerRegistry*, const ObjectDescription& );

/**
  */
  void requestRemainingNodes(GraphDescription& gd)
    throw (const RegistryException&);

private: //session sharing methods
  void requestSharedGraph(GraphDescription& , const RegistryName& )
    throw (const RegistryException&);

  void requestSharedGraph(GraphDescription& , IServerRegistry* registry, const RegistryName& registry_name )
    throw (const RegistryException&);
    
  void requestExclusiveGraph(GraphDescription& gd, const RegistryName& registry_name)
    throw (const RegistryException&);

  list<RegistryName> getBestRegistry(GraphDescription& gd)
    throw (const RegistryException&);
private:
  /** This struct stores the reserved nodes and all need information to release and refresh the
    * reservation.
    */
  struct ReservationStruct {
    ReservationStruct() : proxy_object(0), node_factory(0), node(0), registry(0), reservation_time(0){}
    RegistryID          registry_id;         ///< Stores the RegistryID from the Registry the ClientRegistry reserves the node
    NMMReference        reference;           ///< Stores the reference of the distributed object
    ProxyObject*        proxy_object;        ///< Stores the pointer of the ProxyObject
    NodeFactory*        node_factory;        ///< Stores the pointer to the NodeFactory. This pointer is only used if the Node is reserved by the local RegistryServer.
    Node*               node;                ///< Pointer to the Node. This pointer is only used if the ClientRegistry allocates the node. This pointer is only used if the Node is reserved by the local RegistryServer
    IServerRegistry*    registry;            ///< Pointer to the IServerRegistry from which we reserved the node
    ObjectID            node_id;             ///< ObjectID of the node
    long                reservation_time;    ///< How long the application can use the node.
  };

public:
  /** This struct stores Interfaces to the ServerRegistries used to reserve nodes and
    * the corresponding IPluginManager.
    */
  struct RegistryStruct {
    RegistryStruct() : server_registry(0), plugin_manager(0), is_optimization_allowed(false) {}
    IServerRegistry*      server_registry;                  ///< Interface to the ServerRegistry
    IPluginManager*       plugin_manager;                   ///< Interface to the corresponding PluginManager
    bool                  is_optimization_allowed;          ///< is true if the ClientRegistry can allotcate nodes in local address space
  };


  /* Stores all required information about a session */
  struct ISessionStruct {
    ISessionStruct() : server_registry(0), session(0) {}
    ~ISessionStruct();
    IServerRegistry*   server_registry;         ///< server registry which manages the session
    ISession*          session;                 ///< pointer to the session
    list<INode*>       node_list;               ///< pointer of self allocated nodes
  };

private:
  map<NMMObject*, ReservationStruct>  __reservation_map;    ///< Map with the reserved nodes.
  map<RegistryName, RegistryStruct>   __named_registry;     ///< Map of used ServerRegistries stored in ResgistryStruct with the corresponding IPluginManager.
  IPluginManager_var                  __internal_p_manager; ///< interface to the local PluginManager
  bool                                __is_running;         ///< is true during the ClientRegistry's thread is running
  bool                                __init_transaction;   ///< is true if initTransaction was called. After closeTransaction it is reset to false
  ThreadMutex                         __condition_mutex;    ///< Mutex for the __condition
  ThreadCondition                     __condition;          ///< Condition variable for the automatic refresh process
  long                                __time_interval;      ///< stores interval the run method is triggerd
  map<IServerRegistry*, long>         __locked_registry;    ///< stores the locked registries and time left for the transaction
  ThreadMutex                         __member_mutex;       ///< This Mutex synchronize the internal thread with the application.

  map<NMMObject*, IServerRegistry*>   __object_map;         ///< stores allocated and reserved objects
  map<ProxyObject*, ISessionStruct>   __session_map;        ///< stores requested sessions
  multimap<uint,RegistryName>         __registry_speeds;    ///< map of server-registrys and their available cpu-speeds
};

/**
 * Simple class for locking the registry.
 * Using this class we assure that the registry is unlocked even if something goes wrong.
 */
class RegistryLock {
private:
  class ClientRegistry* reg;
  
public:
  RegistryLock(class ClientRegistry& registry) : reg(&registry) {
    reg->initTransaction();
  }
  
  void unlock() {
    if (reg) {
      reg->releaseTransaction();
      reg = NULL;
    }
  }
  
  ~RegistryLock() {
    unlock();
  }
};

}//end namespace
#endif
