/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_DEVICE_PLUGIN_HPP_
#define _NMM_DEVICE_PLUGIN_HPP_

#include "Plugin.hpp"

#include "nmm/interfaces/device/IDevice.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/ProxyObject.hpp"

namespace NMM {
  
  /**
   *  @brief Device plugin
   *
   *  This is a template class for plugins that use a device. The plugin creates a dummy node
   *  and attempts to call IDevice::setDevice for a set of device names.
   *
   *  The set of device names is the set of all filenames of the form @a device_name + @a i where @a i
   *  is a positive number, possibly including zero. The set may also include the standard device name
   *  which is simply @c device_name. For example, the set of devices may be "/dev/dsp", "/dev/dsp1",
   *  "/dev/dsp2", ... . If the standard device is checked, then @a i starts with 1, otherwise @a i starts
   *  with zero. If device @a n is not found, no check is done for any device with a higher number.
   *  The standard device is always checked first. If it is not available, no checks for other devices
   *  are made.
   *
   *  @sa IDevice
   */
  template<typename NODE>
  class TDevicePlugin
    : public TPlugin<NODE>
  {
  public:
    /**
     *  @brief Construct a TDevicePlugin and set the parameters for the device check
     *  @param device_name The device base name, e.g. "/dev/dsp"
     *  @param check_standard_device Check for the standard device
     *  @param node_name The name of the node
     *  @param max_instances Maximum instances of the node
     */
    TDevicePlugin(const string& device_name, bool check_standard_device, const string& node_name, int max_instance = 1);
    
    /**
     *  @brief Virtual destructor
     */
    virtual ~TDevicePlugin() throw();
    
    virtual void nmmRegistryInit(Registry&) throw(const Exception&);
    
  protected:
    /**
     *  @brief Check for standard auxiliary device(s)
     *  @param i_device Interface for setting the device
     *
     *  This method can be overridden to check for auxiliary devices (such as a mixer device for
     *  a soundcard). Auxiliary devices are optional, so this method must not fail.
     *
     *  The standard auxiliary device is the auxiliary device associated with the standard main device
     *  (for example /dev/mixer for /dev/dsp).
     */
    virtual void checkForStandardAuxDevice(IDevice& i_device) throw();
    
    /**
     *  @brief Check for auxiliary device(s)
     *  @param i_device Interface for setting the device
     *  @param device_num Device number
     *
     *  This method can be overridden to check for auxiliary devices (such as a mixer device for
     *  a soundcard). Auxiliary devices are optional, so this method must not fail.
     *
     *  The device number @c device_num is the number of the main device. This method checks
     *  for auxiliary devices associated with this main device (for example /dev/mixer1 for /dev/dsp1).
     */
    virtual void checkForAuxDevice(IDevice& i_device, unsigned device_num) throw();
    
  private:
    /** Device base name.
     */
    string __device_name;
    
    /** Check for standard device.
     */
    bool __check_standard_device;
  };
  
  template<typename NODE>
  TDevicePlugin<NODE>::TDevicePlugin(const string& device_name, bool check_standard_device, const string& node_name, int max_instance /*= -1*/)
    : TPlugin<NODE>(node_name, max_instance),
      __device_name(device_name),
      __check_standard_device(check_standard_device)
  {}
  
  template<typename NODE>
  TDevicePlugin<NODE>::~TDevicePlugin() throw() 
  {}
  
  template<typename NODE>
  void TDevicePlugin<NODE>::nmmRegistryInit(Registry&) throw(const Exception&) {
    /* we create a NODE and get the INode interface we need */
    NODE node;
    
    /* prepare the INode for registration*/
    INode_var i_node(prepareRegistration(node));
    
    /* we get the IDevice interface to specify the node capabilities on the system */
    IDevice_var i_device;
    i_device.reset(node.template getCheckedInterface<IDevice>());
    
    /* first we try to set the standard device using the IDevice interface */
    if(__check_standard_device) {
      GLOBAL_DEBUG_STREAM("DevicePlugin<" << this->getNodeName() << ">: Checking for standard device " << __device_name << std::endl);
      try {
        i_device->setDevice(__device_name);
      } catch(const DeviceNotFoundException& e) {
        /* if the  device can not be set the setDevice() method throws an exception
           which means for the device that no device is available or it can not be used
           for other reason (e.g. we have no access to the device). If the standard device
           can not be used we assume that no other device is available too and throw an
           exception, that the node can not be used. */
        GLOBAL_DEBUG_STREAM(e << endl);
        throw Exception("No devices available");
      }
      
      /* check for standard auxiliary device */
      checkForStandardAuxDevice(*i_device);
      
      /* if the node is configured, then we create the NodeDescription */
      NodeDescription description;
      createNodeDescription(*i_node.get(), description);
      
      /* we register the node */
      registerNode(description, this->getMaxInstanceCount());
    }
    
    /* if we have found a default device we try to scan for other devices in the same manner*/
    unsigned i = __check_standard_device ? 1 : 0;
    while ( 1 ) {
      GLOBAL_DEBUG_STREAM("DevicePlugin<" << this->getNodeName() << ">: Checking for device " << __device_name << i << std::endl);
      
      /* we assume that the next devices are __device_name + 1, __device_name + 2, ... */
      string next_device_name = __device_name + i;
      
      try {
        i_device->setDevice(next_device_name);
      } catch (const DeviceNotFoundException& e) {
        /* BEGIN> If a device can not be set, then no other devices are
           available and we leave the loop */
        GLOBAL_DEBUG_STREAM(e << endl);
        if(!__check_standard_device) {
          throw Exception("No devices available");
        }
        return;
        /*End< */
      } catch (const WrongDeviceException& e) {
        /* We found the wrong device, so just continue the loop. perhaps we will find the proper device later */
        ++i;
        continue;
      } catch (const Exception &e){
        GLOBAL_DEBUG_STREAM(e << endl);
      } catch (...){
        GLOBAL_DEBUG_STREAM("Unknown Ex" << endl);
      }
      
      /* check for auxiliary device */
      checkForAuxDevice(*i_device, i);
      
      /* if the node is configured we create the NodeDescription */
      NodeDescription description;
      createNodeDescription(*i_node.get(), description);
      
      /* we register the node */
      registerNode(description, this->getMaxInstanceCount());
      ++i;
    }
    /*End< */
  }
  
  template<typename NODE>
  void TDevicePlugin<NODE>::checkForStandardAuxDevice(IDevice&) throw()
  {}
  
  template<typename NODE>
  void TDevicePlugin<NODE>::checkForAuxDevice(IDevice&, unsigned) throw()
  {}
  
}

#endif
