/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_GRAPHDESCRIPTION_HPP
#define NMM_GRAPHDESCRIPTION_HPP

#include "NodeDescription.hpp"

namespace NMM {

class ISession;
class InstreamChannel;

struct GraphEdge : public SerializedValue {
/*! @name Constructor & Destructor
  */
//@{
/** constructor
  */ 
  GraphEdge();
  GraphEdge(int out_id, const string& out_tag, int in_id, const string& in_tag, const Format* format, int id,
            const string& c_buffer_strategy = "", const string& c_event_strategy = "");
  GraphEdge(const GraphEdge&);
  virtual ~GraphEdge();
/** virtual copy constructor
  */
  virtual GraphEdge* copy() const { return new GraphEdge(*this); }

  enum EdgeStatus { UNCONNECTED_IN_ID = -1, UNCONNECTED_OUT_ID = -1, UNDEF_EDGE_ID = -1, HELP_EDGE_ID = -2 };

  enum OverlapStatus { COMPLETE = 0, PARTIAL_COPY = 1, PARTIAL_OUTPUT = 2, NONE = 3  };
  
  GraphEdge& operator=(const GraphEdge&);
  bool isIdenticalOverlap(const GraphEdge& edge) const;
  void registerAttributes();

  bool isOutConnected() const { if(out_id.getValue() == UNCONNECTED_OUT_ID) return false; else return true; }
  bool isInConnected() const  { if(in_id.getValue()  == UNCONNECTED_IN_ID) return false; else return true; }
  bool isInHelper() const     { if(in_id.getValue()  == HELP_EDGE_ID) return true; else return false; }

  void setFormat(const Format*);
  Format* getFormat() { if(format.getValue() != 0) return format.getValue()->getPValue(); else return 0; }

  TValue<int>                 out_id;
  TValue<string>              out_tag;
  TValue<int>                 in_id;
  TValue<string>              in_tag;
  TValue< TValue<Format> * >  format;
  TValue<OverlapStatus>         overlap_type;
  TValue<int>                 q_out_id;
  TValue<int>                 q_in_id;
  TValue<int>                 id;
  TValue<int>                 q_id;
  TValue<int>                 r_id;
  TValue<string>              c_buffer_strategy;       ///< stores the communication strategy used for the connection
  TValue<string>              c_event_strategy;        ///< stores the secondary communication strategy used for the connection
  CEvent                      out_strategy_config;
  CEvent                      in_strategy_config;
  TValue<InstreamChannel*>    instream_channel;        ///< stores the instream channel of a connected edge. 
  static const char* GraphEdge_type_name;
  static const char* GraphEdge_type_ns;
};
OValueStream& operator<<(OValueStream& ov, const GraphEdge::OverlapStatus& eo);
IValueStream& operator>>(IValueStream& iv, GraphEdge::OverlapStatus& eo);


NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<NMM::GraphEdge> & string_map) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<NMM::GraphEdge> & string_map) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<int, list<int> >& string_map) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<int, list<int> > & string_map) ;

}




namespace NMM {

/** \class GraphDescription
  * A GraphDescription describes a graph with nodes that should be connected.
  */
class GraphDescription  : public SerializedValue {
public:
/*! @name Constructor & Destructor
  */
//@{
/** constructor
  */ 
  GraphDescription();

/** destructor
  */
  ~GraphDescription();

/** copy constructor
  */
  GraphDescription(const GraphDescription& );

/** virtual copy constructor
  */
  virtual GraphDescription* copy() const { return new GraphDescription(*this); }
//@}

  bool isIdenticalOverlap(const GraphDescription& gd) const;

  /**
   * Clears all entries of the graph description
   */
  void clear();

  void addNode(NodeDescription& nd);

  void setINode(const NodeDescription& nd, const INode& inode);

        INode* getINode(const NodeDescription& nd);
  const INode* getINode(const NodeDescription& nd) const;

        INode* getINode(const NodeDescription* nd);
  const INode* getINode(const NodeDescription* nd) const;

  const NodeDescription& getFirstNodeDescription()          const;

  const list<int>& getOutEdgeList(int i)            const; 

  const list<int>& getInEdgeList(int i)            const; 

  const NodeDescription& getNodeDescription(unsigned int i)                   const;

  NodeDescription& getNodeDescription(unsigned int i);

  const GraphEdge& getEdge(int i) const { return __edges.getRValue()[i]; }

  GraphEdge& getEdge(int i)  { return __edges.getRValue()[i]; }

  size_t getNumEdges() const { return __edges.getRValue().size(); }

  size_t GraphDescription::addEdgeFromIDs(int nd_from, const string& out_tag, 
					  int nd_to  , const string& in_tag,
					  const Format* format, const string& buffer_strategy, const string& event_strategy);

  size_t addEdge(const NodeDescription& nd_from, const string& out_tag, 
                 const NodeDescription& nd_to  , const string& in_tag, 
                 const Format* format = 0, const string& buffer_strategy = "", const string& event_strategy = "");
  
  size_t addEdge(const NodeDescription& nd_from,  
                 const NodeDescription& nd_to  , const string& in_tag, 
                 const Format* format = 0, const string& buffer_strategy = "", const string& event_strategy = "");
  
  size_t addEdge(const NodeDescription& nd_from, const string& out_tag, 
                 const NodeDescription& nd_to  ,  
                 const Format* format = 0, const string& buffer_strategy = "", const string& event_strategy = "");
  
  size_t addEdge(const NodeDescription& nd_from, 
                 const NodeDescription& nd_to  ,  
                 const Format* format = 0, const string& buffer_strategy = "", const string& event_strategy = "");
  
  void addEdges(const NodeDescription* n1    , const NodeDescription* n2    , const NodeDescription* n3    , 
		const NodeDescription* n4 = 0, const NodeDescription* n5 = 0, const NodeDescription* n6 = 0,
		const NodeDescription* n7 = 0, const NodeDescription* n8 = 0, const NodeDescription* n9 = 0,
		const NodeDescription* n10 = 0,const NodeDescription* n11 = 0,const NodeDescription* n12 = 0);

  size_t addInEdge(const NodeDescription& nd_to, const string& in_tag, 
		 const Format* format = 0);

  size_t addInEdge(const NodeDescription& nd_to,  
		 const Format* format = 0);

  size_t addOutEdge(const NodeDescription& nd_from, const string& out_tag, 
		  const Format* format = 0);

  size_t addOutEdge(const NodeDescription& nd_from,  
		  const Format* format = 0);

  size_t addOutHelpEdge(const NodeDescription& nd_from, const string& out_tag, 
			const Format* format = 0);

  void setOutStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
			    const NodeDescription& nd_to  , const string&  in_tag,
			    const CEvent& out_config);

  void setOutStrategyConfig(const NodeDescription& nd_from, 
			    const NodeDescription& nd_to  , const string&  in_tag,
			    const CEvent& out_config);

  void setOutStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
			    const NodeDescription& nd_to  , 
			    const CEvent& out_config);

  void setOutStrategyConfig(const NodeDescription& nd_from, 
			    const NodeDescription& nd_to  ,
			    const CEvent& out_config);

  void setInStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
			   const NodeDescription& nd_to  , const string&  in_tag,
			   const CEvent& in_config);
  
  void setInStrategyConfig(const NodeDescription& nd_from, 
			   const NodeDescription& nd_to  , const string&  in_tag,
			   const CEvent& in_config);
  
  void setInStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
			   const NodeDescription& nd_to  , 
			   const CEvent& in_config);

  void setInStrategyConfig(const NodeDescription& nd_from, 
			   const NodeDescription& nd_to  ,
			   const CEvent& in_config);

  void removeEdge(const int edge_num);

  void removeSubgraph(const int edge_num);

  void showAll();
  void showEdges();

  const SequenceValue<NodeDescription>& getNodeDescriptionList()   const { return __node_list; }
        SequenceValue<NodeDescription>& getNodeDescriptionList()         { return __node_list; }


  static const char* GraphDescription_type_name;
  static const char* GraphDescription_type_ns;

  void   setComment(const string& comment) { __comment = comment; }
  string getComment() { return __comment; }

  void realizeGraph() throw(const BindingException&, const Exception&);                        

  void realizeHandoverGraph(const GraphDescription& master_gd) throw(const BindingException&, const Exception&);                       
  void realizeSharedGraph() throw(const BindingException&, const Exception&);                        

  void activateGraph();

  void startGraph();

  void stopGraph();

  void unrealizeGraph();
                   
  const SequenceValue<INode >& getINodeList() const { return __interface_list; }
        SequenceValue<INode >& getINodeList()       { return __interface_list; }

  /* returns true if all NodeDescriptions have given sharing type s*/
  bool isSharingType(NodeDescription::SharingType s);

  bool getRequestFlag(const NodeDescription& nd) const;

  void setRequestFlag(const NodeDescription& nd, bool flag);

  bool getReleaseFlag(const NodeDescription& nd) const;

  void setReleaseFlag(const NodeDescription& nd, bool flag);


  void setISession(const ISession& session);

  const ISession* getISession()  const  { return __i_session.getValue(); }
        ISession* getISession()         { return __i_session.getValue(); }

  void setSinkSynchronizerDescription(const ObjectDescription& od);

  ObjectDescription getSinkSynchronizerDescription() { return __sink_synch_descr.getValue(); }

/** This method configures the connection between a ProxyObject and it's NMMObject,
  * if a specific transport strategy is stored in the corresponding NodeDescription.
  */
  void configureOutOfBandBindings();

/** This method connects all sink nodes nodes with the SinkSynchronizer provided by the Session.
  */
  void connectSinkSynchronizer();
  
  /**
   * Enable/disable prefetching for sink nodes
   */
  void setPrefetchingEnabled(bool enable);

/** This method removes all sink nodes nodes from the SinkSynchronizer provided by the Session.
  */
  void disconnectSinkSynchronizer();

  const list<INode*> GraphDescription::getPartialOverlappedNodes();

  const GraphDescription& operator=(const GraphDescription& gd);
private:
  void registerAttributes();
  void realizeCommunicationChannel(CommunicationChannel* c_channel, GraphEdge& edge);

private:
  list<int> __empty_list;

  typedef map<int, list<int> >::iterator InternalIt ;
  typedef map<int, list<int> >::const_iterator CInternalIt ;

  TValue< vector<GraphEdge> > __edges;
  TValue< map<int, list<int> > > __out_edges;
  TValue< map<int, list<int> > > __in_edges;

  SequenceValue< NodeDescription >      __node_list;        ///< stores the NodeDescription of the nodes
  SequenceValue< INode  >               __interface_list;   ///< stores the INode's of allocated nodes
  TValue< vector<bool> >                __request_flags;    ///< the requests falgs specify wheter a node should be requested by server registry or not
  TValue< vector<bool> >                __release_flags;    ///< the release falgs specify wheter a node should be releases or not. If this flag is false, no reference counting is performed

  unsigned int                          __internal_counter; ///< internal counter which is set inside the NodeConfiguration
  TValue<ISession*>                     __i_session;        ///< stores the interface to the session object
  string                                __comment;          ///< stores the comment
  TValue<ObjectDescription>             __sink_synch_descr;///< object description of the sink synchronizer object to use

};

typedef auto_ptr<GraphDescription> GraphDescription_var;


} //end namespace
#endif
