/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_NODEFACTORYSTRUCTS_HPP
#define NMM_NODEFACTORYSTRUCTS_HPP

#include "RegistryStructs.hpp"


namespace NMM {

class IValueStream;
class OValueStream;
class Registry;
class NodeFactory;
/** The NodeFactoryID is used to identify a NodeFactory. To identify a NodeFactory its class name
  * is used.
  */
class NodeFactoryID  :  public SerializedValue
{
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  * \param fac_name Name of the NodeFactory
  */
  NodeFactoryID(const string& fac_name = "");

/** Copy constructor
  */
  NodeFactoryID(const NodeFactoryID& f_id);

/** Virtual copy constructor.
  * This mehtod creates a copy from this object.
  * \param fac_name Name of the NodeFactory
  */
  virtual NodeFactoryID* copy()              const { return new NodeFactoryID(*this); }
//@}

/** @name Access stored information
  */
//@{
/** This method returns the name of the NodeFactory
  */
  inline const string&    getFactoryName()   const {return __fac_name.getRValue();}

/** This method returns the host oname of the NodeFactory
  */
  inline const string&    getHostname()      const {return __hostname.getRValue();}
//@}

/** @name Operator
  */
//@{
/** The assignment operator copies the infromation stored in given f_id to this NodeFactoryID
  */
  NodeFactoryID& operator=(const NodeFactoryID& f_id);

/** The operator== compares the given f_id with this NodeFactoryID and returns true
  * if the name of the NodeFacory are identicall, otherwise false.
  */
  bool operator==(const NodeFactoryID& f_id) const;

/** The operator!= compares the given f_id with this NodeFactoryID and returns false
  * if the name of the NodeFactory are identicall, otherwise true.
  */
  bool operator!=(const NodeFactoryID& f_id) const {return !(*this == f_id);}
//@}

private: //methods
  void registerAttributes();

private: //attributes
  TValue<string>    __hostname;          ///< stores the hostname (for debug reason)
  TValue<string>    __fac_name;          ///< stores the name of the NodeFactory
};

class FactoryTicket  :  public SerializedValue {
friend class Registry;
friend class NodeFactory;
public:
/** @name Consturctor & Destructor
  */
//@{
/** Constructor
  */
  FactoryTicket();

/** Copy-Constructor
  */
  FactoryTicket(const FactoryTicket& );

/** virtual Copy-Constructor
  */
  FactoryTicket* copy()                             const { return new FactoryTicket(*this); }
//@}

/** @name Operator
  */
//@{
/** The assignment operator stores all values from the given ft in this object. If the is_valid-flag
  * is true in the given ft, it is set to false and the is_value-flag of this FacotryTicket is set to
  * true. So the implementation of the assignement operator ensures that a valid FactoryTicket can not
  * be used twice.
  */
  FactoryTicket& operator=(const FactoryTicket& ft);
//@}

/** @name Access FacotryTicket information
  */
//@{
/** This method returns the config event to configure the created node.
  */
  inline const CEvent& getConfigEvent()             const { return __config_event; }

/** This method returns the config event to configure the created node.
  */
  inline       CEvent& getConfigEvent()                   { return __config_event; }

/** This method returns the NodeFactoryID of the NodeFactory that can create the node.
  */
  inline const NodeFactoryID& getNodeFactoryID()    const { return __factory_id; }

/** This method returns the NodeFactoryID of the NodeFactory that can create the node.
  */
  inline       NodeFactoryID& getNodeFactoryID()          { return __factory_id; }

/** This method returns the ObjectID of the node that should be created.
  */
  inline const ObjectID& getObjectID()                  const { return __node_id.getRValue(); }

/** This method returns true if the FactoryTicket can be used to create a node. If the node
  * was created, the FactoryTicket returns false.
  */
  inline bool  isValid()                            const { return __is_valid.getRValue(); }
//@}

/** @name Set FactoryTicket information
  * These methods are for internal use in a Registry.
  */
//@{
/** This method sets the ObjectID of the Node that should be created.
  * \param node_id The ObjectID of the node that should be created.
  */
  inline void  setObjectID(const ObjectID& node_id)           { __node_id.setValue(node_id); }

/** This method sets ConfigEvent which is need to configure the node.
  * \param config_event The CEvent that is send to the node.
  */
  inline void  setConfigEvent(const CEvent& config_event) { __config_event = config_event; }

/** This method sets the NodeFactoryID of the NodeFactory which can create the Node.
  * \id NodeFactoryID The ID of the NodeFactory.
  */
  inline void  setNodeFactoryID(const NodeFactoryID& id)  { __factory_id = id; }
//@}

protected:
/** sets the internal valid-flag to the given value
  * \param value the new value of the is_valid-flag
  */
  void  setValid(bool value)                        const { __is_valid = value; }
private:

/** This method registers all attributes by the base class
  */
  void  registerAttributes();
private:
  CEvent                    __config_event;    ///< stores the configuration of the node.
  NodeFactoryID             __factory_id;      ///< stores the NodeFactoryID of the NodeFactory that can create the node
  TValue<ObjectID>            __node_id;         ///< stores the ObjectID of the node
  mutable TValue<bool>      __is_valid;        ///< is true if the FacotryTicket is valid
};
}//end namespace
#endif
