/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <sys/stat.h>
#include <dirent.h>
#include <iostream>
#include "PluginManager.hpp"
#include "Plugin.hpp"
#include "nmm/utils/utils.hpp"

namespace NMM {

PluginManager::PluginManager()
  : StdNamedObject("PluginManager")
{
  addPluginPath(getNMMPluginDir());
}


PluginManager::~PluginManager() {
  DEBUG_STREAM(">Destroy"<< endl);
  /*Begin> Try to remove all used plugins */
  list<string> tmp_list = getUsedPlugins();
  for (list<string>::iterator it = tmp_list.begin(); it != tmp_list.end(); ++it) {
    try {
      removeLibrary(*it);
    } catch (const Exception& e) {
      ERROR_STREAM("Could not remove the plugin " << *it << endl);
    }
  }
  /*End< */
  DEBUG_STREAM("<Destroy"<< endl);
}

void PluginManager::addPluginPath(const string& path) {
  /*Begin> insert the path if it is not stored already in __plugin_path*/
  if( find(__plugin_path.begin(), __plugin_path.end(), path) == __plugin_path.end() ) {
    __plugin_path.push_back(path);
  } else {
    MESSAGE_STREAM("Plugin path " << path << "already set"<<endl);
  }
  /*End< */
}

int PluginManager::getPluginSize(const string& plugin)            const {
  Plugin** const plugin_array = requestPluginArray( plugin );
  if (! plugin_array )
    return -1;

  int counter = 0;
  while ( plugin_array[counter] != 0 )
    ++counter;
    
  return counter;
}

void PluginManager::loadLibrary(const string& plugin)
  throw (const Exception&)
{
  loadPlugin(plugin);
  activateRegistryPart(plugin);
  activateFactoryPart(plugin);
}


void PluginManager::loadExternalPlugins(const string& path)
    throw (const Exception& )
{
  list<string> file_list = readFolder(path);
  if ( file_list.empty() ) {
    WARNING_STREAM("Directory " << path << " provides no plugins. "<< endl);  
    return;
  }  
  
  /*Begin> Try to load the libraries */
  for (list<string>::iterator it = file_list.begin(); it != file_list.end(); ++it )
  {
    if ( __plugin_map.find( *it ) == __plugin_map.end() ) {
      try {
        PluginStruct ps;
        /* allocate the LTDll object */
        ps.dll_object= new LTDll();
        /* open the library */
        ps.dll_object->open(path+"/"+*it);
        /* Store the plugin object */ 
        __plugin_map.insert(pair<string, PluginStruct>(*it, ps));
      } catch (const Exception& e) {
        WARNING_STREAM ("Could not load external plugin " << *it << endl);
	WARNING_STREAM ("Error message: " << e << endl);        
      }
    } else {
      WARNING_STREAM ("External plugin " << *it << " already loaded " << endl);    
    }
  }
}

void PluginManager::removeExternalPlugins(const string& path)
    throw (const Exception& )
{
  list<string> file_list = readFolder(path);
  if ( file_list.empty() ) {
    WARNING_STREAM("Directory " << path << " provides no plugins to remove. "<< endl);  
    return;
  }  
  
  for ( list<string>::iterator it = file_list.begin(); it != file_list.end(); ++it) {
    map<string, PluginStruct>::iterator p_it = __plugin_map.find(*it);

    /*Begin> Check if the plugin was loaded */
    if (p_it == __plugin_map.end()) {
      WARNING_STREAM("The external plugin with name " << *it << " was not loaded" << endl);
      continue;
    }
    /*End< */  
    
    if (p_it->second.dll_object) {
      try {
        p_it->second.dll_object->close();
      } catch (const Exception& e) {
        WARNING_STREAM("Could not close external library " << p_it->first << endl);  
      }
      delete p_it->second.dll_object;
      __plugin_map.erase(*it);
    }    
  } 
}


void PluginManager::loadPlugin(const string& plugin)
    throw (const Exception&)
{
  if (isPluginUsed(plugin)) {
    MESSAGE_STREAM("Plugin " << plugin << " is already loaded" << endl);
    return;
  }

  list<string> file_list;
  string folder;
  /*Begin> Find the folder for the given library*/
  for (list<string>::iterator it = __plugin_path.begin(); it != __plugin_path.end(); ++it)
  {
    try {
      file_list = readFolder(*it);
      if (find(file_list.begin(), file_list.end(), plugin) != file_list.end()) {
        folder = *it;
        break; //leave the loop
      }
    } catch (const Exception& e) {
      ERROR_STREAM(e << endl);
    }
  }
  /*End< */

  if (folder.empty()) {
    ERROR_STREAM (" Could not find the library for plugin " << plugin << endl);
    return;
  }

  /*Begin> Try to load the library */
  try {
    PluginStruct ps;
    /* allocate the LTDll object */
    ps.dll_object= new LTDll();
    /* open the library */
    ps.dll_object->open(folder+"/"+plugin);

    /* Begin> Search for the plugin array with name nmm_plugins and initialize the plugins */
    ps.plugins = static_cast<Plugin**>(ps.dll_object->resolve(__plugin_symbol));

    /* Begin> Check if the plugin was already loaded. */
    for ( map<string, PluginStruct>::iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
      if ( (it->second.plugins == ps.plugins) && (ps.plugins != 0) )  {
        /* store the name as alias */
        it->second.alias.push_back(plugin);
        /* close the library again */
        ps.dll_object->close();
        /* delete the dll_object */
        delete ps.dll_object;
        return;
      }
    }
    /* End<*/
    
    if (ps.plugins) {
      Plugin** tmp_plugins = ps.plugins;
      unsigned int index = 0;
      while(*tmp_plugins) {
        (*tmp_plugins)->setLibraryName(plugin);
        (*tmp_plugins)->setPluginIndex(index);
        ++index;
        (*tmp_plugins)->nmmInit();
        ++tmp_plugins;
      }
    } else {
      WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
    }
    /*End< */
    __plugin_map.insert(pair<string, PluginStruct>(plugin, ps));
    applyDefinedObjectIDs( plugin );
  }
  /*Begin> Error handling: If an excetpion is thrown, remove the plugin */
  catch (const Exception& e) {
    removeLibrary(plugin);
    throw e;
  }
  /*End< */
  DEBUG_STREAM("Plugin " << plugin << " successfully loaded" << endl);
}

void PluginManager::removeLibrary(const string& name)
  throw (const Exception&)
{
  map<string, PluginStruct>::iterator it = __plugin_map.find(name);

  /*Begin> Check if the plugin is loaded */
  if (it == __plugin_map.end()) {
    WARNING_STREAM("A plugin with name " << name << " was not loaded" << endl);
    return;
  }
  /*End< */


  /*Begin> Call the unregister method for the Registry */
  deactivateRegistryPart(name);
  /*End< */

  /*Begin> Call the unregister method for the NodeFactory */
  deactivateFactoryPart(name);
  /*End< */

  /*Begin> Call the generic unregister method for nmm plugins */
  if (it->second.plugins) {
    Plugin** tmp_plugins = it->second.plugins;
    while(*tmp_plugins) {
      (*tmp_plugins)->nmmFini();
      ++tmp_plugins;
    }
  } else {
    WARNING_STREAM("Library " << name << " has no NMM plugins" << endl);
  }
  /*End< */


  /*Begin> Close the Dll Object and remove it from the map */
  try {
    if (it->second.dll_object) {
      it->second.dll_object->close(); //throws Exception
    }
    else
      GLOBAL_WARNING_STREAM("No DLL-Object available for library " << it->first << ". Maybe it causes a segfault" << endl);
  } catch (const Exception& e) {
    ERROR_STREAM(e << endl);
  }
  delete it->second.dll_object;   //delete the dll_object
  DEBUG_STREAM("Plugin " << it->first << " successfully removed" << endl);
  __plugin_map.erase(it);
  /*End< */
}


void PluginManager::deactivateRegistryPart(const string& plugin)
    throw (const Exception&)
{
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);

  /*Begin> Check if the plugin is loaded */
  if (it == __plugin_map.end()) {
    WARNING_STREAM("A plugin with name " << plugin << " was not loaded" << endl);
    return;
  }
  /*End< */

  /*Begin> Call the unregister method for the Registry */
  if (it->second.reg_active && it->second.plugins && it->second.plugins) {
    Plugin** tmp_plugins = it->second.plugins;
    while(*tmp_plugins) {
      (*tmp_plugins)->nmmRegistryFini(*it->second.reg_active);
      ++tmp_plugins;
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  it->second.reg_active = 0; //reset the reg_active to ensure it is called once only.
  /*End< */
}

void PluginManager::deactivateFactoryPart(const string& plugin)
    throw (const Exception&)
{
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);

  /*Begin> Check if the plugin is loaded */
  if (it == __plugin_map.end()) {
    WARNING_STREAM("A plugin with name " << plugin << " was not loaded" << endl);
    return;
  }
  /*End< */

  /*Begin> Call the unregister method for the NodeFactory */
  if (it->second.fac_active && it->second.plugins && it->second.plugins) {
    Plugin** tmp_plugins = it->second.plugins;
    while(*tmp_plugins) {
      (*tmp_plugins)->nmmFactoryFini(*it->second.fac_active);
      ++tmp_plugins;
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  it->second.fac_active = 0; //reset the fac_active to ensure it is called once only.
  /*End< */
}

bool PluginManager::isPluginUsed(const string& plugin) const {
  /* If the plugin is not in the internal map, it is not used */
  if (__plugin_map.find(plugin) == __plugin_map.end())
    return false;

  return true;
}

bool PluginManager::isFactoryPartActive(const string& plugin)  const {
  map<string, PluginStruct>::const_iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) //plugin not loaded
    return false;

  if (it->second.fac_active)
    return true;

  return false;
}

bool PluginManager::isRegistryPartActive(const string& plugin)  const {
  map<string, PluginStruct>::const_iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) //if plugin not load return false
    return false;

  if (it->second.reg_active)
    return true;

  return false;
}

void PluginManager::activateRegistryPart(const string& plugin, Registry& registry)
  throw (const Exception& )
{
  /*Begin> Check if plugin is load and not active */
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) { //plugin not loaded
    DEBUG_STREAM("Plugin " << plugin << " not loaded" << endl);
    return;
  }

  if (it->second.reg_active) {
    DEBUG_STREAM("Plugin " << plugin << " is already activated by a Registry" << endl);
    return;
  }

  if (__defined_id_map.find(plugin) != __defined_id_map.end() ) {
    WARNING_STREAM(" plugin " << plugin <<" has predeinied IDs -> don't call activate Registry part" << endl);
    return;
  }
  /*End< */

  /* Store the used Registry */
  it->second.reg_active = &registry;
  Plugin** tmp_plugins = it->second.plugins;
  if (tmp_plugins) {
    while(*tmp_plugins) {
      try {
        /* register the plugins with the library */
        (*tmp_plugins)->nmmRegistryInit(registry);
      } catch (const Exception& e) {
        ERROR_STREAM("Could not activate registry part of plugin " << plugin << endl);
      }
      ++tmp_plugins;
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  DEBUG_STREAM("Registry part of plugin " << plugin << " successfully activated" << endl);
  /*End< */
}

void PluginManager::activateFactoryPart(const string& plugin, NodeFactory& factory)
  throw (const Exception&)
{
  /*Begin> Check if plugin is load and not activate */
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) { //plugin not loaded
    DEBUG_STREAM("Plugin " << plugin << " not loaded" << endl);
    return;
  }

  if (it->second.fac_active) {
    DEBUG_STREAM("Plugin " << plugin << " is already activated by a NodeFactory" << endl);
    return;
  }
  /*End< */

  /* Store the used NodeFactory */
  it->second.fac_active = &factory;
  Plugin** tmp_plugins = it->second.plugins;
  if(tmp_plugins) {
    while(*tmp_plugins) {
      try {
        /* register the plugins with a node factory */
        (*tmp_plugins)->nmmFactoryInit(factory);
      } catch (const Exception& e) {
        it->second.fac_active = 0;
        ERROR_STREAM("Could not activate factory part of plugin " << plugin << endl);
        ERROR_STREAM(e << endl);
      }
      ++tmp_plugins;
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  /*End< */
}

list<string> PluginManager::getUsedPlugins() const {
  list<string> tmp_list;
  for (map<string, PluginStruct>::const_iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it)
  {
    tmp_list.push_back(it->first);
  }

  return tmp_list;
}

list<string> PluginManager::getPlugins() const {
  list<string> plugin_list;
  list<string> tmp_list;
  /*Begin> Find the folder for the given plugin*/
  for (list<string>::const_iterator it = __plugin_path.begin(); it != __plugin_path.end(); ++it)
  {
    try {
      tmp_list = readFolder(*it);
      plugin_list.insert(plugin_list.end(), tmp_list.begin(), tmp_list.end());
    } catch (const Exception& e) {
      ERROR_STREAM(e << endl);
    }
  }
  /*End> */
  return plugin_list;
}

list<string> PluginManager::readFolder(const string&  folder) const
  throw (const Exception& )
{
  struct dirent **namelist;
  struct stat buffer;

  list<string> file_list;

  int count = scandir(folder.c_str(), &namelist, 0, alphasort);
  string tmp;
  if (count < 0) {
    /* if the directory can not be accessed throw an exception */
    throw Exception("Scandir for directory " + folder + " failed ");
  } else {
    for (int i = 0; i< count; i++) {
      tmp = folder + "/" + namelist[i]->d_name;
      if (stat(tmp.c_str(), &buffer) == 0) {
        if (S_ISREG(buffer.st_mode)) {
          file_list.push_back(namelist[i]->d_name);
        }
      }
    }
  }
  free(namelist);
  return file_list;
}


Plugin** PluginManager::requestPluginArray(const string& plugin) {
  /* search the library */
  InternalIt it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) {
    for ( map<string, PluginStruct>::iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
      if ( find(it->second.alias.begin(), it->second.alias.end(), plugin ) != it->second.alias.end() )  {
        return it->second.plugins;
      }
    }
    return 0;
  }

  return  it->second.plugins;
}

Plugin** const  PluginManager::requestPluginArray(const string& plugin) const {
  /* search the library */
  CInternalIt it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) {
    for ( map<string, PluginStruct>::const_iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
      if ( find(it->second.alias.begin(), it->second.alias.end(), plugin ) != it->second.alias.end() )  {
        return it->second.plugins;
      }
    }
    return 0;
  }

  return  it->second.plugins;
}

list<ObjectID> PluginManager::getPluginID(const string& plugin, const unsigned int& index ){
  /* Begin> If predefined IDs are available, use them */
  map<string, vector<list<ObjectID> > >::iterator i_it = __defined_id_map.find( plugin );
  if ( i_it != __defined_id_map.end() ) {
    if ( index < i_it->second.size() ) {
      return i_it->second[index];
    }
  }
  
  /* allocate the return value */
  list<ObjectID> node_id;

  /* Begin> If the Plugin is not load, return an empty list */
  InternalIt it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) {        
    return node_id;
  }
  /* End< */

  /* Begin> If no Plugin-Array is available, return an empty list */
  Plugin** plugins = it->second.plugins;
  if (!plugins)
    return node_id;
  /* End< */

  /* insert the ObjectID's of a */
  for (unsigned int i = 0; i< plugins[index]->getSize(); ++i)
    node_id.push_back(plugins[index]->getObjectID(i));

  return node_id;      
}


list<string> PluginManager::getPluginPath()      const {
  return __plugin_path;
}

void PluginManager::activateRegistryPart(const string& plugin)
  throw (const Exception&)
{
  activateRegistryPart(plugin, SLocalRegistry::getInstance());
}

void PluginManager::activateFactoryPart(const string& plugin)
  throw (const Exception&)
{
  activateFactoryPart(plugin, SLocalNodeFactory::getInstance());
}

void PluginManager::defineObjectIDs(const string& plugin, const vector<list<ObjectID> >& id_list) {
  __defined_id_map.insert(pair<string, vector<list<ObjectID> > >(plugin, id_list));
  applyDefinedObjectIDs(plugin);
}

void PluginManager::applyDefinedObjectIDs(const string& plugin) {
  InternalIt p_it = __plugin_map.find(plugin);
  if ( p_it == __plugin_map.end() ) {
    return;
  }

  if ( !p_it->second.plugins ) {
    return;
  }
 
  /* Check if predefined IDs are available and set them */
  map<string, vector<list<ObjectID> > >::iterator it = __defined_id_map.find(plugin);

  if (it != __defined_id_map.end()) {
    for ( unsigned int i = 0; i< it->second.size(); ++i) {
      if (p_it->second.plugins[i] == 0) {
        break;
      }
      for (list<ObjectID>::iterator it2 = it->second[i].begin(); it2 != it->second[i].end(); ++it2) {
        p_it->second.plugins[i]->addObjectID(*it2);
      }
    }
  }
}


const string PluginManager::__plugin_symbol = "nmm_plugins";


}//end namespace
