/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
 * USA
 */

#include <unistd.h>
#include "RegistryStructs.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"

namespace NMM {

UserInfo::UserInfo()
  : SerializedValue("UserInfo", "NMM")
{
  registerAttributes();

  /* Begin> initialize the attributes */
  __username    = SProcessUserInfo::getInstance().getUsername();
  __passwd      = SSecurityCheck::getInstance().getPasswd();
  __domainname  = SProcessUserInfo::getInstance().getDomainname();
  __hostname    = SProcessUserInfo::getInstance().getHostname();
  __pid         = SProcessUserInfo::getInstance().getPid();
  /* End< */
}

UserInfo::UserInfo(const UserInfo& ui)
  : SerializedValue(ui)
{
  registerAttributes();
  assign(ui);
}

void UserInfo::assign(const UserInfo& ui) {
  /* Begin> assign the attributes */
  __username    = ui.getUsername();
  __passwd      = ui.getPasswd();
  __domainname  = ui.getDomainname();
  __hostname    = ui.getHostname();
  __pid         = ui.getPid();
  /* End < */
}

const UserInfo& UserInfo::operator=(const UserInfo& ui)  {
  SerializedValue::operator=(ui);
  assign(ui);
  return *this;
}

void UserInfo::registerAttributes() {
  /* Begin> register the attributes by the baseclass */
  insertArgument(&__hostname);
  insertArgument(&__domainname);
  insertArgument(&__username);
  insertArgument(&__passwd);
  insertArgument(&__pid);
  /* End< */
}

bool UserInfo::operator==(const UserInfo& u_info) const {
  /* Begin> compare all attributes */

  /* compare the hostname */
  if (this->getHostname() != u_info.getHostname())
    return false;

  /* compare the domainname */
  if (this->getDomainname() != u_info.getDomainname())
    return false;

  /* compare the username */
  if (getUsername() != u_info.getUsername())
    return false;

  /* compare process id */
  if (getPid() != u_info.getPid()) {
    return false;
  }
  /* End< */

  return true;
}


ProcessUserInfo::ProcessUserInfo()
  : __username(::getenv("LOGNAME")),
    __hostname(NMM::getHostname()),
    __domainname(NMM::getDomainname()),
    __pid(::getpid())
{
  char* home = getenv("HOME");
  if ( home ) {
    __home_directory = home;
  } else {
    if ( __username == "root" ){
      __home_directory = "/root";
    } else {
      __home_directory = string("/home/") + __username;
    }    
  }
}

RegistryID::RegistryID(Value::MemoryMgm mgm)
    :  SerializedValue("RegistryID", "NMM", mgm), __reference(0)
{
  allocate();
}

RegistryID::RegistryID(const string& reg_name, const NMMReference& reference, Value::MemoryMgm mgm)
    :  SerializedValue("RegistryID", "NMM", mgm), __reg_name(reg_name), __reference(reference)
{
  allocate();
}


RegistryID::RegistryID(const RegistryID& ri)
  : SerializedValue(ri)
{
  allocate();
  *this = ri;
}

void RegistryID::allocate() {
  /* Begin> register the attributes by the baseclass */
  insertArgument(&__reg_name);
  insertArgument(&__reference);
  insertArgument(&__user_info);
  /* End< */
}

bool RegistryID::operator==(const RegistryID& id) const {
  /* Begin> Compare all attributes */

  /* compare the NMMReference */
  if (getReference() != id.getReference())
    return false;

  /* compare the UserInfo */
  if (getUserInfo() != id.getUserInfo())
    return false;
  /* End< */

  return true;
}

NodeInfo::NodeInfo()
  :  SerializedValue(TypeInfo("NodeInfo", "NMM"))
{
  registerAttributes();
}

NodeInfo::NodeInfo(const ObjectID& node_id, const string& lib_name, const string& node_name, int inst_count, int max_count, int nt)
  : SerializedValue(TypeInfo("NodeInfo", "NMM")), __node_id(node_id), __library_name(lib_name),
    __node_name(node_name), __instance_count(inst_count), __max_instance_count(max_count), __node_type(nt)
{
  registerAttributes();
}

NodeInfo::NodeInfo(const NodeInfo& ni)
  : SerializedValue(ni)
{
  registerAttributes();
  *this = ni;
}

void NodeInfo::registerAttributes() {
  /* Begin> register the attributes by the baseclass */
  insertArgument(&__node_id);
  insertArgument(&__library_name);
  insertArgument(&__node_name);
  insertArgument(&__instance_count);
  insertArgument(&__max_instance_count);
  insertArgument(&__node_type);
  /* End< */
}


OValueStream& operator<<(OValueStream& ov, const list<Response>& string_list){
  return writeList<Response>(ov, string_list, "Response");
}
OValueStream& operator<<(OValueStream& ov, const list<RegistryID>& string_list)
{
  return writeList<RegistryID>(ov, string_list, "RegistryID");
}
OValueStream& operator<<(OValueStream& ov, const list<ObjectID>& string_list)
{
  return writeList<ObjectID>(ov, string_list, "ObjectID");
}

OValueStream& operator<<(OValueStream& ov, const pair<RegistryID, list<ObjectID> >& string_list)
{
  return writePair<RegistryID, list<ObjectID> >(ov, string_list, "RegistryID", "list_ObjectID");
}

OValueStream& operator<<(OValueStream& ov, const list<NodeInfo>& string_list) {
  return writeList<NodeInfo>(ov, string_list, "NodeInfo");
}



IValueStream& operator>>(IValueStream& ov, list<Response>& string_list) {
  return readList<Response>(ov, string_list, "Response");
}
IValueStream& operator>>(IValueStream& ov, list<RegistryID>& string_list){
  return readList<RegistryID>(ov, string_list, "RegistryID");
}
IValueStream& operator>>(IValueStream& ov, list<ObjectID>& string_list){
  return readList<ObjectID>(ov, string_list, "ObjectID");
}

IValueStream& operator>>(IValueStream& ov, pair<RegistryID, list<ObjectID> >& string_list){
  return readPair<RegistryID, list<ObjectID> >(ov, string_list, "RegistryID", "list_ObjectID");
}

IValueStream& operator>>(IValueStream& ov,list<NodeInfo>& string_list) {
  return readList<NodeInfo>(ov, string_list, "NodeInfo");
}

/* Begin> register all structs by the ValueFacotry */
TValueConstructor<list<Response> >     list_Response_value("list_Response", "std");
TValueConstructor<list<RegistryID> >   list_RegistryID_value("list_RegistryID", "std");
TValueConstructor<list<ObjectID> >       list_ObjectID_value("list_ObjectID", "std");
TValueConstructor<list<NodeInfo> >     list_NodeInfo_value("list_NodeInfo", "std");
TValueConstructor<pair<RegistryID, list<ObjectID> > > pair_response_value("pair_RegistryID_list_ObjectID", "std");

TConstructor<UserInfo>    __nmm__user_info("UserInfo", "NMM");
TConstructor<RegistryID>  __nmm__registry_id("RegistryID", "NMM");
TConstructor<NodeInfo>    __nmm__node_info("NodeInfo", "NMM");
TConstructor<RegistryException>  __nmm__registry_exception("RegistryException", "NMM");

/* End< */

ostream& operator<<(ostream& o, const RegistryID& reg_id) {
  o << reg_id.getRegistryName () << endl;
  o << reg_id.getReference() << endl;
  o << reg_id.getUserInfo() << endl;
  return o;
}

ostream& operator<<(ostream& o, const UserInfo& ui){
  o << "Host    : " << ui.getHostname() << endl;
  o << "Domain  : " << ui.getDomainname() << endl;
  o << "User    : " << ui.getUsername() << endl;
  o << "Pid     : " << ui.getPid() << endl;
  return o;
}

}//end namespace

