/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_RMMESH_HPP_
#define _NMM_RMMESH_HPP_

#include "RMIndexVertex.hpp"
#include "nmm/base/serialize/ComplexValue.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "Slope.hpp"
#include <map>

namespace NMM {

/**
 * In this mesh, the cpu-usage of a node can be saved.
 * For every format a vertex is created to store usage statistics
 */
class RMMesh : 
    public SerializedValue, 
    public NamedObject  
{

    RMIndexVertex* start;   // pointer to the first index in the mesh

    Slope slopes;

// private methods

    void setPointers(RMIndexVertex* index, RMVertex* vertex , list<string>::iterator act_pos, bool dummy = false);  // set lower/higher pointers of a vertex
    void deleteAll();  // delete the whole mesh

    map<string,QualityParameter>* formatToMap(const Format& description);

    /*!
     * remove a dummy vertex from the mesh
     */
    Result eraseDummyVertex(RMVertex* vertex);

    RMIndexVertex* findFirstNonMatching(RMVertex &vertex);

    float computeUsage(RMVertex &from, RMVertex &to);

//    vector<RMVertex>    vertex_index; // a vector to access vertices by id

    void registerValues();

    typedef list<RMVertex*>::iterator VI;
    typedef list<string>::iterator LI;

  public:

    //! The default Constructor
    RMMesh();
    RMMesh(const list<string>& order);
    RMMesh(const RMMesh& rm);
    ~RMMesh();

    /*!
     * add a new vertex to the mesh
     * if dummy is true the vertex will not be added to the index and no lower/higher-pointers of other vertices will point to it
     * a dummy vertex can be erased
     */
    Result addVertex(const Format* description, const float cpu_usage = 0.0F, const float bandwidth = 0.0F, bool dummy = false);

    /*!
     *
     */
    Result addVertex(RMVertex& vertex);

    /*!
     * get a vertex for the specified format
     * \returns 0 if there is no such vertex
     */
    RMVertex* getVertex(const Format* description);

    /*!
     * get a vertex by its unique id
     */
    const RMVertex* getVertex(unsigned int id) const {
      if (id == 0) {
        WARNING_STREAM2("no vertex with id 0" << endl);
        return 0;
      } else {
        for (list<RMVertex>::const_iterator i = vertex_index.getRValue().begin(); i != vertex_index.getRValue().end(); ++i) {
          if (i->getID() == id) {
            return &(*i);
          }
        }
//        return &(vertex_index.getRValue()[id]);
      }
      return 0;
    }

    /*!
     * \returns the cpu usage of the vertex with specified format
     *          if there is no vertex with this format the cpu-usage is interpolated (and a dummy vertex is added)
     */
    float getUsage(Format* description);

    float getBandwidth(Format* description);

    /*!
     * save the measured points to a file readable by HnSRTree-2.0
     */
    Result savePoints(const string& filename);

    RMMesh*    copy()         const { return new RMMesh(*this); }

    void setFormat(const Format&);
    void getFormat(Format&);

    //debug
    void printIndex();
    void printDimensions();

  private:

    float linearEstimate(float x_to_estimate, QualityParameter x1, float y1, QualityParameter x2, float y2);

    NMM::TValue<list<RMVertex> >   vertex_index;  // a vector to access vertices by id (id 0 is reserved)
    NMM::TValue<list<string> >  dimension_order;  // the sequence the parameters are ordered in
    NMM::TValue<unsigned int>    mesh_dimension;  // dimension of the mesh
    NMM::TValue<Format>             mesh_format;  // the format used to setup the mesh

    void updateAttributes();
    void prepareAttributes();
};

}

#endif
