/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "ResourceManager.hpp"
#include "nmm/interfaces/base/resourcemgr/IResourceCollector.hpp"
#include "nmm/base/ProxyObject.hpp"

using namespace NMM;

// static members
uint ResourceManager::node_counter;
//RMConfigManager ResourceManager::rmcm;

ResourceManager::ResourceManager() : NamedObject("ResourceManager") {
  DEBUG_STREAM("constructing ResourceManager" << endl);
  node_counter = 0;
  iperf_server = false;
  cpu_info.start();
  cpu_num = cpu_info.getCPUNum();

  free_cpu.resize(cpu_num);
  used_cpu.resize(cpu_num);

  has_server_registry = false;
  reservation_pending = false;

  // set free cpu
  for (uint i = 0; i < cpu_num; i++) {
    free_cpu[i] = (cpu_info.getFreeCPU(i) / 1000);
  }
  DEBUG_STREAM2("initializing RMConfigManager" << endl);

  rmcm.load();

  interface_num = rmcm.getInterfaces(); // initialize using 0 network-interfaces

  free_net.resize(interface_num+1);
  used_net.resize(interface_num+1);

  for (uint i = 1; i <= interface_num; i++) {
    free_net[i] = rmcm.getMaxBW(i);
  }

  DEBUG_STREAM("ResourceManager constructed" << endl);
}

ResourceManager::~ResourceManager() {
  free_cpu.clear();
  used_cpu.clear();

  free_net.clear();
  used_net.clear();

  running_nodes.clear();

  cpu_info.stop();

  if (iperf_server) {
    i_perf.stopIPERFServer();
  }

  DEBUG_STREAM2("saving RM-data" << endl);

  rmcm.save();

  DEBUG_STREAM("ResourceManager destructed" << endl);
}

void ResourceManager::setRegistry(IRegistry* ptr) {
  if (! has_server_registry) {
    server_registry.reset(ptr);
  }
  has_server_registry = true;
}

Result ResourceManager::addNode(const INode* node , uint& id, const uint& interface) {
  IResourceCollector_var irc(const_cast<INode*>(node)->getParentObject()->getInterface<IResourceCollector>());
  float node_usage = irc->getActualUsage();
  float node_bandwidth = irc->getActualBandwidth();
//  float node_bandwidth = irc->getActualBufferSize() / irc->getActualUsage();

  bool free_cpu_found = false;
  uint oncpu = 0;
  for (uint i = 0; i < cpu_num; i++) {
    if (free_cpu[i] > node_usage) {
      free_cpu_found = true;
      oncpu = i;
      break;
    }
  }
  if (! free_cpu_found) {
    ERROR_STREAM("no cpu with sufficient resources found" << endl);
    return FAILURE;
  }

  if (free_net[interface] < node_bandwidth) {
    DEBUG_STREAM("remaining bandwidth on interface " << interface << "(" << free_net[interface] <<") not sufficient to run node " << node->getParentObject()->getInterface<INamed>()->getName() << " (" << node_bandwidth << ")"<< endl );
    return FAILURE;
  }

  free_cpu[oncpu] -= node_usage;
	DEBUG_STREAM("cpu " << oncpu << ":  " <<  free_cpu[oncpu] << " left" << endl);
  used_cpu[oncpu] += node_usage;
  free_net[interface] -= node_bandwidth;
	DEBUG_STREAM("interface " << interface << ":  " << free_net[interface] << " left" << endl);
  used_net[interface] += node_bandwidth;

  RMNodeInfo node_info;
  node_info.name = node->getParentObject()->getInterface<INamed>()->getName();
  node_info.on_cpu = oncpu;
  node_info.usage = node_usage;
  node_info.interface = interface;
  node_info.bandwidth = node_bandwidth;
  node_info.id = node_counter++;

  id = node_info.id;

  running_nodes.push_back(node_info);
  DEBUG_STREAM(node_info.name << " should run on cpu " << node_info.on_cpu << endl);

  return SUCCESS;
}

Result ResourceManager::releaseNode(const INode* node,const uint& id) {

  uint oncpu,interface;
  float usage,bw;
  list<RMNodeInfo>::iterator node_pos;
  string name = node->getParentObject()->getInterface<INamed>()->getName();

  bool found = false;
  for (list<RMNodeInfo>::iterator i = running_nodes.begin(); i !=  running_nodes.end(); ++i) {
    if ( (i->name == name) && ( i->interface == interface) && i->id == id) {
      oncpu = i->on_cpu;
      usage = i->usage;
      interface = i->interface;
      node_pos = i;
      found = true;
      break;
    }
  }

  if (! found) {
    ERROR_STREAM(name << " not found for releasing! (on interface "  << interface << ")" << endl);
    return FAILURE;
  }

  running_nodes.erase(node_pos);

  used_cpu[oncpu] -= usage;
  free_cpu[oncpu] += usage;

  if (interface != 0) {
    if ( interface <= interface_num ) {
      used_net[interface-1] -= bw;
      free_net[interface-1] += bw;
    } else {
      ERROR_STREAM("Node on non existing interface" << endl);
    }
  }

  return SUCCESS;
}

void ResourceManager::setNetworkInterfaces(const unsigned int& new_num) {
  if (new_num == interface_num) {
    return;
  }

  free_net.resize(new_num);
  used_net.resize(new_num);

}

void ResourceManager::getCPUSpeed(uint& bogo_mips,float& free) {
  if (reservation_pending) {
    bogo_mips = 0;
    free = 0.0;
  } else {
    bogo_mips = cpu_info.getCPUSpeed();
    free = cpu_info.getTotalFreeCPU();
  }
}

void ResourceManager::getCPUInfo(string& name, uint& bogo_mips, uint& num, float& free) {
  num = cpu_info.getCPUNum();
  bogo_mips = cpu_info.getCPUSpeed();
  name = cpu_info.getCPUType();
  free = cpu_info.getTotalFreeCPU();
}

void ResourceManager::getCPUFree(float& free) {
  free = cpu_info.getTotalFreeCPU();
}

void ResourceManager::getCPULeft(float& left) {
  left = 0.0;
  if (reservation_pending) {
    return;
  }
  for (vector<float>::const_iterator i = free_cpu.begin(); i != free_cpu.end(); ++i) {
    left += *i;
  }
}

void ResourceManager::setCPUUpdateIntervall(const unsigned int& seconds) {
  cpu_info.setUpdateIntervall(seconds);
}

float ResourceManager::testConnection(const string& hostname) {
  return i_perf.testConnection(hostname, 22802 , 2);
}

void  ResourceManager::testConnections(const list<string>& hosts, list<float>& speeds) {

  vector<IPerf*> i_perfs(hosts.size());
  uint num = 0;
  // start all connection tests
  for (list<string>::const_iterator i = hosts.begin(); i != hosts.end(); ++i) {
    DEBUG_STREAM("Testing connection to " << *i << endl);
    i_perfs[num] = new IPerf();
    i_perfs[num++]->startAsyncConnection(*i,22802, 5);
  }
  DEBUG_STREAM("Waiting for answers " << endl);
  // get all answers
  num = 0;
  for (list<string>::const_iterator i = hosts.begin(); i != hosts.end(); ++i) {
    float speed = 0.0;
    i_perfs[num]->getClientSpeed(speed);
    delete i_perfs[num++];
    speeds.push_back(speed);
  }

}


void ResourceManager::startServer() {
  if (! iperf_server) {
    DEBUG_STREAM2("Starting iperf server" << endl);
    i_perf.startIPERFServer(22802);
  }
  iperf_server = true;
}

void ResourceManager::stopServer() {
  if (iperf_server) {
    DEBUG_STREAM2("Stopping iperf server" << endl);
    i_perf.stopIPERFServer();
  }
  iperf_server = false;
}

Result ResourceManager::setHosts(const list<string>& ihosts) {
    hosts = ihosts;
  return SUCCESS;
}

void ResourceManager::setReservationPending() {
  reservation_pending = true;
}

void ResourceManager::commitReservation() {
  reservation_pending = false;
}

void ResourceManager::getCPUDescription(string& description) {
  strstream cpu;
  cpu.clear();
  cpu << cpu_info.getCPUType() <<  cpu_info.getCPUSpeed() << '\0';
  description = cpu.str();
}

string ResourceManager::getMesh(string node_name, string format_id) {
  return rmcm.getMesh(node_name,format_id);
}

list<pair<string,string> > ResourceManager::getMeshes(string node_name) {
  return rmcm.getMeshes(node_name);
}

void ResourceManager::setMesh(string node_name, string format_id, string mesh) {
  rmcm.setMesh(node_name, format_id, mesh);
}

Result ResourceManager::updateNodeInfo(const INode* node) {
  IResourceCollector_var node_rc;
  node_rc.reset(const_cast<INode*>(node)->getParentObject()->getInterface<IResourceCollector>());

  Format node_format;
  node_rc->getActualFormat(node_format);
//  node_rc->updateVertex(node_format, node_rc->getCPUperBuffer(), node_rc->getMaxReceivedBufferSize());
//  DEBUG_STREAM2( "UPDATING : " << node_rc->getCPUperBuffer() << "   " << node_rc->getMaxReceivedBufferSize() << endl);
	float cpu = node_rc->getCPUperBuffer();
	float bw  = node_rc->getActualBandwidth();
  node_rc->updateVertex(node_format, cpu, bw);
  DEBUG_STREAM2( "UPDATING : " << cpu << "   " << bw << endl);

  string s_mesh;
  if (node_rc->getMesh(s_mesh) == SUCCESS) {
    rmcm.setMesh(node->getParentObject()->getInterface<INamed>()->getName() , node_format.getFormatType() , s_mesh);
    rmcm.save();
  }


  return SUCCESS;
}



