/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_VALUESTREAM_HPP_
#define _NMM_VALUESTREAM_HPP_

#include <string>
#include <list>
#include <vector>

#include "nmm/nmm_strstream.hpp"


using namespace std;


namespace NMM {

class Value;

/** This struct stores the type inforamtion of a class. Therfore a string is used to store
  * the type name and a string for the namespace.
  */
struct TypeInfo {
/** @name Constructor & Destructor
  */
//@{
/** Constructor, wich accepts the type name and a namespace.
  */
  explicit TypeInfo(const string& name, const string& name_space) : t_name(name), t_namespace(name_space) {}

/** Constructor, which accepts the type name
  */
  explicit TypeInfo(const string& name) : t_name(name) {}

/** Constructor
  */
  explicit TypeInfo() {}

         bool operator==(const TypeInfo& to) const ;
  inline bool operator!=(const TypeInfo& to) const { return !( *this == to ); }
  
  inline bool operator<(const TypeInfo& to) const {
    return (t_name < to.t_name) || (t_name == to.t_name && t_namespace < to.t_namespace);
  }
  
  string t_name;                ///< stores the type name
  string t_namespace;           ///< stores the namespace
};

/** This class is base class for all OStreams and OValueStreams nad defines operator to write
  * the basic c++ types. Several OStreams can be combined to creat a stack of Streams. All values
  * are forward from stream to stream until the data have the desired format. So an XMLStream
  * can be combined with an ZIP-Stream. First the XMLStream serialize data and the ZIPStream
  * compress the XML-structure.
  */
class OStream  {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  OStream() : __partner_stream(0) {}

/** Destructor
  */
  virtual ~OStream(){};
//@}

/** @name The operators
  */
//@{
/** This method writes a string i to the stream and returns a reference to the stream.
  */ 
  virtual OStream& operator<<(const string&)         = 0;

/** This method writes an u_int16_t i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const u_int16_t&)      = 0;

/** This method writes an int16_t i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const int16_t&   )     = 0;

/** This method writes an u_int32_t i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const u_int32_t&)      = 0;

/** This method writes an int32_t i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const int32_t&   )     = 0;

/** This method writes an u_int64_t i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const u_int64_t& )     = 0;

/** This method writes an int64_t i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const int64_t& )       = 0;

/** This method writes a bool i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const bool& )          = 0;

/** This method writes an unsigned long i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const unsigned long&)  = 0;

/** This method writes a long i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const long& )          = 0;

/** This method writes a double i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const double& )        = 0;

/** This method writes an float i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const float& )         = 0;

/** This method writes an char i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const char& )          = 0;

/** This method writes an unsigned char i to the stream and returns a reference to the stream.
  */
  virtual OStream& operator<<(const unsigned char&)  = 0;

/** This method must be called if no other values are written.
  */
  virtual void writeData() {};
//@}

/** @name Write char-array
  */
//@{
/** This method writes an char buffer of given size to the stream.
  * \param buffer Poitner to the char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const char* buffer, unsigned int size) = 0;

/** This method writes an unsigned char buffer of given size to the stream.
  * \param buffer Poitner to the unsigned char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const unsigned char* buffer, unsigned int size) = 0;
//@}

/** @name Access the Partner stream
  */
//@{
/** This method sets the partner stream, which receives the data
  */
  inline void  setOPartnerStream(OStream* partner_stream) { __partner_stream = partner_stream; }

/** This method returns the  current partner stream
  */
  inline       OStream* getOPartnerStream()               { return __partner_stream; }

/** This method returns the  current partner stream
  */
  inline const OStream* getOPartnerStream() const         { return __partner_stream; }
//@}
private:
  OStream* __partner_stream;              ///< Pointer to the parnter stream
};

/** An OValueStream stores additinoal type infomation that are used for a type safe reconstruction
  * or to create a prototype of the current value before its read from an IValueStream.
  */
class OValueStream  : public OStream {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  OValueStream() {}

/** Destructor
  */
  virtual ~OValueStream() {};

/** virtual copy constructor
  */
  virtual OValueStream* copy() const = 0;
//@}

/** @name The operators
  */
//@{
/** This method writes an u_int16 i to the stream and returns a reference to the stream.
  */ 
  virtual OValueStream& operator<<(const u_int16_t&)      = 0;

/** This method writes an int16 i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const int16_t&  )      = 0;

/** This method writes an u_int32 i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const u_int32_t&)      = 0;

/** This method writes an int32 i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const int32_t&   )     = 0;

/** This method writes an u_int64 i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const u_int64_t& )     = 0;

/** This method writes an int64 i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const int64_t& )       = 0;

/** This method writes a bool i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const bool& )          = 0;

/** This method writes an unsigned long i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const unsigned long&)  = 0;

/** This method writes a long i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const long& )          = 0;

/** This method writes a double i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const double& )        = 0;

/** This method writes a float i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const float& )         = 0;

/** This method writes a char i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const char& )          = 0;

/** This method writes a unsigned char i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const unsigned char&)  = 0;

/** This method writes a string i to the stream and returns a reference to the stream.
  */
  virtual OValueStream& operator<<(const string&)         = 0;
//@}
/** @name Write char-array
  */
//@{
/** This method writes an char buffer of given size to the stream.
  * \param buffer Poitner to the char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const char* buffer, unsigned int size) = 0;

/** This method writes an unsigned char buffer of given size to the stream.
  * \param buffer Poitner to the unsigned char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const unsigned char* buffer, unsigned int size) = 0;
//@}

/** @name Write TypeInfo
  */
//@{
/** This method starts a new compound type.
  * If no such magic number exist, the type information are written.
  * \param type The type name of the struct
  * \param ns namespace of the struct
  */
  virtual void beginType(const string& type, const string& ns = "std") = 0;

/** This method starts a new compound type.
  * If no such magic number exist, the type information are written.
  * \param type The TypeInfo of the struct
  */
  virtual void beginType(const TypeInfo& type);

/** This method set the end flag of a current type and must be called after all attributes of a
  * complex struct are written.
  */
  virtual void endType() = 0;

/** This method returns the type name of this ValueStream
  */
  virtual const string& getOValueStreamType() const = 0;
//@}
private:
  OValueStream(const OValueStream&);
public: //methods

/* Begin> Static converter methods to convert integers and floats to a string */
  inline static string convertToString(const int& i);
  inline static string convertToString(const unsigned int& i)   {return convertToString(reinterpret_cast<const int&>(i));}
  inline static string convertToString(const long& i);
  inline static string convertToString(const unsigned long& i)  {return convertToString(reinterpret_cast<const long&>(i));}
  inline static string convertToString(const char* i)           {return string(i);}
  inline static string convertToString(const unsigned char* i)  {return string(reinterpret_cast<const char*>(i));}
  inline static string convertToString(const double& i)         {strstream is; is << i; string tmp; is >> tmp; return tmp;}
  inline static string convertToString(const float& i)          {strstream is; is << i; string tmp; is >> tmp; return tmp;}
  inline static string convertToString(const u_int64_t& i)      {strstream is; is << i; string tmp; is >> tmp; return tmp;}
  inline static string convertToString(const int64_t& i)        {strstream is; is << i; string tmp; is >> tmp; return tmp;}
  inline static const string& convertToString(const string& i)  {return i;}
/* End< */
};


/** This class is base class for all IStreams and IValueStreams and defines operator to read
  * the basic c++ types. Several IStreams can be combined to create a stack of Streams, so the
  * data can be reconstruct according to the used serialization streams. 
  */
class IStream {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  IStream() : __partner_stream(0) {}

/** Destructor
  */
  virtual ~IStream() {}
//@}

/** @name The operators
  */
//@{
/** This method reads an u_int16 i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(u_int16_t&)       = 0;

/** This method reads an int16 i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(int16_t&)         = 0;

/** This method reads an u_int32 i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(u_int32_t&  )     = 0;

/** This method reads an int32 i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(int32_t&  )       = 0;

/** This method reads an u_int64 i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(u_int64_t&  )     = 0;

/** This method reads an int64 i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(int64_t&  )       = 0;

/** This method reads a bool i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(bool&  )          = 0;

/** This method reads an unsigned long i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(unsigned long&)   = 0;

/** This method reads a long i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(long& )           = 0;

/** This method reads a double i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(double& )         = 0;

/** This method reads a float i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(float& )          = 0;

/** This method reads a char i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(char& )           = 0;

/** This method reads an unsigned char i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(unsigned char&)   = 0;

/** This method reads a string i from the stream and returns a reference to the stream.
  */
  virtual  IStream& operator>>(string&)          = 0;

/** This method can be used to read all need infromation from the partner stream.
  */
  virtual void readData() {}
//@}

/** @name Read char-array
  */
//@{
/** This method reads size chars and store they in the given buffer.
  * \param buffer The buffer to store the read char's. The minimum size of the array must be size
  * \param size The size of char's that should be read
  */
  virtual void readBuffer(char*, unsigned int size)        = 0   ;

/** This method reads size unsigned chars and store they in the given buffer.
  * \param buffer The buffer to store the read unsigned char's. The minimum size of the array must be size
  * \param size The size of char's that should be read
  */
  virtual void readBuffer(unsigned char* buffer, unsigned int size) = 0;
//@}

/** @name Access the Partner stream
  */
//@{
/** This method sets the partner stream, which is used to read the data
  */
  virtual void  setIPartnerStream(IStream* partner_stream) { __partner_stream = partner_stream; }

/** This method returns the  current partner stream
  */
  virtual       IStream* getIPartnerStream()               { return __partner_stream; }

/** This method returns the  current partner stream
  */
  virtual const IStream* getIPartnerStream() const         { return __partner_stream; }

/** This method returns true if no values are inside the OStream.
  */
  virtual bool isEmpty()    const = 0;
//@}
private:
  IStream* __partner_stream;          ///< stores the current partner stream
};

/** An IValueStream reads additinoal type information and can be asked for the current type.
  */
class IValueStream : public IStream {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  IValueStream() {}

/** Destructor
  */
  virtual ~IValueStream() {}

/** Virtual copy constructor
  */
  virtual IValueStream* copy() const = 0;
//@}

/** @name The operators
  */
//@{
/** This method reads an u_int16_t i from the stream and returns a reference to the stream.
  */  
  virtual  IValueStream& operator>>(u_int16_t&)       = 0;

/** This method reads an int16_t i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(int16_t&)         = 0;

/** This method reads an u_int32_t i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(u_int32_t&  )     = 0;

/** This method reads an int32_t i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(int32_t&  )       = 0;

/** This method reads an u_int64_t i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(u_int64_t&  )     = 0;

/** This method reads an int64_t i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(int64_t&  )       = 0;

/** This method reads a bool i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(bool&  )          = 0;

/** This method reads an unsigned long i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(unsigned long&)   = 0;


/** This method reads a long i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(long& )           = 0;

/** This method reads a double i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(double& )         = 0;

/** This method reads a float i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(float& )          = 0;

/** This method reads a char i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(char& )           = 0;

/** This method reads an unsigned char i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(unsigned char&)   = 0;

/** This method reads a string i from the stream and returns a reference to the stream.
  */
  virtual  IValueStream& operator>>(string&)          = 0;
//@}

/** @name Navigate through the attributes
  */
//@{
/** This method accesses the internal attributes of complex structure.
  */
  virtual bool goDown()    = 0;

/** This method leaves the internal attributes of complex structure.
  */
  virtual void goUp()      = 0;

/** This method reads all information from the next element
  */
  virtual bool nextElement() = 0;
//@}

/** @name Navigate through the attributes
  */
//@{
/** This method returns the current type information in a TypeInfo struct
  */
  virtual  const TypeInfo& getCurrentTypeInfo() const = 0;

/** This method returns the type name of the current element. The Stream must not be empty!
  * \sa isEmpty();
  */
  virtual const string& getCurrentType() const = 0;

/** This method returns the namespace of the current element. The Stream must not be empty!
  * \sa isEmpty();
  */
  virtual const string& getCurrentNamespace() const = 0;

/** This method returns the type name of this IValueStream.
  */
  virtual const string& getIValueStreamType() const = 0;
//@}
private:
  IValueStream(const IValueStream&);
};


string OValueStream::convertToString(const int& i) {
  strstream stream;
  stream << i;
  string r_value;
  stream >> r_value;
  return r_value;
}

string OValueStream::convertToString(const long& i)           {
  strstream stream;
  stream << i;
  string r_value;
  stream >> r_value;
  return r_value;
}


/** This class is used in the ValueFactory to create an IValueStream and the corresponding OValueStream.
  * If an instance of this class is created it automatically registers themself by the ValueFactory.
  */
class ValueStreamConstructor {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  * \param i_stream The type name of the IValueStream
  * \param o_stream The type name of the OValueStream
  */
  ValueStreamConstructor(const string& i_stream, const string& o_stream);

/** Destructor
  */
  virtual ~ValueStreamConstructor() {}

/** Copy constructor
  */
  virtual ValueStreamConstructor* copy() const = 0;
//@}

/** @name Factory methods
  */
//@{
/** This method creates an IValueStremam. The caller of this method is responsible to delete the object.
  */
  virtual IValueStream* createIValueStream()     const = 0;

/** This method creates an OValueStremam. The caller of this method is responsible to delete the object.
  */
  virtual OValueStream* createOValueStream()     const = 0;
//@}
private:
  string __i_stream;        ///< type name of the IValueStream
  string __o_stream;        ///< type name of the OValueStream
};


/** This class implements the ValueStreamConstructor using the new operator to create the
  * IValueStream and the corresponding OValueStream.
  */
template<class T, class V>
class TValueStreamConstructor : public ValueStreamConstructor {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  * \param i_stream The type name of the IValueStream
  * \param o_stream The type name of the OValueStream
  */
  TValueStreamConstructor(const string& i_stream, const string& o_stream)
    : ValueStreamConstructor(i_stream, o_stream)
  { }

/** Destructor
  */
  virtual ~TValueStreamConstructor() {}

/** Virtual copy constructor
  */
  virtual ValueStreamConstructor* copy() const { return new TValueStreamConstructor<T,V>(*this); }

//@}
/** @name Factory methods
  */
//@{
/** This method creates an IValueStremam. The caller of this method is responsible to delete the object.
  */
  virtual IValueStream* createIValueStream()     const  { return new T(); }

/** This method creates an OValueStremam. The caller of this method is responsible to delete the object.
  */
  virtual OValueStream* createOValueStream()     const  { return new V(); };
//@}
};

/* Begin> static TypeInfo's for the C++ types */
static const TypeInfo   int_type_info  (  "int");
static const TypeInfo   int16_type_info(  "int16_t");
static const TypeInfo u_int16_type_info("u_int16_t");
static const TypeInfo   int32_type_info(  "int32_t");
static const TypeInfo u_int32_type_info("u_int32_t");
static const TypeInfo   int64_type_info(  "int64_t");
static const TypeInfo u_int64_type_info("u_int64_t");

static const TypeInfo bool_type_info("bool");

static const TypeInfo unsigned_long_type_info("unsigned_long");
static const TypeInfo long_type_info("long");
static const TypeInfo double_type_info("double");
static const TypeInfo float_type_info("float");

static const TypeInfo char_type_info("char");
static const TypeInfo unsigned_char_type_info("unsigned_char");
static const TypeInfo char_ptr_type_info("char_ptr");
static const TypeInfo unsigned_char_ptr_type_info("unsigned_char_ptr");
static const TypeInfo string_type_info("string", "std");
/* End */

/* Begin> size of the the C++ types in bytes */
static const unsigned int size_of_bool        = 1;
static const unsigned int size_of_char        = 1;
static const unsigned int size_of_u_char      = 1;
static const unsigned int size_of_int         = 4;
static const unsigned int size_of_int16_t     = 2;
static const unsigned int size_of_int32_t     = 4;
static const unsigned int size_of_int64_t     = 8;
static const unsigned int size_of_u_int16_t   = 2;
static const unsigned int size_of_u_int32_t   = 4;
static const unsigned int size_of_u_int64_t   = 8;
static const unsigned int size_of_float       = 4;
static const unsigned int size_of_double      = 8;
static const unsigned int size_of_long        = 4;
static const unsigned int size_of_long_long   = 8;
/* End< */

}

#endif // _NMM_ValueStream_HPP_


