/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_BUFFERED_NET_STREAM_HPP
#define NMM_BUFFERED_NET_STREAM_HPP

#include "NetStream.hpp"

namespace NMM {
  
  class Sock;
  
  /**
   *  @brief Buffered network stream
   *
   *  This class implements a buffered network stream. Data is written to a buffer and written
   *  to the network when the buffer overflows. If more than the requested number of bytes can be read,
   *  then the surplus data is stored in a buffer for subsequent read operations to reduce the number
   *  of actual socket reads.
   */
  class BufferedNetStream : public NetStream {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    BufferedNetStream(Sock&);

    /** Destructor
     */
    virtual ~BufferedNetStream();

    /** virtual Copy constructor
     */
    virtual BufferedNetStream* copy() const;
    //@}
    
    /** Write all data to the stream.
     * This method flushes the internal buffer.
     */
    virtual void writeData();

    /** returns true if the internal buffer of this NetStream is empty
      */
    virtual bool isEmpty() const;

    /** resizes the internal sending buffer
      */
    void setSendBuffer(size_t size);

    /** resizes the internal receiving buffer
      */
    void setReceiveBuffer(size_t size);

    /** resizes the internal sending buffer
      */
    size_t getSendBuffer()      const  { return __send_buffer.size(); }

    /** resizes the internal receiving buffer
      */
    size_t getReceiveBuffer()    const  { return __recv_buffer.size(); }
        
  protected:
    /** Internal mehtod to write the date to the network
     */
    void writeValue(const void*, size_t);

    /** Internal mehtod to read the date from the network
     */
    void readValue(void*, size_t);

    /** Flush internal buffers
     */
    void flush();
    
    /** Internal method to write directly to socket
     */
    void internalWriteValue(const void* value, size_t size);
    
    /** Internal method to read directly from socket
     */
    size_t internalReadValue(void* value, size_t size);
    
  private:
    vector<char>  __send_buffer;        ///< temporary send buffer
    size_t        __send_buffer_used;   ///< stores the used send buffer size
    
    vector<char>  __recv_buffer;        ///< temporary recv buffer
    size_t        __recv_buffer_begin;  ///< beginning of used area in receive buffer
    size_t        __recv_buffer_end;    ///< end of used area in receive buffer
  };
}
#endif





