/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_CONTROLLER_HPP_
#define _NMM_CONTROLLER_HPP_

#include "nmm/base/StdNamedObject.hpp"
#include "nmm/base/Node.hpp"
#include "nmm/interfaces/base/sync/IControllerImpl.hpp"

#include "TimedElement.hpp"

namespace NMM {

  class Synchronizer;

  //! This class represents node controllers for synchronization.
  class Controller : public TimedElement,
                     public StdNamedObject,
		     virtual public IControllerImpl {
  public:
    static const IController::Mode DEFAULT_MODE = IController::OFFLINE; 
    //!< The default mode should be offline, so that no buffer can be lost. */
    
  private:
    IController::Mode mode;

    //! Checks with a given timestamp if the related buffer is valid.
    /*!
     * This method must be reimplemented by the subclasses.
     * In the ONLINE mode, the decision what to do with
     * the buffer - discard or use it - will be made here.
     * \param timestamp The timestamp of the buffer to check.
     * \return true if the buffer is valid.
     */
    virtual bool checkBufferValid(const Timestamp timestamp) = 0;
    
  protected:
    /** true, if controller should discard buffers if necessary. */
    bool discard_buffers;

    /** true, if all buffers should be regarded as invalid */
    bool invalidate_buffers;

    /** true, invalidate_buffers was set to different value */
    bool invalidate_buffers_changed;

    /** timestamp starting from which buffers are regarded as invalid if invalidate_buffers == true */
    Timestamp invalidate_buffers_ts;
    
    /** locks code in isBufferValid vs. setBuffersInvalid */
    ThreadMutex invalidate_mutex;

  public:
    //! Constructor which just sets the mode and no parent node.
    /*!
     * The mode of the controller is set to mode, the name is set to 
     * "Controller".
     */
    Controller(const IController::Mode mode);

    //! The main constructor. (Arguments: parent node and mode.)
    /*!
     * The mode is set to mode, the name is set to the name of the parent
     * node + " Controller". If parent = 0, the name is simply set to
     * "Controller" without a node id.
     */
    Controller(const Node* parent = 0, const IController::Mode mode = DEFAULT_MODE);

    //! The main destructor.
    virtual ~Controller();

    //! This method is part of the interface between controller and node.
    /*!
     * This method is called by source nodes when they have produced a new
     * buffer and by sink nodes when they have dequeued the next buffer.
     * If the controller is in the OFFLINE mode, the method always will 
     * return with true, and each buffer is recognized as valid.
     * If the controller is in the ONLINE mode, the checkBufferValid - 
     * method is called. This method will be reimplemented in the 
     * subclasses.
     * \param timestamp The timestamp of the newly produced or discarded 
     *                  buffer.
     * \return true if this buffer should sent through the rest of the 
     *              pipeline, false if the buffer should be discarded.
     */
    bool isBufferValid(const Timestamp& timestamp);
    
  protected:
    //! Sets the mode of the controller.
    /*!
     * \param mode The new mode of the controller.
     * \return SUCCESS if the mode is valid.
     */
    virtual Result setMode(const IController::Mode& mode);

    /**
     * Permit discarding buffers.
     * Set the flag that allows or disallows discarding buffers.
     * Set by default. In ONLINE mode, the controller may discard buffers if this flag is set.
     */
    virtual Result permitDiscardBuffers(const bool& value);

    /**
     * The controller is told to pause the node via this method
     */
    virtual Result pause() = 0;

    /**
     * The controller is told to wake up the node via this method
     */
    virtual Result wakeup() = 0;

    /**
     * The controller is told to reset the node via this method
     */
    virtual Result reset() = 0;

    /**
     * Sets whether buffers should always be regarded as invalid (and therefore not presented)
     * or if the usual checks should be performed to check whether a buffer is valid or not.
     * This allows to enable/disable a node.
     * @param status true for marking all buffers as invalid
     */
    virtual void setBuffersInvalid(const bool& status);

    /**
     * Sets whether buffers should always be regarded as invalid (and therefore not presented)
     * or if the usual checks should be performed to check whether a buffer is valid or not.
     * This allows to enable/disable a node starting from a particular time.
     * @param status true for marking all buffers as invalid
     * @param status_time timestamp from which on the new status should be active
     */
    virtual void setBuffersInvalidTimed(const bool& status, const Timestamp& status_time);

    /**
     * Can be used to estimate the delay between method call and execution
     * @return the system time when the method call was executed
     */
    virtual Time ping();

  };
}

#endif // _NMM_CONTROLLER_HPP_



