/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_LOCAL_TIME_HPP
#define _NMM_LOCAL_TIME_HPP

#include "Types.hpp"
#include "nmm/base/serialize/ComplexValue.hpp"

#include <ctime>

namespace NMM {
  
  /**
   *  @brief Local time
   *
   *  Represents broken down calendar time. This is a wrapper of the C struct 'tm'.
   */
  class LocalTime : public SerializedValue {
  public:
    /**
     *  @brief Constructor
     *  @param time A time value
     *  @param local Create local time
     *
     *  Creates the broken down time representation of the given @a time.
     *  If @a local is true, the broken down time is in local time. Otherwise it is in GMT.
     */
    LocalTime(const Time& time, bool local=true);
    
    /**
     *  @brief Constructor
     *  @param tm_struct C structure representing broken down time
     *
     *  Creates the broken down time directly from the C structure @a tm_struct.
     */
    LocalTime(const struct ::tm& tm_struct);
    
    /**
     *  @brief Copy constructor
     */
    LocalTime(const LocalTime&);
    
    /**
     *  @brief Get second
     *  @return The second of the minute (0~59)
     */
    unsigned getSecond() const;
    
    /**
     *  @brief Get minute
     *  @return The minute of the hour (0~59)
     */
    unsigned getMinute() const;

    /**
     *  @brief Get hour
     *  @return The hour of the day (0~23)
     */
    unsigned getHour() const;

    /**
     *  @brief Get day of month
     *  @return The day of the month (1~number of days for this month)
     */
    unsigned getDayOfMonth() const;

    /**
     *  @brief Get day of week
     *  @return The day of the month (0~6 where 0=Sunday)
     */
    unsigned getDayOfWeek() const;

    /**
     *  @brief Get day of year
     *  @return The day of the month (0~number of days for this year-1)
     */
    unsigned getDayOfYear() const;
    
    /**
     *  @brief Get month
     *  @return The month of the year (0~11)
     */
    unsigned getMonth() const;
    
    /**
     *  Get GMT offset.
     *  @return The timezone as GMT offset in seconds
     *
     *  Returns the time zone that was used to compute
     *  this broken-down time value, including any adjustment for
     *  daylight saving; it is the number of seconds that you must
     *  add to UTC to get local time.  You can also think of this as
     *  the number of seconds east of UTC.  For example, for U.S.
     *  Eastern Standard Time, the value is `-5*60*60'.
     */
    unsigned getGMTOffset() const;
    
    /**
     *  @brief Get year.
     *  @return The year
     */
    unsigned getYear() const;
    
    /**
     *  @brief Check for daylight savings time
     *  @return Daylight savings time flag
     *
     *  This is a flag that indicates whether Daylight Saving Time is
     *  (or was, or will be) in effect at the time described.
     */
    bool isDaylightSavingsTime() const;
    
    /**
     *  @brief Get time zone
     *  @return Name of time zone
     *
     *  This field is the name for the time zone that was used to
     *  compute this broken-down time value.
     */
    string getTimeZone() const;
    
    /**
     *  @brief Reset time
     *  @param tm_struct C structure representing broken down time
     *
     *  Creates the broken down time directly from the C structure @a tm_struct.
     */
    void reset(const struct ::tm& tm_struct);
    
    /**
     *  @brief Reset time
     *  @param time A time value
     *  @param local Create local time
     *
     *  Creates the broken down time representation of the given @a time.
     *  If @a local is true, the broken down time is in local time. Otherwise it is in GMT.
     */
    void reset(const Time& time, bool local=true);
    
    /**
     *  @brief Get string representation
     *  @param format String describing format
     *
     *  Generate a string representation of this time. The format of the representation
     *  is determined by the @a format string. The string is interpreted in the same way
     *  as the format argument of the ISO C99 function strftime. See 'info libc' for details.
     */
    string toString(const string& format) const;
    
    /**
     *  @brief Get type info for this class
     */
    static TypeInfo getStaticTypeInfo();
    
  private:
    friend class TValue<LocalTime>;
    friend class TValue<LocalTime*>;
    friend class TConstructor<LocalTime>;
    
    /**
     *  @brief Default constructor
     */
    LocalTime();
    
    /**
     *  @brief Register attributes
     */
    void registerAttributes();
    
    /**
     *  @brief Update attributes
     *
     *  Updates the internal C structure after deserialization.
     */
    virtual void updateAttributes();
    
  private:
    /** Second.
     */
    TValue<u_int8_t>  __second;
    
    /** Minute.
     */
    TValue<u_int8_t>  __minute;
    
    /** Hour.
     */
    TValue<u_int8_t>  __hour;
    
    /** Day of month.
     */
    TValue<u_int8_t>  __day_of_month;
    
    /** Day of week.
     */
    TValue<u_int8_t>  __day_of_week;
    
    /** Day of year.
     */
    TValue<u_int8_t>  __day_of_year;
    
    /** Month of year.
     */
    TValue<u_int8_t>  __month;
    
    /** GMT offset.
     */
    TValue<u_int8_t>  __gmt_offset;
    
    /** Year.
     */
    TValue<unsigned>  __year;
    
    /** Daylight savings time flag.
     */
    TValue<bool>      __is_dst;
    
    /** Time zone name.
     */
    TValue<string>    __time_zone;
    
    /** Internal C structure.
     */
    struct ::tm         __tm_struct;
  };
  
  inline unsigned LocalTime::getSecond() const {
    return __second.getRValue();
  }
  
  inline unsigned LocalTime::getMinute() const {
    return __minute.getRValue();
  }
  
  inline unsigned LocalTime::getHour() const {
    return __hour.getRValue();
  }
  
  inline unsigned LocalTime::getDayOfMonth() const {
    return __day_of_month.getRValue();
  }
  
  inline unsigned LocalTime::getDayOfWeek() const {
    return __day_of_week.getRValue();
  }
  
  inline unsigned LocalTime::getDayOfYear() const {
    return __day_of_year.getRValue();
  }
  
  inline unsigned LocalTime::getMonth() const {
    return __month.getRValue();
  }
  
  inline unsigned LocalTime::getGMTOffset() const {
    return __gmt_offset.getRValue();
  }
  
  inline unsigned LocalTime::getYear() const {
    return __year.getRValue();
  }
  
  inline bool LocalTime::isDaylightSavingsTime() const {
    return __is_dst.getRValue();
  }
  
  inline string LocalTime::getTimeZone() const {
    return __time_zone.getRValue();
  }
  
}

#endif
