/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_SINK_CONTROLLER_HPP_
#define _NMM_SINK_CONTROLLER_HPP_

#include "Controller.hpp"
#include "nmm/interfaces/base/sync/ISinkControllerImpl.hpp"

namespace NMM {
  class GenericSyncSinkNode;
  
  //! This class represents a synchronization controller for sink nodes.
  /*!
   *  In addition to the isBufferValid - method from controller, a 
   *  sinkcontroller has a waitToPresent - method which allows it to
   *  tell the sink node how long it should wait before presenting the 
   *  next buffer.
   */
  class SinkController : public Controller,
                         virtual public ISinkControllerImpl {
  protected:
    bool     block_for_sync_reset; //!< Tells the controller whether to block its parent during syncReset
    bool     notify_in_time;    //!< Tells the controller to notify its first buffer in time
    bool     first_buffer_in_time; //!< True for the first buffer(s) in the stream that did not arrive in time
    bool     wait_to_present;   //!< Tells the controller to calculate interval to wait 
    Interval max_average_delay; //!< If the average latency is greater than theo_latency
                                //!< + max_average_delay, the synchronizer is informed.
    Interval max_delay;         //!< If this value is exceeds from a single buffer,
                                //!< the synchronizer should be informed.
    Interval theo_latency;      //!< The theoretical latency set by the synchronizer.
    Interval default_theo_latency; //!< The default value for theoretical latency if not set by the synchronizer.
    Interval average_latency;   //!< The average latency of incoming buffers (computed by
                                //!< the sink controller).
    Interval min_wait_interv;   //!< Stores the value set by setMinWaitInterval    

    int average_count;          //!< The number of buffers for the average 
                                //!< latency computation.
    int n;                      //!< The counter for the average latency computation.  
    int max_disc;               //!< The number of maximal discarded buffers.
    int disc_count;             //!< The number of actual discarded buffers.
    
    bool first_buffer;          //!< True for the first buffer in a stream.
    
    int sink_id;                //!< ID of this controller at the synchronizer
    
    Timestamp old_ts;           //!< The last buffer's timestamp.
    
    GenericSyncSinkNode* parent_node;
    ISinkSynchronizer* synchronizer;
    
    //! The states of the SinkController.
    /*! 
     *  The states are used to avoid problems with the sync_reset-event and multiple
     *  streams. If an sync_reset-event is sent and a demux distributes it to 
     *  multiple streams, the new events will not arrive at the sink nodes in the 
     *  same moment. 
     */
    enum State {
      WORKING,           //!< The default state. Synchronization is active.
      WAIT_FOR_RESET,    //!< The sink controller waits for a sync_reset event at the
                         //!< sink node, i.e. a call of the reset-method.
      WAIT_FOR_RESTART,  //!< The controller has received a sync_reset and waits to be restarted by the synchronizer
      PAUSE
    };
    
    //! The actual state of the SinkController.
    State state;
    
    //! Mutex to guarantee a consistent state variable.
    ThreadMutex state_mutex;
    
    //! Initialize Variables for both constructors
    void init_variables();
    
    //! Resets the parameters of average_latency computation.
    virtual void reset_average();
    
  public:
    //! Constructor that only sets the mode.
    /*!
     *  The name is "Controller" as default.
     */
    SinkController(const IController::Mode mode);
    
    //! The main constructor.
    /*!
     *  Sets the name and the mode, default values are "Controller" and
     *  DEFAULT_MODE (OFFLINE).
     */
    SinkController(GenericSyncSinkNode* parent = 0, 
		   const IController::Mode mode = Controller::DEFAULT_MODE);
    
    //! The main destructor.
    virtual ~SinkController();
    
     //! This method is by a Controller::isBufferValid - call. 
     /*!
      *  It checks if the buffer should be prepared for presenatation or discarded. 
      */
     virtual bool checkBufferValid(const Timestamp timestamp);
     
     //! Tells the node how long it should wait until presenting the next buffer.
     /*!
      *  This method is called by the GenericSyncSinkNode to get to know how long
      *  it has to wait until the optimal time for presentation of this buffer 
      *  will be reached.
      *  \param timestamp The timestamp of the buffer that should be presented.
      *  \return An interval which represents the time to wait.
      */
     virtual Interval waitToPresent(const Timestamp timestamp); 

    /**
     * Resets the internal parameters.
     *
     *  This method is called by the GenericSyncSinkNode when a sync_reset-event
     *  occurs. The internal parameters of synchronization are reinitialized.
     *  It also changes the state of the controller
     *          from WORKING        to RESET   or
     *          from WAIT_FOR_RESET to WORKING.
     *  The reset-method of the synchronizer (if there is one) is called.
     */
    virtual Result reset();

    virtual Result pause();

    virtual Result wakeup();

  protected:
    /** Tells the controller the maximal delay.
     *
     * With this method you can tell the controller which delay of an
     * incoming buffer should be accepted.
     * In the isBufferValid - method, the controller computes a 
     * theoretical presentation time for this buffer. If the real time
     * in this moment is greater than this theoretical presentation time
     * plus the maximal delay, the buffer will be recognized as invalid.
     * @param max_delay The maximal delay that will be allowed.
     * @return SUCCESS if the delay could be set (if it is greater than
     *                 or equal zero),
     *         FAILURE otherwise.
     */
    virtual Result setMaxDelay(const Interval& value);

    /**
     * If the average latency exceeds this value, the synchronizer gets to know 
     * that.
     */
    virtual Result setMaxAverageDelay(const Interval& value);

    /**
     * Sets the default theoretical latency.
     *
     * This method is called by the synchronizer to tell the controllers the
     * default theoretical latency of the pipeline.
     */
    virtual Result setDefaultTheoLatency(const Interval& value);

    /**
     * Used for the communication with the synchronizer.
     *
     * This method is called by the synchronizer to tell the controllers the
     * theoretical latency of the pipeline.
     */
    virtual Result setTheoLatency(const Interval& value);

    /**
     * Sets the synchronizer which refers to the stream.
     * @param sync synchronizer
     * @param id id of this controller at the synchronizer
     */
    virtual Result setSinkSynchronizer(const ISinkSynchronizer* value, const int& id);

    /**
     * Sets the number of buffers used to compute the average latency.
     * A small average_count means that the synchronizer is told the actual 
     * average_latency of incoming buffers very often. That means the
     * synchronization reacts very quickly (and this reaction may be visible)
     * For this reason higher values (such as 50 for audio and 20 for video) 
     * are recommended.
     */
    virtual Result setAverageCount(const int& value);

    /**
     * Determines the length of possible gaps if buffers arrive too late. 
     *
     * Sets the maximal number of buffers that can be discarded before at least
     * one buffer has to be presented.
     */
    virtual Result setMaxDiscardingCount(const int& value);

    /**
     * This method is called by the AudioVideoSynchronizer.
     *
     *  It changes the state 
     *          from WORKING to WAIT_FOR_RESET or
     *          from RESET   to WORKING.
     */
    virtual Result waitForReset();

    /**
     * Restarts the SinkController.
     * The state is changed from RESET to WORKING.
     */
    virtual Result restart();

    /**
     * Tells the node to wait for the time stamp to become valid
     * or to simply present the buffer when it arrives
     * @param enable true to enable waiting, false to disable
     */
    virtual Result enableWaitToPresent(const bool& enable);

    /**
     * Sets the minimum interval that a sink controller will
     * return when it is asked how long the node should wait
     * until it presents the current buffer (usually by the
     * method waitToPresent()). If the interval to wait is 
     * smaller than the interval set with this method, the
     * sink controller should return a zero interval.
     */
    virtual Result setMinWaitInterval(const Interval& min_wait);

    /**
     * Set whether the sink controller should notify its
     * sink synchronizer if the first buffer after a syncReset
     * arrived in time (in regard to the theoretical latency)
     * The notification is done by calling ISinkSynchronizer::setFirstBufferInTimeAverageLatency
     * @param enable true if the controller should notify 
     *               its synchronizer
     */
    virtual void enableFirstBufferInTime(const bool& enable);

    /**
     * Enables/disables the blocking of the controller (and
     * therefore the blocking of its parent node) when waiting
     * for restart() after having received a syncReset.
     * Blocking is enabled per default.
     * @param enable true to enable blocking; else false
     */
    virtual void enableBlocking(const bool& enable);
  };

}

#endif // _NMM_SINK_CONTROLLER_HPP_



