
#ifndef NMM_IDVDImpl_HPP
#define NMM_IDVDImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IDVD.hpp"

#include "nmm/interfaces/device/IDeviceImpl.hpp"

namespace NMM
{
  /**
   *  @brief DVD Interface
   *
   *  This interface is used for direct navigation on DVDs. It can be used for determining the number
   *  of chapters, titles and view angles and for selecting which title, chapter and view angle to play.
   *
   *  The interface is derived from IDevice to provide methods for getting and setting the DVD device.
   */

  class IDVDImpl  : virtual public IDeviceImpl { 
  public:
    IDVDImpl() { addInterface(new IDVD(*this)); registerEvents(); };

    virtual ~IDVDImpl() {};

  protected:
    /**
     *  @brief Select chapter, title and view angle
     *  @param title Title to play
     *  @param chapter Chapter to play
     *  @param angle View angle to play
     *  @return SUCCESS if the method succeeds, otherwise FAILURE
     *
     *  This method sets the @a title, @a chapter and view @a angle to play. The title must be between 1 and getTitles()
     *  inclusive, the chapter must be between 1 and getChapters(title) inclusive and the view angle must be between
     *  1 and getAngles(title) inclusive. Otherwise the method fails.
     */

    virtual Result select(const unsigned int& title, const unsigned int& chapter, const unsigned int& angle) = 0;
    /**
     *  @brief Get current title
     *  @return The current title number
     *
     *  This method returns the number of the title currently being played. If no title is being played, or the
     *  current title is invalid, the result is zero.
     */

    virtual unsigned int getCurrentTitle() const = 0;
    /**
     *  @brief Get current chapter
     *  @return The current chapternumber
     *
     *  This method returns the number of the chapter currently being played. If no chapter is being played, or the
     *  current chapter is invalid, the result is zero.
     */

    virtual unsigned int getCurrentChapter() const = 0;
    /**
     *  @brief Get number of titles
     *  @return The number of titles on the DVD
     *
     *  This method returns the number of titles on the DVD, which is also the highest possible title number.
     */

    virtual unsigned int getTitles() const = 0;
    /**
     *  @brief Get number of chapters in a title
     *  @param title Title number
     *  @return The number of chapters in the title
     *
     *  This method returns the number of chapters in the given @a title, which is also the highest possible chapter
     *  number for that title. The title number must be between 1 and getTitles() inclusive.
     */

    virtual unsigned int getChapters(const unsigned int& title) const = 0;
    /**
     *  @brief Get number of view angles in a title
     *  @param title Title number
     *  @return The number of angles in the title
     *
     *  This method returns the number of angles in the given @a title, which is also the highest possible angle
     *  number for that title. The title number must be between 1 and getTitles() inclusive.
     */

    virtual unsigned int getAngles(const unsigned int& title) const = 0;
    /**
     *  @brief Send a flush event
     *
     *  This method causes the ndoe to send a flush event.
     *  Currently, applications using the DVDNavReadNode do this explicitly before changing the chapter.
     *  If this is really necessary, the node should do it itself, but it has to be tested.
     */

    virtual void sendFlush() = 0;
    /**
     *  @brief Reset the DVD
     *
     *  Reset the DVD to the beginning.
     */

    virtual void reset() = 0;


  private:
    void registerEvents();
  };

}

#endif
