
#ifndef NMM_IWidgetFactory_HPP
#define NMM_IWidgetFactory_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "IWidget.hpp"
#include "IWidgetComposite.hpp"
#include "IProgressBar.hpp"
#include "ITextView.hpp"
#include "ITimeView.hpp"
#include "IImageView.hpp"
#include "IRadioButtons.hpp"
#include "IButtons.hpp"
#include "IBorderDecorator.hpp"
#include "IHeaderDecorator.hpp"
#include "IPlaneDecorator.hpp"
#include "IScrollDecorator.hpp"

namespace NMM
{
  /**
   * Widget factory.
   * This interface, which is implemented by the OSD manager, is used
   * for creating widget objects. I wrote a seperate interface for this for the
   * following reasons:
   * - It has a lot of methods only for creating widgets that would inflate IOSDManager
   * - It depends on all the widget interfaces. I don't want IOSDManager to include them all
   */

  class IWidgetFactory : virtual public Interface { 
  public:
    IWidgetFactory() {  setTypeInfo(TypeInfo(IWidgetFactory::IWidgetFactory_type, "NMM")); };

    IWidgetFactory(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(IWidgetFactory::IWidgetFactory_type, "NMM")); setEventReceiver(parent); };

    virtual ~IWidgetFactory() {};

    static const char* IWidgetFactory_type;

    virtual  string getInterfaceType() const { return IWidgetFactory_type; }

    Interface* copy() const { return new IWidgetFactory(*this); }
    static const char* createWidgetComposite_event;
    static const char* createProgressBar_event;
    static const char* createTextView_event;
    static const char* createTimeView_event;
    static const char* createImageView_event;
    static const char* createRadioButtons_event;
    static const char* createButtons_event;
    static const char* createBorderDecorator_event;
    static const char* createHeaderDecorator_event;
    static const char* createPlaneDecorator_event;
    static const char* createScrollDecorator_event;
    /**
     * Create a widget composite.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual IWidgetComposite* createWidgetComposite()
      throw(const Exception&);
    /**
     * Create a progress bar.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual IProgressBar* createProgressBar()
      throw(const Exception&);
    /**
     * Create a text view.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual ITextView* createTextView()
      throw(const Exception&);
    /**
     * Create a time view.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual ITimeView* createTimeView()
      throw(const Exception&);
    /**
     * Create an image view.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual IImageView* createImageView()
      throw(const Exception&);
    /**
     * Create radio buttons.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual IRadioButtons* createRadioButtons()
      throw(const Exception&);
    /**
     * Create buttons.
     * Creates a new widget and returns an interface for it.
     * @return an interface for this widget
     */

    virtual IButtons* createButtons()
      throw(const Exception&);
    /**
     * Create border decorator.
     * Creates a new widget and returns an interface for it.
     * @param internal the internal widget of the decorator
     * @return an interface for this widget
     */

    virtual IBorderDecorator* createBorderDecorator(const IWidget& internal)
      throw(const Exception&);
    /**
     * Create border decorator.
     * Creates a new widget and returns an interface for it.
     * @param internal the internal widget of the decorator
     * @return an interface for this widget
     */

    virtual IHeaderDecorator* createHeaderDecorator(const IWidget& internal)
      throw(const Exception&);
    /**
     * Create plane decorator.
     * Creates a new widget and returns an interface for it.
     * @param internal the internal widget of the decorator
     * @return an interface for this widget
     */

    virtual IPlaneDecorator* createPlaneDecorator(const IWidget& internal)
      throw(const Exception&);
    /**
     * Create scroll decorator.
     * Creates a new widget and returns an interface for it.
     * @param internal the internal widget of the decorator
     * @return an interface for this widget
     */

    virtual IScrollDecorator* createScrollDecorator(const IWidget& internal)
      throw(const Exception&);


  };

  typedef auto_ptr<IWidgetFactory> IWidgetFactory_var;




}

#endif
