/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
 * USA.
 */

#include "ConsoleProducer.hpp"


#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/base/IReceiver.hpp"

#include "nmm/utils/xml/xml++.hpp"

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <iostream>

namespace NMM {
  
  ConsoleProducer::ConsoleProducer() {
    initialized = false;
  }
  
  ConsoleProducer::~ConsoleProducer() {
    deinitTTY();
  }
  
  Result ConsoleProducer::initTTY() {
    GLOBAL_DEBUG_STREAM("ConsoleProducer: initTTY" << endl);
    if (initialized) {
      return SUCCESS;
    }
    
    /* poll stdin */
    setFileDescriptor(0);
    
    struct termios tty;
    
    /* get the current attributes of stdin */
    if (tcgetattr(0, &tty) < 0) {
      GLOBAL_ERROR_STREAM("ConsoleProducer: Could not get tty attributes!" << endl);
      throw Exception("ConsoleProducer: Could not get tty attributes!");
      return FAILURE;
    }
    
    savetty = tty;
    
    tty.c_lflag &= ~(ICANON|ECHO); /* no character echo also */
    tty.c_cc[VMIN] = 1;    /* limit: read at least 1 char */
    tty.c_cc[VTIME] = 0;   /* limit: read with no timeout */
    
    /* set the new attributes for stdin (now!) */
    if (tcsetattr(0, TCSANOW, &tty) < 0) {
      GLOBAL_ERROR_STREAM("ConsoleProducer: Could not set tty attributes!" << endl);
      throw Exception("ConsoleProducer: Could not set tty attributes!");
      return FAILURE;
    }
    
    initialized = true;
    return SUCCESS;
  }
  
  Result ConsoleProducer::deinitTTY() {
    GLOBAL_DEBUG_STREAM("ConsoleProducer: deinitTTY" << endl);
    if (!initialized) {
      return SUCCESS;
    }
    initialized = false;

    /* restore the old attributes for stdin (now!) */
    if (tcsetattr(0, TCSANOW, &savetty) < 0) {
      GLOBAL_ERROR_STREAM("ConsoleProducer: Could not set tty attributes!" << endl);
      return FAILURE;
    }
    
    return SUCCESS;
  }
  
  void ConsoleProducer::addKey(unsigned int key, string event) {
    if (key >= keymap.size()) {
      unsigned int fill = key - keymap.size();
      for (unsigned int i=0; i<fill; ++i) {
        keymap.push_back("");
      }
      keymap.push_back(event);
    } else {
      keymap[key] = event;
    }
  }
  
  string ConsoleProducer::getKeyMapping(unsigned int key) {
    if (key >= keymap.size()) {
      return "";
    } else {
      return keymap[key];
    }
  }
  
  Result ConsoleProducer::readConfig(string fname) {
    /* open XML document */
    XMLTree* xml = new XMLTree( fname );
    
    /* get pointer to XML root */
    XMLNode* rootnode = xml->root();
    if (rootnode == NULL) {
      GLOBAL_ERROR_STREAM("ConsoleProducer: readConfig(" << fname << "): no XML root node found" << endl);
      return FAILURE;
    }
    
    if (rootnode->name() != "configuration") {
      GLOBAL_ERROR_STREAM("ConsoleProducer: readConfig(" << fname << "): no <configuration> tag found" << endl);
      return FAILURE;
    }
    
    XMLNode* node1;
    
    /* parse the keyboard configuration */
    XMLNodeList nlist;
    XMLNodeIterator iter, stop;
    
    nlist = rootnode->children("keyboard");
    iter = nlist.begin();
    if ( iter == nlist.end() ) {
      GLOBAL_DEBUG_STREAM("ConsoleProducer: readConfig(" << fname << "): no <keyboard> entry found in configuration" << endl);
      return FAILURE;
    }
    
    node1 = *iter;
    
    /* Get a list with all <event> entries */
    nlist = node1->children( "event" );

    iter = nlist.begin();
    stop = nlist.end();
    
    /* Process each <event> entry */
    while ( iter != stop ) {
      XMLProperty* prop;
      XMLNode* node = *iter;
      iter++;
      
      /* Try to find <event> entry with "key" attribute */
      string key;
      prop = node->property( "key" );
      if ( prop ) key = prop->value();
      
      /* now the cdata-string of <event> tag containing the event-string */
      string event = node->getData();
      GLOBAL_DEBUG_STREAM("Found Key: " << key << " -> " << event << endl);
      
      int k;
      /* All "key" attributes with length=1 will correspond to the key itself */
      if ( key.length() == 1 ) {
        k = key[0];
        addKey(k, event);
      } else {  /* All other keys are keywords for the corresponding keys (like F1, SHIFT, ...) */
        GLOBAL_WARNING_STREAM("ConsoleProducer: can't handle key " << key << " << mapped to event " << event << endl);
        // fw: don't know how to read these special keys from cin (there must be another way to do this)
//         if ( key == "F1" )  add_key( XK_F1 , event );
//         if ( key == "F2" )  add_key( XK_F2 , event );
//         if ( key == "F3" )  add_key( XK_F3 , event );
//         if ( key == "F4" )  add_key( XK_F4 , event );
//         if ( key == "F5" )  add_key( XK_F5 , event );
//         if ( key == "F6" )  add_key( XK_F6 , event );
//         if ( key == "F7" )  add_key( XK_F7 , event );
//         if ( key == "F8" )  add_key( XK_F8 , event );
//         if ( key == "F9" )  add_key( XK_F9 , event );
//         if ( key == "F10" ) add_key( XK_F10 , event );
//         if ( key == "F11" ) add_key( XK_F11 , event );
//         if ( key == "F12" ) add_key( XK_F12 , event );

//         if ( key == "LEFT" )  add_key( XK_Left , event );
//         if ( key == "RIGHT" ) add_key( XK_Right , event );
//         if ( key == "UP" )    add_key( XK_Up , event );
//         if ( key == "DOWN" )  add_key( XK_Down , event );

//         if ( key == "ESC" )  add_key( XK_Escape , event );
//         if ( key == "RETURN" )  add_key( XK_Return , event );

//         if ( key == "PAGEUP"   ) { add_key ( XK_Prior , event ); }
//         if ( key == "PAGEDOWN" ) { add_key ( XK_Next , event ); }
//         if ( key == "HOME"     ) { add_key ( XK_Home , event ); }
//         if ( key == "END"      ) { add_key ( XK_End , event ); }
//         if ( key == "INSERT"   ) { add_key ( XK_Insert , event ); }
//         if ( key == "DELETE"   ) { add_key ( XK_Delete , event ); }

//         /* KP- means KeyPad keys */
//         if ( key == "KP-DIV"   ) { add_key ( XK_KP_Divide , event ); }
//         if ( key == "KP-MUL"   ) { add_key ( XK_KP_Multiply , event ); }
//         if ( key == "KP-SUB"   ) { add_key ( XK_KP_Subtract , event ); }
//         if ( key == "KP-ADD"   ) { add_key ( XK_KP_Add , event ); }
//         if ( key == "KP-ENTER" ) { add_key ( XK_KP_Enter , event ); }
      }
    }
    
    return SUCCESS;
  }
  
  void ConsoleProducer::handleInput() {
    char buf[256];
    int count = read(0, buf, 256);
    
    /* parse escape sequences */
    if (buf[0] == 27 || count > 1) {
      // XXX: fw: not yet...
      return;
    }
    
    /* Simulate a key press followed by a key release */
    string key = getKeyMapping(buf[0]);
    auto_ptr<Event> event;
    
    event.reset(IDisplayEvents::create_keyPressed(key));
    sendEvent(*event);
    
    event.reset(IDisplayEvents::create_keyReleased(key));
    sendEvent(*event);
  }
  
  void ConsoleProducer::start() {
    initTTY();
    
    InputProducer::start();
  }

  void ConsoleProducer::stop() {
    InputProducer::stop();
    
    deinitTTY();
  }
  
}

