/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_XPRODUCER_HPP_
#define _NMM_XPRODUCER_HPP_

#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define  XK_MISCELLANY
#include <X11/keysymdef.h>

#include "nmm/utils/thread/Thread.hpp"
#include "nmm/base/EventDispatcher.hpp"
#include "nmm/base/GenericProducer.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "XVirtualImg.hpp"

#include "nmm/base/IReceiver.hpp"

#define BUF_LEN 80
#define KEY_NUM 0x10000

struct button_coordinate_type {
  int x1;
  int y1;
  int x2;
  int y2;
  char *event;
};

typedef struct button_coordinate_type button_coordinate;

namespace NMM {
  
  /**
   * X11 Producer.
   * Produces NMM events based on X11 events.
   */
  class XProducer
    : public InputProducer
  {

  private:
    
    /**
     * Internal event receiver.
     * An event receiver which forwards events to an external event dispatcher.
     * Temporary, becomes obsolete as soon as EventDispatcher implements IEventReceiver.
     */
    class InternalReceiver
      : public IEventReceiver
    {
    private:
      EventDispatcher* dispatcher;
      
    public:
      InternalReceiver();
      
      void connectTo(EventDispatcher* dispatcher);
      
      Result receiveEvent(const CEvent&, const IEventSender* = 0 );
      Result receiveEvent(const Event&, const IEventSender* = 0 );
      
      Result receiveEvent(const CEvent&, const IEventSender* = 0 ) const;
      Result receiveEvent(const Event&, const IEventSender* = 0 ) const;
    };
    
    
    Display                   *display_id;	  // X-Display ID
    
    char                      *keyevent[KEY_NUM]; // (Mapping table) Keymapping to event-strings
    vector<button_coordinate> bc;                 // (Mapping table) for button coordinates
    
    bool                      x_auto_repeat;      // XAutoRepeat (keyboard repeat)
    
    bool SendingMouseEventsDirect; //instead of checking them against given Rectangles in mmbox e.g.
    bool UsingKeySymKeyEvents;        //instead of Events with strings
    
    XVirtualImg*              xvirtualimg;
    
    InternalReceiver          internal_receiver;
    
    /**
     * The event-handling function
     * Events can be received by all registered input-devices
     * (like X-Display, Console, mouse, ...)
     */
    void handleInput();
    
    /**
     * Delete all Key-Mappings
     */
    void delete_keyevents();

    /**
     * Add key with code k and corresponding event to mapping table
     */
    bool add_key( int k, string event);

    /**
     * Delete all mouse coordinates-Mappings
     */
    void delete_mousecoordinates();

    /**
     * Add mouse coordinates and corresponding event to mapping table
     */
    bool add_mousecoordinates( int x1, int y1, int x2, int y2, string event);

    /**
     * Returns event, if given point is within a rectangle in mapping table
     */
    char *checkcoordinate( int x, int y );

      

  public:
    enum { F1=0, F2, F3, F4, F5, F6, F7, F8, F9, F10, F11, F12,
	   LEFT, RIGHT, UP, DOWN, ESCAPE, RETURN, PAGEUP, PAGEDOWN,
	   HOME, END, INSERT, DELETE };
	   
     XProducer(XVirtualImg* xvi);
    ~XProducer();

    /**
     * This method returns whether XAutoRepeat is set to false
     * while its X-Window has focus
     */
    bool isXAutoRepeat();

    /**
     * This method sets the XAutoRepeat for
     * its X-Window.
     * If true, the keyboard autorepeat is disabled while
     * the X-Window is in focus and turned on again when
     * losing focus.
     */
    void setXAutoRepeat(const bool&);

    /**
     * Registration of X-Display Input-device
     */
    //void registerXDisplayNode(XDisplayNode *);

    /**
     * Registration of X-Display Input-device
     * This function registers the given Display-ID to the class.
     * Events that occure in the given Display will be handled by
     * the getEvent() function
     */
    void registerDisplay( Display* );

    /**
     * Reads a XML-configuration file.
     * In the configuration file are informations how events
     * will be mapped to the corresponding event-strings
     */
    bool readConfig( string );
    
    /**
     * retuns pressed Key (only usefull, if no receiver is connected)
     */
    bool getKey(char&);
      
      
    /**
     * Switches between processing only Mouseclicks, which are in predefined Screenrectangles or sending all MouseEvents.
     */
    void setSendingMouseEventsDirect(bool smed);
      
    /**
     * retuns Value of boolean Variable SendingMouseEventsDirect
     */
    bool isSendingMouseEventsDirect();

    /**
     * Switches between sending KeyEvents with strings or with KeySym representing the key triggering the event.
     */
    void setUsingKeySymKeyEvents(bool ukke);
      
    /**
     * retuns Value of boolean Variable UsingKeysymKeyEvents
     */
    bool isUsingKeySymKeyEvents();
    
    /**
     * Connect to event dispatcher.
     * Creates an internal event receiver with this dispatcher and connects to it.
     * Temporary, becomes obsolete as soon as EventDispatcher implements IEventReceiver.
     */
    void connectTo(EventDispatcher* dispatcher);
    
  };

}

#endif
