/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "TimerEvents.hpp"

namespace NMM {

  TimerProducer::TimerProducer(){
    cond = new ThreadCondition(mutex);
    
    new_time = 0;
  }
  
  TimerProducer::~TimerProducer(){
    delete cond;
  }
  
  void TimerProducer::start() {
    MutexGuard guard(mutex);
    activateThread();
  }
  
  void TimerProducer::stop() {
    cond->notify();
    deactivateThread();
  }
  
  void TimerProducer::registerTimer( int sec, int min, int h, int day, int mon, int year, Event* e ) {
    struct tm abs_time;
    
    abs_time.tm_sec  = sec;
    abs_time.tm_min  = min;
    abs_time.tm_hour = h;
    abs_time.tm_mday = day;
    abs_time.tm_mon  = mon-1;
    abs_time.tm_year = year-1900;
    
    registerTimer( mktime( &abs_time ), e );
  }

  void TimerProducer::registerTimer( time_t t, Event* e ) {
    if ( queue.empty() ) {
      // if queue is empty, set new timer and enqueue event
      te.set( t, e );
      queue.push( te );
      setTimer( t );
    } else {
      // if queue is not empty...
      te = queue.top();
      
      //...check if new timer must be set, if event should be scheduled before the earliest one
      if ( t < te.getTime() ) setTimer( t );
      
      // enqueue event
      te.set( t, e );
      queue.push( te );
    }
  }
  
  void TimerProducer::setTimer( time_t time ) {
    // new_time is used in timerThread to wait
    new_time = time;

    // awake waiting timerThread
    cond->notify();
  }
  
  void TimerProducer::run(){
    while( isRunning() ) {
      MutexGuard guard(mutex);
      
      // if time to wait is set to zero, wait infinitely
      if ( new_time == 0 ) cond->wait();
      else {
        // wait until timer expired or new timer is set
        if ( cond->timedWait( new_time, 0 ) != SUCCESS ) {
          if ( !queue.empty() ) {
	    // process event in queue
            te = queue.top();
	    queue.pop();
            
            sendEvent( *te.getEvent() );
            
	    // set new timer
            if ( !queue.empty() ) {
              te = queue.top();
	      setTimer( te.getTime() );
	    } else setTimer( 0 );
          }
        }
      }
    } // while running
  }

} // NMM

