/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVGXercescDatabase.hpp"

#include <fstream>
#include <unistd.h>
#include <time.h>
#include <regex.h>
#include <stdio.h>
#include <sys/stat.h>

#include "nmm/base/sync/LocalTime.hpp"
#include "nmm/base/sync/Clock.hpp"
#include "nmm/base/ProxyObject.hpp"
 
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

//Xerces-C-Header
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLChar.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/parsers/AbstractDOMParser.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>

namespace NMM{

  TVGXercescDatabase::TVGXercescDatabase(ITVGuide* itvg)
    : workDir("./"),
      directory(),
      insertdirectory(),
      getMutex(),
      searchMutex(),
      insertMutex(),
      fileAccessMutex(),
      termUpdate(false),
      doNamespaces(false),
      doSchema(false),
      schemaFullChecking(false),
      gOutputEncoding(XMLString::transcode("UTF-8")),
      gSplitCdataSections(true),
      gDiscardDefaultContent(true),
      gUseFilter(false),
      gFormatPrettyPrint(false),
      gWriteBOM(false)
  {
    
    itvguide.reset(itvg);
    // Initialize the XML4C system
    try{
      XMLPlatformUtils::Initialize();
      

      //Instanciate the DOMParser
      static const XMLCh gLS[] = { chLatin_L, chLatin_S, chNull };
      impl = DOMImplementationRegistry::getDOMImplementation(gLS);
      parser = NULL;
      //insertParser = NULL;
      /*
      parser->setFeature(XMLUni::fgDOMNamespaces, doNamespaces);
      parser->setFeature(XMLUni::fgXercesSchema, doSchema);
      parser->setFeature(XMLUni::fgXercesSchemaFullChecking, schemaFullChecking);
      parser->setFeature(XMLUni::fgDOMValidateIfSchema, true);
      // enable datatype normalization - default is off
      parser->setFeature(XMLUni::fgDOMDatatypeNormalization, true);
      */
      //Instanciate DOMWriter
      writer = ((DOMImplementationLS*)impl)->createDOMWriter();
      writer->setEncoding(gOutputEncoding);
    
      if (writer->canSetFeature(XMLUni::fgDOMWRTSplitCdataSections, gSplitCdataSections))
        writer->setFeature(XMLUni::fgDOMWRTSplitCdataSections, gSplitCdataSections);
      if (writer->canSetFeature(XMLUni::fgDOMWRTDiscardDefaultContent, gDiscardDefaultContent))
        writer->setFeature(XMLUni::fgDOMWRTDiscardDefaultContent, gDiscardDefaultContent);
      if (writer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, gFormatPrettyPrint))
        writer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, gFormatPrettyPrint);
      if (writer->canSetFeature(XMLUni::fgDOMWRTBOM, gWriteBOM))
        writer->setFeature(XMLUni::fgDOMWRTBOM, gWriteBOM);

      //initialize the Transcoder to transcode to ISO-8859-1
      transcoder = new XercesTranscoder();

    }catch (const XMLException& toCatch){
      GLOBAL_ERROR_STREAM("Error during initialization! :\n"
           << toCatch.getMessage() << endl);
    }catch(...){
      GLOBAL_ERROR_STREAM("strange things are going on" << endl);

    }


    priorityMap[ProgramDescription::DVB] = 10;
    priorityMap[ProgramDescription::NXTVEPG] = 9;
    priorityMap[ProgramDescription::XMLTV] = 5;
    priorityMap[ProgramDescription::UNKNOWN] = 0;
    epsilon = 0;
    directory.setRootDir(workDir);
    insertdirectory.setRootDir(workDir);
    readChannels();   
    readCategories();

  }//end Constructor


  TVGXercescDatabase::~TVGXercescDatabase(){
    saveChannels();
    delete impl;
    delete writer;    
    delete transcoder;
    XMLPlatformUtils::Terminate();
  }

  Result TVGXercescDatabase::insertChannel(const string &id, const string &name){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase : insertChannel( " << id << " , " << name << ")"<< endl);

    ChannelDescription ins;
    ins.setID(id);
    ins.addChannelName(name);

      
    //check for aliases and add them to ChannelDescription
    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      string filename = getNMMResourceDir() + "/tvguide/channelAliases.xml";
      MutexGuard famg(fileAccessMutex);
      doc = parser->parseURI(filename.c_str());

    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("insertChannel: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("insertChannel: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("insertChannel: general Exception: "  << endl);
    }
    
    if(doc){
      DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
      if(root){
        for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){            
          if(child->getNodeType() == DOMNode::ELEMENT_NODE){
            char *nodename = XMLString::transcode(child->getNodeName());
            if(XMLString::compareString(nodename, "channel") == 0){
              for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){            
                if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                  char *alias = XMLString::transcode(i->getTextContent());
                  string tmpName = name; transform(tmpName);
                  string tmpAlias = string(alias); transform(tmpAlias);
                  if(tmpAlias == tmpName){
                    for(DOMNode* j = child->getFirstChild(); j != NULL; j = j->getNextSibling()){
                      if(!XMLString::isAllWhiteSpace(j->getTextContent())){
                        char *newalias = XMLString::transcode(j->getTextContent());
                        ins.addChannelName(string(newalias));
                        XMLString::release(&newalias);
                      }
                    }
                    break;
                  }
                  XMLString::release(&alias);
                }
              }
            }
            XMLString::release(&nodename);
          }//end if("channel") 
        }//end for(child)
      }//end if(root)
    }//end if(doc)
    

    if(channels.empty()){
      channels.push_back(ins);
      insertdirectory.makeDir(insertdirectory.getCurDir() +"/" + id);
    }else{
      list<ChannelDescription>::iterator it = channels.begin();
      list<ChannelDescription>::iterator insertPos = channels.end();

      while(it != channels.end()){
        ChannelDescription cur = *it;

        //don't insert a channel twice
        if(cur.getID() == id){ //same id?
          return FAILURE;
        }else{ //other ID but same name? happens, if having >= 2 sources
          list<string> names = cur.getChannelNames();
          list<string>::iterator iter = names.begin();
          while(iter != names.end()){
            string s = *iter; transform(s);
            string tmpname = name; transform(tmpname);
            iter++;
            if(s == tmpname){
              return FAILURE;
            }
          }
        }//end if(cur.getID())
        
        //insert channel in appropriate place
        if(cur.getID() > id){
          insertPos = it;
        }
         
        it++;
        
      }//end while

      
      //if we have come so far, there is no duplicate channel, and we can insert the new one
      channels.insert(insertPos, ins);
      insertdirectory.makeDir(insertdirectory.getCurDir() +"/" + id);
      //and send an event, that we have a new channel

    }
    saveChannels();
    return SUCCESS;
  }

  Result TVGXercescDatabase::insertChannelDescription(const ChannelDescription &cd){
    if(channels.empty()){
      channels.push_back(cd);
      insertdirectory.makeDir(insertdirectory.getCurDir() +"/" + cd.getID());
    }else{

      list<ChannelDescription>::iterator insertPos = channels.begin();
      for(list<ChannelDescription>::iterator i = channels.begin(); i != channels.end(); i++){

        //don't insert a channel twice
        if(i->getID() == cd.getID()){ //same id?

          return FAILURE;
        }else{ //other ID but same name? happens, if having >= 2 sources
          list<string> names = i->getChannelNames();
          for(list<string>::iterator j = names.begin(); j != names.end(); j++){
            list<string> insNames = cd.getChannelNames();
            for(list<string>::iterator k = insNames.begin(); k != insNames.end(); k++){
              string s1 = *j; transform(s1);
              string s2 = *k; transform(s2);
              if(s1 == s2){
                return FAILURE;
              }
            }
          }
        }//end if(cur.getID())
        
        //insert channel in appropriate place
        if(i->getID() > cd.getID()){
          insertPos = i;
        }
      }//end for(i)

      //if we have come so far, there is no duplicate channel, and we can insert the new one

      //check for aliases and add them to ChannelDescription
      DOMDocument *doc = NULL;
      try{
        parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
        parser->resetDocumentPool();
        string filename = getNMMResourceDir() + "/tvguide/channelAliases.xml";

        MutexGuard famg(fileAccessMutex);
        doc = parser->parseURI(filename.c_str());

      }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("insertChannel: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
      }catch(const DOMException &e){
        GLOBAL_ERROR_STREAM("insertChannel: DOMException: " << XMLString::transcode(e.msg) << endl);
      }catch (...){
        GLOBAL_ERROR_STREAM("insertChannel: general Exception: "  << endl);
      }

      
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){            
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *nodename = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(nodename, "channel") == 0){
                for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){            
                  if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                    char *alias = XMLString::transcode(i->getTextContent());
                    string tmpName = cd.getChannelNames().front(); transform(tmpName);
                    string tmpAlias = string(alias); transform(tmpAlias);
                    if(tmpAlias == tmpName){
                      for(DOMNode* j = child->getFirstChild(); j != NULL; j = j->getNextSibling()){            
                        if(!XMLString::isAllWhiteSpace(j->getTextContent())){
                          char *newalias = XMLString::transcode(j->getTextContent());
                          const_cast<ChannelDescription&>(cd).addChannelName(string(newalias));
                          XMLString::release(&newalias);
                        }
                      }
                      break;
                      
                      XMLString::release(&alias);
                    }
                  }
                }
              }
              XMLString::release(&nodename);
            }//end if("channel") 
          }//end for(child)
        }//end if(root)
      }//end if(doc)
      
      channels.insert(insertPos, cd);
      insertdirectory.makeDir(insertdirectory.getCurDir() +"/" + cd.getID());
      //and send an event, that we have a new channel

    }
    saveChannels();
    return SUCCESS;
  }

  void TVGXercescDatabase::setDirectory(const string &path){
    workDir = path;
    directory.setRootDir(workDir);
    insertdirectory.setRootDir(workDir);
    readChannels();
    readCategories();
  }

  list<string> TVGXercescDatabase::getCategories() throw (const NMM::Exception&){
    return categories;
  }

  string TVGXercescDatabase::getDirectory()throw (const NMM::Exception&){
    return workDir;
  }

  list<ChannelDescription> TVGXercescDatabase::getChannelsList()throw (const NMM::Exception&){
    return channels;
  }  

  string TVGXercescDatabase::getChannelById(const string &id){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase: getChannelById( " << id << ")" << endl);

    list<ChannelDescription>::iterator it = channels.begin();
    while(it != channels.end()){
      ChannelDescription cur = *it;
      
      if(cur.getID() == id){
        return cur.getChannelNames().front();
      }else{
        it++;
      }
    }
    return string("");
  }  

  string TVGXercescDatabase::getChannelByName(const string &name){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase: getChannelByName( " << name << ")" << endl);

    MutexGuard mg(getMutex);
    
    //check for aliases if no match is found so far
    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      string filename = getNMMResourceDir() + "/tvguide/channelAliases.xml";
      MutexGuard famg(fileAccessMutex);
      doc = parser->parseURI(filename.c_str());
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("insertChannel: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("insertChannel: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("insertChannel: general Exception: "  << endl);
    }


    for( list<ChannelDescription>::iterator it = channels.begin(); it != channels.end(); it++){
      ChannelDescription cur = *it;  
      list<string> channelNames = cur.getChannelNames();
      for(list<string>::iterator nameIt = channelNames.begin(); nameIt != channelNames.end(); nameIt++){
        string n; n = *nameIt;
        string tmp = name;
        transform(n);
        transform(tmp);
        if(n == tmp){
          return cur.getID();
        }

      }//end while nameIt

      for(list<string>::iterator nameIt = channelNames.begin(); nameIt != channelNames.end(); nameIt++){
        string n; n = *nameIt; transform(n);
        string tmp = name; transform(tmp);

        if(doc){
          DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
          if(root){
            for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){            
              if(child->getNodeType() == DOMNode::ELEMENT_NODE){
                char *nodename = XMLString::transcode(child->getNodeName());
                if(XMLString::compareString(nodename, "channel") == 0){
                  bool aliasFound = false;
                  bool correctChannel = false;
                  for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){            
                    if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                      char *alias = XMLString::transcode(i->getTextContent());
                      string tmpAlias = string(alias); transform(tmpAlias);
                      if(tmpAlias == tmp){
                        aliasFound = true;
                      }
                      if(tmpAlias == n){
                        correctChannel = true;
                      }

                      XMLString::release(&alias);
                    }
                  }
                  if(aliasFound && correctChannel){
                    XMLString::release(&nodename);
                    return cur.getID();
                  }
                }
                XMLString::release(&nodename);
              }//end if("channel") 
            }//end for(child)
          }//end if(root)
        }//end if(doc)
      }
    }//end while it

    return string("");
  }  


  Result TVGXercescDatabase::insertEvent(const ProgramDescription &pd){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase::insertEvent(...)" << endl);

    MutexGuard mg(insertMutex);

    string pdStartTime = transformToDatabaseTime(pd.getStartTime());
    string pdEndTime   = transformToDatabaseTime(pd.getEndTime());

    if(pdStartTime.empty()) {
      return FAILURE;
    }
    if(getChannelById(pd.getChannelID()) == ""){
      GLOBAL_WARNING_STREAM("Event not inserted. No channel-ID given" << endl);
      return FAILURE;
    }

    string start = pdStartTime.substr(0,8);
    string chan =  pd.getChannelID();
    string filename = workDir + "/" + chan + "/" + start + ".xml" ;


    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
  
    //Check, if directory exists. if not, create it
    if(!getChannelById(chan).empty()){
      if(!insertdirectory.changeDir(chan)){
        insertdirectory.makeDir(workDir +"/" + chan);
        insertdirectory.setRootDir(workDir);
      }else{
        insertdirectory.setRootDir(workDir);
      }
    }

    try{
      XMLPlatformUtils::Initialize();
    }catch (const XMLException& toCatch){
      GLOBAL_ERROR_STREAM("Error during initialization! :\n"
           << toCatch.getMessage() << endl);
      return FAILURE;
    }catch(...){
      GLOBAL_ERROR_STREAM("strange things are going on" << endl);
      return FAILURE;
    }

    DOMBuilder *insertParser = NULL;

    DOMDocument *doc = NULL;
    insertParser = NULL;
    try{
      insertParser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      insertParser->resetDocumentPool();
      insertParser->setProperty(XMLUni::fgXercesScannerName, (void*)XMLUni::fgWFXMLScanner);
      {
        MutexGuard famg(fileAccessMutex);
        doc = insertParser->parseURI(filename.c_str());
      }

      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          DOMNamedNodeMap *attributes = NULL;
          string curProgramStart;
          string curProgramStop;
          unsigned int curSourceType = 0;
          
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){              
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "programme") == 0){
                XMLString::release(&name);
                attributes = child->getAttributes();
                
                if(attributes->getLength() > 0){
                  DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStart = string(value);
                    XMLString::release(&value);
                  }
                  attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("stop"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStop = string(value);
                    XMLString::release(&value);
                  }
                  attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("sourcetype"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curSourceType = atoi(value);
                    XMLString::release(&value);
                  }
                }
                
                //get the time 
                string nextProgramStart;
                //get start-time of next program
                DOMNode *nextChild = child->getNextSibling();
                if(nextChild){
                  if(nextChild->getNodeType() == DOMNode::ELEMENT_NODE){
                    char *nextName = XMLString::transcode(child->getNodeName());
                    if(XMLString::compareString(nextName, "programme") == 0){
                      DOMNamedNodeMap *attributes = child->getAttributes();
                      if(attributes->getLength() > 0){
                        DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                        if(attributeNode){
                          char *value = XMLString::transcode(attributeNode->getValue());
                          string s(value);
                          nextProgramStart = stringToTime(s);
                          XMLString::release(&value);
                        }
                        
                      }
                    }
                    XMLString::release(&nextName);
                  }
                }
                
                if(compareTimes( pdStartTime, curProgramStart) == 0){
                    //test, if the start-time is the same (+- epsilon)
                    //if yes, test the title-string of aproximate equality
                    //if equal, and extended Description is lager, replace or add the description and additional values not already available.
                    list<string> lt = pd.getTitles();
                    string s1("");
                    for(list<string>::iterator i = lt.begin(); i != lt.end(); i++){
                      s1 += *i;
                    }
                    
                    string s2("");
                    for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){
                      if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                        char *nodename = XMLString::transcode(i->getNodeName());
                        if(XMLString::compareString(nodename, "title") == 0){
                          s2 += transcoder->transcodeTo(i->getTextContent());
                        }
                        XMLString::release(&nodename);
                      }
                    }                   


                    if(compareStrings(s1, s2)){
                      
                      mergeInformation(child, pd, doc);
                    }  
                    //cerr << "equal ST " << pdStartTime << " " <<  curProgramStart << " " << s1 << " | " << s2 << endl;

                    //update the DB according to priority
                    unsigned int sourcetmp = pd.getSourceType();
                    if(priorityMap[(ProgramDescription::sourceType_t)sourcetmp] 
                       >= priorityMap[(ProgramDescription::sourceType_t)curSourceType]){
                      //cerr << "replace " << pd.getTitles().front() << endl;
                      DOMNode *replaceNode = toDOMNode(pd, doc);
                      doc->getDocumentElement()->replaceChild(replaceNode, child);
                      //Now, we have to update the database. 
                      //adjust end-time of previous program and following program. But only, if they overlap!
                      
                      //update the endtime of previous program, if necessary
                      updatePreviousEndTime(replaceNode, pdStartTime, pd);
                      
                      //update the starttime of next program, if necessary
                      updateNextStartTime(doc, replaceNode, pdEndTime, pd);
                      
                      {
                        MutexGuard famg(fileAccessMutex);
                        XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
                        writer->writeNode(ft, *doc);
                        delete ft; ft = NULL;
                      }

                      insertParser->release();
                      XMLPlatformUtils::Terminate();

                      return SUCCESS;
                         
                    }
                    insertParser->release();
                    XMLPlatformUtils::Terminate();

                    return FAILURE;
                                        
                  }
                if(
                   (compareTimes(curProgramStart, pdStartTime) < 0 //6
                    && compareTimes(pdEndTime, curProgramStop) < 0)
                   ||
                   (compareTimes(curProgramStart, pdStartTime) > 0  //3 4 5 
                    && compareTimes(pdStartTime, curProgramStop) > 0)
                   || 
                   (compareTimes(curProgramStart, pdEndTime) > 0  //1 2 3
                    && compareTimes(pdEndTime, curProgramStop) >= 0)
                   ){
                  //Here, we have to do a little more. We have to update the database and remove overlapping, old entries 
                  //depending on the source 
                  list<string> lt = pd.getTitles();
                  string s1("");
                  for(list<string>::iterator i = lt.begin(); i != lt.end(); i++){
                    s1 += *i;
                    }
                  
                  string s2("");
                  for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){
                    if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                      char *nodename = XMLString::transcode(i->getNodeName());
                      if(XMLString::compareString(nodename, "title") == 0){
                        s2 += transcoder->transcodeTo(i->getTextContent());
                      }
                      XMLString::release(&nodename);
                    }
                  }
                  
                  
                  //compute, how much they overlap
                  double overlap = 0.0;
                  
                  if(compareTimes(curProgramStart, pdStartTime) > 0
                     && compareTimes(pdEndTime, curProgramStop) >= 0){
                    overlap = 100.0;
                  }else if(compareTimes(curProgramStart, pdStartTime) < 0){
                    overlap = (double)(stringToTime(pdEndTime) -stringToTime(curProgramStart))
                      / (double)(stringToTime(curProgramStop) - stringToTime(curProgramStart)) * 100.0;
                  }else{
                    overlap = (double)(stringToTime(curProgramStop) - stringToTime(pdStartTime))
                      / (double)(stringToTime(curProgramStop) - stringToTime(curProgramStart)) * 100.0;
                  }
                  //if overlap is not at least 80%, try following program
                  if(overlap >=  80) {
                    //cerr << "overlap > 80&" << endl;
                  
                    if(compareStrings(s1, s2)){ 
                      //cerr << "duplicate" << endl;
                      //?? update the information acording to priorityMap. 
                      
                      mergeInformation(child, pd, doc);
                      
                      unsigned int sourcetmp = pd.getSourceType();
                      if(priorityMap[(ProgramDescription::sourceType_t)sourcetmp] 
                         >= priorityMap[(ProgramDescription::sourceType_t)curSourceType]){
                        
                        //cerr << "higher prio, so update" << endl;
                        
                        DOMNamedNodeMap *attr = child->getAttributes();
                        if(attr){
                          DOMAttr *attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("start"));
                          if(attrNode){
                            ((DOMElement*)child)->setAttribute(XMLString::transcode("start"),  XMLString::transcode(pdStartTime.c_str()));
                            
                          }
                          attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("stop"));
                          if(attrNode){
                            ((DOMElement*)child)->setAttribute(XMLString::transcode("stop"),  XMLString::transcode(pdEndTime.c_str()));
                            
                          }
                          
                          DOMNode *previousChild = child->getPreviousSibling();
                          if(previousChild){
                            attr = previousChild->getAttributes();
                            attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("stop"));
                            if(attrNode){
                              string et(XMLString::transcode(attrNode->getValue()));
                              if(compareTimes(et, pdStartTime) < 0){
                                ((DOMElement*)previousChild)->setAttribute(XMLString::transcode("stop"),  XMLString::transcode(pdStartTime.c_str()));
                              }                          
                            }
                          }
                          
                          //update the starttime of next program, if necessary
                          updateNextStartTime(doc, child, pdEndTime, pd);
                          
                        }
                        
                        {
                          MutexGuard famg(fileAccessMutex);
                          XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
                          writer->writeNode(ft, *doc);
                          delete ft; ft = NULL;
                        }

                        insertParser->release();

                        XMLPlatformUtils::Terminate();
                        //cerr << "duplicate" << endl;
                        return FAILURE;
                      } 
                    }else{
                      //cerr << "overlapping, but not the same" << endl;
                      //cerr << "so check the source" << endl;
                      
                      unsigned int sourcetmp = pd.getSourceType();
                      if(priorityMap[(ProgramDescription::sourceType_t)sourcetmp] 
                         >= priorityMap[(ProgramDescription::sourceType_t)curSourceType]){
                        
                        //cerr << "pd has higher or equal priority, so replace" << endl;
                        DOMNode *replaceNode = toDOMNode(pd, doc);
                        doc->getDocumentElement()->replaceChild(replaceNode, child);
                        //Now, we have to update the database. 
                        //adjust end-time of previous program and following program. But only, if they overlap!
                        
                        //update the endtime of previous program, if necessary
                        updatePreviousEndTime(replaceNode, pdStartTime, pd);
                        
                        //update the starttime of next program, if necessary
                        updateNextStartTime(doc, replaceNode, pdEndTime, pd);
                        //cerr << "REPLACE"<<endl;
                        
                        {
                          MutexGuard famg(fileAccessMutex);
                          XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
                          writer->writeNode(ft, *doc);
                          delete ft; ft = NULL;
                        }

                        insertParser->release();
                        XMLPlatformUtils::Terminate();

                        return SUCCESS;
                      }else{
                        //cerr << "pd has lower priority, so discard" << endl;
                        insertParser->release();
                        XMLPlatformUtils::Terminate();

                        return FAILURE;
                      }
                      
                    }
                    //cerr << "unkown1" << endl;
                    insertParser->release();
                    XMLPlatformUtils::Terminate();
                    return FAILURE;
                  }//end if overlap
                }

                if( compareTimes(curProgramStart, pdStartTime ) < 0 ){
                  //here, we have to update the times too
                  DOMNode *toAdd = toDOMNode(pd, doc);
                  //cerr << "INSERT BEFORE" << endl;
                  doc->getDocumentElement()->insertBefore(toAdd, child);
                  
                  //update the endtime of previous program, if necessary
                  updatePreviousEndTime(toAdd, pdStartTime, pd);

                  //update the starttime of next program, if necessary
                  updateNextStartTime(doc, toAdd, pdEndTime, pd);

                  //add the category to the interal list
                  if(!pd.getCategory().empty()){
                    list<string> cl = pd.getCategory();
                    for(list<string>::iterator i = cl.begin(); i != cl.end(); i++){
                      categories.push_back(*i);
                    }
                    categories.sort();
                    categories.unique();
                    saveCategories();
                    
                  }
                  //write the new xml-tree to file
                  
                  {
                    MutexGuard famg(fileAccessMutex);
                    XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
                    writer->writeNode(ft, *doc);
                    delete ft; ft = NULL;
                  }

                  insertParser->release();
                  XMLPlatformUtils::Terminate();
                  return SUCCESS;       
                  }
                }
            }
          }
          
          
          //This node can be appended to the XMLTree
            
          DOMNode *previousChild = doc->getDocumentElement()->getLastChild();              
          if(previousChild){
            DOMNamedNodeMap *attr = previousChild->getAttributes();
            DOMAttr *attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("sourcetype"));
            if(attrNode){
              curSourceType = atoi(XMLString::transcode(attrNode->getValue()));
            }
          }
          DOMNode *toAdd = toDOMNode(pd, doc);
          doc->getDocumentElement()->appendChild(toAdd);
          
          unsigned int sourcetmp = pd.getSourceType();
          if(priorityMap[(ProgramDescription::sourceType_t)sourcetmp] 
             >= priorityMap[(ProgramDescription::sourceType_t)curSourceType]){
            
            //update the endtime of previous program, if necessary
            updatePreviousEndTime(toAdd, pdStartTime, pd);
          }
          
          //add the category to the interal list
          if(!pd.getCategory().empty()){
            list<string> cl = pd.getCategory();
              for(list<string>::iterator i = cl.begin(); i != cl.end(); i++){
                categories.push_back(*i);
              }
              categories.sort();
              categories.unique();
              saveCategories();
          }
          //write the new xml-tree to file
          {
            MutexGuard famg(fileAccessMutex);
            XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
            writer->writeNode(ft, *doc);
            delete ft; ft = NULL;
          }
          insertParser->release();
          XMLPlatformUtils::Terminate();

          return SUCCESS;    
        }
      }else{
      }
      insertParser->release();
      XMLPlatformUtils::Terminate();

    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("insertEvent: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("insertEvent: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("insertEvent: Exception: " << e << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("insertEvent: general Exception: " << endl);
    }
    

    //if opening the file failed or the file is damaged, write a new aone
    try{
      //GLOBAL_DEBUG_STREAM("Creating File: " << filename << endl);
      DOMDocument *doc = impl->createDocument(0,
                                              XMLString::transcode("tv"),
                                              0);
      DOMNode *toAdd = toDOMNode(pd, doc);
      doc->getDocumentElement()->appendChild(toAdd);
      //add the category to the interal list
      if(!pd.getCategory().empty()){
        list<string> cl = pd.getCategory();
        for(list<string>::iterator i = cl.begin(); i != cl.end(); i++){
          categories.push_back(*i);
        }
        categories.sort();
        categories.unique();
        saveCategories();
      }
      
      MutexGuard famg(fileAccessMutex);
      XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
      writer->writeNode(ft, *doc);
      delete ft; ft = NULL;
      delete doc; doc = NULL;
      
      return SUCCESS;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("insertEvent: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("insertEvent: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("insertEvent: Exception: " << e << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("insertEvent: general Exception: " << endl);
    }     

    return FAILURE;
   
  }

  ProgramDescription TVGXercescDatabase::getEvent(const string &start, const string &end, const string &id){
    MutexGuard mg(getMutex);

    string realStart = transformToDatabaseTime(start);

    string filename = workDir + "/" + id + "/" + realStart.substr(0,8) + ".xml" ;
    ProgramDescription p;
    
    if(!fileExists(filename)) return p;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return p;
    }


    time_t startTime = 0;
    time_t stopTime  = 0;

    try{
      startTime = stringToTime(transformToDatabaseTime(start));
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("start-time is in wrong format " << endl);
      return p;
    }

    try{
      stopTime  = stringToTime(transformToDatabaseTime(end));
    }catch(const Exception &e){
      GLOBAL_WARNING_STREAM("stop-time is in wrong format " << endl);
    }


    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      {
        MutexGuard famg(fileAccessMutex);
        doc = parser->parseURI(filename.c_str());
      }
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            time_t curProgramStart = 0;
            time_t curProgramStop  = 0;
            
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "programme") == 0){
                
                DOMNamedNodeMap *attributes = child->getAttributes();
                
                if(attributes->getLength() > 0){
                  DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStart = stringToTime(string(value));
                  }
                  attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("stop"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStop = stringToTime(string(value));
                  }
                  
                }

                if((curProgramStart <= startTime  && curProgramStop >= stopTime && stopTime != 0) 
                   || (curProgramStart >= startTime ) && (curProgramStop <= stopTime )){
                  p = toProgramDescription(child);
                  return p;
                }
              }else{
                continue;
              }   
            }
          } 
        }
      }
      parser->release();;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("getEvent: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("getEvent: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("getEvent: Unknown Exception" << endl);
    }
      
    return p;
  }

  list<ProgramDescription> TVGXercescDatabase::getEventsOnDayOnChannel(const string &day, const string &chan){

    list<ProgramDescription> l;
    string filename = workDir + "/" + chan + "/" + day.substr(0,8) + ".xml" ;

    if(!fileExists(filename)) return l;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return l;
    }

    MutexGuard mg(getMutex);
    
    string tz("+0000");
    try{
      tz = day.substr(10, 5);
    }catch(...){}
    
    string mytime = transformToDatabaseTime(day.substr(0,8) + "0000 " + tz);
    ProgramFilter pf;
    pf.setStartTime(mytime); 
    mytime = transformToDatabaseTime(day.substr(0,8) + "2359 " + tz);
    pf.setEndTime(mytime); 
    pf.addChannel(chan);

    return search(pf);

  }

  list<ProgramDescription> TVGXercescDatabase::search(const ProgramFilter &pf){
    GLOBAL_DEBUG_STREAM("TVGXercescDatabase: search(...)" << endl);

    MutexGuard mg(searchMutex);

    Clock clock;

    //read all specified subdirs (= channels)
    list<string> cl;
    if(pf.getChannels().empty()){
      for(list<ChannelDescription>::iterator it = channels.begin(); it != channels.end(); it++){
        cl.push_back(it->getID());
      }
    }else{
      cl = pf.getChannels();
    }
    
    list<ProgramDescription> l; //returnlist

    if(pf.getStartTime().find("Now") != string::npos || pf.getEndTime().find("Now") != string::npos){
      for(list<string>::iterator subdirIt = cl.begin(); subdirIt != cl.end(); subdirIt++){
        usleep(500);
        ProgramDescription p = getCurrentEventOnChannel(*subdirIt);
        if(!p.empty()) {
          regex_t exp;
          bool noreturn = false;
          //Check Searchphrase
          if(!(regcomp(&exp, pf.getSearchPhrase().c_str(), REG_EXTENDED | REG_ICASE))){
            if(pf.getSearchPhrase() == ""){
            }else{
              if(regexec(&exp, p.getDescription().c_str() ,0,NULL,0) == REG_NOMATCH){
                list<string> titles = p.getTitles();
                for(list<string>::iterator i = titles.begin(); i != titles.end(); i++){
                  string s = *i;
                  if(regexec(&exp, s.c_str() ,0,NULL,0) == REG_NOMATCH){
                    noreturn = true;
                  }
                }
              }
            }
          }
          //Filter by categories if necessary
          if(!pf.getCategories().empty()){
            list<string> catl = pf.getCategories();
            list<string> pcatl = p.getCategory();
            noreturn = true;
            for(list<string>::iterator i = catl.begin(); i != catl.end(); i++){
              for(list<string>::iterator j = pcatl.begin(); j != pcatl.end(); j++){
                string s1 = *i;
                string s2 = *j;
                if(s1 == s2){
                  noreturn = false;
                }
              }
            }
          }
          if(!noreturn){
            l.push_back(p);
          }
        }
      }
      return l;
    }else if(pf.getStartTime().find("Next") != string::npos || pf.getEndTime().find("Next") != string::npos){
      for(list<string>::iterator subdirIt = cl.begin(); subdirIt != cl.end(); subdirIt++){
        usleep(500);
        ProgramDescription p = getFollowingEventOnChannel(*subdirIt);
        if(!p.empty()) {
          regex_t exp;
          bool noreturn = false;
          //Check Searchphrase
          if(!(regcomp(&exp, pf.getSearchPhrase().c_str(), REG_EXTENDED | REG_ICASE))){
            if(pf.getSearchPhrase() == ""){
            }else{
              if(regexec(&exp, p.getDescription().c_str() ,0,NULL,0) == REG_NOMATCH){
                list<string> titles = p.getTitles();
                for(list<string>::iterator i = titles.begin(); i != titles.end(); i++){
                  string s = *i;
                  if(regexec(&exp, s.c_str() ,0,NULL,0) == REG_NOMATCH){
                    noreturn = true;
                  }
                }
              }
            }
          }
          //Filter by categories if necessary
          if(!pf.getCategories().empty()){
            list<string> catl = pf.getCategories();
            list<string> pcatl= p.getCategory();
            noreturn = true;
            for(list<string>::iterator i = catl.begin(); i != catl.end(); i++){
              for(list<string>::iterator j = pcatl.begin(); j != pcatl.end(); j++){
                string s1 = *i;
                string s2 = *j;
                if(s1 == s2){
                  noreturn = false;
                }
              }
            }
          }
          if(!noreturn)l.push_back(p);
        }
      }
      return l;
    }


    //Time time0 = clock.getTime();
    string startTime = transformToDatabaseTime(pf.getStartTime());    
    string endTime   = transformToDatabaseTime(pf.getEndTime());

    //iteratore over all given subdirs
    for(list<string>::iterator subdirIt = cl.begin(); subdirIt != cl.end(); subdirIt++){
      //Time time1 = clock.getTime();
      if(directory.changeDir(*subdirIt)){

        //compute file to read
        vector<DirReader::directory_t> files = directory.getRegFiles();
        for(vector<DirReader::directory_t>::iterator i = files.begin(); i != files.end(); i++){
          string dirFile = i->name.substr(0,8);
          string startFile = startTime.substr(0,8);
          string endFile   = endTime.substr(0,8);

          string file;
          if(startTime == "" && endTime == ""){
            //neither start- nor end-time given
            file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else if(startTime != "" && endTime != ""){ 
            //start- and end-time given
            startFile = startTime.substr(0,8);
            endFile   = endTime.substr(0,8);
            if(atoi(startFile.c_str()) <= atoi(dirFile.c_str()) && atoi(dirFile.c_str()) <= atoi(endFile.c_str()))
              file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else if(startTime != ""){
            //only start-time given
            startFile = startTime.substr(0,8);
            if(atoi(startFile.c_str()) <= atoi(dirFile.c_str()))
              file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else if(endTime != ""){
            //only end-time given
            endFile   = endTime.substr(0,8);
            if(atoi(dirFile.c_str()) <= atoi(endFile.c_str()))
              file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else{
            continue;
          }
          if(file.empty()) continue;

          DOMDocument *doc = NULL;
          try{
            parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
            parser->resetDocumentPool();
            {
              MutexGuard famg(fileAccessMutex);
              doc = parser->parseURI(file.c_str());
            }

            if(doc){
              DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
              if(root){
                for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){

                  if(child->getNodeType() == DOMNode::ELEMENT_NODE){
                    char *name = XMLString::transcode(child->getNodeName());
                    if(XMLString::compareString(name, "programme") == 0){
                      DOMNamedNodeMap *attributes = child->getAttributes();
                      
                      if(attributes->getLength() > 0){
                        //check start-time
                        DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                        if(attributeNode){
                          char *value = XMLString::transcode(attributeNode->getValue());
                          if(stringToTime(string(value)) < stringToTime(startTime)
                             || (stringToTime(string(value)) > stringToTime(endTime) && !endTime.empty())
                             ){
                            continue;
                          }
                          XMLString::release(&value);
                        }
                        
                      }

                      //check the searchPhrase by using regular expressions
                      regex_t exp;
                      int title_res = pf.getSearchPhrase().empty() ? 0 : -1;
                      int desc_res = pf.getSearchPhrase().empty() ? 0 : -1;
                      int cat_res = pf.getCategories().empty() ? 0 : -1;
                      
                      if((!pf.getSearchPhrase().empty() || !pf.getCategories().empty())
                         && !(regcomp(&exp, pf.getSearchPhrase().c_str(), REG_EXTENDED | REG_ICASE))){
                        //in titles
                        for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){                          
                          if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                            char *name = XMLString::transcode(i->getNodeName());
                            if(XMLString::compareString(name, "title") == 0){
                              if(!pf.getSearchPhrase().empty()){
                                string value = transcoder->transcodeTo(i->getTextContent());
                                if(regexec(&exp, value.c_str() ,0,NULL,0) != REG_NOMATCH){
                                  //if we find a match, we set a flag
                                  title_res = 0;
                                } 
                              }else{
                                title_res = 0;
                              }
                            }else if(XMLString::compareString(name, "desc") == 0){
                                //in extended description
                              if(!pf.getSearchPhrase().empty()){
                                string value = transcoder->transcodeTo(i->getTextContent());
                                if(regexec(&exp, value.c_str() ,0,NULL,0) != REG_NOMATCH){
                                  //if we find a match in description, we set a flag
                                  desc_res = 0;
                                }
                              }else{
                                desc_res = 0;
                              }
                            }else if(XMLString::compareString(name, "category") == 0){
                              if(!pf.getCategories().empty()){
                                list<string> catl = pf.getCategories();
                                for(list<string>::iterator j = catl.begin(); j != catl.end(); j++){
                                  if(transcoder->transcodeTo(i->getTextContent()) == *j){
                                    cat_res = 0;
                                  }
                                }
                              }
                            }
                            XMLString::release(&name);
                          }
                        }
                      }else{
                      }

                      //if flags are -1, then no match was found at all, and we can skip the rest
                      if((!pf.getSearchPhrase().empty()) && title_res == -1 && desc_res == -1) continue;   
                      //check for Category
                      if(!pf.getCategories().empty() && cat_res == -1) continue;
                    }
                  }
                  
                  ProgramDescription p;
                  p = toProgramDescription(child);
                  
                  l.push_back(p);
                }//for(DOMNode child)
              }//if(root)
            }//if(doc)
            parser->release();;
          }catch(const XMLException &e){
            GLOBAL_ERROR_STREAM("search: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
          }catch(const DOMException &e){
            GLOBAL_ERROR_STREAM("search: DOMException: " << XMLString::transcode(e.msg) << endl);
          }catch (...){
            GLOBAL_ERROR_STREAM("search: general Exception: " << endl);
          }
        
        }//end for(files)
      }//end if(changeDir)
    //Time time3 = clock.getTime();
    //cerr << "Time for Reading 1 of  " << cl.size()  << " channels (" << *subdirIt << ") : " << time3 - time1 << endl;
    directory.setRootDir(workDir);
    }//end for(channels)
    
    
        
    //Time time4 = clock.getTime();
    //cerr << "Time for Search: " << time4 - time0 << endl; 
    return l;
  }
  
  void TVGXercescDatabase::addExtDescription(const string& start, const string& end, const string& id, const string& desc){
    GLOBAL_DEBUG_STREAM(" TVGXercescDatabase: addExtendedDescription(...)" << endl);

    string filename = workDir + "/" + id + "/" + start.substr(0,8) + ".xml" ;

    if(!fileExists(filename)) return;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    MutexGuard mg(insertMutex);
    try{
      XMLPlatformUtils::Initialize();
    }catch (const XMLException& toCatch){
      GLOBAL_ERROR_STREAM("Error during initialization! :\n"
           << toCatch.getMessage() << endl);
      return;
    }catch(...){
      GLOBAL_ERROR_STREAM("strange things are going on" << endl);
      return ;
    }

    DOMBuilder *insertParser = NULL;

    DOMDocument *doc = NULL;
    try{
      insertParser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      insertParser->resetDocumentPool();
      insertParser->setProperty(XMLUni::fgXercesScannerName, (void*)XMLUni::fgWFXMLScanner);
      {
        MutexGuard famg(fileAccessMutex);
        doc = insertParser->parseURI(filename.c_str());
      }
    
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            time_t startTime = stringToTime(transformToDatabaseTime(start));
            time_t endTime = 0;
            if(!end.empty()){
              endTime = stringToTime(transformToDatabaseTime(end));
            }
            time_t curProgramStart = 0;
            time_t curProgramEnd = 0;
            
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "programme") == 0){
                
                DOMNamedNodeMap *attributes = child->getAttributes();
                
                if(attributes->getLength() > 0){
                  DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStart = stringToTime(string(value));
                    XMLString::release(&value);
                  }
                  attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("stop"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramEnd = stringToTime(string(value));
                    XMLString::release(&value);
                  }
                }


                if((curProgramStart == startTime ) && (curProgramEnd == endTime)){
                  DOMElement *add = doc->createElement(XMLString::transcode("desc"));
                  add->setTextContent(transcoder->transcodeFrom(desc));

                  child->appendChild(add);
                  
                  {
                    MutexGuard famg(fileAccessMutex);
                    XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
                    writer->writeNode(ft, *doc);
                    delete ft; ft = NULL;
                  }
                  insertParser->release();
                  XMLPlatformUtils::Terminate();

                  return;
                }              
              }
            }
          }
        }
      }
      insertParser->release();
      XMLPlatformUtils::Terminate();

    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("addExtendedDescription: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("addExtendedDescription: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("addExtendedDescription: general Exception" << endl);    
    }

  }

  ProgramDescription TVGXercescDatabase::getCurrentEventOnChannel(const string& channel){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase: getCurrentEventOnChannel( " << channel << ")" << endl);

    ProgramDescription p;
          
      
    time_t now = time(NULL);;

    //time_t now = time(NULL);
    static char tmp[25];
    strftime(tmp, sizeof(tmp), "%Y%m%d%H%M", gmtime(&now));

    string startTime(tmp);
    string day = string(tmp).substr(0,8);
      
    string filename = workDir + "/" + channel + "/" + day + ".xml" ;
    
    if(!fileExists(filename)) return p;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return p;
    }
    
    MutexGuard mg(getMutex);
    
    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      {
        MutexGuard famg(fileAccessMutex);
        doc = parser->parseURI(filename.c_str());
      }

      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            string curProgramStart;
            string curProgramStop;
            
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "programme") == 0){
                
                DOMNamedNodeMap *attributes = child->getAttributes();
                
                if(attributes->getLength() > 0){
                  DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStart = string(value);
                    XMLString::release(&value);
                  }
                  attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("stop"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStop = string(value);
                    XMLString::release(&value);
                  }
                  
                }
                
                if(compareTimes(curProgramStart, startTime) >= 0
                   && compareTimes(curProgramStop, startTime)<=0){
                  p = toProgramDescription(child);
                  return p;
                }
              }
            } 
          } 
        }
      }
      parser->release();
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("XMLException: " << e.getMessage() << endl);
    }catch(const DOMException &e){
    }catch (...){}
    
    return p;
  }
  
  ProgramDescription TVGXercescDatabase::getFollowingEventOnChannel(const string& channel){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase: getCurrentFollowingOnChannel( " << channel << ")" << endl);

    ProgramDescription p;
          
      
    time_t now = time(NULL);
    static char tmp[25];
    strftime(tmp, sizeof(tmp), "%Y%m%d%H%M", gmtime(&now));
    string day = string(tmp).substr(0,8);
    
    string startTime = string(tmp);
    string filename = workDir + "/" + channel + "/" + day + ".xml" ;
    
    if(!fileExists(filename)) return p;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return p;
    }
    
    MutexGuard mg(getMutex);
    
    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      {
        MutexGuard famg(fileAccessMutex);
        doc = parser->parseURI(filename.c_str());
      }

      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            string curProgramStart ;
            string curProgramStop  ;
            
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "programme") == 0){
                
                DOMNamedNodeMap *attributes = child->getAttributes();
                
                if(attributes->getLength() > 0){
                  DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("start"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStart = string(value);
                    XMLString::release(&value);
                  }
                  attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("stop"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    curProgramStop = string(value);
                    XMLString::release(&value);
                  }
                  
                }
                
                if(compareTimes(curProgramStart, startTime) < 0 ){
                  //DOMNode* nextChild = child->getNextSibling();
                  //if(nextChild){
                  //p = toProgramDescription(nextChild);
                  p = toProgramDescription(child);
                  return p;
                }
              }
              XMLString::release(&name);

            }
          }
           
          //if we can't selcet the enxt child, then the current show is the last show on this day
          //so take a look at the next day
          now += (60 * 60 * 24);
          static char tmp[25];
          strftime(tmp, sizeof(tmp), "%Y%m%d%H%M", localtime(&now));
          day = string(tmp).substr(0,8);
          
          filename = workDir + "/" + channel + "/" + day + ".xml" ;
          
          parser->resetDocumentPool();
          {
            MutexGuard famg(fileAccessMutex);
            doc = parser->parseURI(filename.c_str());
          }
          if(doc){
            DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
            if(root){
              for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){                          
                if(child->getNodeType() == DOMNode::ELEMENT_NODE){
                  char *name = XMLString::transcode(child->getNodeName());
                  if(XMLString::compareString(name, "programme") == 0){
                    //we rely on the order in the file
                    p = toProgramDescription(child);
                    return p; 
                  }
                }
              }
            }
          }
        }
      }
    
   

      parser->release();;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("XMLException: " << e.getMessage() << endl);
    }catch(const DOMException &e){
    }catch (...){}
    
    return p;
    
  }
  
  
  
  /*Begin***************** Private Helpers *******************************/
  void TVGXercescDatabase::readChannels(){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase::readChannels()" << endl);

    string filename = workDir + "/channels.xml";

    if(!fileExists(filename)) return;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    //MutexGuard mg(tvmutex);

    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      {
        MutexGuard famg(fileAccessMutex);
        doc = parser->parseURI(filename.c_str());
      }
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          channels.clear();
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "channel") == 0){
                ChannelDescription c;
                DOMNamedNodeMap *attributes = child->getAttributes();
                
                if(attributes->getLength() > 0){
                  DOMAttr *attributeNode = (DOMAttr*) attributes->getNamedItem(XMLString::transcode("id"));
                  if(attributeNode){
                    char *value = XMLString::transcode(attributeNode->getValue());
                    c.setID(string(value));
                    XMLString::release(&value);
                  }
                }
                for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){
                  if(i->getNodeType() == DOMNode::ELEMENT_NODE){
                    char *displayname = XMLString::transcode(i->getNodeName());
                    if(XMLString::compareString(displayname, "display-name") == 0){
                      if(!XMLString::isAllWhiteSpace(i->getTextContent()))
                        c.addChannelName(transcoder->transcodeTo(i->getTextContent()));
                    }
                  }
                }
                channels.push_back(c);
              }
            }
          }
        }else{}
      }
      parser->release();;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("readChannels: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("readChannels: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("readChannels: gerneal Exception: " << endl);
    }

  }
   
    
  void TVGXercescDatabase::saveChannels(){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase: saveChannels()" << endl);

    string filename = workDir + "/channels.xml";

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    try{
      DOMDocument *doc = impl->createDocument(0,
                                              XMLString::transcode("tv"),
                                              0);
      for(list<ChannelDescription>::iterator i = channels.begin(); i != channels.end(); i++){
        DOMNode *toAdd = toDOMNode(*i, doc);
        doc->getDocumentElement()->appendChild(toAdd);
      }

      MutexGuard famg(fileAccessMutex);
      XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
      writer->writeNode(ft, *doc);

      delete ft; ft = NULL;
      delete doc; doc = NULL;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("saveChannels: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("saveChannels: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("saveChannels: gerneal Exception: " << endl);
    }

  }//end saveChannels()
  
  void TVGXercescDatabase::readCategories(){
    //GLOBAL_DEBUG_STREAM("TVGXercescDatabase: readCategories()" << endl);

    string filename = workDir + "/categories.xml";

    if(!fileExists(filename)) return;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }
    //MutexGuard mg(tvmutex);

    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->resetDocumentPool();
      {
        MutexGuard famg(fileAccessMutex);
        doc = parser->parseURI(filename.c_str());
      }
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          categories.clear();
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "category") == 0){
                categories.push_back(transcoder->transcodeTo(child->getTextContent()));
              }
            }
          }
        }
      }
      categories.sort();
      categories.unique();
      parser->release();;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("readCategories: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("readCategories: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("readCategories: general Exception: " << endl);
    }

  }
  void TVGXercescDatabase::saveCategories(){
    try{

    string filename = workDir + "/categories.xml";

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }
    //MutexGuard mg(tvmutex);

      DOMDocument *doc = impl->createDocument(0,
                                              XMLString::transcode("tv"),
                                              0);
      for(list<string>::iterator i = categories.begin(); i != categories.end(); i++){
        DOMElement *toAdd = doc->createElement(XMLString::transcode("category"));
        DOMText *catText = doc->createTextNode(transcoder->transcodeFrom(*i));
        toAdd->appendChild(catText);
        doc->getDocumentElement()->appendChild(toAdd);
      }
      
      MutexGuard famg(fileAccessMutex);
      XMLFormatTarget *ft = new LocalFileFormatTarget(XMLString::transcode(filename.c_str()));
      writer->writeNode(ft, *doc);

      delete ft; ft = NULL;
      delete doc; doc = NULL;
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("saveCategories: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("saveCategories: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("saveCategories: Exception: " << e << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("saveCategories: general Exception: " << endl);
    }

  }

  bool TVGXercescDatabase::compareStrings(const string s1, const string s2){
    
    string regexString("");
    string testString("");
    //use the smaller string as pattern
    if(s1.size() < s2.size()){
      regexString = s1;
      testString = s2;
    }else{
      regexString = s2;
      testString = s1;
    }

    //TODO: get rid of special signs like !;*&%

    regex_t exp;
    if(!(regcomp(&exp, regexString.c_str(), REG_EXTENDED | REG_ICASE))){
      if(regexec(&exp, testString.c_str() ,0,NULL,0) == REG_NOMATCH){
        return false;
      }else{
        return true;
      }
    }
    return false;
  }
  int TVGXercescDatabase::compareTimes(string t1, string t2){
    time_t time1 = stringToTime(t1);
    time_t time2 = stringToTime(t2);

    if(time1 == time2){
      return 0;
    }else if(time1 > time2){
      return -1;
    }else if (time1 < time2){
      return +1;
    }
    
    return 0;
  }

  /*End******************** Private Helpers ******************************/

  DOMNode* TVGXercescDatabase::toDOMNode(ProgramDescription pd, DOMDocument *doc){
    //DOMElement* root = doc->getDocumentElement();

    DOMElement* add  = doc->createElement(XMLString::transcode("programme"));
    add->setAttribute(XMLString::transcode("start"), XMLString::transcode(transformToDatabaseTime(pd.getStartTime()).c_str()));
    add->setAttribute(XMLString::transcode("stop"),  XMLString::transcode(transformToDatabaseTime(pd.getEndTime()).c_str()));
    add->setAttribute(XMLString::transcode("channel"), XMLString::transcode(pd.getChannelID().c_str()));
    add->setAttribute(XMLString::transcode("sourcetype"), XMLString::transcode(itoa(pd.getSourceType()).c_str()));
    list<string> titleList = pd.getTitles();
    for(list<string>::iterator i = titleList.begin(); i != titleList.end(); i++){
      DOMElement *addTitle = doc->createElement(XMLString::transcode("title"));
      addTitle->setTextContent(transcoder->transcodeFrom(*i));
      add->appendChild(addTitle);
    }
    
    DOMElement *addDesc = doc->createElement(XMLString::transcode("desc"));
    addDesc->setAttribute(XMLString::transcode("lang"), XMLString::transcode("de"));
    if(!pd.getDescription().empty()){
      addDesc->setTextContent(transcoder->transcodeFrom(pd.getDescription()));
      add->appendChild(addDesc);
    }
    list<string>catList = pd.getCategory();
    for(list<string>::iterator i = catList.begin(); i != catList.end(); i++){
      DOMElement *addCategory = doc->createElement(XMLString::transcode("category"));
      addCategory->setTextContent(transcoder->transcodeFrom(*i));
      add->appendChild(addCategory);
    }

    return add;    
  }
  
  
  ProgramDescription TVGXercescDatabase::toProgramDescription(DOMNode *node){
    ProgramDescription pd;

    if(node->hasAttributes()) {
      // get all the attributes of the node
      DOMNamedNodeMap *pAttributes = node->getAttributes();
      int nSize = pAttributes->getLength();
      if(nSize > 0){
        DOMAttr *pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("start"));
        char *name = NULL;

        if(pAttributeNode){
          name = XMLString::transcode(pAttributeNode->getValue());
          //XERCES_STD_QUALIFIER cout << "  setStartTime(" << name << ")" << endl;;
          pd.setStartTime(string(name));
          XMLString::release(&name);
        }
        
        pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("stop"));
        if(pAttributeNode){
          name = XMLString::transcode(pAttributeNode->getValue());
          //XERCES_STD_QUALIFIER cout << "  setEndTime  (" << name << ")" << endl;;
          pd.setEndTime(string(name));
          XMLString::release(&name);
        }
        
        pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("channel"));
        if(pAttributeNode){
          name = XMLString::transcode(pAttributeNode->getValue());
          //XERCES_STD_QUALIFIER cout << "  setChannelID(" << name << ")" << endl;;
          list<ChannelDescription>::iterator it = channels.begin();
          while(it != channels.end()){
            ChannelDescription cur = *it;
            
            if(cur.getID() == string(name)){
              pd.setChannelDescription(*it);
              break;
            }else{
              it++;
            }
          }
          XMLString::release(&name);
        }
        
        pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("sourcetype"));
        if(pAttributeNode){
          name = XMLString::transcode(pAttributeNode->getValue());
          //XERCES_STD_QUALIFIER cout << "  setEndTime  (" << name << ")" << endl;;
          pd.setSourceType(atoi(name));
          XMLString::release(&name);
        }

        pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("duration"));
        if(pAttributeNode){
          name = XMLString::transcode(pAttributeNode->getValue());
          //XERCES_STD_QUALIFIER cout << "  setDuration(" << name << ")" << endl;;
          pd.setDuration(atoi(name));
          XMLString::release(&name);
        }
        
        for (DOMNode* child = node->getFirstChild(); child != 0; child=child->getNextSibling()){
          if (child->getNodeType() == DOMNode::ELEMENT_NODE){
            char *name = XMLString::transcode(child->getNodeName());
            
            if(XMLString::compareString(name, "title") == 0){
              //evaluate title-tag
              DOMNode* text = child->getFirstChild();
              if(text){
                if(text->getNodeType() == DOMNode::TEXT_NODE){
                  //XERCES_STD_QUALIFIER cout << endl << "  addTitle(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                  pd.addTitle(transcoder->transcodeTo(text->getTextContent()));
                }
              }
            }

            if(XMLString::compareString(name, "desc") == 0){
              //evaluate description              
              DOMNode* text = child->getFirstChild();
              if(text){
                if(text->getNodeType() == DOMNode::TEXT_NODE){
                  //XERCES_STD_QUALIFIER cout << "  addDescription(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                  string tmp = pd.getDescription();
                  pd.setDescription(tmp + " " + transcoder->transcodeTo(text->getTextContent()));
                }
              }
              
            }

            if(XMLString::compareString(name, "category") == 0){              
              //evaluate description              
              DOMNode* text = child->getFirstChild();
              if(text){
                if(text->getNodeType() == DOMNode::TEXT_NODE){
                  pd.addCategory(transcoder->transcodeTo(text->getTextContent()));
                  //XERCES_STD_QUALIFIER cout << "  addCategory(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                }
              }
            }
            if(XMLString::compareString(name, "duration") == 0){
              //evaluate duration
              DOMNamedNodeMap *childAttributes = child->getAttributes();
              int childSize = childAttributes->getLength();
              if(childSize > 0){
                DOMAttr *childAttributeNode =
                  (DOMAttr*) childAttributes->getNamedItem(XMLString::transcode("unit"));
                if(childAttributeNode){
                  char *lang = xercesc::XMLString::transcode(childAttributeNode->getValue());
                  //XERCES_STD_QUALIFIER cout << "    lang=" << lang << endl;
                  XMLString::release(&lang);
                }
              }
              DOMNode* text = child->getFirstChild();
              if(text){
                if(text->getNodeType() == DOMNode::TEXT_NODE){
                  char *duration = XMLString::transcode(text->getNodeValue());
                  //XERCES_STD_QUALIFIER cout << "  addDuration(" << duration << ")" << endl;
                  pd.setDuration(atoi(duration));
                  XMLString::release(&duration);
                }
              }
            }
            XMLString::release(&name);
          }
        }
      }
    }
    return pd;
  }

  DOMNode* TVGXercescDatabase::toDOMNode(ChannelDescription cd, DOMDocument *doc){
    //DOMElement* root = doc->getDocumentElement();
    
    DOMElement* add  = doc->createElement(XMLString::transcode("channel"));
    add->setAttribute(XMLString::transcode("id"), XMLString::transcode(cd.getID().c_str()));
    
    list<string> nameList = cd.getChannelNames();
    for(list<string>::iterator i = nameList.begin(); i != nameList.end(); i++){
      DOMElement *addname = doc->createElement(XMLString::transcode("display-name"));
      DOMText *nameText = doc->createTextNode(XMLString::transcode(i->c_str()));
      addname->appendChild(nameText);
      add->appendChild(addname);
    }
    
    return add;
  }

  ChannelDescription TVGXercescDatabase::toChannelDescription(DOMNode* node){
    ChannelDescription cd;
    if(node->hasAttributes()) {
      // get all the attributes of the node
      DOMNamedNodeMap *pAttributes = node->getAttributes();
      int nSize = pAttributes->getLength();
      if(nSize > 0){
        DOMAttr *pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("id"));
        // get attribute name
        if(pAttributeNode){
          char *name = XMLString::transcode(pAttributeNode->getValue());
          cd.setID(string(name));
          XERCES_STD_QUALIFIER cout << "  setChannelID(" << name << ")" << endl;;
          XMLString::release(&name);
          
          for (DOMNode* child = node->getFirstChild(); child != 0; child=child->getNextSibling()){
            if (child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *name = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(name, "display-name") == 0){
                DOMNode* text = child->getFirstChild();
                if(text){
                  if(text->getNodeType() == DOMNode::TEXT_NODE){
                    char *channel = XMLString::transcode(text->getNodeValue());
                    cd.addChannelName(string(channel));
                    XERCES_STD_QUALIFIER cout << "  addChannelName(" << channel << ")" << endl;
                    XMLString::release(&channel);
                  }
                }
              }
              XMLString::release(&name);
            }
          }
        }
      }
    }
    return cd;
  }


  string TVGXercescDatabase::transformToDatabaseTime(string s){
    try{
      if(s.empty()) return s;

      time_t localTime = time(NULL);
      localtime(&localTime); //Just to be sure, that external time-variables of time.h are initialized
      
      time_t xmlTime = stringToTime(s);
      unsigned int pos = s.find_first_of(' ');
      int xmlDifference = 0;
      if(pos != string::npos){
        //if no timezone-info is available, assume UTC
        xmlDifference = atoi(s.substr(pos+1,5).c_str())/100;
      }
      char timeString[25];
      //Compute the difference between the xml-Timezone and our timezone and update the outputtime accordingly
      xmlTime +=   (-(xmlDifference*3600));// + (-1*timezone));
      double timeDifference = 0;//-1*timezone/60/60 * 100; //in hours
      struct tm *xmltime_tm = localtime(&xmlTime);
      if(xmltime_tm->tm_isdst >= 0){
        //xmlTime += ((xmltime_tm->tm_isdst + xmlDifference) * 3600);
        //timeDifference += (xmltime_tm->tm_isdst * 100);
      }
      strftime(timeString, sizeof(timeString), "%Y%m%d%H%M", localtime(&xmlTime));
      sprintf(timeString, "%s %0+5.0f", timeString, timeDifference);
      return string(timeString);

    }catch (...){
      return s;
      //throw WrongTimeFormatException();
    }    
  }

  
  void TVGXercescDatabase::mergeInformation(DOMNode* child, ProgramDescription pd, DOMDocument *doc){

    string description("");
    for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){
      if(i->getNodeType() == DOMNode::ELEMENT_NODE){
        char *nodename = XMLString::transcode(i->getNodeName());
        if(XMLString::compareString(nodename, "desc") == 0){
          description += transcoder->transcodeTo(i->getTextContent());
        }
        XMLString::release(&nodename);
      }
    }
    
    if(description.size() < pd.getDescription().size()){
      //update Information
      //remove the old Description
      for(DOMNode* i = child->getFirstChild(); i != NULL; i = i->getNextSibling()){
        if(i->getNodeType() == DOMNode::ELEMENT_NODE){
          char *nodename = XMLString::transcode(i->getNodeName());
          if(XMLString::compareString(nodename, "desc") == 0){
            child->removeChild(*(&i));
          }
          XMLString::release(&nodename);
        }
      }
      //and add the new one
      DOMElement *newDesc = doc->createElement(XMLString::transcode("desc"));
      DOMText *desctext = doc->createTextNode(transcoder->transcodeFrom(pd.getDescription()));
      newDesc->appendChild(desctext);
      child->appendChild(newDesc);
    }
    //if the new information has an end-time, add it, if not already available
    if(!pd.getEndTime().empty()){
      DOMNamedNodeMap *attr = child->getAttributes();
      DOMAttr *attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("stop"));
      if(attrNode){
        //Check for empty string
        if(XMLString::isAllWhiteSpace(attrNode->getValue())){
          ((DOMElement*)child)->setAttribute(XMLString::transcode("stop"),  XMLString::transcode(transformToDatabaseTime(pd.getEndTime()).c_str()));
        }
      }else{
        ((DOMElement*)child)->setAttribute(XMLString::transcode("stop"),  XMLString::transcode(transformToDatabaseTime(pd.getEndTime()).c_str()));
      }
    }
    
    //XMLNodeList curCategory = curProgram->children("category");
    list<string>catList = pd.getCategory();
    for(list<string>::iterator i = catList.begin(); i != catList.end(); i++){
      bool addCategory = true;
      for (DOMNode* j = child->getFirstChild(); j != 0; j=j->getNextSibling()){
        char *nodename = XMLString::transcode(j->getNodeName());
        if(XMLString::compareString(nodename, "category") == 0){
          if(transcoder->transcodeTo(j->getTextContent()) == *i){
            addCategory=false;
            continue;
          }
        }
        XMLString::release(&nodename);
        
      }
      if(addCategory){
        DOMElement *addCategory = doc->createElement(XMLString::transcode("category"));
        addCategory->setTextContent(transcoder->transcodeFrom(*i));
        child->appendChild(addCategory);
        
        categories.push_back(*i);
        categories.sort();
        categories.unique();
        saveCategories();
      }
    }
  }
  
  void TVGXercescDatabase::updateNextStartTime(DOMDocument * /*doc*/, DOMNode* child, string pdEndTime, const ProgramDescription &pd){
    DOMNode *nextChild = child->getNextSibling();
    if(nextChild){
      //get the starttime of the next node
      DOMNamedNodeMap *attr = nextChild->getAttributes();
      DOMAttr *attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("start"));
      if(attrNode){
        string st(XMLString::transcode(attrNode->getValue()));
        if(compareTimes(st, pdEndTime) > 0){
          unsigned int curSourceType = 0;
          //determine the sourcetype
          attr = child->getAttributes();
          if(attr){
            attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("sourcetype"));
            if(attrNode){
              char *value = XMLString::transcode(attrNode->getValue());
              curSourceType = atoi(value);
              XMLString::release(&value);
            }
          }

          unsigned int sourcetmp = pd.getSourceType();
          if(priorityMap[(ProgramDescription::sourceType_t)sourcetmp] 
             >= priorityMap[(ProgramDescription::sourceType_t)curSourceType]){      
            attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("stop"));
            if(attrNode){
              string et(XMLString::transcode(attrNode->getValue()));
              /*
              bool leave = false;
              while(compareTimes(pdEndTime, et) <= 0 && !leave){
                cerr << "remove child updatenextTime "<< st << " " << et << endl;
                doc->getDocumentElement()->removeChild(nextChild);
                nextChild = child->getNextSibling();
                if(nextChild == NULL) {
                  leave=true;
                }else{
                  //update the time-string to compare
                  attr = nextChild->getAttributes();
                  attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("start"));
                  st = string(XMLString::transcode(attrNode->getValue()));
                  attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("stop"));
                  et = string(XMLString::transcode(attrNode->getValue()));
                }
              }
              */
            }
            if(nextChild && (!pdEndTime.empty())  && compareTimes(pdEndTime, st) > 0){
              //cerr << "1" << endl;
              ((DOMElement*)nextChild)->setAttribute(XMLString::transcode("start"),  XMLString::transcode(pdEndTime.c_str()));
            }
          }else{
            if(compareTimes(pdEndTime, st) > 0){
              //cerr << "2" << endl;
              ((DOMElement*)child)->setAttribute(XMLString::transcode("stop"),  XMLString::transcode(st.c_str()));
            }
          }                    
        }
      }
    }
  }

  void TVGXercescDatabase::updatePreviousEndTime(DOMNode* child, string pdStartTime, const ProgramDescription &pd){
    DOMNode *previousChild = child->getPreviousSibling();
    if(previousChild){
      DOMNamedNodeMap *attr = previousChild->getAttributes();
      DOMAttr *attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("stop"));
      if(attrNode){
        string et(XMLString::transcode(attrNode->getValue()));
        if(compareTimes(et, pdStartTime) < 0){
          unsigned int sourcetmp = pd.getSourceType();
          unsigned int curSourceType = 0;
          
          attr = child->getAttributes();
          if(attr){
            attrNode = (DOMAttr*) attr->getNamedItem(XMLString::transcode("sourcetype"));
            if(attrNode){
              char *value = XMLString::transcode(attrNode->getValue());
              curSourceType = atoi(value);
              XMLString::release(&value);
            }
          }

          if(priorityMap[(ProgramDescription::sourceType_t)sourcetmp] 
             >= priorityMap[(ProgramDescription::sourceType_t)curSourceType]){
            if(compareTimes(pdStartTime, et) > 0){
              //cerr << "3" << endl;
              ((DOMElement*)previousChild)->setAttribute(XMLString::transcode("stop"),  
                                                         XMLString::transcode(pdStartTime.c_str()));
            }
          }else{
            //            if(!et.empty())
            //cerr << "4" << endl;
            ((DOMElement*)child)->setAttribute(XMLString::transcode("start"),  XMLString::transcode(et.c_str()));
          }                          
        }
      }
    }
  }

  bool TVGXercescDatabase::fileExists(string file){
    struct stat info;
    if (stat(file.c_str(),&info) == 0) {
      return true;
    }
    return false;

  }

  
  /*Start******************** XercesTranscoder ******************************/
  XercesTranscoder::XercesTranscoder(){
    fgUnicodeForm = NULL;
    transcoder = XMLPlatformUtils::fgTransService->
      makeNewTranscoderFor("ISO-8859-1", res , 1024 );//, XMLPlatformUtils::fgMemoryManager);
  }
  XercesTranscoder::XercesTranscoder(string encoding){
    fgUnicodeForm = NULL;
    transcoder = XMLPlatformUtils::fgTransService->
      makeNewTranscoderFor(encoding.c_str(), res , 1024 );//, XMLPlatformUtils::fgMemoryManager);
  }
  XercesTranscoder::~XercesTranscoder(){}
  
  string XercesTranscoder::transcodeTo(const XMLCh* s){
    try{
      if(transcoder){
        unsigned int charsEaten = 0;
        unsigned int len = XMLString::stringLen(s);
        unsigned char myByte[8*len];
        unsigned int ui = transcoder->
          transcodeTo((const XMLCh*) s,
                      (unsigned int) len,
                      (XMLByte*) myByte,
                      (unsigned int) 8*len,
                      charsEaten,
                      XMLTranscoder::UnRep_Throw);
        
        myByte[ui]='\0';
        string retVal = string((char*)myByte);
        //XERCES_STD_QUALIFIER cout << "  (" << retVal << ")" << endl;
        return retVal;
      }
    }catch(...){
      GLOBAL_ERROR_STREAM("an error occured in transcodeTo" << endl);
    }
    
    GLOBAL_ERROR_STREAM("transcoding to failed" << endl);
    return string(XMLString::transcode(s));
  }

  XMLCh* XercesTranscoder::transcodeFrom(string  s){
    if(transcoder){
      unsigned int charsEaten = 0;
      unsigned int len = s.length();

      if(fgUnicodeForm) delete [] fgUnicodeForm;

      fgUnicodeForm = new XMLCh[8*len];
      const char *desc = s.c_str();
      unsigned char* c = new unsigned char[1024];
      memset(c, 0, 1024);

      unsigned int ui = transcoder->
        transcodeFrom((const XMLByte*) desc,
                      len,
                      fgUnicodeForm,
                      (unsigned int) 1024,
                      charsEaten,
                      c);
      
      fgUnicodeForm[ui]='\0';
      delete [] c;
      return fgUnicodeForm;
    }
    
    GLOBAL_ERROR_STREAM("trancodeFrom failed" << endl);
    return XMLString::transcode(s.c_str());
  }
  /*End******************** XercesTranscoder ******************************/


}//end namespace NMM

