/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "TVGuidePeerConnector.hpp"

#include "nmm/utils/utils.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

//for Benchmarks
#include "nmm/base/sync/LocalTime.hpp"
#include "nmm/base/sync/Clock.hpp"

#include <unistd.h>
#include <stdio.h>

namespace NMM{

  TVGuidePeerConnector::TVGuidePeerConnector(ITVGuide* itvg)
    : running(true),
      insertHandlerMutex(),
      insertHandlerCondition(insertHandlerMutex),
      requestHandlerMutex(),
      requestHandlerCondition(requestHandlerMutex),
      insertHandlerIdCounter(0),
      requestHandlerIdCounter(0)
  {
    localTVGuide = itvg;
    
    Thread t = getThread();
    t.setMinPriority();
  }

  TVGuidePeerConnector::~TVGuidePeerConnector(){
    running = false;
    deactivateThread();

    //notify waiting insert-handlers
    insertHandlerMutex.lock();
    insertHandlerCondition.notify();
    insertHandlerMutex.unlock();
    
    //notify waitung request-handlers
    requestHandlerMutex.lock();
    requestHandlerCondition.notify();
    requestHandlerMutex.unlock();
  }

  bool TVGuidePeerConnector::getChannels(){
    list<Interface*> peers;
    try{
      //peers.push_back(this->getCheckedInterface<ITVGuidePeerObject>());
      return getChannels(peers);
    }catch(const Exception &e){
      cerr << "can't request my own interface" << endl;
    }
    return false;

  }

  bool TVGuidePeerConnector::getChannels(list<Interface*> peers){
    MutexGuard mg(distMutex);

    if(!hosts.empty()){
      while(running){
        requestHandlerMutex.lock();
        if(requestHandlers.size() <= MAX_REQUEST_HANDLERS){
          requestHandlerIdCounter++;
          requestHandlers[requestHandlerIdCounter] = new TVGuidePeerRequestHandler(requestHandlerIdCounter,
                                                                                   this,
                                                                                   localTVGuide,
                                                                                   hosts,
                                                                                   peers);
          requestHandlerMutex.unlock();
          return true;
        }
        GLOBAL_DEBUG_STREAM("TVGuidePeerConnector::getChannels() - waitung for free slot" << endl);
        requestHandlerCondition.wait();
        requestHandlerMutex.unlock();
        
      }
    }
    return true;
  }


  void TVGuidePeerConnector::verify(string day, string channel, list<ProgramDescription> verifyList){
    MutexGuard mg(distMutex);


    time_t now = time(NULL); 
    struct tm *t = localtime(&now); //make sure, the timezone-vaiable is initialized
    tzset();
    char tmp[6];
    double d = -1*timezone/60/60*100;
    if(daylight !=0 && t->tm_isdst) d += 100;
    sprintf(tmp, " %0+5.0f", d);
    string tz = string(tmp);

    //make a list with our ITVGuidePeerObject-interface a parameter for search(...)
    // Don't need to add my interface because the requesthandler will do this for us
    list<Interface*> theStartInterface;

    if(verifyList.empty()){
      ProgramFilter pf;

      //adjust the times
      string startTime = day.substr(0, 8) + "0000" + tz;
      string endTime   = day.substr(0, 8) + "2400" + tz;
            
      pf.setStartTime(startTime);
      pf.setEndTime(endTime);
      pf.addChannel(channel);

      //start a search
      search(pf, theStartInterface);
      
    }else{
      ProgramDescription pd = localTVGuide->getEvent(day.substr(0, 8) + "0000" + tz, day.substr(0, 8) + "0000" + tz, channel);
      string lastEndTime;
      if(pd.empty()){
        //lastEndTime = day.substr(0, 8) + "0000" + tz;
      }else{
        pd.localizeTimes();
        lastEndTime = pd.getEndTime();
      }
      
      ProgramFilter pf;
      for(list<ProgramDescription>::iterator i = verifyList.begin(); i != verifyList.end(); i++){
        i->localizeTimes();
        if((lastEndTime.substr(0,12) < i->getStartTime().substr(0,12)) && !lastEndTime.empty()){
          pf.setStartTime(lastEndTime);
          pf.setEndTime(i->getStartTime());
          pf.addChannel(channel);
          //start a search
          search(pf, theStartInterface);
        }
        lastEndTime = i->getEndTime();
      }//end for(verifyList)

      list<ProgramDescription>::iterator last = verifyList.end();
      last--;
      lastEndTime = last->getEndTime();
      if(lastEndTime.substr(0,12) <=  day.substr(0, 8) + "2400"){
        pf.setStartTime(lastEndTime);
        pf.setEndTime(day.substr(0, 8) + "2400" + tz);
        pf.addChannel(channel);
        
        //start a search
        search(pf, theStartInterface);
      }
    }//end if(verifyList.empty())
  }

  void TVGuidePeerConnector::verify(string /*start*/, string /*end*/, string channel, list<ProgramDescription> verifyList){
    MutexGuard mg(distMutex);

    string lastEndTime;

    //make a list with our ITVGuidePeerObject-interface a parameter for search(...)
    // Don't need to add my interface because the requesthandler will do this for us
    list<Interface*> theStartInterface;

    for(list<ProgramDescription>::iterator i = verifyList.begin(); i != verifyList.end(); i++){
      if(lastEndTime.empty()){
        lastEndTime = i->getEndTime();
      }else{
        if(lastEndTime != i->getStartTime()){
          ProgramFilter pf;
          pf.setStartTime(lastEndTime);
          pf.setEndTime(i->getStartTime());
          pf.addChannel(channel);
          //start a search
          search(pf, theStartInterface);
        }
        lastEndTime = i->getEndTime();
      }
    }//end for(verifyList)
  }

  void TVGuidePeerConnector::getCurrentEventOnChannel(const string& channel){
    MutexGuard mg(distMutex);

    ProgramFilter pf;
    pf.setStartTime("Now");
    pf.addChannel(channel);
    //make a list with our ITVGuidePeerObject-interface a parameter for search(...)
    // Don't need to add my interface because the requesthandler will do this for us
    list<Interface*> theStartInterface;
    //start a search
    search(pf, theStartInterface);
  }

  void TVGuidePeerConnector::getFollowingEventOnChannel(const string& channel){
    MutexGuard mg(distMutex);

    ProgramFilter pf;
    pf.setStartTime("Next");
    pf.addChannel(channel);
    //make a list with our ITVGuidePeerObject-interface a parameter for search(...)
    // Don't need to add my interface because the requesthandler will do this for us
    list<Interface*> theStartInterface;
    //start a search
    search(pf, theStartInterface);

  }

  void TVGuidePeerConnector::search(ProgramFilter pf, list<Interface*>  peers){
    
    if(hosts.empty()) return;
    

    while(running){
      requestHandlerMutex.lock();
      if(requestHandlers.size() <= MAX_REQUEST_HANDLERS){
        pf.addPeerHost(getHostname());
        
        requestHandlerIdCounter++;
        requestHandlers[requestHandlerIdCounter] = new TVGuidePeerRequestHandler(requestHandlerIdCounter,
                                                                                 this,
                                                                                 localTVGuide,
                                                                                 hosts,
                                                                                 pf, 
                                                                                 peers);
        requestHandlerMutex.unlock();
        return;
      }
      GLOBAL_DEBUG_STREAM("TVGuidePeerConnector::search() - waitung for free slot" << endl);
      requestHandlerCondition.wait();
      requestHandlerMutex.unlock();

    }

    
  }

  string TVGuidePeerConnector::getChannelById(string id){
    ITVGuide_var remoteTVGuide(NULL);
    string result;
    
    for(list<TCPAddress>::iterator i = hosts.begin(); i != hosts.end(); i++){
      ServiceProxy service_proxy;
      
      try{
        service_proxy.connectTo(*i, "TVGuide");
        remoteTVGuide.reset(service_proxy.getInterface<ITVGuide>());
      } catch (const Exception &e){
        GLOBAL_ERROR_STREAM("TVGuidePeerConnector::getChannelByID() - Could not connect to session location: " <<" because of " << e << endl);
        continue;
      }
      
      result = remoteTVGuide->getChannelById(id);

      if(!result.empty()){
        return result;
      }
    }
    return result;
  }



  void TVGuidePeerConnector::run(){
  }
  
  void TVGuidePeerConnector::insertEvents(list<ProgramDescription> pdlist){
    MutexGuard mg(distMutex);
    
    if(pdlist.empty()) return;

    while(running){
      insertHandlerMutex.lock();
      if(insertHandlers.size() <= MAX_INSERT_HANDLERS){
        //now insert the events found
        insertHandlerIdCounter++;
        insertHandlers[insertHandlerIdCounter] = new TVGuidePeerInsertHandler(insertHandlerIdCounter, 
                                                                              this, 
                                                                              localTVGuide, 
                                                                              pdlist);
        insertHandlerMutex.unlock();
        return;
      }
      GLOBAL_DEBUG_STREAM("TVGuidePeerConnector::insertEvents() - waitung for free slot" << endl);
      insertHandlerCondition.wait();
      insertHandlerMutex.unlock();
    }
  }

  void TVGuidePeerConnector::addHost(TCPAddress adr){
    requestHostsForChannels.push_back(adr);
    hosts.push_back(adr);
  }
  void TVGuidePeerConnector::clearHosts(){
    requestHostsForChannels.clear();
    hosts.clear();
  }

  void TVGuidePeerConnector::insertHandlerDone(unsigned int id){
    insertHandlerMutex.lock();

    if(insertHandlers[id]){
      delete insertHandlers[id];
      insertHandlers.erase(id);
    }

    //notify waiting insert-handlers
    insertHandlerCondition.notify();
    insertHandlerMutex.unlock();
  }
  void TVGuidePeerConnector::requestHandlerDone(unsigned int id){
    requestHandlerMutex.lock();

    if(requestHandlers[id]){
      delete requestHandlers[id];
      requestHandlers.erase(id);
    }

    //notify waiting request-handlers
    requestHandlerCondition.notify();
    requestHandlerMutex.unlock();
  }

  Result TVGuidePeerConnector::peerResult(const ProgramDescriptionList& pdlist, 
                                          const int& id, 
                                          const list<Interface*>& history){
    
    insertEvents(pdlist.getList());
    
    if(!history.empty()){
      list<Interface*> newHistory = history;
      ITVGuidePeerObject *peer = dynamic_cast<ITVGuidePeerObject*>(newHistory.back());
      newHistory.pop_back();
      
      peer->peerResult(pdlist, id, newHistory);
    }

    return SUCCESS;
  }

  Result TVGuidePeerConnector::peerChannelResult (const ChannelDescriptionList& cdlist, 
                                                  const list<Interface*>& history){

    MutexGuard mg(distMutex);
    
    if(cdlist.getList().empty()) return FAILURE;

    while(running){
      insertHandlerMutex.lock();
      if(insertHandlers.size() <= MAX_INSERT_HANDLERS){
        //now insert the events found
        insertHandlerIdCounter++;
        insertHandlers[insertHandlerIdCounter] = new TVGuidePeerInsertHandler(insertHandlerIdCounter, 
                                                                              this, 
                                                                              localTVGuide, 
                                                                              cdlist.getList());
        insertHandlerMutex.unlock();
        break;
      }
      GLOBAL_DEBUG_STREAM("TVGuidePeerConnector::peerChannelResult() - waitung for free slot" << endl);
      insertHandlerCondition.wait();
      insertHandlerMutex.unlock();
    }

    //notify last peer
    if(!history.empty()){
      list<Interface*> newHistory = history;
      ITVGuidePeerObject *peer = dynamic_cast<ITVGuidePeerObject*>(newHistory.back());
      newHistory.pop_back();
      peer->peerChannelResult(cdlist, newHistory);
    }

    return SUCCESS;
  }

  TCPAddress TVGuidePeerConnector::getLastChannelHost(){
    MutexGuard mg(distMutex);

    if(!requestHostsForChannels.empty()){
      TCPAddress adr = requestHostsForChannels.front();
      requestHostsForChannels.pop_front();
      return adr;
    }else{
      throw(Exception("No hosts left"));
    }
  }

  /* **************************
      TVGuidePeerInserthandler
     ************************** */

  TVGuidePeerInsertHandler::TVGuidePeerInsertHandler(unsigned int id,
                                                     TVGuidePeerConnector *peerConnector,
                                                     ITVGuide *localTVGuide, 
                                                     list<ProgramDescription> programme)
    : mode(PROGRAM), 
      localtvg(localTVGuide),
      insertPDList(programme),
      myID(id),
      parentPeer(peerConnector)
  {
    activateThread();
  }
  TVGuidePeerInsertHandler::TVGuidePeerInsertHandler(unsigned int id,
                                                     TVGuidePeerConnector *peerConnector,
                                                     ITVGuide *localTVGuide, 
                                                     list<ChannelDescription> channels)
    : mode(CHANNEL), 
      localtvg(localTVGuide),
      insertCDList(channels),
      myID(id),
      parentPeer(peerConnector)
  {
    activateThread();
  }
  
  TVGuidePeerInsertHandler::~TVGuidePeerInsertHandler(){
    deactivateThread();
  }

  
  void TVGuidePeerInsertHandler::run(){

    Thread t = getThread();
    t.setMinPriority();

    switch(mode){
    case CHANNEL:
      {
        Result res = FAILURE;
        for(list<ChannelDescription>::iterator i = insertCDList.begin(); i != insertCDList.end(); i++){
          if(localtvg->insertChannelDescription(*i) == SUCCESS && res == FAILURE){
            res = SUCCESS;
          }
        }//end for
        
        if(res) {
          try{
            localtvg->getParentObject()->getCheckedInterface<ITVGuideUpdateFinished>()->newChannel();
          }catch(const Exception &e){
            GLOBAL_WARNING_STREAM("TVGuidePeerInsertHandler::run() - can't notify tvguide: " << e << endl); 
          }
        }
      }
      break;
    case PROGRAM:
      {
        Result res = FAILURE;
        string localID;
        Clock clock;
        Time t0 = clock.getTime();
        for(list<ProgramDescription>::iterator i = insertPDList.begin(); i != insertPDList.end(); i++){
          string channelID = i->getChannelID();
          string remoteName = localtvg->getChannelById(channelID);
          if(!remoteName.empty()){
            localID = localtvg->getChannelByName(remoteName);
            if(!localID.empty()){
              i->setChannelID(localID);
              if(localtvg->insertEvent(*i) == SUCCESS && res == FAILURE){
                res = SUCCESS;
              }
            }
          }
        }//end for
        Time t1 = clock.getTime();
        cerr << "Time for inserting " << insertPDList.size() << " events: " << t1-t0 << endl;
        if(res) {
          try{
            localtvg->getParentObject()->getCheckedInterface<ITVGuideUpdateFinished>()->newShow(localID);
          }catch(const Exception &e){
            GLOBAL_WARNING_STREAM("TVGuidePeerConnector::run() - can't notify tvguide: " << e << endl); 
          }
        }
      }
      break;
    default:
      break;
    }

    //tell the PeerConnector that I'm done
    parentPeer->insertHandlerDone(myID);

  }//end run()
    
  /* *********************************
      TVGuideRequestHandler
     ********************************* */

  TVGuidePeerRequestHandler::TVGuidePeerRequestHandler(unsigned int id, 
                                                       TVGuidePeerConnector *peerConnector,
                                                       ITVGuide* tvguide,
                                                       list<TCPAddress>& hosts, 
                                                       ProgramFilter request, 
                                                       list<Interface*> history)
    : mode(PROGRAM), 
      myID(id),
      parentPeer(peerConnector),
      localTVGuide(tvguide),
      requestHosts(hosts),
      currentRequest(request),
      lastHistory(history)
  {
    activateThread();
  }

  TVGuidePeerRequestHandler::TVGuidePeerRequestHandler(unsigned int id, 
                                                       TVGuidePeerConnector *peerConnector,
                                                       ITVGuide* tvguide,
                                                       list<TCPAddress>& hosts, 
                                                       list<Interface*> history)
    : mode(CHANNEL), 
      myID(id),
      parentPeer(peerConnector),
      localTVGuide(tvguide),
      requestHosts(hosts),
      lastHistory(history)
  {
    activateThread();
  }

  TVGuidePeerRequestHandler::~TVGuidePeerRequestHandler(){
    //cerr << "destroy request " << myID << endl;
    deactivateThread();
  }

  void TVGuidePeerRequestHandler::run(){

    Thread t = getThread();
    t.setMinPriority();

    ITVGuide_var remoteTVGuide(NULL);

    switch(mode){
    case CHANNEL:
      {
        ITVGuide_var remoteTVGuide(NULL);
        try{
          while(true){
            ServiceProxy service_proxy;
            TCPAddress i = parentPeer->getLastChannelHost();
          
            try{
              service_proxy.connectTo(i, "TVGuide");
              remoteTVGuide.reset(service_proxy.getInterface<ITVGuide>());
              
              //query
              lastHistory.push_back(parentPeer->getCheckedInterface<ITVGuidePeerObject>());
              remoteTVGuide->getParentObject()->getCheckedInterface<ITVGuidePeerSearch>()
                ->peerChannelSearch(lastHistory);        
              
            } catch (const BindingException& e) {
              GLOBAL_ERROR_STREAM("TVGuidePeerConnector::getChannels() - Could not connect to session location: " <<" because of " << e << endl);
              continue;
            }catch (const Exception& e) {}
          }
        }catch(const Exception &e){
          //cerr << e << endl;
        }
      }
      break;
    case PROGRAM:
      {
        //Query known TVGuides
        for(list<TCPAddress>::iterator i = requestHosts.begin(); i != requestHosts.end(); i++){
          ServiceProxy service_proxy;
          
          //If we should query a alraedy Queried TVGuide, skip the host
          if(currentRequest.hasPeerHost(i->getHostname())) continue;
          
          
          //connect
          try{
            service_proxy.connectTo(*i, "TVGuide");
            remoteTVGuide.reset(service_proxy.getInterface<ITVGuide>());
            currentRequest.addPeerHost(i->getHostname());
          } catch (const Exception &e){
            GLOBAL_ERROR_STREAM("TVGuidePeerrequestHandler::run() - Could not connect to " 
                                << i->getHostname()<< ":" << i->getPort() << " : " << e << endl);
            continue;
          }
          
          
          //query the correct channels for the search-request from remote host
          list<string> channels = currentRequest.getChannels();
          for(list<string>::iterator j = channels.begin(); j != channels.end(); j++){
            string channelName = localTVGuide->getChannelById(*j);
            if(!channelName.empty()){
              string remoteID = remoteTVGuide->getChannelByName(channelName);
              if(!remoteID.empty()){
                currentRequest.addChannel(remoteID);
              }
            }
          }
          
          try{
            //make the query
            lastHistory.push_back(parentPeer->getCheckedInterface<ITVGuidePeerObject>());
            remoteTVGuide->getParentObject()->getCheckedInterface<ITVGuidePeerSearch>()
              ->peerSearch(currentRequest, lastHistory);        
          }catch(const Exception &e){
            GLOBAL_ERROR_STREAM("TVGuidePeerRequestHandler: querying remote host failed: " << e  << endl);
          }
        }        
      }
    
    default:
      break;
    }

    //tell the PeerConnector that I'm done
    parentPeer->requestHandlerDone(myID);

  }
}
