/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "NxtvepgPlugin.hpp"

#include <fstream>

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

#include "nmm/base/registry/ReferenceManager.hpp"

#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/utils/utils.hpp"
#include "nmm/utils/DirReader.hpp"

#include "nmm/utils/fork/Fork.hpp"

//Xerces-C-Header
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMImplementationRegistry.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMError.hpp>
#include <xercesc/dom/DOMLocator.hpp>
#include <xercesc/dom/DOMNamedNodeMap.hpp>
#include <xercesc/dom/DOMAttr.hpp>
#include <xercesc/parsers/AbstractDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
namespace NMM{
  

  static NxtvepgPlugin * nxtvepgplugin = new NxtvepgPlugin();

  NxtvepgPlugin::NxtvepgPlugin()
    : TVGuidePlugin("NxtvepgPlugin"),
      daemonHost("localhost"),
      daemonPort(3187),
      period("d1"),
      dbdir("/usr/tmp/nxtvepg"),
      nxtvepgTmpFile("/tmp/nxtvepg.xml"),
      language("de"),
      termUpdate(false),
      useTvsplit(false),
      tmpDir("/tmp/mmboxnxtvepg/"),
      update_timer(this, &NxtvepgPlugin::restartUpdate)
  {
   
    configFile = getNMMResource("tvguide") + "/NxtvepgPluginConf.xml";
    readPluginConfig();

  }

  NxtvepgPlugin::~NxtvepgPlugin(){}
  
  void NxtvepgPlugin::setLocation(const string &loc){
    daemonHost = loc;
  }
  void NxtvepgPlugin::setPort(const int&p){
    daemonPort = p;
  }
  
  void NxtvepgPlugin::setConfigFile(const string &file){
    configFile = file;
    readPluginConfig();
  }
  string NxtvepgPlugin::getConfigFile(){
    return configFile;
  }

  void NxtvepgPlugin::addProvider(const string& p){
    provider.push_back(p);
    provider.sort();
    provider.unique(); //make sure, providers are unique in this list
  }
  list<string> NxtvepgPlugin::getProvider(){
    return provider;
  }
  
  void NxtvepgPlugin::setDBDir(const string& db){
    dbdir = db;
  }
  string NxtvepgPlugin::getDBDir(){
    return dbdir;
  }

  void NxtvepgPlugin::setLocale(const string& l){
    language = l;
  }
  string NxtvepgPlugin::getLocale(){
    return language;
  }

  void NxtvepgPlugin::setPeriod(const string &p){
    period = p;
  }
  string NxtvepgPlugin::getPeriod(){
    return period;
  }

  Result NxtvepgPlugin::doInit(){

    //setup timerproducer for triggered updates
    timer = new TimerProducer();
    timer->connectTo(this); 
    timer->start();
    registerEvent("NEXTNXTVEPGUPDATE", &update_timer);

    static const XMLCh gLS[] = { chLatin_L, chLatin_S, chNull };
    impl = DOMImplementationRegistry::getDOMImplementation(gLS);
    parser = NULL;

    transcoder = new XercesTranscoder();

    int res = SFork::getInstance().executeCommand("nxtvepg -help 2>/dev/null");

    if(res == 256){
      return SUCCESS;
    }else{
      return FAILURE;
    }
  }

  Result NxtvepgPlugin::doDeinit(){
    //timer->stop();
    //timer->disconnect();
    //removeEvent("NEXTNXTVEPGUPDATE");
    return SUCCESS;
  }

  Result NxtvepgPlugin::doStop(){
    termUpdate = true;
    if(useTvsplit) SFork::getInstance().executeCommand("killall -9 tv_split 2>/dev/null");
    writePluginConfig();
    return SUCCESS;
  }

  Result NxtvepgPlugin::doStart(){
    termUpdate = false;
    
    string command = "nxtvepg -dump xml";
    if(!dbdir.empty())
      command += " -dbdir " + dbdir;

    if(!provider.empty()) command += " -provider";
    for(list<string>::iterator i = provider.begin(); i != provider.end(); i++){
      command += " " + *i;
    }

    
    if(useTvsplit){
      command += " | tv_split --output " + tmpDir + "/%Y%m%d-%channel.xml 2>/dev/null";
    }else{
      command += " > " + nxtvepgTmpFile;
    }
    

    DirReader dir;
    dir.setRootDir("/");
    if(!dir.changeDir(tmpDir)){
      dir.makeDir(tmpDir);
    }

   
    if(true){ //nxtvepg executed successfully
      writePluginConfig();
      if(useTvsplit){
        dir.changeDir(tmpDir);	  

        if(dir.getRegFiles().size() <= 10){
          GLOBAL_DEBUG_STREAM(" Getting nxtvepg-data with command " << command << endl);
          SFork::getInstance().executeCommand(command.c_str());
        }else{
          GLOBAL_DEBUG_STREAM("enough files available. no need to start program" << endl);
        }

        dir.changeDir(tmpDir);	  
	vector<DirReader::directory_t> files = dir.getRegFiles();
	for(vector<DirReader::directory_t>::iterator i = files.begin(); i != files.end(); i++){
          if(termUpdate) return SUCCESS;
	  DirReader::directory_t file = *i;
	  int resXMLTV = readXMLTV(file.path + file.name);
	  if(resXMLTV == 0){
	    string rmComm = "rm " + file.path + file.name;
	    SFork::getInstance().executeCommand(rmComm.c_str()); 
	  }else if(resXMLTV == -2){
            //just delete the file because it can't be opened. maybe it is corrupt
	    string rmComm = "rm " + file.path + file.name;
	    SFork::getInstance().executeCommand(rmComm.c_str()); 
          }
	}
      }else{
        GLOBAL_DEBUG_STREAM(" Getting nxtvepg-data witch command " << command << endl);
        int res = SFork::getInstance().executeCommand(command.c_str());
        if(res == 0){
          readXMLTV(nxtvepgTmpFile);
        }else{
          return FAILURE;
        }
      }
    }
 
    //since all the available data is grabbed, we setup a timer to start the next update in the future
    //Setup period
    //compute time for an update
    int updateInterval = 0;
    
    if(termUpdate) return SUCCESS;

    try{
      if(period != ""){
        updateInterval = atoi((period.substr(1).c_str()));
        if(period.substr(0,1) == "d"){
          //days
          updateInterval *= 3600*24;
        }else if (period.substr(0,1) == "h"){
          //hours
          updateInterval*=3600;
        }else if (period.substr(0,1) == "m"){
          //minutes
          updateInterval*=60;
        }else if (period.substr(0,1) == "s"){
          //seconds
        }else{
          //default is days
          updateInterval *= 3600*24;
          GLOBAL_ERROR_STREAM("period-String empty. This should not hapen, since period is initialized with d1" << endl);
        }
      }//end if interval!=""
    }catch(const Exception &e){
      updateInterval *= 3600*24;
      GLOBAL_ERROR_STREAM("Error parsing period" << e << endl);
    }
    
    if(termUpdate) return SUCCESS;
    
    GLOBAL_DEBUG_STREAM("period: " << period << " updateInterval: " << updateInterval << endl);
    Event* e = new Event("NEXTNXTVEPGUPDATE");
    timer->registerTimer(time(NULL)+updateInterval, e);


    return stop();

  }

  /*******************************
   *   EventListener-Function    *
   ******************************* */

  Result NxtvepgPlugin::restartUpdate(){
    GLOBAL_DEBUG_STREAM("restart Nxtvepg-Update" << endl);
    start(); 
    return SUCCESS;
  }


  /* ********************
   *   Private Methods   *
   *********************** */  
  void NxtvepgPlugin::readPluginConfig(){
    GLOBAL_DEBUG_STREAM("Reading NxtvepgPlugin-Config" << endl);


    XMLTree* xml  = new XMLTree(configFile);
    XMLNode* root = NULL;
    if(xml) {
      root = xml->root();
    }else{
      GLOBAL_ERROR_STREAM("Unable to open NxtvepgPlugin-Config-File " << configFile << endl);
      return;
    }

    if(root){
      /*
      //get lastUpdate
      XMLNodeList lastUpdate = root->children("lastUpdate");
      if(!lastUpdate.empty()){
        XMLNode* update = lastUpdate.front();
        XMLProperty* dateProp = update->property("date");
        if(dateProp) lastUpdateDate = dateProp->value();
        XMLProperty* chanProp = update->property("channel");
        if(chanProp) lastUpdateChannel = chanProp->value();
      }
      */
      //get period
      XMLNodeList periodList = root->children("period");
      if(!periodList.empty()){
        XMLNode* periodNode = periodList.front();
        XMLProperty* periodProp = periodNode->property("data");
        if(periodProp) period = periodProp->value();
      }

      //get provider
      XMLNodeList providerList = root->children("provider");
      for(XMLNodeIterator i = providerList.begin(); i != providerList.end(); i++){
        XMLNode* p = *i;
        XMLProperty* providerProp = p->property("hex");
        if(providerProp){
          provider.push_back(providerProp->value());
          provider.sort();
          provider.unique();
        }
      }
      //get dbdir
      XMLNodeList dbList = root->children("dbdir");
      if(!dbList.empty()){
        XMLNode* dbNode = dbList.front();
        XMLProperty* dbProp = dbNode->property("value");
        if(dbProp) dbdir = dbProp->value();
      }

      //get language
      XMLNodeList langList = root->children("language");
      if(!langList.empty()){
        XMLNode* langNode = langList.front();
        XMLProperty* langProp = langNode->property("value");
        if(langProp) language = langProp->value();
      }

    }else{
      GLOBAL_ERROR_STREAM("Unable to open NxtvepgPlugin-Config-File " << configFile << endl);
      return;
    }
   
    delete xml;

  }
  void NxtvepgPlugin::writePluginConfig(){
    GLOBAL_DEBUG_STREAM("Writing NxtvepgPlugin-Config" << endl);
    
    ofstream outfile(configFile.c_str());
    if(!outfile.fail()){

      outfile << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>" << endl;
      //outfile << "<!DOCTYPE xmltvplugin SYSTEM \"xmltv.dtd\">" << endl;

      outfile << "<nxtvepgplugin>" << endl;
      //outfile << "<lastUpdate date=\"" << lastUpdateDate << "\" channel=\"" << lastUpdateChannel << "\"></lastUpdate>" << endl;
      outfile << "<period data=\"" << period << "\"/>" << endl;
      
      for(list<string>::iterator i = provider.begin(); i != provider.end(); i++){
        outfile << "<provider hex=\"" << *i << "\"></provider>" << endl;
      }
      if(!dbdir.empty()){
        outfile << "<dbdir value=\"" << dbdir << "\"></dbdir>" << endl;
      }
      outfile << "</nxtvepgplugin>" << endl;
    }else{
      GLOBAL_ERROR_STREAM("Could not open NxtvepgPlugin-Configuration-File " << configFile << endl);
    }
    outfile.close();
    
  }



  /** read XML-Files **/
  int NxtvepgPlugin::readXMLTV(const string& file){
    if(termUpdate) return 0;
    
    if(!tvgobject.get()) {
      GLOBAL_ERROR_STREAM("No TVGuide-Object available" << endl);
      return -1;
    }

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(file)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << file << ", check .nmmrc for options" << endl);
      return -1;
    }

    try{
      XMLPlatformUtils::Initialize();
    }catch (const XMLException& toCatch){
      GLOBAL_ERROR_STREAM("Error during initialization! :\n"
           << toCatch.getMessage() << endl);
    }catch(...){
      GLOBAL_ERROR_STREAM("strange things are going on" << endl);

    }

    //read the source-file and build the XML-Tree
    DOMDocument *doc = NULL;
    try{

      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->setProperty(XMLUni::fgXercesScannerName, (void*)XMLUni::fgWFXMLScanner);
      parser->resetDocumentPool();
      doc = parser->parseURI(file.c_str());
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          map<string, string> channelMapping;
          DOMNode* text = NULL; //DOMNode to evalute text-content
          DOMNamedNodeMap *pAttributes = NULL; //attribute-List of a node
          DOMAttr *pAttributeNode = NULL; //1 attribute of a node

          //process channels
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            if(termUpdate) {
              XMLPlatformUtils::Terminate(); 
              return 0;
            }

            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *nodename = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(nodename, "channel") == 0){
                bool insertChannel = true;
                // get all the attributes of the node
                pAttributes = child->getAttributes();
                int nSize = pAttributes->getLength();
                if(nSize > 0){
                  pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("id"));
                  // get attribute name
                  if(pAttributeNode){
                    char *value = XMLString::transcode(pAttributeNode->getValue());
                    string tmpName = tvgobject->getChannelById(string(value)); //to search DB only once
                    if(tmpName != ""){
                      //channelId already exists in database
                      channelMapping[string(value)] = string(value);
                      continue;
                    }
                    for (DOMNode* i = child->getFirstChild(); i != NULL; i=i->getNextSibling()){
                      if (i->getNodeType() == DOMNode::ELEMENT_NODE){
                        char *displayname = XMLString::transcode(i->getNodeName());
                        if(XMLString::compareString(displayname, "display-name") == 0){
                          //test, if a channel with this name already exists
                          string tmpId = tvgobject->getChannelByName(string(XMLString::transcode(i->getTextContent())));//to search DB only once
                          if(tmpId != ""){
                            channelMapping[string(value)] = tmpId;
                            insertChannel = false;
                            continue;
                          }
                          XMLString::release(&displayname);  
                        }
                        
                        //if we come so far, the channel does not exist in the DB and we have to add it
                        if(insertChannel){
                          string s = string(XMLString::transcode(i->getTextContent()));
                          tvgobject->insertChannel(string(value),s);
                          channelMapping[string(value)] = string(value);
                        }
                      }   
                    }
                    XMLString::release(&value);
                  }
                }
              }else if(XMLString::compareString(nodename, "programme") == 0){
                ProgramDescription *pd = new ProgramDescription(ProgramDescription::NXTVEPG);
                if(child->hasAttributes()) {
                  // get all the attributes of the node
                  pAttributes = child->getAttributes();
                  int nSize = pAttributes->getLength();
                  if(nSize > 0){
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("start"));
                    char *name = NULL;
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setStartTime(" << name << ")" << endl;;
                      pd->setStartTime(string(name));
                      XMLString::release(&name);
                    }
                    
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("stop"));
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setEndTime  (" << name << ")" << endl;;
                      pd->setEndTime(string(name));
                      XMLString::release(&name);
                    }
                    
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("channel"));
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setChannelID(" << name << ")" << endl;;
                      pd->setChannelID(channelMapping[string(name)]);
                      XMLString::release(&name);
                    }
                    
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("duration"));
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setDuration(" << name << ")" << endl;;
                      pd->setDuration(atoi(name));
                      XMLString::release(&name);
                    }
                    
                    for (DOMNode* i = child->getFirstChild(); i != NULL; i=i->getNextSibling()){
                      if (child->getNodeType() == DOMNode::ELEMENT_NODE){
                        name = XMLString::transcode(i->getNodeName());
                        
                        if(XMLString::compareString(name, "title") == 0){
                          //evaluate title-tag
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              //XERCES_STD_QUALIFIER cout << endl << "  addTitle(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                              pd->addTitle(transcoder->transcodeTo(text->getTextContent()));
                            }
                          }
                        }
                        
                        if(XMLString::compareString(name, "desc") == 0){
                          //evaluate description              
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              //XERCES_STD_QUALIFIER cout << "  addDescription(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                              pd->setDescription(transcoder->transcodeTo(text->getTextContent()));
                            }
                          }
                          
                        }
                        
                        if(XMLString::compareString(name, "category") == 0){              
                          //evaluate description              
                          DOMNamedNodeMap *childAttributes = i->getAttributes();
                          if(childAttributes){
                            if(childAttributes->getLength() > 0){
                              DOMAttr *childAttributeNode = (DOMAttr*) childAttributes->getNamedItem(XMLString::transcode("lang"));
                              if(childAttributeNode){
                                char *lang = XMLString::transcode(childAttributeNode->getValue());
                                if(string(lang) == language){
                                  text = i->getFirstChild();
                                  if(text){
                                    if(text->getNodeType() == DOMNode::TEXT_NODE){
                                      pd->addCategory(transcoder->transcodeTo(text->getTextContent()));
                                      //XERCES_STD_QUALIFIER cout << "  addCategory(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                                    }
                                  }
                                }
                              }else{
                                text = i->getFirstChild();
                                if(text){
                                  if(text->getNodeType() == DOMNode::TEXT_NODE){
                                    pd->addCategory(transcoder->transcodeTo(text->getTextContent()));
                                    //XERCES_STD_QUALIFIER cout << "  addCategory(" << XMLString::transcode(text->getTextContent()) << ")"$
                                  }
                                }
                              }
                            }
                          }
                        }
                        if(XMLString::compareString(name, "duration") == 0){
                          //evaluate duration
                          DOMNamedNodeMap *childAttributes = i->getAttributes();
                          int childSize = childAttributes->getLength();
                          if(childSize > 0){
                            DOMAttr *childAttributeNode =
                              (DOMAttr*) childAttributes->getNamedItem(XMLString::transcode("unit"));
                            if(childAttributeNode){
                              char *lang = xercesc::XMLString::transcode(childAttributeNode->getValue());
                              //XERCES_STD_QUALIFIER cout << "    lang=" << lang << endl;
                              XMLString::release(&lang);
                            }
                          }
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              char *duration = XMLString::transcode(text->getNodeValue());
                              //XERCES_STD_QUALIFIER cout << "  addDuration(" << duration << ")" << endl;
                              pd->setDuration(atoi(duration));
                              XMLString::release(&duration);
                            }
                          }
                        }
                        XMLString::release(&nodename);
                      }
                    }
                  }
                }
                //insert event to TVGuide-Object
                tvgobject->insertEvent(*pd);
                delete pd;
                usleep(500);
              }              
            }
            if(termUpdate) {
              XMLPlatformUtils::Terminate(); 
              return 0;
            }
            
          }
        }//if(root)
      }//end if(doc)

      parser->release();
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("NxtvepgPlugin: saveCategories: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
      GLOBAL_ERROR_STREAM("NxtvepgPlugin: saveCategories: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
      GLOBAL_ERROR_STREAM("NxtvepgPlugin: saveCategories: general Exception: " << endl);
    }

    if(!termUpdate) termUpdate=false;
    XMLPlatformUtils::Terminate(); 
    return 0;
  }
  


}
