/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "ConflictChecker.hpp"

#include "nmm/interfaces/misc/tvguide/ProgramDescription.hpp"
#include "nmm/utils/utils.hpp"

namespace NMM{
  TVTimerConflictChecker::TVTimerConflictChecker(){}
  TVTimerConflictChecker::~TVTimerConflictChecker(){}
  
  Result TVTimerConflictChecker::checkForConflicts(TVTimerEntry &entry, list<TVTimerEntry> &listToCheck){

   for(list<TVTimerEntry>::iterator i = listToCheck.begin(); i!=listToCheck.end(); i++ ) {

     //Don't check timers that are already recorded. This could happen if you start instant-recording with EPG-informations
     if(i->getStatus() == TVTimerEntry::RECORDED || i->getStatus() == TVTimerEntry::RECORDED_WITH_ERROR){
       continue;
     }

     try{
       //special case: no endtime is given. so record until end of time and (disk-)space
       if(entry.getEndTime() == 0){
         if( (i->getStartTime() <= entry.getStartTime()) && (entry.getStartTime() < i->getEndTime())){
           if(i->getChannelName() == entry.getChannelName()){
             throw TimerAlreadyProgrammedException(i->getID());
           }else{
             return FAILURE;
           }
         }else{
           continue;
         }
       }


       Result res = checkTimes(i->getStartTime(), i->getEndTime(), entry.getStartTime(), entry.getEndTime());
       if(res == FAILURE) {
         if(i->getChannelName() != entry.getChannelName()){
           entry.addConflict(i->getID());
           return FAILURE;
         }else{
           continue;
         }
       }
       
       if(i->getPeriod() == TVTimerEntry::DAILY){
         //cerr << "DAILY" << endl;
         time_t iStart     = iStart = i->getStartTime();
         time_t entryStart = entryStart =entry.getStartTime();
         time_t iEnd       = i->getEndTime();
         time_t entryEnd   = entry.getEndTime();
         
         struct tm *t = localtime(&iStart);
         struct tm iStart_tm    = *t;
         t = localtime(&iEnd);
         struct tm iEnd_tm      = *t;
         t = localtime(&entryStart);
         struct tm entryStart_tm= *t;
         t = localtime(&entryEnd);
         struct tm entryEnd_tm  = *t;       
         
         iStart_tm.tm_year = entryStart_tm.tm_year;
         iStart_tm.tm_mon  = entryStart_tm.tm_mon;
         iStart_tm.tm_mday = entryStart_tm.tm_mday;
         
         iEnd_tm.tm_year = entryEnd_tm.tm_year;
         iEnd_tm.tm_mon  = entryEnd_tm.tm_mon;
         iEnd_tm.tm_mday = entryEnd_tm.tm_mday;
         
         iStart = mktime(&iStart_tm);
         iEnd = mktime(&iEnd_tm);
         
         Result res = checkTimes(iStart, iEnd, entryStart, entryEnd);
         if(res == FAILURE) {
           if(i->getChannelName() != entry.getChannelName()){
             entry.addConflict(i->getID());
             return FAILURE;
           }else{
             continue;
           }
         }
       }else if(i->getPeriod() == TVTimerEntry::WEEKLY){
         //cerr << "WEEKLY" << endl;
         time_t iStart     = i->getStartTime();
         time_t iEnd       = i->getEndTime();
         time_t entryStart = entry.getStartTime();
         time_t entryEnd   = entry.getEndTime();
         
         struct tm *t = localtime(&iStart);
         struct tm iStart_tm    = *t;
         t = localtime(&iEnd);
         struct tm iEnd_tm      = *t;
         t = localtime(&entryStart);
         struct tm entryStart_tm= *t;
         t = localtime(&entryEnd);
         struct tm entryEnd_tm  = *t;
         
         //make the dates equal to compare only the time
         iStart_tm.tm_year = entryStart_tm.tm_year;
         iStart_tm.tm_mon  = entryStart_tm.tm_mon;
         iStart_tm.tm_mday = entryStart_tm.tm_mday;
         
         if(iStart_tm.tm_wday == entryStart_tm.tm_wday){
           iEnd_tm.tm_year = entryEnd_tm.tm_year;
           iEnd_tm.tm_mon = entryEnd_tm.tm_mon;
           iEnd_tm.tm_mday = entryEnd_tm.tm_mday;
           
           iStart = mktime(&iStart_tm);
           iEnd   = mktime(&iEnd_tm);
           
           Result res = checkTimes(iStart, iEnd, entryStart, entryEnd);
           if(res == FAILURE) {
             if(i->getChannelName() != entry.getChannelName()){
               entry.addConflict(i->getID());
               return FAILURE;
             }else{
               continue;
             }
           }
         }
         
       }else if(i->getPeriod() == TVTimerEntry::MONTHLY){
         //cerr << "MONTHLY" << endl;
         time_t iStart     = i->getStartTime();
         time_t iEnd       = i->getEndTime();
         time_t entryStart = entry.getStartTime();
         time_t entryEnd   = entry.getEndTime();
         
         struct tm *t = localtime(&iStart);
         struct tm iStart_tm    = *t;
         t = localtime(&iEnd);
         struct tm iEnd_tm      = *t;
         t = localtime(&entryStart);
         struct tm entryStart_tm= *t;
         t = localtime(&entryEnd);
         struct tm entryEnd_tm  = *t;
         
         //make the dates equal to compare only the time
         iStart_tm.tm_year = entryStart_tm.tm_year;
         iStart_tm.tm_mon  = entryStart_tm.tm_mon;
         
         iEnd_tm.tm_year = entryEnd_tm.tm_year;
         iEnd_tm.tm_mon = entryEnd_tm.tm_mon;
         
         iStart = mktime(&iStart_tm);
         iEnd   = mktime(&iEnd_tm);
         
         Result res = checkTimes(iStart, iEnd, entryStart, entryEnd);
         if(res == FAILURE) {
           if(i->getChannelName() != entry.getChannelName()){
             entry.addConflict(i->getID());
             return FAILURE;
           }else{
             continue;
           }
         }
         
       }else if(i->getPeriod() == TVTimerEntry::WEEKDAYS){
         //cerr << "WEEKDAYS" << endl;
         time_t iStart     = i->getStartTime();
         time_t iEnd       = i->getEndTime();
         time_t entryStart = entry.getStartTime();
         time_t entryEnd   = entry.getEndTime();
         
         struct tm *t = localtime(&iStart);
         struct tm iStart_tm    = *t;
         t = localtime(&iEnd);
         struct tm iEnd_tm      = *t;
         t = localtime(&entryStart);
         struct tm entryStart_tm= *t;
         t = localtime(&entryEnd);
         struct tm entryEnd_tm  = *t;
         
         //make the dates equal to compare only the time
         iStart_tm.tm_year = entryStart_tm.tm_year;
         iStart_tm.tm_mon  = entryStart_tm.tm_mon;
         iStart_tm.tm_mday = entryStart_tm.tm_mday;
         
         if((1 <= entryStart_tm.tm_wday) && (entryStart_tm.tm_wday <= 5)){
           iEnd_tm.tm_year = entryEnd_tm.tm_year;
           iEnd_tm.tm_mon = entryEnd_tm.tm_mon;
           iEnd_tm.tm_mday = entryEnd_tm.tm_mday;
           
           iStart = mktime(&iStart_tm);
           iEnd   = mktime(&iEnd_tm);
           
           Result res = checkTimes(iStart, iEnd, entryStart, entryEnd);
           if(res == FAILURE) {
             if(i->getChannelName() != entry.getChannelName()){
               entry.addConflict(i->getID());
               return FAILURE;
             }else{
               continue;
             }
           }
         }
       }else if(i->getPeriod() == TVTimerEntry::WEEKENDS){
         //cerr << "WEEKENDS" << endl;
         time_t iStart     = i->getStartTime();
         time_t iEnd       = i->getEndTime();
         time_t entryStart = entry.getStartTime();
         time_t entryEnd   = entry.getEndTime();
         
         struct tm *t = localtime(&iStart);
         struct tm iStart_tm    = *t;
         t = localtime(&iEnd);
         struct tm iEnd_tm      = *t;
         t = localtime(&entryStart);
         struct tm entryStart_tm= *t;
         t = localtime(&entryEnd);
         struct tm entryEnd_tm  = *t;
         
         //make the dates equal to compare only the time
         iStart_tm.tm_year = entryStart_tm.tm_year;
         iStart_tm.tm_mon  = entryStart_tm.tm_mon;
         iStart_tm.tm_mday = entryStart_tm.tm_mday;
         
         if(0 == entryStart_tm.tm_wday || 6 == entryStart_tm.tm_wday){
           iEnd_tm.tm_year = entryEnd_tm.tm_year;
           iEnd_tm.tm_mon = entryEnd_tm.tm_mon;
           iEnd_tm.tm_mday = entryEnd_tm.tm_mday;
           
           iStart = mktime(&iStart_tm);
           iEnd   = mktime(&iEnd_tm);
           
           Result res = checkTimes(iStart, iEnd, entryStart, entryEnd);
           if(res == FAILURE) {
             if(i->getChannelName() != entry.getChannelName()){
               entry.addConflict(i->getID());
               return FAILURE;
             }else{
               continue;
             }
           }
         }
       }else if(i->getPeriod() == TVTimerEntry::RANDOM_ONLY_CHANNEL
                || i->getPeriod() == TVTimerEntry::RANDOM
                || i->getPeriod() == TVTimerEntry::RANDOM_ONLY_CHANNEL_AND_TIME){

         list<ProgramDescription> nextTimers = i->getNextTimers();
         //cerr << "RANDOM " << nextTimers.size() << endl;         
         for(list<ProgramDescription>::iterator j = nextTimers.begin(); j != nextTimers.end(); j++){
           if(checkTimes(stringToTime(j->getStartTime()) ,
                         stringToTime(j->getEndTime()),
                         entry.getStartTime(),
                         entry.getEndTime())
              == FAILURE){
             if(i->getChannelName() != entry.getChannelName()){
               entry.addConflict(i->getID());
               return FAILURE;
             }else{
               continue;
             }
           }else{
           }
         }
       }
     }catch(const TimerAlreadyProgrammedException& e){
       if(i->getChannelName() != entry.getChannelName()){
         entry.addConflict(i->getID());
         return FAILURE;
       }else{
         throw TimerAlreadyProgrammedException(i->getID());
       }
     }catch(const Exception &e){
       entry.addConflict(i->getID());
       return FAILURE;
     }
   }
   
   return SUCCESS;
   
  }

  Result TVTimerConflictChecker::checkTimes(time_t start1, time_t end1, time_t start2, time_t end2)throw (TimerAlreadyProgrammedException){

    //if the timer is the same
    if(start1 <= start2 && start2 <= end1 && start1 <= end2 && end2 <= end1){
      throw TimerAlreadyProgrammedException();    
    }    

    //For already recording infinite timers. new timers will result in conflict or find a new TVCard
    if(end1 == 0) return FAILURE;

    if(start1 <= start2 && start2 < end1 ){
      return FAILURE;
    }
    if(start1 < end2   && end2 <=  end1){
      return FAILURE;
    }
    if(start1 >= start2 && end2 >=  end1){
      return FAILURE;
    }
    if(start1 <= start2 && end2 <=  end1){
      return FAILURE;
    }    
    
    
    return SUCCESS;
  }


}


