/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <errno.h>

#include "RawWriteNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/registry/Plugin.hpp"


NMM_REGISTER_PLUGIN(NMM::RawWriteNode, "RawWriteNode")

namespace NMM {
  RawWriteNode::RawWriteNode(const char* name,
			     const StreamQueue::Mode queue_mode,
			     const int queue_size) : 
    GenericSinkNode(name, queue_mode, queue_size ),
    file_stream()
  {
    write_mode  = IFileModeHandler::FILE_PER_STREAM;
    file_format = IFileModeHandler::DATA_ONLY;
    filename_mode = IFileModeHandler::FILENAME_ONLY;
    file_name   = "";
    buf_count   = 0;
  };

  RawWriteNode::~RawWriteNode() 
  {}

  void RawWriteNode::setWorkingDirectory(const string&) {
    // XXX: not supported
  }
  
  string RawWriteNode::getWorkingDirectory() {
    // XXX: not supported
    return "";
  }
  
  void RawWriteNode::setFileExtension(const string&) {
    // XXX: not supported
  }
  
  string RawWriteNode::getFileExtension() {
    // XXX: not supported
    return "";
  }
  
 void RawWriteNode::setLoop(const bool&) {
   /// XXX: not supported
  }

  void RawWriteNode::setFormat(const IFileModeHandler::FileFormat& format)
  {
    file_format = format;
  }
  
  IFileModeHandler::FileFormat RawWriteNode::getFormat() {
    return file_format;
  }

  void RawWriteNode::setMode(const IFileModeHandler::FileMode& mode)
  {
    write_mode = mode;
  }
  
  IFileModeHandler::FileMode RawWriteNode::getMode() {
    return write_mode;
  }

  void RawWriteNode::setFilenameMode(const IFileModeHandler::FilenameMode& mode)
  {
    filename_mode = mode;
  }

  IFileModeHandler::FilenameMode RawWriteNode::getFilenameMode() {
    return filename_mode;
  }

  void RawWriteNode::setFilename(const string &name) {
    if(write_mode == IFileModeHandler::FILE_PER_STREAM) {
      if(isConstructed()) {
	file_name = name;
      }      
      else {
	ERROR_STREAM("Node not in CONSTRUCTED state, setFileName illegal in current state and with FILE_PER_STREAM mode.");
      }
    }
    else {
      file_name_mutex.lock();
      file_name = name;
      file_name_mutex.unlock();
    }
  }

  Result RawWriteNode::doInit() 
  {
    Format* in_format = getOwnInputProperty()->addNewFormat(Format::wild_wild_type);
    getOwnInputProperty()->setDefaultFormat(in_format);

    return SUCCESS;
  }

  Result RawWriteNode::doInitOutput() 
  {
    if (!getInputFormat()) {
      ERROR_STREAM("InputFormat is NULL!" << endl);
      return FAILURE;
    }
    Format* in_format = new Format(*getInputFormat());
    getOwnWorkingInputProperty()->addNewFormat(in_format);
    getOwnWorkingInputProperty()->setDefaultFormat(in_format);

    if(filename_mode != IFileModeHandler::TIMESTAMP_AS_NAME &&
       write_mode == IFileModeHandler::FILE_PER_STREAM) {
      
      /* Security check */
      if (!SSecurityCheck::getInstance().checkFilenameForWriting(file_name)) {
        ERROR_STREAM("Security check not passed for file " << file_name << ", check .nmmrc for options" << endl);
        getOwnWorkingInputProperty()->erase();
        return FAILURE;
      }
      
#if (__GNUC__ >= 3)
      file_stream.open(file_name.c_str(), ios_base::out | ios_base::trunc);//, 00644);
#else
      file_stream.open(file_name.c_str(), ios::out | ios::trunc, 00644);
#endif
      if(file_stream.fail()) {
	ERROR_STREAM( "Error while opening file " << file_name << endl);
        getOwnWorkingInputProperty()->erase();
	return FAILURE;
      }
      else {
	DEBUG_STREAM( "Opened file " << file_name << endl);
      }
    }

    return SUCCESS;
  }

  Buffer *RawWriteNode::processBuffer(Buffer *buffer) 
  {
    MutexGuard m(file_name_mutex);
    int write_status;
    int write_attempt;

    if(write_mode == IFileModeHandler::FILE_PER_FRAME ||
       filename_mode == IFileModeHandler::TIMESTAMP_AS_NAME) { // create new file
      
      string tmpfname = file_name;
      if (filename_mode == IFileModeHandler::FILENAME_BUFCOUNT) tmpfname= file_name + buf_count;
      else if (filename_mode == IFileModeHandler::TIMESTAMP_AS_NAME) {
	tmpfname = buffer->getTimestamp().sync_time.getPrettyString();
      }
      
      // file_name_mutex.lock();
      
      /* Security check */
      if (!SSecurityCheck::getInstance().checkFilenameForWriting(file_name)) {
        ERROR_STREAM("Security check not passed for file " << file_name << ", check .nmmrc for options" << endl);
      }
      
#if (__GNUC__ >= 3)
      file_stream.open(file_name.c_str(), ios_base::out | ios_base::trunc);//, 00644);
#else
      file_stream.open(file_name.c_str(), ios::out | ios::trunc, 00644);
#endif
      // file_name_mutex.unlock();

      if(file_stream.fail()) {
	ERROR_STREAM( "Error while opening file " << tmpfname << endl);
      } else {
	DEBUG_STREAM( "Opened file " << tmpfname << endl);
      }
    }
		
    if(file_format == IFileModeHandler::DATA_ONLY) {
      write_attempt = buffer->getUsed();
      file_stream.write(buffer->getData(), buffer->getUsed());
      write_status = buffer->getUsed();
    }
    if(file_format == IFileModeHandler::SIZE_DATA) {
      write_attempt = sizeof(size_t) + buffer->getUsed();
      size_t used = buffer->getUsed();
      //file_stream.write(&used, sizeof(size_t));
      file_stream.write(reinterpret_cast<char*>(&used), sizeof(size_t));
      write_status = sizeof(size_t);
      file_stream.write(buffer->getData(), buffer->getUsed());
      write_status += buffer->getUsed();
    }
    if(file_format == IFileModeHandler::SIZE_TIMESTAMP_DATA) {
      write_attempt = sizeof(size_t) + sizeof(Timestamp) + buffer->getUsed();
      size_t used = buffer->getUsed();
      Timestamp ts  = (buffer->getTimestamp());
      //file_stream.write(&used, sizeof(size_t));
      file_stream.write(reinterpret_cast<char*>(&used), sizeof(size_t));
      write_status = sizeof(size_t);
      //file_stream.write(&ts, sizeof(Timestamp));
      file_stream.write(reinterpret_cast<char*>(&ts), sizeof(Timestamp));
      write_status += sizeof(Timestamp);
      file_stream.write(buffer->getData(), buffer->getUsed());
      write_status += buffer->getUsed();
      ERROR_STREAM("Writer: IFileModeHandler::SIZE_TIMESTAMP_DATA used=" << used << " Timestamp=" << ts << " Data" << endl);
    }

    if(write_status == write_attempt) {
      DEBUG_STREAM( "Wrote " << write_status << " bytes. " << endl);
    }
    else {
      if(write_status < 0) {
	ERROR_STREAM( "Error Number " << errno << " while writing file " << endl);
      }
      else {
	ERROR_STREAM( "Tried to write " << write_attempt << " bytes, but could only write " << write_status << endl);
      }
    }

    if(write_mode == IFileModeHandler::FILE_PER_FRAME) {
      file_stream.close();
      //file_name_mutex.unlock();
    }

    buf_count++;
    buffer->release();
    return 0;
  }

  Result RawWriteNode::doDeinitOutput() 
  {
    if(write_mode == IFileModeHandler::FILE_PER_STREAM) {
      file_stream.close();
      DEBUG_STREAM("Successfully closed file" << endl);
    }
    return SUCCESS;
  }

}
