/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "MagickReadNode.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "nmm/interfaces/base/IFormatChange.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <fstream>
#include <string>

NMM_REGISTER_PLUGIN(NMM::MagickReadNode, "MagickReadNode")

namespace NMM{

  MagickReadNode::MagickReadNode(const char* name)
  : GenericSourceNode(name),
    __buffer_size(0),
    __end_track(false),
    __format_changed(false)
  {
  }

  MagickReadNode::~MagickReadNode()
  {
  }

  Result MagickReadNode::doStart()  {
    return SUCCESS;
  }

  Result MagickReadNode::doStop() {
    return SUCCESS;
  }


  Result MagickReadNode::doInit() {
    // Set the supported formats.
    // The formats rgb24 with 24 and 32 bits are supported.
    Format* rgb_format = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    rgb_format->addWildcard(Format::x_resolution_param);
    rgb_format->addWildcard(Format::y_resolution_param);
    rgb_format->addWildcard(Format::ratio_x_param);
    rgb_format->addWildcard(Format::ratio_y_param);
    rgb_format->addStringParamValue(Format::channelorder_param,"rgb");
    rgb_format->addStringParamValue(Format::endian_param, "little");
    rgb_format->addStringParamValue(Format::colorspace_param,"rgb24");
    rgb_format->addIntParamValue(Format::bitperpixel_param,32);
    rgb_format->addStringParamValue(Format::format_param, "packet");
    rgb_format->addWildcard(Format::framerate_param);

    // The default format has 32 bits.
    getOwnOutputProperty()->setDefaultFormat(rgb_format);

    rgb_format = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    rgb_format->addWildcard(Format::x_resolution_param);
    rgb_format->addWildcard(Format::y_resolution_param);
    rgb_format->addWildcard(Format::ratio_x_param);
    rgb_format->addWildcard(Format::ratio_y_param);
    rgb_format->addStringParamValue(Format::channelorder_param,"rgb");
    rgb_format->addStringParamValue(Format::endian_param, "little");
    rgb_format->addStringParamValue(Format::colorspace_param,"rgb24");
    rgb_format->addIntParamValue(Format::bitperpixel_param,24);
    rgb_format->addStringParamValue(Format::format_param, "packet");
    rgb_format->addWildcard(Format::framerate_param);


    return SUCCESS;
  } // doInit

  Result MagickReadNode::doInitOutput() 
  {
    if(!SSecurityCheck::getInstance().checkFilenameForReading(__filename)) {
      ERROR_STREAM("Security check not passed for file " << __filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }

    // Store old resolution.
    unsigned int width = __image.columns();
    unsigned int height = __image.rows();

    // Read the image file at this point to be able to set the output format.
    try {
      __image.read(__filename);
    } 
    catch (const Magick::ErrorFileOpen& e)
    {
      ERROR_STREAM(e.what() << endl);
      return FAILURE;
    }
    catch (const Magick::Exception& e)
    {
      // magick warnings...
      ERROR_STREAM(e.what() << endl);
    }
    catch (...)
    {
      ERROR_STREAM("Unknown exception." << endl);
    }

    // Check if format has changed.
    if (width != __image.columns() || height != __image.rows())
    {
      __format_changed = true;
    }


    DEBUG_STREAM("Reading file " << __filename << endl);

    setProducingFlag(true);

    // Initialize output format depending on the image resolution.
    pair<int, int> resolution = getImageResolution();

    if(!isOutputConnected("default")) {
      __format_changed = false;
    }

    Format* rgb_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
    rgb_format->addIntParamValue(Format::x_resolution_param, resolution.first);
    rgb_format->addIntParamValue(Format::y_resolution_param, resolution.second);
    rgb_format->addIntParamValue(Format::ratio_x_param, resolution.first);
    rgb_format->addIntParamValue(Format::ratio_y_param, resolution.second);
    rgb_format->addStringParamValue(Format::channelorder_param,"rgb");
    rgb_format->addStringParamValue(Format::endian_param, "little");
    rgb_format->addStringParamValue(Format::colorspace_param,"rgb24");
    rgb_format->addIntParamValue(Format::bitperpixel_param,32);
    rgb_format->addStringParamValue(Format::format_param, "packet");
    rgb_format->addFloatParamValue(Format::framerate_param, 1.0);
    getOwnWorkingOutputProperty()->setDefaultFormat(rgb_format);

    rgb_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
    rgb_format->addIntParamValue(Format::x_resolution_param, resolution.first);
    rgb_format->addIntParamValue(Format::y_resolution_param, resolution.second);
    rgb_format->addIntParamValue(Format::ratio_x_param, resolution.first);
    rgb_format->addIntParamValue(Format::ratio_y_param, resolution.second);
    rgb_format->addStringParamValue(Format::channelorder_param,"rgb");
    rgb_format->addStringParamValue(Format::endian_param, "little");
    rgb_format->addStringParamValue(Format::colorspace_param,"rgb24");
    rgb_format->addIntParamValue(Format::bitperpixel_param,24);
    rgb_format->addStringParamValue(Format::format_param, "packet");
    rgb_format->addFloatParamValue(Format::framerate_param, 1.0);


    return SUCCESS;
  } // doInitOutput

  Result MagickReadNode::doActivate() {
    // Initialize internal variables.

    // Try to get the output format.
    const Format* out_format = getOutputFormat();

    if (out_format == NULL) {
      return FAILURE;
    }

    // Try to get the number of bits per pixel from the ouput format.
    try {
      __bit_per_pixel = out_format->getIntValue(Format::bitperpixel_param);
    } catch (...) {
      return FAILURE;
    }

    pair<int, int> resolution = getImageResolution();

    // Calculate the buffer size.
    __buffer_size = resolution.first * resolution.second * (__bit_per_pixel >> 3);

    return SUCCESS;
  } // doActivate


  void MagickReadNode::setFilename(const string& filename)
  {
    __filename = filename;

  }

  Message* MagickReadNode::processBuffer(Buffer*)
  {
    if (__format_changed) {
      __format_changed = false;
      setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());
      DEBUG_STREAM("Send setFormat " << endl << *getOwnWorkingOutputProperty()->getDefaultFormat() << endl);
      return new CEvent(IFormatChange::create_setFormat(*getOutputFormat()));
    }

    if (__end_track) {
      __end_track = false;
      setProducingFlag(false);

      DEBUG_STREAM("Sending endTrack..." << endl);
      Event e(ITrack::endTrack_event);
      return new CEvent(e);
    }

    // Create a new NMM buffer.
    Buffer* buffer = getNewBuffer(__buffer_size);

    buffer->setUsed(__buffer_size);

    // Write the pixel data into the buffer, depending on the number of bits per pixel.
    if (__bit_per_pixel == 24) {
      __image.write(0, 0, __image.columns(), __image.rows(), "BGR",
		    Magick::CharPixel, buffer->getData());
    }
    else {
      __image.write(0, 0, __image.columns(), __image.rows(), "BGRA",
		    Magick::CharPixel, buffer->getData());
    }

    // Disable the producing flag.
    __end_track = true;

    // Finally return the buffer.
    return buffer;
  } // processBuffer

  pair<int, int> MagickReadNode::getImageResolution() {
    // Return the resolution by calling the corresponding Magick::__Image methods.
    return make_pair(__image.columns(), __image.rows());
  }
} //end namespace

