/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "BufferDropNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include <stdio.h>
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::BufferDropNode, "BufferDropNode")

namespace NMM {

  BufferDropNode::BufferDropNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericFilterNode(name, qmode, qsize),
    drop_num(0),
    drop_denom(0),
    drop_ratio(0),
    ratio_counter(0)
  {
    buffer_to_drop = 0;
    buffer_count = 0;
  }
  
  void BufferDropNode::setBufferToDrop(const unsigned int &buffer_num) 
  {
    buffer_to_drop = buffer_num;
  }
  
  Result BufferDropNode::doInit() 
  {
    Format* any_format_i = new Format("*/*");
    getOwnInputProperty()->addNewFormat(any_format_i);
    
    Format* any_format_o = new Format("*/*");
    getOwnOutputProperty()->addNewFormat(any_format_o);
    
    return SUCCESS;
  }
  
  Result BufferDropNode::doInitOutput() {
    /* get input format */
    if (!getInputFormat()) {
      GLOBAL_ERROR_STREAM("No Input Format! FAILURE" << endl);
      return FAILURE;
    }

    Format* i_format = new Format(*getInputFormat());
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    Format* o_format = new Format(*i_format);
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Message* BufferDropNode::processBuffer(Buffer *in) 
  {
    if(buffer_to_drop > 0) { 
      ++buffer_count;
      if(buffer_count == buffer_to_drop) {
	buffer_count = 0;
	in->release();
	return 0;
      }
    }

    else if (drop_num > 0) {
      ++buffer_count;
      
      int64_t old_integer_part = ratio_counter.getNumerator() / ratio_counter.getDenominator();
      ratio_counter += drop_ratio;
      int64_t new_integer_part = ratio_counter.getNumerator() / ratio_counter.getDenominator();
      if (old_integer_part < new_integer_part) {
        DEBUG_STREAM("DropDecision: " << buffer_count << "." << " Dropping!" << endl);
        in->release();
        in = NULL;
      } else {
        DEBUG_STREAM("DropDecision: " << buffer_count << "." << " Leaving!" << endl);
      }

      if (buffer_count == drop_denom) {
        buffer_count = 0;
      }
    }

    return in;
  }

  void BufferDropNode::setDropRatio(const unsigned int& numerator, const unsigned int& denominator) {
    if ((denominator == 0) || (numerator == 0) || (numerator > denominator)) {
      ERROR_STREAM("setDropRatio() Invalid Args: " << numerator << " / " << denominator << endl);
      return;
    }

    if (buffer_to_drop > 0) {
      buffer_to_drop = 0;
    }

    drop_num = numerator;
    drop_denom = denominator;

    drop_ratio = Rational(drop_num, drop_denom);
    drop_ratio.reduce();

    buffer_count = 0;
  }
}
