/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *                    Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SVGEncodeNode.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/sync/StreamTimer.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include <stdio.h>
#include <sstream>
#include <math.h>

NMM_REGISTER_PLUGIN(NMM::SVGEncodeNode, "SVGEncodeNode")

namespace NMM{

  Point::Point(const float& x1, const float& y1, DiffTimer* timer) {
    x = x1;
    y = y1;
    time = timer->getDiff();
  }
  
  Point::Point(const float& x1, const float& y1){
    x = x1;
    y = y1;
    time=0;
  }

  Point::Point(){
    x=0;
    y=0;
    time=0;
  }

  Point::~Point(){
  }

  int Point::getX() const {
    return static_cast<int>(x);
  }

  int Point::getY() const {
    return static_cast<int>(y);
  }
  
  float Point::getX_float() const {
    return x;
  }

  float Point::getY_float() const {
    return y;
  }

  int Point::distX(const Point& p2) const {
    return (abs(static_cast<int>(p2.getX() - x)));
  }

  int Point::distY(const Point& p2) const {
    return (abs(static_cast<int>(p2.getY() - y)));
  }
  
  float Point::dist_float(const Point& p2) const {
    double tmp = (abs(p2.getY() - getY()))^2+(abs(p2.getX() - getX()))^2;
    return sqrt(tmp); 
  }

  Point Point::add(const Point& p2) const {
    return Point(getX_float()+p2.getX_float(),getY_float()+p2.getY_float());
  }

  Point Point::mult_c(const float& c) const {
    return Point(getX_float()*c,getY_float()*c);
  }

  float Point::getTime() const {
    return time;
  }
  
  SVGObject::SVGObject(Point& p) : NamedObject("SVGObject"){
    startpoint = p;
    red = 0;
    green = 0;
    blue = 0;
    isSent = false;
    isFinished = false;
    stroke_width = 2;
  }
  
  SVGObject::SVGObject(Point& point, int r, int g, int b) : NamedObject("SVGObject"){
    startpoint = point;
    red = r;
    green = g;
    blue = b;
    isSent = false;
    isFinished = false;
    stroke_width = 2;
  }
  
  SVGObject::SVGObject(Point& point, DiffTimer* svgtimer, int r = 0, int g = 0, int b = 0, int stroke = 2) : NamedObject("SVGObject"){
    startpoint = point;
    red = r;
    green = g;
    blue = b;
    isSent = false;
    isFinished = false;
    timer = svgtimer;
    stroke_width = stroke;
  }

  SVGObject::~SVGObject() {
  }

  void SVGObject::finish(Point&) {
  }

  bool SVGObject::getIsFinished() {
    return isFinished;
  }

  void SVGObject::send() {
    isSent = true;
  }
  
  bool SVGObject::getIsSent() {
    return isSent;
  }

  void SVGObject::addPoint(Point&){  
  }
  
  string SVGObject::getSVGColorString() const {
    string colorstring("");
    if (!isFinished) {
      colorstring = "rgb(64,64,64)";
    } else {
      colorstring = string("rgb(") + red + "," + green + "," + blue + ")";
    }
    return colorstring;
  }
  
  void SVGObject::setColor(int r, int g, int b) {
    bool valid = true;

    if (r >= 0 && r <= 255) {
      red = r;
    } else {
      valid = false;
      DEBUG_STREAM("Red value not in range 0 to 255.");
    }
    
    if (valid && g >= 0 && g <= 255) {
      green = g;
    } else {
      valid = false;
      DEBUG_STREAM("Green value not in range 0 to 255.");
    }

    if (valid && b >= 0 && b <= 255) {
      blue = b;
    } else {
      DEBUG_STREAM("Blue value not in range 0 to 255.");
    }
  }
  
  SVGLine::SVGLine(Point& point) : SVGObject(point){
    isFinished=false;
    endpoint=point;
  }

  SVGLine::SVGLine(Point& point, int r, int g, int b) : SVGObject(point, r, g, b){
    isFinished=false;
    endpoint=point;
  }

  SVGLine::SVGLine(Point& point, DiffTimer* timer, int r = 0, int g = 0, int b = 0, int stroke = 2) : SVGObject(point, timer, r, g, b, stroke){
    isFinished=false;
    endpoint=point;
  }

  SVGLine::~SVGLine(){
  }

  void SVGLine::finish(Point& point){
    isFinished = true;
    endpoint = point;
  }

  void SVGLine::addPoint(Point& point){
    endpoint = point;
  }
  
  string SVGLine::getSVGString() const{ 
    string svgstring("");   
    svgstring += string("<line x1=\"") + startpoint.getX() + "\" y1=\"" + startpoint.getY() + "\" ";
    svgstring += string("x2=\"") + endpoint.getX() + "\" y2=\"" + endpoint.getY() + "\" ";
    svgstring += string("style=\"stroke:") + getSVGColorString() + ";stroke-width:" + stroke_width + ";\"/>\n";
    return svgstring;
  }

  string SVGLine::getAnimatedSVGString(){   
    stringstream  svgstring(stringstream::in | stringstream::out);
    int x1 = startpoint.getX();
    int y1 = startpoint.getY();
    int x2 = endpoint.getX();
    int y2 = endpoint.getY();
    float dur = endpoint.getTime()-startpoint.getTime();
        
    svgstring << string("<!-- Begin of an SVG-Line from (") << x1 << "," << y1 << ") to (" << x2 << "," << y2 << ") -->\n";
    
    svgstring << string("<line x1=\"") << x1 << "\" y1=\"" << y1 << "\" ";
    svgstring << string("x2=\"") << x1 << "\" y2=\"" << y1 << "\" ";
    svgstring << string("style=\"stroke:") << getSVGColorString() << ";stroke-width:" << stroke_width << ";\">\n";

    svgstring << string("\t<animate attributeName=\"x2\" attributeType=\"XML\"\n");
    svgstring << string("\t\tbegin=\"") << startpoint.getTime() << "s\" dur=\"" << dur << "s\"";
    svgstring << string(" from=\"") << x1 << "\" to=\"" << x2 << "\" fill=\"freeze\" />\n";

    svgstring << string("\t<animate attributeName=\"y2\" attributeType=\"XML\"\n");
    svgstring << string("\t\tbegin=\"") << startpoint.getTime() << "s\" dur=\"" << dur << "s\"";
    svgstring << string(" from=\"") << y1 << "\" to=\"" << y2 << "\" fill=\"freeze\" />\n";

    svgstring << string("</line>\n\n");
    return svgstring.str();
  }

  void SVGLine::setPathNumber(int) {
    GLOBAL_DEBUG_STREAM("Dummy-method without any function. Method for setting the pathnumber in SVGFreehand" << endl);
  }
  
  SVGRectangle::SVGRectangle(Point& point) : SVGObject(point) {
    endpoint=point;
    width=0;
    height=0;
  }

  SVGRectangle::SVGRectangle(Point& point, int r, int g, int b) : SVGObject(point, r, g, b) {
    endpoint=point;
    width=0;
    height=0;
  }
  
  SVGRectangle::SVGRectangle(Point& point, DiffTimer* timer, int r = 0, int g = 0, int b = 0, int stroke = 2) : SVGObject(point, timer, r, g, b, stroke) {
    endpoint=point;
    width=0;
    height=0;
  }
  
  SVGRectangle::~SVGRectangle(){
  }
  
  void SVGRectangle::finish(Point& point){
    isFinished = true;
    endpoint = point;
    width = startpoint.distX(point);
    height = startpoint.distY(point);
  }

  void SVGRectangle::addPoint(Point& point){
    endpoint = point;
    width = startpoint.distX(point);
    height = startpoint.distY(point);
  }
  
  string SVGRectangle::getSVGString() const{
    string svgstring("");

    int x1 = 0, y1 = 0;
    if (startpoint.getX() <= endpoint.getX()) {
      x1 = startpoint.getX();
    } else {
      x1 = endpoint.getX();
    }

    if (startpoint.getY() <= endpoint.getY()) {
      y1 = startpoint.getY();
    } else {
      y1 = endpoint.getY();
    }
    
    svgstring += string("<rect x=\"") + x1 + "\" y=\"" + y1 + "\" ";
    svgstring += string("width=\"") + width + "\" height=\"" + height;
    svgstring += string("\" style=\"stroke:") + getSVGColorString() + ";stroke-width:" + stroke_width + ";fill:none;\"/>\n";
    
    return svgstring;
  }

  string SVGRectangle::getAnimatedSVGString(){
    
    stringstream  svgstring(stringstream::in | stringstream::out);
    int x1 = 0, y1 = 0;
    if (startpoint.getX() <= endpoint.getX()) {
      x1 = startpoint.getX();
    } else {
      x1 = endpoint.getX();
    }

    if (startpoint.getY() <= endpoint.getY()) {
      y1 = startpoint.getY();
    } else {
      y1 = endpoint.getY();
    }
    
    float dur = endpoint.getTime()-startpoint.getTime();
        
    svgstring << string("<!-- Begin of an SVG-Rectangle from (") << x1 << "," << y1 << ") with width=";
    svgstring << width << " and height=" << height << " -->\n";
    
    svgstring << string("<rect x=\"") << x1 << "\" y=\"" << y1 << "\" ";
    svgstring << string("width=\"0\" height=\"0\" ");
    svgstring << string("style=\"stroke:") << getSVGColorString() << ";stroke-width:" << stroke_width << ";fill:none;\">\n";

    svgstring << string("\t<animate attributeName=\"width\" attributeType=\"XML\"\n");
    svgstring << string("\t\tbegin=\"") << startpoint.getTime() << "s\" dur=\"" << dur << "s\"";
    svgstring << string(" from=\"0\" to=\"") << width << "\" fill=\"freeze\" />\n";

    svgstring << string("\t<animate attributeName=\"height\" attributeType=\"XML\"\n");
    svgstring << string("\t\tbegin=\"") << startpoint.getTime() << "s\" dur=\"" << dur << "s\"";
    svgstring << string(" from=\"0\" to=\"") << height << "\" fill=\"freeze\" />\n";

    svgstring << string("</rect>\n\n");
    return svgstring.str();
  }

  void SVGRectangle::setPathNumber(int) {
    GLOBAL_DEBUG_STREAM("Dummy-method without any function. Method for setting the pathnumber in SVGFreehand" << endl);
  }

  SVGFreehand::SVGFreehand(Point& point) : SVGObject(point){
    addPoint(point);
  }

  SVGFreehand::SVGFreehand(Point& point, int r, int g, int b) : SVGObject(point, r, g, b){
    addPoint(point);
  }

  SVGFreehand::SVGFreehand(Point& point, DiffTimer* timer, int r = 0, int g = 0, int b = 0, int stroke = 2) : SVGObject(point, timer, r, g, b, stroke){
    addPoint(point);
  }

  SVGFreehand::~SVGFreehand(){
  }

  void SVGFreehand::finish(Point& point){
    isFinished = true;
    addPoint(point);
  }

  void SVGFreehand::addPoint(Point& point){
    points.push_back(point);
    if (points.size()==3) {                            // if there are three points 
      Point mirrored = mirror(points[1], points[0]);   // the second point is mirrored on the first
      points.insert(points.begin(), mirrored);
      calculateBezier();                               // then you have four points to calcuate the Bezier curve between the two points
    } else if (points.size()>3) {                      // if there are more than three points
      calculateBezier();                               // you have at least four points to calculate the Bezier curve
    }
  }
  
  Point SVGFreehand::mirror(Point source, Point mirror) {
    int xm = 2*mirror.getX() - source.getX();
    int ym = 2*mirror.getY() - source.getY();
    return Point(xm,ym);
  }

  int SVGFreehand::bezier1(int x0, int x1, int x2) {
    return (-x0 + 6*x1 + x2)/6;
  }

  int SVGFreehand::bezier2(int x1, int x2, int x3) {
    return (x1 + 6*x2 - x3)/6;
  }

  void SVGFreehand::calculateBezier(){
    int last = points.size()-1;
    if (points.size()>=4) {
      Point p0 = points[last-3];
      Point p1 = points[last-2];
      Point p2 = points[last-1];
      Point p3 = points[last];
      Point b1(bezier1(p0.getX(),p1.getX(),p2.getX()), bezier1(p0.getY(),p1.getY(),p2.getY()));
      Point b2(bezier2(p1.getX(),p2.getX(),p3.getX()), bezier2(p1.getY(),p2.getY(),p3.getY()));
      beziers.push_back(b1);
      beziers.push_back(b2);
    }
  }

  string SVGFreehand::getSVGString() const{
    string svgstring("");
  
    svgstring += string("<path d=\"M") + startpoint.getX() + "," + startpoint.getY() + "\n";
    for (unsigned int i=2; i < points.size()-1; i++) {
      svgstring += string("C") + beziers[2*i-4].getX() + "," + beziers[2*i-4].getY();
      svgstring += string(" ") + beziers[2*i-3].getX() + "," + beziers[2*i-3].getY();
      svgstring += string(" ") + points[i].getX() + "," + points[i].getY() + "\n";
    }
    svgstring += string("\" style=\"stroke:") + getSVGColorString() + ";stroke-width:" + stroke_width + ";fill:none\"/>\n";
    return svgstring;
  }
  
  string SVGFreehand::getAnimatedSVGString() {
    string svgstring("");
    stringstream svgstream(stringstream::in | stringstream::out);
    float start = startpoint.getTime();
    float dur = points[points.size()-1].getTime() - start;
    
    svgstring += string("\n<!-- Begin of an SVG-Freehand -->\n");
    svgstring += string("<path id=\"path") + pathnumber + "\" d=\"M" + startpoint.getX() + "," + startpoint.getY() + "\n";

    for (unsigned int i=2; i < points.size()-1; i++) {
      svgstring += string("C") + beziers[2*i-4].getX() + "," + beziers[2*i-4].getY();
      svgstring += string(" ") + beziers[2*i-3].getX() + "," + beziers[2*i-3].getY();
      svgstring += string(" ") + points[i].getX() + "," + points[i].getY() + "\n";
    }

    // string will not use floats for values so I use a stringstream
    svgstream << string("") << string("\" style=\"stroke:") << getSVGColorString() << ";stroke-width:" << stroke_width;
    svgstream << ";fill:none\">" << endl;
    svgstream << string("<animate id=\"pathAnim") << pathnumber << "\" attributeName=\"stroke-dashoffset\"";
    svgstream << string(" from=\"0\" to=\"0\" dur=\"") << dur << "s\" begin=\"" << start << "s\" fill=\"freeze\"";
    
    svgstring += svgstream.str();
    svgstring += string(" keySplines=\"0.3 0 0.7 1\" calcMode=\"spline\"/>\n");
    svgstring += string("</path>\n\n");

    return svgstring;
  }

  void SVGFreehand::setPathNumber(int number) {
    pathnumber = number;
  }

  SVGPointer::SVGPointer(Point& point) : SVGObject(point){
    addPoint(point);
  }

  SVGPointer::SVGPointer(Point& point, int r, int g, int b) : SVGObject(point, r, g, b){
    addPoint(point);
  }

  SVGPointer::SVGPointer(Point& point, DiffTimer* timer, int r = 0, int g = 0, int b = 0, int stroke = 0) : SVGObject(point, timer, r, g, b, stroke){
    addPoint(point);
  }

  SVGPointer::~SVGPointer(){
  }

  void SVGPointer::finish(Point& point){
    isFinished = true;
    addPoint(point);
  }

  void SVGPointer::addPoint(Point& point){
    points.push_back(point);
    if (points.size()==3) {                            // if there are three points 
      Point mirrored = mirror(points[1], points[0]);   // the second point is mirrored on the first
      points.insert(points.begin(), mirrored);
      calculateBezier();                               // then you have four points to calcuate the Bezier curve between the two points
    } else if (points.size()>3) {                      // if there are more than three points
      calculateBezier();                               // you have at least four points to calculate the Bezier curve
    }
  }
  
  Point SVGPointer::mirror(Point source, Point mirror) {
    int xm = 2*mirror.getX() - source.getX();
    int ym = 2*mirror.getY() - source.getY();
    return Point(xm,ym);
  }

  int SVGPointer::bezier1(int x0, int x1, int x2) {
    return (-x0 + 6*x1 + x2)/6;
  }

  int SVGPointer::bezier2(int x1, int x2, int x3) {
    return (x1 + 6*x2 - x3)/6;
  }

  void SVGPointer::calculateBezier(){
    int last = points.size()-1;
    if (points.size()>=4) {
      Point p0 = points[last-3];
      Point p1 = points[last-2];
      Point p2 = points[last-1];
      Point p3 = points[last];
      Point b1(bezier1(p0.getX(),p1.getX(),p2.getX()), bezier1(p0.getY(),p1.getY(),p2.getY()));
      Point b2(bezier2(p1.getX(),p2.getX(),p3.getX()), bezier2(p1.getY(),p2.getY(),p3.getY()));
      beziers.push_back(b1);
      beziers.push_back(b2);
    }
  }

  string SVGPointer::getAnimatedSVGString() {
    stringstream svgstream(stringstream::in | stringstream::out);
    float start = startpoint.getTime();
    float end   = points[points.size()-1].getTime() - 0.2;
    float dur   = end - start;
    
    svgstream << endl << "<!-- begin of a pointer movement -->" << endl;
    svgstream << "<path d=\"M 0 0 L 6 20 L 15 16 Z\" style=\"visibility: hidden;\">" << endl;
    svgstream << "\t<set attributeName=\"visibility\" attributeType=\"CSS\"" << endl;
    svgstream << "\tbegin=\"" << start << "s\" dur=\"0.1s\" to=\"visible\" fill=\"freeze\"/>" << endl;
    svgstream << "\t<animateMotion path=\"" << endl;
    svgstream << "\tM " << startpoint.getX() << "," << startpoint.getY() << endl;

    for (unsigned int i=2; i < points.size()-1; i++) {
      svgstream << "C" << beziers[2*i-4].getX() << "," << beziers[2*i-4].getY();
      svgstream << " " << beziers[2*i-3].getX() << "," << beziers[2*i-3].getY();
      svgstream << " " << points[i].getX()      << "," << points[i].getY() << endl;
    }
    svgstream << "\" begin=\"" << start << "s\" dur=\"" << dur << "s\" fill=\"freeze\" />" << endl;
    svgstream << "\t<set attributeName=\"visibility\" attributeType=\"CSS\"" << endl;
    svgstream << "\tbegin=\"" << end+0.2 << "s\" dur=\"0.1s\" to=\"hidden\" fill=\"freeze\"/>" << endl;
    svgstream << "</path>" << endl << endl;
    return svgstream.str();
  }

  string SVGPointer::getSVGString() const{
    GLOBAL_DEBUG_STREAM("Dummy-method without any function. Method for getting svgstring in other SVGObjects" << endl);
    return string("");
  }

  void SVGPointer::setPathNumber(int){
    GLOBAL_DEBUG_STREAM("Dummy-method without any function. Method for setting the pathnumber in SVGFreehand" << endl);
  }

  SVGEncodeNode::SVGEncodeNode(string fn, const char* name) : GenericSyncSourceNode(name)  {

    // this node shall not be the master for source synchronization
    setSyncMaster(false);
    // this node does not produce at a constant rate
    setConstantRateProducer(false);
    isMouseButtonPressed = false;
    filename = fn;
    pix_count = 0;
    pixel_taken = 5;

    /* setting blue color */
    red = 0; green = 0; blue = 255;
    stroke_width = 2;
    width = 0;
    height = 0;
    path_counter = 0;
    actualObject = 0;
    timer = new DiffTimer();
    object_iter = objects.begin();
    svgstream = 0;
    svginput = 0;
    withFile = true;
    addedObject = false;

    streamTimer = new StreamTimer(this, StreamTimer::REAL_TIME);

    DEBUG_STREAM("End > Create SVGEncodeNode" << endl);
  }
  
  SVGEncodeNode::SVGEncodeNode(const char* name) : GenericSyncSourceNode(name)  {

    // this node shall not be the master for source synchronization
    setSyncMaster(false);
    // this node does not produce at a constant rate
    setConstantRateProducer(false);
    
    isMouseButtonPressed = false;
    filename = string("");
    pix_count = 0;
    pixel_taken = 5;

    /* setting blue color */
    red = 0; green = 0; blue = 255;
    stroke_width = 2;
    width = 0;
    height = 0;
    path_counter = 0;
    actualObject = 0;
    timer = new DiffTimer();
    object_iter = objects.begin();
    svgstream = 0;
    svginput = 0;
    withFile = true;
    addedObject = false;

    streamTimer = new StreamTimer(this, StreamTimer::REAL_TIME);

    DEBUG_STREAM("End > Create SVGEncodeNode" << endl);
  }
  
  SVGEncodeNode::~SVGEncodeNode() {
    svgstream = 0;
    if (timer) delete timer;
    if (streamTimer) delete streamTimer;
  }

  Result SVGEncodeNode::doInit() {
    return SUCCESS;
  };

  Result SVGEncodeNode::doInitOutput() {
    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("text/svg");
    
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    if (withFile)
      newFile();
 
    // reset the stream timer
    streamTimer->reset();

    return SUCCESS;
  };

  // virtual functions inherited from GenericSyncSourceNode
  Time SVGEncodeNode::getLocalTime(u_int64_t)
  {
    // as SVGEncodeNode is not a constant rate producer, getLocalTime should not be called
    // and does not have to deliver a useful value.
    return Time(0,0);
  }

  Message* SVGEncodeNode::produceBuffer(Buffer* ) {
    Buffer* outBuf = 0;
    SVGObject* actual = 0;
    string svgstring = string("");
    int buffer_size = 0;
    vector<SVGObject*>::iterator position;  
    vector<SVGObject*>::iterator end_pos;
    end_pos = object_iter;
    for (position = objects.begin(); position != end_pos; position++) {
      mutex.lock();
      actual = *position;
      svgstring += string(actual->getSVGString());
      mutex.unlock();
    }
    mutex.lock();
    
    if (actualObject){
      svgstring += actualObject->getSVGString();
    };
    
    mutex.unlock();
    buffer_size = svgstring.size();
    if ((outBuf = getNewBuffer( buffer_size )) != 0) {
      memcpy(outBuf->getData(), svgstring.c_str(), buffer_size );
      outBuf->setUsed(buffer_size);
      // Set the correct time stamp for the buffer
      streamTimer->setTimestamp(outBuf); 
      if(!addedObject) {
        setProducingFlag(false);
      } else {
        addedObject = false;
      };
      return(outBuf);
    }
    
    return 0;
  };
  
  Result SVGEncodeNode::doDeactivate() {
    if (withFile)
      writeFile();
    return SUCCESS;
  };

  void SVGEncodeNode::setColor(const int& r, const int& g, const int& b) {
    red = r;
    green = g;
    blue = b;
  }

  void SVGEncodeNode::setDimension(const int& width_, const int& height_) {
    width = width_;
    height = height_;
  }

  void SVGEncodeNode::newFile() {
    filename = string("") + timer->getTime().getPrettyString();
    string complete_filename = filename + ".svg";
    svgstream = new ofstream(complete_filename.c_str(), ios::trunc | ios::out);
    
    if (svgstream->good()) { 
      *svgstream << "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>" << endl;
      *svgstream << "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20001102//EN\"" <<endl;
      *svgstream << "\"http://www.w3.org/TR/2000/CR-SVG-20001102/DTD/svg-20001102.dtd\">" << endl << endl;
      *svgstream << "<svg xml:space=\"preserve\" width=\"" << width <<"\" height=\"" << height << "\" ";
      *svgstream << "onload=\"init(evt)\">" << endl << endl;
      *svgstream << "<script type=\"text/javascript\"><![CDATA[" << endl;
      *svgstream << "function init(evt) {" << endl;
      *svgstream << "  svgdoc=evt.getTarget().getOwnerDocument();" << endl;
      *svgstream << "  for(var i=1; i<" << path_counter+1 << "; i++) {" << endl;
      *svgstream << "    setAnimations(i);" << endl;
      *svgstream << "  }" << endl;
      *svgstream << "}" << endl << endl;
      *svgstream << "function setAnimations(number) {" << endl;
      *svgstream << "  path=svgdoc.getElementById('path'+number);" << endl;
      *svgstream << "  pathAnim=svgdoc.getElementById('pathAnim'+number);" << endl;
      *svgstream << "  trackLength=Math.round(path.getTotalLength()+2);" << endl;
      *svgstream << "  path.getStyle().setProperty('stroke-dasharray',trackLength.toString()+\" \"+trackLength.toString());" << endl;
      *svgstream << "  path.getStyle().setProperty('stroke-dashoffset',trackLength.toString());" << endl;
      *svgstream << "  pathAnim.setAttribute('from',trackLength.toString());" << endl;
      *svgstream << "  pathAnim.setAttribute('values',trackLength.toString()+';0');" << endl;
      *svgstream << "}" << endl;
      *svgstream << "]]>" << endl << " </script>" << endl;
    } else {
      ERROR_STREAM("An error occured while opening the file" << endl);
    };
    svgstream->flush();
  }

  void SVGEncodeNode::writeFile() {
    if (!objects.empty()) {
      vector<SVGObject*>::iterator get_object;
      get_object = objects.begin();
      while (get_object != object_iter) {
	SVGObject* actual = *get_object;
	if (actual){
	  *svgstream << actual->getAnimatedSVGString();
	  delete actual;
	  actual = 0;
	  get_object++;
	}
      }
      objects.clear();
      object_iter = objects.begin();
      timer->reset();
      path_counter = 0;
            
      svgstream->flush();
      closeFile(); 
    } else {
      DEBUG_STREAM("No objects found. No objects in file created!" << endl);
    }
  }

  void SVGEncodeNode::closeFile() {
    *svgstream << "</svg>" << endl;   // This is the closing tag in a svg file
    svgstream->close();
    delete svgstream;
    svgstream = 0;
  }

  int SVGEncodeNode::getPixelTaken(){
    return pixel_taken;
  }
  
  void SVGEncodeNode::setPixelTaken(const int& pt){
    pixel_taken = pt;
  }

  void SVGEncodeNode::setWithFile(const bool& withFile_) {
    withFile = withFile_;
  }

  void SVGEncodeNode::setStrokeWidth(const int& stroke) {
    stroke_width = stroke;
  }

  void SVGEncodeNode::drawSVGLine(const int& x, const int& y, const int& buttonmask) {
    Point actualPoint(x, y, timer);

    // if mousebutton wasn't pressed and now is pressed 
    if (!isMouseButtonPressed && buttonmask != 0) {       
      isMouseButtonPressed = true;
      actualObject = new SVGLine(actualPoint, timer, red, green, blue, stroke_width);
    } 

    // if mousebutton was pressed and now is already pressed
    else if (isMouseButtonPressed && buttonmask != 0) { 
      mutex.lock();
      actualObject->addPoint(actualPoint);
      mutex.unlock();
    } 

    // if mousebutton was pressed and now is released
    else if (isMouseButtonPressed && !buttonmask) {     
      isMouseButtonPressed = false;
      mutex.lock();
      actualObject->finish(actualPoint);
      mutex.unlock();
      if (object_iter != objects.end()) {
	deleteObjects(object_iter, objects.end());
	objects.erase(object_iter, objects.end());
      }
      if (actualObject) {
	objects.push_back(actualObject);
	object_iter = objects.end();
      }
      actualObject = 0;
    }

    if(!getProducingFlag()){
      setProducingFlag(true);
    }else{
      addedObject = true;
    }
  }

  void SVGEncodeNode::drawSVGRectangle(const int& x, const int& y, const int& buttonmask) {
    Point actualPoint(x, y, timer);
    
    // if mousebutton wasn't pressed and now is pressed
    if (!isMouseButtonPressed && buttonmask != 0) {       
      isMouseButtonPressed = true;
      actualObject = new SVGRectangle(actualPoint, timer, red, green, blue, stroke_width);
    } 

    // if mousebutton was pressed and now is already pressed
    else if (isMouseButtonPressed && buttonmask != 0) {
      mutex.lock();
      actualObject->addPoint(actualPoint);
      mutex.unlock();
    } 

    // if mousebutton was pressed and now is released
    else if (isMouseButtonPressed && !buttonmask) {     
      isMouseButtonPressed = false;
      mutex.lock();
      actualObject->finish(actualPoint);
      mutex.unlock();
      if (object_iter != objects.end()) {
	deleteObjects(object_iter, objects.end());
	objects.erase(object_iter, objects.end());
      }
      if (actualObject) {
	objects.push_back(actualObject);
	object_iter = objects.end();
      }
      actualObject = 0;
    }

    if(!getProducingFlag()){
      setProducingFlag(true);
    }else{
      addedObject = true;
    }
  }

  void SVGEncodeNode::drawSVGFreehand(const int& x, const int& y, const int& buttonmask) {
    Point actualPoint(x, y, timer);

    // if mousebutton wasn't pressed and now is pressed
    if (!isMouseButtonPressed && buttonmask != 0) {
      isMouseButtonPressed = true;
      actualObject = new SVGFreehand(actualPoint, timer, red, green, blue, stroke_width);
      pix_count++;
    } 

    // while mousebutton is pressed
    else if (isMouseButtonPressed && buttonmask != 0) {
      // not every pixel is taken
      if ( (pix_count % pixel_taken) == 0 ) {             
	mutex.lock();
	actualObject->addPoint(actualPoint);
	mutex.unlock();
      }
      pix_count++;
    } 

    // if mousebutton was pressed and now is released
    else if (isMouseButtonPressed && !buttonmask) {     
      isMouseButtonPressed = false;
      mutex.lock();
      actualObject->finish(actualPoint);
      mutex.unlock();
      if (object_iter != objects.end()) {
	deleteObjects(object_iter, objects.end());
	objects.erase(object_iter, objects.end());
      }
      if (actualObject) {
	path_counter++;
	if (withFile) {
	  actualObject->setPathNumber(path_counter);
	}
	objects.push_back(actualObject);
	object_iter = objects.end();
      }
      actualObject = 0;
      pix_count = 0;
    }

    if(!getProducingFlag()){
      setProducingFlag(true);
    }else{
      addedObject = true;
    }
  }

  void SVGEncodeNode::drawSVGPointer(const int& x, const int& y, const int& buttonmask) {
    Point actualPoint(x, y, timer);
    
    // if mousebutton wasn't pressed and now is pressed
    if (!isMouseButtonPressed && buttonmask != 0) {       
      isMouseButtonPressed = true;
      actualObject = new SVGPointer(actualPoint, timer);
      pix_count++;
    } 

    // while mousebutton is pressed
    else if (isMouseButtonPressed && buttonmask != 0) { 
      // not every pixel is taken
      if ( (pix_count % pixel_taken) == 0 ) {             
	mutex.lock();
	actualObject->addPoint(actualPoint);
	mutex.unlock();
      }
      pix_count++;
    } 

    // if mousebutton was pressed and now is released
    else if (isMouseButtonPressed && !buttonmask) {     
      isMouseButtonPressed = false;
      mutex.lock();
      actualObject->finish(actualPoint);
      mutex.unlock();
      if (actualObject) {
	objects.push_back(dynamic_cast<SVGPointer*>(actualObject));
	object_iter = objects.end();
      }
      actualObject = 0;
      pix_count = 0;
    }

    //setProducingFlag(true);
  }

  void SVGEncodeNode::newAnnotation(){
    if (!objects.empty() && withFile) {
      writeFile();
      newFile();
    }
    setProducingFlag(true);
  }
  
  void SVGEncodeNode::clearAnnotation(){
    clearHistory();
    setProducingFlag(true);
    if (withFile)
      newFile();
  }

  void SVGEncodeNode::undo(){
    if (!objects.empty() && object_iter != objects.begin()) {
      object_iter--;
      
      // Check wether the object is an SVGFreehand-Object
      SVGObject* pathtest = *object_iter;
      string svgstring = pathtest->getSVGString();
      int found = svgstring.find(string("<path"));
      
      if ( found >= 0 ) {
	path_counter--;
      }
    } else {
      WARNING_STREAM("There are no objects to delete. History is empty." << endl);
    }
    
    setProducingFlag(true);
  }

  void SVGEncodeNode::redo() {
    if (!(object_iter == objects.end())) {
      
      // Check wether the object is an SVGFreehand-Object
      SVGObject* pathtest = *object_iter;
      string svgstring = pathtest->getSVGString();
      int found = svgstring.find(string("<path"));
      
      if ( found >= 0 ) {
	path_counter++;
      }

      object_iter++;
    } else {
      WARNING_STREAM("There are no more Objects to restore. End of history is reached." << endl);
    }

    setProducingFlag(true);
  }
  
  void SVGEncodeNode::repaint() {
    setProducingFlag(true);
  }

  char* SVGEncodeNode::readFile(const char* filename){
    svginput = new ifstream(filename, ios::in);
    if (svginput->good()){
      svginput->seekg (0, ios::end);
      int length = svginput->tellg();
      svginput->seekg (0, ios::beg);
      char *buffer = new char [length];
      
      // read data as a block:
      svginput->read (buffer,length);
      
      return buffer;
    } else {
      ERROR_STREAM("Could not open desired SVGFile \"" << filename <<"\"" << endl);
      return 0;
    };
  }
  
  void SVGEncodeNode::clearHistory() {
    // deleting all objects in vector 
    for (unsigned int i=0; i < objects.size(); i++) {
      SVGObject* actual = objects[i];
      if (actual){
	delete actual;
	actual = 0;
      }
    }
    objects.clear();
    object_iter = objects.begin();
    path_counter = 0;

    // deleting all pointer in vector
    /*for (unsigned int i=0; i < pointer.size(); i++){
      SVGPointer* actualPointer = pointer[i];
      if (actualPointer) {
	delete actualPointer;
	actualPointer = 0;
      }
    }
    pointer.clear();
    */
  }

  void SVGEncodeNode::deleteObjects(vector<SVGObject*>::iterator start, vector<SVGObject*>::iterator end) {
    // deleting all unnecessary SVGObjects in memory
    while (start!= end) {
      delete *start;
      start++;
    }
  }

}
