/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "OSDNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <stdio.h>
#include <iostream>
#include <assert.h>

#ifndef NULL
#define NULL 0
#endif

NMM_REGISTER_PLUGIN(NMM::OSDNode, "OSDNode")

namespace NMM {
  OSDNode::OSDNode(const char *name /* = "OSDNode" */,
		   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
		   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size) {

    osd_font=NULL;
    osd_progbar_type=-1;
    osd_progbar_value=100;   // 0..255
    dxs = 0;
    dys = 0;
    
    osd_cond = new ThreadCondition(osd_mutex);	
    new_text_flag = false;
  }

  OSDNode::~OSDNode() { 
    delete osd_cond;
  }

  void OSDNode::setupOutputFormats(Property* property) {
    NMM::Format *o_format;
    o_format=property->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addStringParamValue(Format::colorspace_param,"rgba32");
    o_format->addStringParamValue(Format::channelorder_param,"bgra");
    o_format->addIntParamValue(Format::bitperpixel_param,32);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::format_param,"packet");
    
    o_format=property->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addStringParamValue(Format::colorspace_param,"yva12");
    o_format->addStringParamValue(Format::channelorder_param,"yvua");
    o_format->addIntParamValue(Format::bitperpixel_param,20);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::format_param,"planar");
    getOwnOutputProperty()->setDefaultFormat(o_format);
  }
  
  Result OSDNode::doInit() {
    setupOutputFormats(getOwnOutputProperty());
    return SUCCESS;
  }

  Result OSDNode::doInitOutput() {
    setupOutputFormats(getOwnWorkingOutputProperty());
    return SUCCESS;
  }
  
  Result OSDNode::doActivate() {
    int frame_size;
    
    const Format* o_format = getOutputFormat();
    if (o_format == 0)
      {
	ERROR_STREAM( "Invalid output format (NULL-ptr)" << endl);
	return FAILURE;
      }

    string colorspace = o_format->getStringValue(Format::colorspace_param);
    image_width  = o_format->getIntValue(Format::x_resolution_param);
    image_height = o_format->getIntValue(Format::y_resolution_param);

    if(colorspace == "rgba32") {
      color = RGBA32;
      buff_size = image_width*image_height*4;
    }
     
    if(colorspace == "yva12") {
      frame_size = image_width * image_height;
      buff_size = 2*frame_size + (frame_size/4)*2; // 4Y + 1U + 1V + Alpha
      color = YVA12;
    }

    if ((image_width <= 0) || (image_height <= 0)) {
      return FAILURE;     
    }
    
    return SUCCESS;
  }
  
  Buffer *OSDNode::processBuffer(Buffer *) {
    MutexGuard m(osd_mutex);

    Buffer *out = 0;
    
    while ( new_text_flag == false ) {
      DEBUG_STREAM("Waiting..." << endl);
      osd_cond->wait();
      DEBUG_STREAM("awake" << endl );
    }
    
    if ((out = getNewBuffer( buff_size )) != 0) {
      out->setUsed( buff_size ); 
      memset(out->getData(), 0, buff_size);
               
      curr_buffer = out; 
      draw_text();
      new_text_flag = false;
      return out;
    }
    else {
      new_text_flag = false;
      return 0;
    }
  }

  Result OSDNode::setText( const string& text ) {
    osd_mutex.lock();
    osd_text = text;
    new_text_flag = true;
    osd_cond->notify();
    osd_mutex.unlock();
    
    return SUCCESS;
  }
  
  Result OSDNode::setFont(const string& font_name, const float& font_factor ) {
    osd_font = read_font_desc( font_name.c_str(), font_factor, 0 );
    if( !osd_font ) return FAILURE;
    
    return SUCCESS;
  }
  
  void OSDNode::draw_alpha_rgb32(int w,int h, unsigned char* src, unsigned char *srca, int srcstride, unsigned char* dstbase,int dststride){
  
    int y;
    for(y=0;y<h;y++){
      register int x;
      for(x=0;x<w;x++){
	if(srca[x]){
	  dstbase[4*x+0]=src[x];
	  dstbase[4*x+1]=src[x];
	  dstbase[4*x+2]=src[x];
	  dstbase[4*x+3]=255-srca[x];
	}
      }
      src+=srcstride;
      srca+=srcstride;
      dstbase+=dststride;
    }
    return;
  }
  
  void OSDNode::draw_alpha_yv12(int w,int h, unsigned char* src, unsigned char *srca, int srcstride, unsigned char* dstbase,int dststride){
  
    int y;
    unsigned char *alpha;
    int alpha_offset;
    
    alpha_offset = image_width*image_height;
    alpha_offset = alpha_offset + (alpha_offset/4)*2; 
    alpha = dstbase + alpha_offset;
    
    for(y=0;y<h;y++){
      register int x;
      for(x=0;x<w;x++){
	if(srca[x]) { 
	  dstbase[x]= src[x];
	  alpha[x] = 255-srca[x];
	}
      }
      src+=srcstride;
      srca+=srcstride;
      dstbase+=dststride;
      alpha+=dststride;
    }
    return;
  }

  
  
  void OSDNode::draw_alpha(int x0,int y0, int w,int h, unsigned char* src, unsigned char *srca, int stride) { 
    switch( color ) {
    case RGBA32:
      draw_alpha_rgb32(w,h,src,srca,stride,(unsigned char *) curr_buffer->getData()+image_width*4*y0+(x0*4),image_width*4);
      break;
      
    case YVA12:
      draw_alpha_yv12(w,h,src,srca,stride,(unsigned char *) curr_buffer->getData()+image_width*y0+x0,image_width);
      break;
    }
  }
  
  void OSDNode::draw_text_osd(){
    int len=osd_text.size();
    int j;
    int x;
    int y;
    
    x = dxs;
    y = dys;
   
    for(j=0;j<len;j++){
      int c=osd_text[j];
      int font=osd_font->font[c];
      if(font>=0)
        draw_alpha(x,y,
		   osd_font->width[c],
		   osd_font->pic_a[font]->h,
		   osd_font->pic_b[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->w);
      x+=osd_font->width[c]+osd_font->charspace;
    }
  }

  void OSDNode::draw_text_progbar(){
    int i;
    int y=dys/2;
    int x;
    int c,font;
    int width=(dxs*2/3-osd_font->width[0x10]-osd_font->width[0x12]);
    int elems=width/osd_font->width[0x11];
    int mark=(osd_progbar_value*(elems+1))>>8;
    x=(dxs-width)/2;

    c=osd_progbar_type;font=osd_font->font[c];
    if(osd_progbar_type>0 && font>=0)
      draw_alpha(x-osd_font->width[c]-osd_font->spacewidth,y,
		 osd_font->width[c],
		 osd_font->pic_a[font]->h,
		 osd_font->pic_b[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->w);

    c=OSD_PB_START;font=osd_font->font[c];
    if(font>=0)
      draw_alpha(x,y,
		 osd_font->width[c],
		 osd_font->pic_a[font]->h,
		 osd_font->pic_b[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->w);
    x+=osd_font->width[c];

    for(i=0;i<elems;i++){
      c=(i<mark)?OSD_PB_0:OSD_PB_1;font=osd_font->font[c];
      if(font>=0)
	draw_alpha(x,y,
		   osd_font->width[c],
		   osd_font->pic_a[font]->h,
		   osd_font->pic_b[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->w);
      x+=osd_font->width[c];
    }

    c=OSD_PB_END;font=osd_font->font[c];
    if(font>=0)
      draw_alpha(x,y,
		 osd_font->width[c],
		 osd_font->pic_a[font]->h,
		 osd_font->pic_b[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->w);
  }

  void OSDNode::draw_text(){
    if(!osd_font) return; // no font

    if(!osd_text.empty()){
      draw_text_osd();
    }
    
    if(osd_progbar_type >= 0) {
      draw_text_progbar();
    }
    
  }
  
}
