/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "OSDOverlayNode.hpp"
#include <signal.h>
#include <sys/time.h>
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#ifndef NULL
#define NULL 0
#endif

NMM_REGISTER_PLUGIN(NMM::OSDOverlayNode, "OSDOverlayNode")

namespace NMM {

  //****************************** OSDtem ********************************************

  OSDOverlayNode::OSDItem::OSDItem() 
  { 
    depth = width = height = x_offset = y_offset = 0; 
    enabled = true;
  }

  //****************************** OSDText ********************************************
  OSDOverlayNode::OSDText::OSDText() 
  {
    current_data = 0;
    image_width = 0;
    image_height = 0;
    color = 0;

    osd_font= 0;
    osd_text= new char[256];
    osd_progbar_type= -1;
    osd_progbar_value= 0;   // 0..255
    dxs = 0;
    dys = 0;
  }

  OSDOverlayNode::OSDText::~OSDText() 
  {
    delete[] osd_text;
  }

  Result OSDOverlayNode::OSDText::setFont(font_desc_t* osd_font)
  {
    this->osd_font = osd_font;
    return SUCCESS;
  }

  Result OSDOverlayNode::OSDText::setText(const string text )
  {    
    const char *s = text.c_str();
    memcpy(osd_text, s, strlen(s) + 1);
    osd_progbar_type= -1;
    return SUCCESS;
  }

  Result OSDOverlayNode::OSDText::setProgressBarValue(const int value ) 
  {    
    osd_progbar_type= 1;
    osd_progbar_value= value;   

    return SUCCESS;
  }

  Result OSDOverlayNode::OSDText::draw(char *data, int color, size_t width, size_t height)
  {
    current_data = data;
    dxs = x_offset;
    dys = y_offset;
    image_width = width;
    image_height = height;
    this->color = color;

    if(!osd_font) return FAILURE; // no font

    if(osd_text){

      if(osd_progbar_type >= 0) {
	draw_text_progbar();
      }
      else {
	GLOBAL_DEBUG_STREAM("DRAW TEXT " << osd_text << " " << color << " " << image_width << " " << image_height << endl);
	draw_text_osd();
      }
    }

    return SUCCESS;
  }

  void OSDOverlayNode::OSDText::draw_alpha(int x0,int y0, int w,int h, unsigned char* src, unsigned char *srca, int stride)
  {
    int NO_BLENDING = 0;
    int REPLACE_BLENDING = 1;
    int x, y, a;
    
    unsigned char *dst_y = (unsigned char *) current_data + image_width*y0 + x0;
    unsigned char *dst_u = (unsigned char *) current_data + image_width*image_height + (image_width/2)*(y0/2) + (x0/2);
    unsigned char *dst_v = (unsigned char *) dst_u + (image_width*image_height)/4;
    
    int dst_y_stride = image_width;
    int dst_u_stride = image_width/2;
    int dst_v_stride = dst_u_stride;
    int xdiv2 = 0;

    for(y = 0; y < h; y++) {
      for(x = 0; x < w; x++) {
	// here, srca[x] means alpha and alpha == 0 means no foreground
	// and alpha == 1 means just take foreground ( = the letters)
	// and alpha > 1 means perform alpha blending with background
	if(srca[x] != NO_BLENDING) {
	  if(srca[x] == REPLACE_BLENDING) {
	    dst_y[x] = src[x];
	    if((x % 2) == 0) {
	      xdiv2 = x/2;
	      dst_u[xdiv2] = 128;
	      dst_v[xdiv2] = 128;
	    }
	  }
	  else {
	    // alpha of foreground 
	    a = 255 - (int) srca[x];
	    // take the foreground with its alpha and the background with (256 - foreground alpha) which is srca[x]
	    dst_y[x] = (unsigned char) ((a * (int) src[x] + (int) srca[x] * (int) dst_y[x]) >> 8); 
	    if((x % 2) == 0) {
	      xdiv2 = x/2;
	      dst_u[xdiv2] = (unsigned char) ((a * (int) 128 + (int) srca[x] * (int) dst_u[xdiv2]) >> 8); 
	      dst_v[xdiv2] = (unsigned char) ((a * (int) 128 + (int) srca[x] * (int) dst_v[xdiv2]) >> 8); 
	    }
	  }
	}
      }
      src += stride;
      srca += stride;
      dst_y += dst_y_stride;
      if((y % 2) == 1) {
	dst_u += dst_u_stride;
	dst_v += dst_v_stride;
      }
    }
  }
  
  void OSDOverlayNode::OSDText::draw_text_osd()
  {
    int len=strlen(osd_text);
    int j;
    int x;
    int y;

    x = dxs;
    y = dys;

    for(j=0;j<len;j++){
      int c=osd_text[j];
      int font=osd_font->font[c];
      if(font>=0) {
        draw_alpha(x,y,
		   osd_font->width[c],
		   osd_font->pic_a[font]->h,
		   osd_font->pic_b[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->w);
      }
      x+=osd_font->width[c]+osd_font->charspace;
    }
  }

  void OSDOverlayNode::OSDText::draw_text_progbar()
  {
    int i;
    int y=dys;
    int x;
    int c,font;
    int width=(dxs*2/3-osd_font->width[0x10]-osd_font->width[0x12]);
    int elems=width/osd_font->width[0x11];
    int mark=(osd_progbar_value*(elems+1))>>8;
    x=(dxs-width)/2;

    c=OSD_PB_START;font=osd_font->font[c];
    if(font>=0)
      draw_alpha(x,y,
		 osd_font->width[c],
		 osd_font->pic_a[font]->h,
		 osd_font->pic_b[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->w);
    x+=osd_font->width[c];

    for(i=0;i<elems;i++){
      c=(i<mark)?OSD_PB_0:OSD_PB_1;font=osd_font->font[c];
      if(font>=0)
	draw_alpha(x,y,
		   osd_font->width[c],
		   osd_font->pic_a[font]->h,
		   osd_font->pic_b[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->bmp+osd_font->start[c],
		   osd_font->pic_a[font]->w);
      x+=osd_font->width[c];
    }

    c=OSD_PB_END;font=osd_font->font[c];
    if(font>=0)
      draw_alpha(x,y,
		 osd_font->width[c],
		 osd_font->pic_a[font]->h,
		 osd_font->pic_b[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->bmp+osd_font->start[c],
		 osd_font->pic_a[font]->w);
  }

  //****************************** OSDImagePNG ****************************************

  OSDOverlayNode::OSDImagePNG::OSDImagePNG()
  {
    png_rows = 0;
    png_width = 0;
    png_height = 0;
    yuv_buffer = 0;

    image_width = 0; 
    image_height = 0;
    color = 0;
    yuv_size = 0;
  }

  OSDOverlayNode::OSDImagePNG::~OSDImagePNG()
  {
    if(yuv_buffer) delete yuv_buffer;
  }

  Result OSDOverlayNode::OSDImagePNG::draw(char *data, int color, size_t width, size_t height)
  {
    image_width = width;
    image_height = height;
    this->color = color;
    GLOBAL_DEBUG_STREAM( "DRAW image (widtch, height): (" << width << ", " <<
			 height << ")" << endl);
    paintPNGImageAlphaYVA12(data, x_offset, y_offset);

    return SUCCESS;
  }

  Result OSDOverlayNode::OSDImagePNG::load(const string url) {
    return preparePNGImage(url.c_str());
  }

  /* ==== Internal methods follow ==== */

  void OSDOverlayNode::OSDImagePNG::rgb2yuv( unsigned char R, unsigned char G, unsigned char B,
					  double *Y, double *U, double *V ) {
    *Y = R *  0.299 + G *  0.587 + B *  0.114;
    *U = R * -0.169 + G * -0.332 + B *  0.500 + 128;
    *V = R *  0.500 + G * -0.419 + B * -0.0813 + 128;
  }

  void OSDOverlayNode::OSDImagePNG::convertPNGImageYUV() {

    int plane_size = png_width * png_height; 

    // 1 y + 1/4 u + 1/4 v + 1 alpha
    yuv_size   = plane_size + plane_size/2 + plane_size;
    yuv_buffer = new unsigned char[yuv_size];
    
    unsigned char* y_plane = yuv_buffer;
    unsigned char* u_plane = yuv_buffer + plane_size;
    unsigned char* v_plane = u_plane + plane_size/4;
    
    //GLOBAL_DEBUG_STREAM("MSATEST plane_size: " << plane_size << endl);
    //GLOBAL_DEBUG_STREAM("MSATEST image_size: " << png_width << "x" << png_height << endl);

    unsigned char *rgba1, *yuv1;
    unsigned char *rgba2, *yuv2;
    unsigned char *rgba3, *yuv3;
    unsigned char *rgba4, *yuv4;
    double y1,u1,v1;
    unsigned char r1, r2, r3, r4, g1, g2, g3, g4, b1, b2, b3, b4;

    for (unsigned row=0; row < png_height/2; row++) {
      rgba1 =(unsigned char*) png_rows[2*row];
      rgba2 = rgba1 + 4;
      rgba3 =(unsigned char*) png_rows[2*row + 1];
      rgba4 = rgba3 + 4;
      yuv1  = y_plane + 2*row * png_width;
      yuv2  = yuv1 + 1;
      yuv3  = y_plane + (2*row + 1) * png_width;
      yuv4  = yuv3 + 1;

      for (unsigned col=0; col < png_width/2; col++) {
	b1 = rgba1[0];
	g1 = rgba1[1];
	r1 = rgba1[2];
	rgb2yuv( r1, g1, b1, &y1, &u1, &v1 );
	*yuv1 = (unsigned char) y1;

	b2 = rgba2[0];
	g2 = rgba2[1];
	r2 = rgba2[2];
	rgb2yuv( r2, g2, b2, &y1, &u1, &v1 );
	*yuv2 = (unsigned char) y1;

	b3 = rgba3[0];
	g3 = rgba3[1];
	r3 = rgba3[2];
	rgb2yuv( r3, g3, b3, &y1, &u1, &v1 );
	*yuv3 = (unsigned char) y1;

	b4 = rgba4[0];
	g4 = rgba4[1];
	r4 = rgba4[2];
	rgb2yuv( r4, g4, b4, &y1, &u1, &v1 );
	*yuv4 = (unsigned char) y1;

	r1 = (r1+r2+r3+r4)/4;
	g1 = (g1+g2+g3+g4)/4;
	b1 = (b1+b2+b3+b4)/4;
	rgb2yuv( r1, g1, b1, &y1, &u1, &v1 );
	*u_plane = (unsigned char) u1;
	*v_plane = (unsigned char) v1;
	u_plane++; v_plane++;

	rgba1 += 8;
	rgba2 = rgba1 + 4;
	rgba3 += 8;
	rgba4 = rgba3 + 4;
	yuv1  += 2;
	yuv2  = yuv1 + 1;
	yuv3  += 2;
	yuv4  = yuv3 + 1;
      }
    }

    
    // copy alpha from rgb image to end of yuv_buffer
    unsigned char *yuv_alpha = yuv_buffer + plane_size + plane_size/2;
    unsigned char *alpha = 0;
    for (unsigned row=0; row < png_height; row++) {
      alpha =(unsigned char*) png_rows[row];
      for (unsigned col=0; col < png_width; col++) {
	*yuv_alpha = alpha[3];
	alpha += 4;
	yuv_alpha++;
      }
    }    
  }

  Result OSDOverlayNode::OSDImagePNG::preparePNGImage(const char *filename) {

    // SEE THE LIBPNG-EXAMPLE.c FOR DETAILS!

    png_structp png_ptr;
    png_infop info_ptr;
    //unsigned int sig_read = 0;
    //png_uint_32 width, height;
    int bit_depth, color_type, interlace_type;
    FILE *fp;
    
    png_rows = NULL;
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    if ((fp = fopen(filename, "rb")) == NULL) {
      GLOBAL_ERROR_STREAM("Could not open file " << filename << endl);
      return FAILURE;
    }
    
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
				     NULL, NULL, NULL);
    
    if (png_ptr == NULL) {
      fclose(fp);
      return FAILURE;
    }
    
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
      fclose(fp);
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      return FAILURE;
    }

    if (setjmp(png_ptr->jmpbuf)) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);
      return FAILURE;
    }

    png_init_io(png_ptr, fp);

    png_read_info(png_ptr, info_ptr);
    
    png_get_IHDR(png_ptr, info_ptr, &png_width, &png_height,
		 &bit_depth, &color_type,
		 &interlace_type, NULL, NULL);

    
    if (bit_depth == 8 && color_type == PNG_COLOR_TYPE_RGB) {
      png_set_filler(png_ptr, 0x00, PNG_FILLER_AFTER);
      png_set_bgr(png_ptr);
      png_read_update_info(png_ptr, info_ptr);
    }

    png_rows = new png_bytep[png_height];

    for (unsigned int row = 0; row < png_height; row++) {
      png_rows[row] = new png_byte[png_get_rowbytes(png_ptr, info_ptr)];
    }
    
    png_read_image(png_ptr, png_rows);
    
    png_read_end(png_ptr, info_ptr);
    
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    
    fclose(fp);
    
    // Finnally get the YUV-Copy of the PNG-Image done.
    convertPNGImageYUV();
    
    return SUCCESS;
  }
  
  void OSDOverlayNode::OSDImagePNG::paintPNGImageAlphaYVA12(char *current_data, int x0, int y0) {
    int w = png_width;
    int h = png_height;
    int plane_size = png_width * png_height; 
    unsigned char *src_y = yuv_buffer;
    unsigned char *src_u = yuv_buffer + plane_size;
    unsigned char *src_v = src_u + plane_size/4;
    unsigned char *src_a = src_v + plane_size/4;
    int src_u_stride = png_width/2;
    int src_v_stride = src_u_stride;
    int src_y_stride = png_width;
    int src_a_stride = png_width;

    int NO_BLENDING = 0;
    int REPLACE_BLENDING = 255;
    int x, y;
    
    int dst_y_stride = image_width;
    int dst_u_stride = image_width/2;
    int dst_v_stride = dst_u_stride;
    unsigned char *dst_y = (unsigned char *) current_data + image_width*y0 + x0;
    unsigned char *dst_u = (unsigned char *) current_data + image_width*image_height + (image_width/2)*(y0/2) + (x0/2);
    unsigned char *dst_v =  dst_u + (image_width*image_height)/4;
    int xdiv2 = 0;

    for(y = 0; y < h; y++) {
      for(x = 0; x < w; x++) {
	// here, src_a[x] means alpha and alpha == 0 means no foreground
	// and alpha == 255 means just take foreground ( = the letters)
	// and alpha > 1 and alpha < 255 means perform alpha blending with background
	if(src_a[x] != NO_BLENDING) {
	  if(src_a[x] == REPLACE_BLENDING) {
	    dst_y[x] = src_y[x];
	    if((x % 2) == 0) {
	      xdiv2 = x/2;
	      dst_u[xdiv2] = src_u[xdiv2];
	      dst_v[xdiv2] = src_v[xdiv2];
	    }
	  }
	  else {
	    
	    // alpha of foreground 
	    // a = (int) src_a[x];
	    // take the foreground with its alpha and the background with (256 - foreground alpha) which is src_a[x]
	    dst_y[x] = (unsigned char) (( (int) src_a[x] * ( (int) src_y[x] - (int) dst_y[x] ) ) / 255 + (int) dst_y[x]);
	    
	    if((x % 2) == 0) {
	      xdiv2 = x/2;
	      dst_u[xdiv2] = 
	        (unsigned char) (( (int) src_a[x] * ( (int) src_u[xdiv2] - (int) dst_u[xdiv2] ) ) / 255 + (int) dst_u[xdiv2]);
	      dst_v[xdiv2] = 
	        (unsigned char) (( (int) src_a[x] * ( (int) src_v[xdiv2] - (int) dst_v[xdiv2] ) ) / 255 + (int) dst_v[xdiv2]);
	    } 
	    
	    
	  }
	}
      }
      src_y += src_y_stride;
      src_a += src_a_stride;
      dst_y += dst_y_stride;
      if((y % 2) == 1) {
	dst_u += dst_u_stride;
	dst_v += dst_v_stride;
	src_u += src_u_stride;
	src_v += src_v_stride;
      }
    }
  }
  
  unsigned char* OSDOverlayNode::OSDImagePNG::getYUVData() {
    return yuv_buffer;
  }

  int OSDOverlayNode::OSDImagePNG::getYUVSize() {
    return yuv_size;
  }
  
  // Tried to implement some assembler code... didn't work
  /*
  unsigned short alphaMMXmul_const1[4] = {256,256,256,256};
  unsigned alphaMMXmul_0[2] = {1,1};
  
  void MixAlphaMMX32(void *dest,const void *src,unsigned len,unsigned opacity) {
    
    __asm__(
	   "mov edi,dest
	    mov ebx,src
	    mov             ecx,len
	    mov             edx,opacity

	    movzx   eax,dl\n
	    movq    mm7,[alphaMMXmul_const1]
	    
	    shl     eax,16
	    add     eax,edx
	    mov     [alphaMMXmul_0],eax
	    mov     [alphaMMXmul_0 + 4],eax
	    movq    mm6,[alphaMMXmul_0]     ; mm6(X) = alpha (4 words)
	    pxor    mm5,mm5
	    psubusw mm7,mm6                 ; mm7(Y) = 256 ? alpha (4 words)
	    
	    ALIGN 16
	    MixAlphaMMX32_MainLoop:
	    
	    movd           mm0,[edi]       ; mm0(A) = 0 0 0 0 | 0 Ra Ga Ba
	    add            edi,4
	    movd           mm1,[ebx]       ; mm1(B) = 0 0 0 0 | 0 Rb Gb Bb
	    add            ebx,4
	    punpcklbw      mm0,mm5         ; mm0 = 0 0 0 Ra | 0 Ga 0 Ba
	    punpcklbw      mm1,mm5         ; mm1 = 0 0 0 Rb | 0 Gb 0 Bb
	    pmullw         mm0,mm6         ; mm0 = 0 Ra*X | Ga*X Ba*X
	    pmullw         mm1,mm7         ; mm1 = 0 Rb*Y | Gb*Y Bb*Y
	    paddusw        mm0,mm1         ; mm0 = 0 Ra*X+Rb*y | Ga*X+Gb*y Ba*X+Bb*Y
	    psrlw          mm0,8           ; mm0 = 0 0 0 Rc | 0 Gc 0 Bc
	    packuswb       mm0,mm0         ; mm0 = 0 0 0 0 | 0 Rc Gc Bc
	    movd           [edi-4],mm0
	    dec            ecx
	    jnz            MixAlphaMMX32_MainLoop
	    emms"
	    );
	    }*/


  //****************************** OSDOverlayNode ****************************************
  OSDOverlayNode* OSDOverlayNode::osd_node;

  OSDOverlayNode::OSDOverlayNode(const char *name,
		     const StreamQueue::Mode queue_mode,
		     const int queue_size )
    : GenericFilterNode(name, queue_mode, queue_size) {

    lastBuffer = 0;
    paused = false; // Would it be right to set paused to true,
    // and set paused to false in doStart()?

    font = 0;

    image_width = 0; 
    image_height = 0;
    color = 0;

    // setup a pointer to this node (needed to auto-clear osd)
    osd_node = this;

    // register Events
    //    registerEvent("item_enabled", new TEDObject2<OSDOverlayNode,string,bool>(this, &OSDOverlayNode::setItemEnabled));
    //registerEvent("playback_paused", new TEDObject1<OSDOverlayNode, bool>(this, &OSDOverlayNode::doPause));
    //registerEvent("resolution_changed", new TEDObject2<OSDOverlayNode, int,int>(this, &OSDOverlayNode::eventResolutionChanged));
    //registerEvent("update",  new TEDObject0<OSDOverlayNode>(this, &OSDOverlayNode::update));
  }

  OSDOverlayNode::~OSDOverlayNode() 
  {
    for(items_t::iterator i = items.begin(); i != items.end(); ++i)
      delete i->second;
  }

  Result OSDOverlayNode::doInit() {
    // initialize formats
    NMM::Format *myInputFormat;
    myInputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"yv12");
    myInputFormat->addStringParamValue(Format::channelorder_param,"yvu");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,12);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::format_param,"planar");

    NMM::Format *myOutputFormat;
    myOutputFormat=getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"yv12");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"yvu");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,12);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::format_param,"planar");

    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);

    return SUCCESS;
  }

  Result OSDOverlayNode::doInitOutput() {
    const Format *in_format = getInputFormat();
    if (in_format == NULL) {
      GLOBAL_ERROR_STREAM("Input isn't connected, so InitOutput failes - so sorry"<<endl);
      return FAILURE;
    }
    if (in_format->getStringValue(Format::colorspace_param)!="yv12") {
      GLOBAL_ERROR_STREAM("Wront Colorspace - so sorry"<<endl);
      return FAILURE;
    }

    // set working format
    // this is equal to input format for all filter nodes
    Format* in_f = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(in_f);
    getOwnWorkingInputProperty()->setDefaultFormat(in_f);
    Format* out_f = new Format(*in_format);
    getOwnWorkingOutputProperty()->addNewFormat(out_f);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
    
    in_f->setIOPartner(out_f);
    out_f->setIOPartner(in_f);

    return SUCCESS;
  }

  Result OSDOverlayNode::doActivate() 
  {
    int frame_size;

    const Format *in_f = getInputFormat();
    if (in_f == NULL) {
      return FAILURE; // error, not without a format!
    }

    string colorspace = in_f->getStringValue(Format::colorspace_param);
    image_width = in_f->getIntValue(Format::x_resolution_param);
    image_height = in_f->getIntValue(Format::y_resolution_param);

    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "Invalid Image-width or -height" << endl);
      return FAILURE;     
    }
    
    if(colorspace == "yv12") {
      frame_size = image_width * image_height;
      color = YVA12;
    }
    else {
      ERROR_STREAM( "Invalid ColorSpace! DoActivate will fail!" << endl);
      return FAILURE;
    }

    return SUCCESS;
  }

  Buffer* OSDOverlayNode::processBuffer(Buffer *in_buffer)
  {
    osd_mutex.lock();

    Buffer *out_buffer;
    // Decide whether are triggered to foreground
    if (getWorkingFlag())
    {
      if(in_buffer != 0) {
	in_buffer->release();
	DEBUG_STREAM("HAPPENED!"<<endl);
      }

      DEBUG_STREAM("getWorkingFlag is true" << endl);
      setWorkingFlag(false);
      if (lastBuffer != 0 )
      {
	out_buffer = getNewBuffer(lastBuffer->getSize());
	*out_buffer = *lastBuffer;
      }
      else
      {
	// this is needed if we got a static background from the beginning
        osd_mutex.unlock();
	return 0;
      }
    }
    else // triggered to background and new background arrived
    {
      DEBUG_STREAM("getWorkingFlag is false" << endl);
      in_buffer = in_buffer->getWriteableInstance();

      if (lastBuffer != 0)
      {
	DEBUG_STREAM("REF COUNT before release ! " << lastBuffer->getCount() << endl);
	lastBuffer->release();
	lastBuffer = 0;
      }

      if (paused)
      {
	// create copy called lastBuffer from current buffer called in_buffer
	lastBuffer = getNewBuffer(in_buffer->getSize());
	*lastBuffer = *in_buffer;
      }

      // Send in_buffer as out_buffer
      out_buffer = in_buffer;
    }

    for(items_t::iterator i = items.begin(); i != items.end(); ++i) {
      if(i->second->enabled) {
	DEBUG_STREAM("DRAWING ... " << i->first << endl);
	i->second->draw( out_buffer->getData(), color, image_width, image_height );
	DEBUG_STREAM("DRAWED ... " << endl);
      }
    }
    osd_mutex.unlock();

    return out_buffer;
  }

  /* ==== Internal methods follow ==== */

  Result OSDOverlayNode::addText(const string& id, const string& text, const size_t& x_offset, const size_t& y_offset)
  {
    //osd_mutex.lock();

    OSDText *t = new OSDText();
    t->x_offset = x_offset;
    t->y_offset = y_offset;
    t->setFont(font);
    t->setText(text);

    items[id] = t;

    //osd_mutex.unlock();
    return SUCCESS;
  }

  Result OSDOverlayNode::printText(const string& id, const string& text, const size_t& x_offset, const size_t& y_offset, const int& seconds)
  {
    osd_mutex.lock();
    items_t::iterator ii;
    OSDText *ptr;

    ii = items.find(id);

    if( ii != items.end() ) {
      ptr = dynamic_cast<OSDText *>(ii->second);
      ptr->setText(text);
      ptr->x_offset = x_offset;
      ptr->y_offset = y_offset;
      ptr->enabled = true;
      // setup system-timer to auto-clear osd
      if (seconds == 0) {
  	    signal(SIGALRM,SIG_IGN);
	    } else {
  	    signal(SIGALRM,&SIGALRM_handler);
      }
	    struct itimerval value,ovalue;
      value.it_value.tv_sec = value.it_interval.tv_sec = seconds;
      value.it_value.tv_usec = value.it_interval.tv_usec = 0;
      setitimer(ITIMER_REAL,&value,&ovalue);
      // system-timer set
      osd_mutex.unlock();
      return SUCCESS;
    }
    else {
      addText(id, text, x_offset, y_offset);
      osd_mutex.unlock();
      return FAILURE;
    }
  }

  Result OSDOverlayNode::removeText(const string& id)
  {
    osd_mutex.lock();
    items_t::iterator ii;
    OSDText *ptr;

    ii = items.find(id);

    if( ii != items.end() ) {
      ptr = dynamic_cast<OSDText *>(ii->second);
      delete ptr;

      items.erase( ii );
      osd_mutex.unlock();
      return SUCCESS;
    }
    else {
      osd_mutex.unlock();
      return FAILURE;
    }
  }

  Result OSDOverlayNode::addProgessBar(const string& id, const int& value, const int& x_offset, const int& y_offset)
  {
    osd_mutex.lock();

    OSDText *t = new OSDText();
    t->x_offset = x_offset;
    t->y_offset = y_offset;
    t->setFont(font);
    t->setProgressBarValue(value);

    items[id] = t; 

    osd_mutex.unlock();
    return SUCCESS;
  }

  Result OSDOverlayNode::addImage(const string& id, const string& url, const int& x_offset, const int& y_offset)
  {
    osd_mutex.lock();
    
    OSDImagePNG *t = new OSDImagePNG();
    t->x_offset = x_offset;
    t->y_offset = y_offset;
    t->load(url);

    items[id.c_str()] = t; 

    osd_mutex.unlock();
    return SUCCESS;
  }

  Result OSDOverlayNode::setItemOffset(const string& id, const int& x_offset, const int& y_offset)
  {
    osd_mutex.lock();
    OSDItem *item = items.find(id)->second;
    if(item) {
      item->x_offset = x_offset ; 
      item->y_offset = y_offset ; 
      osd_mutex.unlock();
      return SUCCESS;
    }
    else {
      osd_mutex.unlock();
      return FAILURE;
    }
  }

  Result OSDOverlayNode::setItemText(const string& id, const string& text)
  {
    osd_mutex.lock();
    items_t::iterator ii;
    OSDText *ptr;

    //OSDText *ptr = dynamic_cast<OSDText *>(items.find(id)->second);

    ii = items.find(id);

    if( ii != items.end() ) {
      ptr = dynamic_cast<OSDText *>(ii->second);
      ptr->setText(text);
      osd_mutex.unlock();
      return SUCCESS;
    }
    else {
      osd_mutex.unlock();
      return FAILURE;
    }
  }

  Result OSDOverlayNode::setItemEnabled(const string& id, const bool& enable)
  {
    items_t::iterator ii;

    osd_mutex.lock();

    ii = items.find(id);

    if( ii != items.end() ) {
      ii->second->enabled = enable;
      DEBUG_STREAM("setItemEnabled: Item " << id << (enable?" enabled":" disabled") <<  endl);
    }

    osd_mutex.unlock();
    
    return SUCCESS;
  }

  Result OSDOverlayNode::disableItems()
  {
    osd_mutex.lock();

    for(items_t::iterator i = items.begin(); i != items.end(); ++i)
      i->second->enabled = false;

    osd_mutex.unlock();

    return SUCCESS;
  }

  Result OSDOverlayNode::update()
  {
    osd_mutex.lock();
   // new_text_flag = true;
    if (paused) setWorkingFlag(true); //if paused, set Flag to get into processBuffer again
    osd_mutex.unlock();

    DEBUG_STREAM("DRAW update" << endl);
    return SUCCESS;
  }

  Result OSDOverlayNode::setFont(const string& font_name, const float& font_factor )
  {
    osd_mutex.lock();
    if(! font ) {
      font = read_font_desc(font_name.c_str(), font_factor, 0 );
      if( !font ) {
	      osd_mutex.unlock();
	      return FAILURE;
      }
    }
    osd_mutex.unlock();
    return SUCCESS;
  }

  Result OSDOverlayNode::doPause(const bool& paused_) {
    DEBUG_STREAM("doPause" << paused_ << endl );
    osd_mutex.lock();
    paused = paused_;
    osd_mutex.unlock();
    return SUCCESS;
  }
  
  Result OSDOverlayNode::setFormat(const Format& format) {
    image_width = format.getIntValue(Format::x_resolution_param);
    image_height = format.getIntValue(Format::y_resolution_param);
    
    // XXX: reset working output property
    
    // we can safely return SUCCESS because this node's output format equals its input format
    return SUCCESS;
  }
  
  void OSDOverlayNode::SIGALRM_handler(int) {
    if (osd_node)
      osd_node->disableItems(); // clear OSD
    signal(SIGALRM,SIG_IGN); // ignore SIGALRM
  }
}
