/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
 * USA.
 */


#include "RTJPEGEncodeNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::RTJPEGEncodeNode, "RTJPEGEncodeNode")

namespace NMM {

  RTJPEGEncodeNode::RTJPEGEncodeNode(const char* name,
				     const StreamQueue::Mode queue_mode,
				     const int queue_size) :
    GenericProcessorNode(name, queue_mode, queue_size ) {

    // Setup benchmark
    benchmark.setNamedObject(this);
    benchmark.setEnabled(true);

    quality = 80;
    quality_size = sizeof(unsigned char);

    mode = IFrameEncodeMode::INTRA_FRAME;
    mode_size = sizeof(IFrameEncodeMode::Mode);

    rtjpeg_struct = new RTjpeg_struct();

    rtjpeg_tables = new long unsigned int[128];
    rtjpeg_tables_size = 128 * sizeof(long unsigned int);
  };

  RTJPEGEncodeNode::~RTJPEGEncodeNode() {
    delete rtjpeg_struct;
    delete rtjpeg_tables;
  }

  Result RTJPEGEncodeNode::doInit() {
    NMM::Format *myInputFormat;
    myInputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"yuv422");
//    myInputFormat->addStringParamValue(Format::colorspace_param,"yuv420"); //according to the RTjpeg manual the format is YUV420P
    myInputFormat->addStringParamValue(Format::channelorder_param,"yuv");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::format_param, "planar");
    getOwnInputProperty()->setDefaultFormat(myInputFormat);

    NMM::Format *myOutputFormat;
    myOutputFormat=getOwnOutputProperty()->addNewFormat("video/rtjpeg");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"yuv422");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"yuv");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
    getOwnOutputProperty()->setDefaultFormat(myOutputFormat);
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }

  Result RTJPEGEncodeNode::doInitOutput() {
    const Format *in_f = getInputFormat();
    
    if (in_f == NULL) {
      ERROR_STREAM( "Format is null" << endl);
      return FAILURE; // error, not without a format!
    }

    int x_resolution = in_f->getIntValue(Format::x_resolution_param);
    int y_resolution = in_f->getIntValue(Format::y_resolution_param);
    int ratio_x = in_f->getIntValue(Format::ratio_x_param);
    int ratio_y = in_f->getIntValue(Format::ratio_y_param);
    float framerate = in_f->getFloatValue(Format::framerate_param);

    Format* i_format = new Format( *in_f );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);

    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/rtjpeg");
    o_format->addIntParamValue(Format::x_resolution_param, x_resolution);
    o_format->addIntParamValue(Format::y_resolution_param, y_resolution);
    o_format->addIntParamValue(Format::ratio_x_param, ratio_x);
    o_format->addIntParamValue(Format::ratio_y_param, ratio_y);
    o_format->addStringParamValue(Format::colorspace_param,"yuv422");
    o_format->addStringParamValue(Format::channelorder_param,"yuv");
    o_format->addIntParamValue(Format::bitperpixel_param,16);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);

    return SUCCESS;
  }

  Result RTJPEGEncodeNode::doActivate() {
    const Format *in_f = getInputFormat();

    if (in_f == NULL) {
      ERROR_STREAM( "Format is null" << endl);
      return FAILURE; // error, not without a format!
    }

    image_width = in_f->getIntValue(Format::x_resolution_param);
    image_height = in_f->getIntValue(Format::y_resolution_param);

    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "image size is wrong: " << endl);
      return FAILURE; 
    }

    // Setup rtjpeg
    rtjpeg_struct_mutex.lock();
    setupRTJPEG();
    rtjpeg_struct_mutex.unlock();

    return SUCCESS;
  }

  void RTJPEGEncodeNode::setupRTJPEG() {
    RTjpeg_init_compress(rtjpeg_tables, image_width, image_height, quality, rtjpeg_struct);
    if(mode == IFrameEncodeMode::INTER_FRAME) {
      DEBUG_STREAM("INTER_FRAME_COMPRESSION" << endl);
      RTjpeg_init_mcompress(rtjpeg_struct);
    }
    else {
      DEBUG_STREAM("INTRA_FRAME_COMPRESSION" << endl);
    }
  }

  Buffer* RTJPEGEncodeNode::processBuffer(Buffer *in_buffer) {
    if (in_buffer == 0)
      return 0;

    Buffer *out_buffer = getNewBuffer(image_width*image_height*2 + quality_size + mode_size + rtjpeg_tables_size, in_buffer);
	
    if (out_buffer) {
      rtjpeg_struct_mutex.lock();

      // set first data entry to compression quality
      memcpy(out_buffer->getData(), &quality, quality_size);
      
      // next, set compression mode
      memcpy(out_buffer->getData() + quality_size, &mode, mode_size);

      // copy tables as 'header' of the data of the out_buffer
      memcpy(out_buffer->getData() + quality_size + mode_size, rtjpeg_tables, rtjpeg_tables_size);

      if(mode == IFrameEncodeMode::INTER_FRAME) {
	out_buffer->setUsed( quality_size + mode_size + rtjpeg_tables_size + 
			     RTjpeg_mcompressYUV422((__rtjpeg__s8 *) out_buffer->getData() 
						                     + quality_size 
                                                                     + mode_size 
                                                                     + rtjpeg_tables_size,
						    (__rtjpeg__u8 *) in_buffer->getData(), 2, 1, rtjpeg_struct));
      }
      else {
	out_buffer->setUsed( quality_size + mode_size + rtjpeg_tables_size + 
			     RTjpeg_compressYUV422 ((__rtjpeg__s8 *) out_buffer->getData() 
						                     + quality_size 
						                     + mode_size
                                                                     + rtjpeg_tables_size, 
						    (__rtjpeg__u8 *) in_buffer->getData(), rtjpeg_struct));
      }
      rtjpeg_struct_mutex.unlock();
			       
      benchmark.benchmark(out_buffer->getUsed());

      in_buffer->release();

      return out_buffer; 
    }
    else {
      in_buffer->release();
      DEBUG_STREAM( "Could not allocate out Buffer!" << endl);
      return 0;
    }
  }

  void RTJPEGEncodeNode::setQuality(const float& value) {
    int q = (int) (value * 255.0f);
    if(q != quality) {
      rtjpeg_struct_mutex.lock();
      quality = q;
      setupRTJPEG();
      rtjpeg_struct_mutex.unlock();
    }
  }
  
  float RTJPEGEncodeNode::getQuality() {
    return (float) quality / 255.0f;
  }
  
  void RTJPEGEncodeNode::setBitrate(const int&) {
    return;
  }
  
  int RTJPEGEncodeNode::getBitrate() {
    return -1;
  }
  
  void RTJPEGEncodeNode::setFrameEncodeMode(const IFrameEncodeMode::Mode& m) {
    if(m != mode) {
      rtjpeg_struct_mutex.lock();
      mode = m;
      if(isActivated() || isStarted()) { // must at least be ACTIVATED, otherwise width and height are not set
	setupRTJPEG();
      }
      rtjpeg_struct_mutex.unlock();
    }
  }

  IFrameEncodeMode::Mode RTJPEGEncodeNode::getFrameEncodeMode() const {
    return mode;
  }
}
