/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_IVTVREADNODE_HPP_
#define _NMM_IVTVREADNODE_HPP_ _NMM_IVTVREADNODE_HPP_

#include "nmm/interfaces/video/tv/IIVTVImpl.hpp"
#include "nmm/interfaces/video/tv/ITVCardImpl.hpp"
#include "nmm/interfaces/device/IDeviceImpl.hpp"

#include "nmm/interfaces/file/IBufferSizeImpl.hpp"

#include "nmm/base/GenericSourceNode.hpp"

#include "nmm/base/sync/StreamTimer.hpp"

#include "nmm/interfaces/video/tv/IChannelInfoImpl.hpp"

#include <inttypes.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "videodev2.h"
#include "ivtv-ext-api.h"

#include "FrequencyTable.hpp"

namespace NMM {
  
  /** IVTV Read Node.
   */
  class IVTVReadNode :
    public GenericSourceNode,
    public IBufferSizeImpl,
    public ITVCardImpl,
    public IIVTVImpl,
    public IChannelInfoImpl
  {
  public:
    /** @name Constructor & Destductor
     */
    //@{
    /** Constructor.
     */
    IVTVReadNode(const char *name = "IVTVReadNode",
                 StreamQueue::Mode = StreamQueue::MODE_SUSPEND,
                 int = StreamQueue::DEFAULT_SIZE);
    
    /** Virtual destructor.
     */
    virtual ~IVTVReadNode();
    //@}
    
  protected:
    /** @name GenericNode Methods
     */
    //@{
    virtual Result doInit();
    virtual Result doDeinit();
    virtual Result doInitOutput();
    virtual Result doActivate();
    
    virtual Message* processBuffer(Buffer*);
    //@}
    
    /** @name IBufferSize interface
     */
    //@{
    /** Set buffer size.
     */
    virtual void setBufferSize(const int& value);
    
    /** Get buffer size.
     */
    virtual int getBufferSize();
    //@}
    
    /** @name ITVCard Interface
     */
    //@{
    /** Set device.
     */
    virtual void setDevice(const string& value) throw(const DeviceNotFoundException&, const WrongDeviceException&);
    
    /** Get device.
     */
    virtual string getDevice() const;
    
    /** Set resolution.
     */
    virtual void setResolution(const int& x, const int& y) throw(const Exception&);

    /** Get resolution.
     */
    virtual void getResolution(int& x, int& y) throw(const Exception&);
    
    /** Set channel.
     */
    virtual Result setChannel(const int& channel);
    
    /** Get channel.
     */
    virtual Result getChannel(int& channel, string& name);
    //@}
    
    
    /** @name Audio Inputs
     */
    //@{
    /** This method returns the number of available inputs.
     */
    virtual unsigned int getAudioCount() const;

    /** This method returns the properties of the specified input.
     * @param index Audio index, ranges from 0 to getAudioCount()-1
     */
    virtual TVAudio getAudio(const unsigned int& index) const;
    
    /** This method returns the list of available inputs.
     */
    virtual list<TVAudio> getAudioList() const;
    
    /** This method returns the current input.
     */
    virtual unsigned int getCurrentAudio() const;
    
    /** This method sets the current input.
     * @param index Audio index, ranges from 0 to getAudioCount()-1
     */
    virtual void setCurrentAudio(const unsigned int& index);
    //@}
    
    
    /** @name Inputs
     */
    //@{
    /** This method returns the number of available inputs.
     */
    virtual unsigned int getInputCount() const;

    /** This method returns the properties of the specified input.
     * @param num Input number, ranges from 0 to getInputCount()-1
     */
    virtual TVInput getInput(const unsigned int& num) const;
    
    /** This method returns the list of available inputs.
     */
    virtual list<TVInput> getInputList() const;
    
    /** This method returns the current input.
     */
    virtual unsigned int getCurrentInput() const;

    /** This method sets the current input.
     * @param num Input number, ranges from 0 to getInputCount()-1
     */
    virtual void setCurrentInput(const unsigned int& num);
    //@}
    
    
    /** @name Standards
     */
    //@{
    /** This method returns the identifier of the standard which matches the given name.
     * If no such standard is found, zero is returned.
     */
    virtual unsigned int getStandardByName(const string& name) const;
    
    /** This method returns the properties of the specified standard.
     * @param num Standard number
     */
    virtual TVStandard getStandard(const unsigned int& num) const;
    
    /** This method returns the list of supported standards.
     */
    virtual list<TVStandard> getStandardList() const;
    
    /** This method returns the standard set for the current input.
     */
    virtual unsigned int getCurrentStandard() const;
    
    /** This method sets the standard for the current input
     * @param num Standard number, ranges from 0 to getStandardCount()-1
     */
    virtual void setCurrentStandard(const unsigned int& standard);
    
    /** This method sets the standard for the current input by name.
     */
    virtual void setStandardByName(const string& name);
    //@}
    
    /** @name Tuners
     */
    //@{
    /** This method returns the number of tuners.
     */
    virtual unsigned int getTunerCount() const;
    
    /** This method returns information about a tuner.
     * @param index Index of tuner, ranges from 0 to getTunerCount()-1
     */
    virtual TVTuner getTuner(const unsigned int& index) const;
    
    /** This method returns the list of tuners.
     */
    virtual list<TVTuner> getTunerList() const;
    
    /** This method returns the current frequency of a tuner.
     * @param tuner Number of the tuner
     * @return Current frequency of the tuner in MHz
     */
    virtual float getFrequency(const unsigned int& tuner) const;
    
    /** This method sets the frequency of a tuner.
     * @param tuner Number of the tuner
     * @param frequency The frequency to set in MHz
     */
    virtual void setFrequency(const unsigned int& tuner, const float& frequency);
    //@}
    
    /** @name Codecs
     */
    //@{
    /** This method returns the codec parameters.
     */
    virtual IVTVCodec getCodecParams() const;
    
    /** This method sets the codec parameters.
     */
    virtual void setCodecParams(const IVTVCodec& codec);

    /** This method sets the GOP size
     */
    virtual void setGOPSize(const unsigned int& value);

    /** This method sets the number of B-Frames
     */
    virtual void setBFrames(const unsigned int& value);

    /** This method sets the GOP closure mode of the codec (1=close GOP, 0=open GOP)
     */
    virtual void setGOPClosure(const unsigned int& value);

    /** This method sets inverse telecine (1=ON, 0=OFF)
     */
    virtual void setPulldown(const unsigned int& value);

    /** This method sets bitrate
     */
    virtual void setBitrate(const unsigned int& value);

    /** This method sets peak bitrate (must be greater than bitrate)
     */
    virtual void setPeakBitrate(const unsigned int& value);

    /** This method sets the Streamtype
     *   IVTV_STREAM_PS     = 0
     *   IVTV_STREAM_TS     = 1
     *   IVTV_STREAM_MPEG1  = 2
     *   IVTV_STREAM_PES_AV = 3
     *   IVTV_STREAM_PES_V  = 5
     *   IVTV_STREAM_PES_A  = 7
     *   IVTV_STREAM_DVD    = 10
    */
    virtual void setStreamType(const unsigned int& value);

    virtual void setBitrateMode(const unsigned int& value);
    virtual void setDNRMode(const unsigned int& value);
    virtual void setDNRType(const unsigned int& value);
    virtual void setDNRSpatial(const unsigned int& value);
    virtual void setDNRTemporal(const unsigned int& value);
    virtual void setAspect(const unsigned int& value);
    //@}
    
    /** @name Controls.
     */
    //@{
    /** This method gets the current brightness.
     */
    virtual unsigned int getBrightness() const;
    
    /** This method gets the current contrast.
     */
    virtual unsigned int getContrast() const;
    
    /** This method gets the current saturation.
     */
    virtual unsigned int getSaturation() const;
    
    /** This method gets the current hue.
     */
    virtual int getHue() const;
    
    /** This method gets the current volume.
     */
    virtual unsigned int getVolume() const;
    
    /** This method gets the mute status.
     */
    virtual bool getMute() const;
    
    /** This method sets the brightness.
     * @param brightness Brightness (0 to 255)
     */
    virtual void setBrightness(const unsigned int& brightness);
    
    /** This method sets the contrast.
     * @param contrast Contrast (0 to 127)
     */
    virtual void setContrast(const unsigned int& contrast);
    
    /** This method sets the saturation.
     * @param saturation Saturation (0 to 127)
     */
    virtual void setSaturation(const unsigned int& saturation);
    
    /** This method sets the hue.
     * @param hue Hue offset (-128 to 127)
     */
    virtual void setHue(const int& hue);
    
    /** This method sets the volume.
     * @param volume Volume (0 to 65535)
     */
    virtual void setVolume(const unsigned int& volume);
    
    /** This method sets the mute state.
     */
    virtual void setMute(const bool& mute);
    //@}
    
    /** @name Frequency tables
     */
    //@{
    /** This method returns the available frequency tables.
     */
    virtual list<string> getFrequencyTableList() const;
    
    /** This method returns the current frequency table.
     */
    virtual string getFrequencyTable() const;
    
    /** This method sets the current frequency table.
     * @param name Name of the frequency table
     */
    virtual void setFrequencyTable(const string& table);

    /**This method sets the input/output-value in MSP-matrix
     * @param input input-value
     * @param output output-value
     */
    virtual void setInputOutput(const int& input, const int& output);

  private:
    /** This method sets the value of a control.
     */
    void setControl(unsigned int id, unsigned int value, const string& name);
    
    /** This method returns the value of a control.
     */
    unsigned int getControl(unsigned int id, const string& name) const;

    /** This method sets the Framerate
     */
    void setFramerate(const unsigned int& value);

    /** This method initializes the frequency tables.
     */
    void initFrequencyTables()
      throw (const Exception&);

    //IChannelInfo interface
    virtual Result hasChannel(int& , const string& ){return FAILURE;}
    virtual TVChannelList getChannelList();
    virtual void setChannelConfigFile(const string &filename);
  private:
    TVChannelList channels;
    string channelConfigFile;
    void loadChannelconfig();

  private:
    typedef vector<TVInput>                 Inputs;
    typedef Inputs::iterator                InputIterator;
    typedef Inputs::const_iterator          InputConstIterator;
    
    typedef vector<TVAudio>                 AudioInputs;
    typedef AudioInputs::iterator           AudioInputIterator;
    typedef AudioInputs::const_iterator     AudioInputConstIterator;
    
    typedef vector<TVTuner>                 Tuners;
    typedef Tuners::iterator                TunerIterator;
    typedef Tuners::const_iterator          TunerConstIterator;
    
    typedef map<unsigned int, TVStandard>   Standards;
    typedef Standards::iterator             StandardIterator;
    typedef Standards::const_iterator       StandardConstIterator;
    
    /* Begin> Device */
    string                      __device;            ///< Device filename
    int                         __fd;                ///< File descriptor to read from device
    /* End< */
    
    /* Begin> Information */
    AudioInputs                 __audio_inputs;      ///< Available audio inputs
    Inputs                      __inputs;            ///< Available inputs
    Tuners                      __tuners;            ///< Available tuners
    Standards                   __standards;         ///< Supported standards
    /* End< */
    
    /* Begin> Output */
    ssize_t                     __buffer_size;       ///< Read buffer size
    bool                        __do_flush;          ///< Need to flush the stream
    StreamTimer                 __stream_timer;      ///< Stream timer
    /* End< */
    
    /* Begin> Channels */
    unsigned int                __cur_channel;       ///< Current channel
    string                      __cur_freqtable;     ///< Current frequency table
    map<string, FrequencyTable> __frequencies;       ///< Frequency tables
    /* End< */
  };
  
}

#endif
