/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_SIGNALREACTOR_HPP
#define NMM_SIGNALREACTOR_HPP

#include "nmm/utils/Singleton.hpp"
#include "nmm/utils/thread/ThreadedObject.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <list>
#include <vector>
#include <queue>

namespace NMM {

class SignalHandler;

/**
  */
class SignalReactor : public NamedObject, public ThreadedObject{
public:
  friend class CreateUsingNew<SignalReactor>;
  typedef void (*signal_handler) (int);
  /** This enum defines the Posix Signal with prefix 'SR_'. They have NOT the same value as
    * the posix signal and are enumerated from 0 to SR_SIGTTOU. So this values can be used to
    * iterate through all Signals
    */
  enum PosixSignal {
       SR_SIGHUP = 0, SR_SIGINT, SR_SIGQUIT, SR_SIGILL, SR_SIGABRT, SR_SIGFPE,
       SR_SIGKILL, SR_SIGSEGV, SR_SIGPIPE, SR_SIGALRM, SR_SIGTERM, SR_SIGUSR1,
       SR_SIGUSR2, SR_SIGCHLD, SR_SIGCONT, SR_SIGSTOP, SR_SIGTSTP, SR_SIGTTIN,
       SR_SIGTTOU
  };

  enum SignalAction { SIG_IGNORE = 0, SIG_DEFAULT};

  ~SignalReactor();

/** registers to all Signal
  */
  void registerSignal(SignalHandler* sig_handler );

/** Signature of sighandler_t is defined as 'void (sighandler_t*)(int)'
  */
  void registerSignal(signal_handler sig_handler);

/**
  */
  inline void registerSignal(PosixSignal signal, SignalHandler* sig_handler )   { handler_array[signal]->object_list.push_front(sig_handler); }

/** Signature of sighandler_t is defined as 'void (sighandler_t*)(int)'
  */
  inline void registerSignal(PosixSignal signal, signal_handler sig_handler)    { handler_array[signal]->method_list.push_front(sig_handler); }

/** removes the SignalHandler form a specific signal
  */
  inline void unregisterSignal(PosixSignal signal, SignalHandler* sig_handler)  { handler_array[signal]->object_list.remove(sig_handler); }

/** removes the SignalHandler form a specific signal
  */
  inline void unregisterSignal(PosixSignal signal, signal_handler sig_handler)  { handler_array[signal]->method_list.remove(sig_handler); }

/** removes from all signals
  */
  void unregisterSignal(SignalHandler* sig_handler);

/** removes from all signals
  */
  void unregisterSignal(signal_handler sig_handler);

/** returns a human readable description of a Signal
  */
  inline const char* getSignalDescription(PosixSignal signal) const {return description_array[signal];}

/** returns a 0 terminated char array with the name of the Signal. For example
  * the return value of SR_SIGSEGV is "SIGSEGV"
  */
  inline const char* getSignalName(PosixSignal signal)        const {return name_array[signal];}

/** This method allows to set/ignore the default signal handling after all other
  * objects are informed. For example if a SIGSEGV happens all registered objects are
  * informed and after this the default action exit is called
  */
  void setAction(PosixSignal signal, SignalAction action);

/** disables the SignalReactors method and resets it to action
  */
  void disableSignal(PosixSignal signal, SignalAction action);

/** enables the SignalReactors method and sets the action to action
  */
  void enableSignal(PosixSignal signal, SignalAction action);
protected:
  virtual void run();
  inline PosixSignal toPosixSignal(int i) const {return converter_array[i];}

private: //methods
  SignalReactor();
  SignalReactor(const SignalReactor& ) ;

  void pushSignal(PosixSignal);
  PosixSignal popSignal();
  void processSignal(PosixSignal signal);

private: //structs
  struct SigHandlerStruct {
    SigHandlerStruct() { default_handler = 0; default_action = SIG_DEFAULT;}
    list<SignalHandler*> object_list;
    list< void(*)(int) > method_list;
    signal_handler default_handler;
    SignalAction default_action;
  };

private: //member
  vector<SigHandlerStruct*> handler_array;
  vector<PosixSignal>       converter_array;
  queue<PosixSignal>        signal_queue;

  ThreadMutex mutex;
  ThreadMutex queue_mutex;
  ThreadCondition* condition;

  friend void generic_handler(int i);
  friend void generic_exit(int i);

  bool is_running;
  static const char* description_array[];
  static const char* name_array[];
  static const int   signal_array[];
  static SignalReactor* local_instance; //pointer for fast access
};


/* Typdef to access the NetReactor */
typedef Singleton<SignalReactor, CreateUsingNew> SSignalReactor;



class SignalHandler {
public:
  virtual void receiveSignal(SignalReactor::PosixSignal)                              = 0;
  virtual void disableSignal(SignalReactor::PosixSignal, SignalReactor::SignalAction) = 0;
  virtual void enableSignal(SignalReactor::PosixSignal, SignalReactor::SignalAction)  = 0;
};
}//end namespace
#endif
