/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_SOCK_HPP
#define NMM_SOCK_HPP NMM_SOCK_HPP

#include "NetExceptions.hpp"
#include "INetAddr.hpp"

using namespace std;

namespace NMM {

  /** Base class for sockets
   */
  class Sock {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Virtual destructor
     */
    virtual ~Sock() {}
    //@}
    
    /** @name Addresses
     */
    //@{
    /** Set address of peer.
     * This method should only be called by a SockConnector or SockAcceptor.
     * @addr Pointer to peer address
     */
    virtual void setPeerAddress(const INetAddr& addr) = 0;

    /** Returns the peer address of a socket.
     * @return Peer address of this socket
     */
    virtual const INetAddr& getPeerAddress() const = 0;
    
    /** Returns the local address of the socket.
     * @return Local address of this socket
     */
    virtual INetAddr getLocalAddress() const = 0;
    //@}
    
    /** @name Socket Handling
     */
    //@{
    /** Assigns a new socket. The old socket is closed if Sock is owner of it
     * @param socket The new socket
     * @param owner Specify if Sock is responsible for the new socket
     */
    virtual void setSocket(int socket, bool owner = false) = 0;
    
    /** returns the current socket.
     * @return current socket
     */
    virtual int  getSocket()  const = 0;
    
    /** Closes the socket.
     */
    virtual void close() = 0;
    //@}
    
    /** @name Transmission Methods
     */
    //@{
    /** Receives data from a connection. This method uses the recv(2) system call.
     * @param buffer Buffer where the received data should be stored
     * @param len Maximum number of bytes to receive
     * @param flags Flags, as specified in recv(2)
     * @return Actual number of bytes received
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    virtual ssize_t recv (void* buffer, size_t len, int flags) throw (const TransmissionException&) = 0;
    
    /** Sends data over a connection. This method uses the send(2) system call.
     * @param buffer Buffer containing the data to be sent
     * @param len Maximum number of bytes to send
     * @param flags Flags, as specified in send(2)
     * @return Actual number of bytes sent
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    virtual ssize_t send (const void* buffer, size_t len, int flags) throw (const TransmissionException&) = 0;
    
    /** Reads data from the socket.
     * This method seems to be obsolete and was only used in the WM-Client example
     * @param buffer Buffer where the read data should be stored
     * @param len Maximum number of bytes to read
     * @return Actual number of bytes read
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    virtual ssize_t read (char* buffer, size_t len) throw (const TransmissionException&) = 0;
    
    /** Writes data to the socket.
     * This method seems to be obsolete and was only used in the WM-Client example
     * @param buffer Buffer containing the data to be written
     * @param len Maximum number of bytes to write
     * @return Actual number of bytes written
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    virtual ssize_t write(const char* buffer, size_t len) throw (const TransmissionException&) = 0;    
    
    /** Read data from the socket, treating it like a stream.
     * This method reads data until len bytes have been read or the connection has been closed.
     * @param buffer Buffer where the read data should be stored
     * @param len Maximum number of bytes to read
     * @param max_packet_size Maximum size of a single packet, 0 to use protocol default
     * @exception TransmissionException if the transmission fails or a socket error occurs
     * @exception ConnectionClosed if the connection was closed before len bytes have been read
     */
    virtual void readStream(char* buffer, size_t len,  size_t max_packet_size = 0)
      throw (const TransmissionException&, const ConnectionClosed&) = 0;
    
    /** Write data from the socket, treating it like a stream.
     * This method tries to write exactly len bytes of data.
     * @param buffer Buffer containing the data to be written
     * @param len Maximum number of bytes to write
     * @param max_packet_size Maximum size of a single packet, 0 to use protocol default
     * @exception TransmissionException if the transmission fails or a socket error occurs
     * @exception ConnectionClosed if the connection was closed before len bytes have been written
     */
    virtual void writeStream(const char* buffer, size_t length, size_t max_packet_size = 0)
      throw (const TransmissionException&, const ConnectionClosed&) = 0;
    //@}
    
  protected:
    /** Create TransmissionException for reading.
     * Creates and throws a TransmissionException based on the given errno value.
     * @param my_errno The errno value after the failed attempt to read from a socket
     */
    static void createReadTransmissionException (int my_errno);
    
    /** Create TransmissionException for writing.
     * Creates and throws a TransmissionException based on the given errno value.
     * @param my_errno The errno value after the failed attempt to write to a socket
     */
    static void createWriteTransmissionException(int my_errno);	
  };
}

#endif //NMM_SOCK_HPP
