/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SockConnector.hpp"
#include "nmm/utils/utils.hpp"

#include <netdb.h>       // for getprotobyname
#include <errno.h>       // for errno
#include <netinet/tcp.h> // for TCP_NODELAY

#include <iostream>

#include "SockStream.hpp"
#include "SockDGram.hpp"

namespace NMM {

  /****************** class SockConnector ******************/
  SockConnector::SockConnector() {
  }

  SockConnector::~SockConnector() {
  }

  Sock* SockConnector::connect(const INetAddr& addr, const char* proto) throw(const ConnectionException&) {
    /* request a protoent-struct for the given protocol */
    struct protoent* protocol = getprotobyname(proto);
    if (protocol == 0) {
      /* If an error occurs create an exception */
      throw ConnectionException("Unknown Protocol", 0, ConnectionException::connect);
    }
    
    /* Initialize the socket */
    int new_socket = -1;
    Sock* new_sock = 0;
    
    if(addr.getTransportProtocol() == "tcp") {
      /* create a new socket */
      new_socket = ::socket(PF_INET, SOCK_STREAM, protocol->p_proto);
      
      /* test the new socket and create an exception if an error occurs */
      if (new_socket == -1) {
        /* If an error occurs create an exception */
        createSocketException(errno);
      }
      
      /* connect to the given address */
      if (::connect(new_socket, addr.getAddress(), addr.getSize()) == -1) {
        /* If an error occurs create an exception */
        createConnectionException(errno);
      }
      
      /* Create a SockStream that can be used for this connection */
      new_sock = new SockStream(new_socket, false);
      
      /* Set peer address of SockStream */
      new_sock->setPeerAddress(addr);
    } else if(addr.getTransportProtocol() == "udp") {
//       /* create a new socket */
//       new_socket = socket(AF_INET, SOCK_DGRAM, protocol->p_proto);
      
//       /* test the new socket and create an exception if an error occurs */
//       if (new_socket == -1) {
//         /* If an error occurs create an exception */
//         createSocketException(errno);
//       }
      
      /* Create a SockDGram that can be used for this connection */
      new_sock = new SockDGram();
      
      /* Set peer address of SockDGram */
      new_sock->setPeerAddress(addr);
    }
    else {
      throw ConnectionException("Unknown transport protocol", 0, ConnectionException::connect);
    }
    return new_sock;

  }


  void SockConnector::connect(Sock& sock, const INetAddr& addr, const char* proto) throw(const ConnectionException&) {
    /* request a protoent-struct for the given protocol */
    struct protoent* protocol = getprotobyname(proto);
    if (protocol == 0) {
      /* create a new socket */
      throw ConnectionException("Unknown Protocol", 0, ConnectionException::connect);
    }

    /* Initialize the socket */
    int new_socket = -1;
    
    /* handle "tcp" connection */
    if(addr.getTransportProtocol() == "tcp") {
      /* create a new socket */
      new_socket = ::socket(PF_INET, SOCK_STREAM, protocol->p_proto);
      
      /* test the new socket and create an exception if an error occurs */
      if (new_socket == -1) {
        /* If an error occurs create an exception */
        createSocketException(errno);
      }
      
      /* test the new socket and create an exception if an error occurs */
      if (::connect(new_socket, addr.getAddress(), addr.getSize()) == -1) {
        /* If an error occurs create an exception */
        createConnectionException(errno);
      }
      
      /* Begin> Set TCP_NODELAY */
      int value = true;
      int error = setsockopt (new_socket, IPPROTO_TCP, TCP_NODELAY, &value, sizeof (value));
      if (error == -1) {
        GLOBAL_ERROR_STREAM("Could not disable tcp delay" << endl);
      }
      /* End< */
      
      /* Set socket */
      sock.setSocket(new_socket, true);
      
      /* Set peer address  */
      sock.setPeerAddress(addr);
    }
    
    else if(addr.getTransportProtocol() == "udp") {
//       new_socket = socket(AF_INET, SOCK_DGRAM, protocol->p_proto);
      
//       /* test the new socket and create an exception if an error occurs */
//       if (new_socket == -1) {
//         /* If an error occurs create an exception */
//         createSocketException(errno);
//       }
      
//       /* Set socket */
//       sock.setSocket(new_socket, true);

      int on = 1;
      bool isMulticastAddress = IN_MULTICAST(addr.getIpAddress());

      bool isBroadCastAddress = (((in_addr_t)addr.getIpAddress() & 0x000000FF) ==  0x000000FF);

      new_socket = ::socket(AF_INET, SOCK_DGRAM, protocol->p_proto);

      if (new_socket == -1)
        createSocketException(errno);

      if (isMulticastAddress) {
	if ( setsockopt (new_socket, SOL_SOCKET, SO_REUSEADDR, ( const char* ) &on, sizeof ( on ) ) == -1 ) {
	  GLOBAL_ERROR_STREAM2("SO_REUSEADDR failed in SockAcceptor" << endl);
	}
	else {
	  GLOBAL_DEBUG_STREAM2("SO_REUSEADDR set for SockAcceptor" << endl);
	}
      }
      
      if (isBroadCastAddress) {
        if (setsockopt(new_socket, SOL_SOCKET, SO_BROADCAST, (const char* ) &on, sizeof(on)) == -1) {
          GLOBAL_ERROR_STREAM2("SO_BROADCAST failed in SockAcceptor" << endl);
        }
        else {
          GLOBAL_DEBUG_STREAM2("SO_BROADCAST set for SockAcceptor"<<endl);
        }
      }
      
      if (::connect(new_socket, addr.getAddress(), addr.getSize()) == -1)
        createConnectionException(errno);

      if (isMulticastAddress) {
	struct ip_mreq mreq;
	mreq.imr_multiaddr.s_addr = htonl(addr.getIpAddress());
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	
	setsockopt(new_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
      }

      /* Set socket */
      sock.setSocket(new_socket, true);

      /* Set peer address  */
      sock.setPeerAddress(addr);
    }
      
    else {
      throw ConnectionException("Unknown transport protocol", 0, ConnectionException::connect);
    }

  }
  
  void SockConnector::createSocketException(int my_errno) {
    switch(my_errno) {
    case EPROTONOSUPPORT :
      throw ConnectionException("The protocol type or the specified protocol is not supported within this domain.", my_errno, ConnectionException::socket);

    case ENFILE :
      throw ConnectionException("Not enough kernel memory to allocate a new socket structure.", my_errno, ConnectionException::socket);

    case EMFILE :
      throw ConnectionException("Process file table overflow.", my_errno, ConnectionException::socket);

    case EACCES :
      throw ConnectionException("Permission to create a socket of the specified type and/or protocol is denied.", my_errno, ConnectionException::socket);

    case ENOBUFS:
      throw ConnectionException("Insufficient memory is available. The socket cannot be created until sufficient resources are freed.", my_errno, ConnectionException::socket);

    case ENOMEM :
      throw ConnectionException("Insufficient memory is available. The socket cannot be created until sufficient resources are freed.", my_errno, ConnectionException::socket);

    case EINVAL :
      throw ConnectionException("Unknown protocol, or protocol family not available.", my_errno, ConnectionException::socket);
    }
  }

  void SockConnector::createShutdownException(int my_errno) {
    switch(my_errno) {
    case EBADF    :
      throw ConnectionException("Value is not a valid descriptor.", my_errno, ConnectionException::shutdown);

    case ENOTSOCK :
      throw ConnectionException("Value is a file, not a socket.", my_errno, ConnectionException::shutdown);

    case ENOTCONN :
      throw ConnectionException("The specified socket is not connected.", my_errno, ConnectionException::shutdown);

    default:
      throw ConnectionException("Unknown Error", my_errno, ConnectionException::shutdown);
    }
  }



  void SockConnector::createConnectionException(int my_errno) {
    switch(my_errno) {
    case EBADF   :
      throw ConnectionException("Bad descriptor.", my_errno, ConnectionException::connect );

    case EFAULT  :
      throw ConnectionException("The socket structure address is outside the user's address space.", my_errno, ConnectionException::connect );

    case ENOTSOCK:
      throw ConnectionException("The descriptor is not associated with a socket.", my_errno, ConnectionException::connect );

    case EISCONN :
      throw ConnectionException("The socket is already connected.", my_errno, ConnectionException::connect );

    case ECONNREFUSED :
      throw ConnectionException("Connection refused at server.", my_errno, ConnectionException::connect );

    case ETIMEDOUT    :
      throw ConnectionException("Timeout while attempting connection. The server may be too busy  to  accept  new connections.", my_errno, ConnectionException::connect );

    case ENETUNREACH  :
      throw ConnectionException("Network is unreachable.", my_errno, ConnectionException::connect );

    case EADDRINUSE   :
      throw ConnectionException("Address is already in use.", my_errno, ConnectionException::connect );

    case EINPROGRESS  :
      throw ConnectionException("The  socket  is non-blocking and the connection cannot be completed immediately.", my_errno, ConnectionException::connect );

    case EALREADY     :
      throw ConnectionException("The socket is non-blocking and a previous connection attempt has  not  yet  been completed.", my_errno, ConnectionException::connect );

    case EAGAIN       :
      throw ConnectionException("No  more  free  local  ports  or  insufficient entries in the routing cache.", my_errno, ConnectionException::connect );

    case EAFNOSUPPORT :
      throw ConnectionException("The  passed  address  didn't  have  the  correct address family in its sa_family field.", my_errno, ConnectionException::connect );

    case EACCES :
      throw ConnectionException("The user tried to connect to a  broadcast  address  without  having  the  socket \
               broadcast flag enabled or the connection request failed because of a local fire-\
               wall rule.", my_errno, ConnectionException::connect );

    case EPERM  :
      throw ConnectionException("The user tried to connect to a  broadcast  address  without  having  the  socket \
               broadcast flag enabled or the connection request failed because of a local fire-\
               wall rule.", my_errno, ConnectionException::connect );
    default :
      throw  ConnectionException("Unknown error", my_errno, ConnectionException::connect );
    }
  }

}
