/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                     
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/*
 * xml++.cpp based on xml++.cc from libxml++.
 * libxml++ is copyright (C) 2000 by Ari Johnson, and
 * are covered by the GNU Lesser General Public License.
 */

#include "xml++.hpp"

static XMLNode *readnode(xmlNodePtr);
static void writenode(xmlDocPtr, XMLNode *, xmlNodePtr, int = 0);

XMLTree::XMLTree(const XMLTree& tree) {
  *this = tree;
}


const XMLTree& XMLTree::operator=(const XMLTree& tree) {
  delete _root;
  _root = 0;
  if (tree.root())
    _root = new XMLNode(*tree.root());

  _filename     = tree.filename();
  _compression  = tree.compression();
  _initialized  = true;

  return *this;
}


bool XMLTree::operator==(const XMLTree& tree) {
  if (tree.root()) {
    if (root()) {
      return (*root() == *tree.root());
    } else {
      return false;
    }
  } else {
    if(!root()) //both have no root node so they are equal
      return true;
  }
  return false;
}


XMLTree::~XMLTree() {
  delete _root;
}

XMLNode* XMLTree::set_root(XMLNode* n) {
 delete _root;
 return _root = n;
}

int XMLTree::set_compression(int c) {
  if(c > 9)
    c = 9;

  if(c < 0)
    c = 0;

  _compression = c;

  return _compression;
}

bool XMLTree::read(void) {
  xmlDocPtr doc;

  if(_root) {
    delete _root;
    _root = NULL;
  }

  xmlKeepBlanksDefault(0);

  doc = xmlParseFile(_filename.c_str());
  if(!doc) {
    _initialized = false;
    return false;
  }

  _root = readnode(xmlDocGetRootElement(doc));
  xmlFreeDoc(doc);
  _initialized = true;

  return true;
}

bool XMLTree::read_buffer(const string &buffer) {
  xmlDocPtr doc;

  _filename = "";

  if(_root) {
    delete _root;
    _root = NULL;
  }

  doc = xmlParseMemory((char *) buffer.c_str(), buffer.length());
  if(!doc) {
    _initialized = false;
    return false;
  }

  _root = readnode(xmlDocGetRootElement(doc));
  xmlFreeDoc(doc);
  _initialized = true;

  return true;
}

bool XMLTree::write(void) const {
  xmlDocPtr doc;
  XMLNodeList children;
  int result;

  xmlKeepBlanksDefault(0);
  doc = xmlNewDoc((xmlChar *) "1.0");
  xmlSetDocCompressMode(doc, _compression);
  writenode(doc, _root, doc->children, 1);
  result = xmlSaveFormatFile(_filename.c_str(), doc, 1);
  xmlFreeDoc(doc);

  if(result == -1)
    return false;

  return true;
}

string XMLTree::write_buffer(void) const {
  string retval;
  char *ptr;
  int len;
  xmlDocPtr doc;
  XMLNodeList children;

  xmlKeepBlanksDefault(0);
  doc = xmlNewDoc((xmlChar *) "1.0");
  xmlSetDocCompressMode(doc, _compression);
  writenode(doc, _root, doc->children, 1);
  xmlDocDumpMemory(doc, (xmlChar **) &ptr, &len);
  xmlFreeDoc(doc);

  retval = ptr;

  free(ptr);

  return retval;
};

XMLNodeList XMLTree::children(const string& /*name*/) const {
  if (root())
    return root()->children();

  return XMLNodeList(); //returns an empty list if no root node available
}

XMLNode::XMLNode(const string &n)
 : _name(n), _content(string()) {
  if(_name.empty())
    _initialized = false;
  else
    _initialized = true;
}

XMLNode::XMLNode(const string &n, const string &c)
 : _name(n), _content(c) {
  _initialized = true;
}


XMLNode::XMLNode(const XMLNode &node) {
  *this = node;
}



const XMLNode& XMLNode::operator=(const XMLNode& node) {
  erase(); //remove all old nodes and properties

  _name = node.name();
  set_content(node.content());

  XMLNodeConstIterator curnode;
  XMLPropertyConstIterator curprop;

  for(curprop = node.properties().begin(); curprop != node.properties().end(); ++curprop)
    add_property((*curprop)->name(), (*curprop)->value());

  for(curnode = node.children().begin(); curnode != node.children().end(); ++curnode)
    add_child(new XMLNode(**curnode));

  return *this;
}
bool XMLNode::operator==(const XMLNode& ) {
  return false;
}

XMLNode::~XMLNode() {
  erase();
}

void XMLNode::erase() {
  XMLNodeIterator curchild;
  XMLPropertyIterator curprop;

  for(curchild = _children.begin(); curchild != _children.end(); ++curchild)
    delete *curchild;

  for(curprop = _proplist.begin(); curprop != _proplist.end(); ++curprop)
    delete *curprop;
}

const string & XMLNode::set_content(const string &c) {
  _content = c;

  return _content;
}

XMLNodeList XMLNode::children(const string &n) const {
  XMLNodeList retval;

  if(n.length() == 0)
    return _children;

  for(XMLNodeConstIterator cur = _children.begin(); cur != _children.end(); ++cur) {
    /*Begin> If name matches, insert the child */
    if((*cur)->name() == n)
      retval.push_back(*cur);
    /*End< */

    /*Begin> Else search the rest of the children for a given name*/
    else {
      XMLNodeList tmp_list = (*cur)->children(n);
      retval.insert(retval.end(), tmp_list.begin(), tmp_list.end());
    }
    /*End< */
  }

  return retval;
}

XMLNode *XMLNode::add_child(const string &n) {
  XMLNode *tmp;

  tmp = new XMLNode(n);

  return add_child(tmp);
}

XMLNode *XMLNode::add_child(XMLNode *n) {
  if(!n)
    return NULL;

  _children.insert(_children.end(), n);
  return n;
}

XMLNode *XMLNode::add_child(const string &n, int pos) {
  XMLNode *tmp;
  
  tmp = new XMLNode(n);

  if(!tmp)
    return NULL;
  XMLNodeIterator i = _children.begin();
  advance(i, pos);

  _children.insert(i, tmp);
  return tmp;
}

XMLNode *XMLNode::add_content(const string &c) {
  XMLNode *tmp;

  tmp = new XMLNode(string(), c);

  return add_child(tmp);
}

string XMLNode::getData() {
  XMLNodeList pcdata = children("text");
  XMLNodeIterator i = pcdata.begin();
  if ( i != pcdata.end() ) {
    XMLNode* i_data = *i;
    return( i_data->content() );
  }
  return ("");
}

void XMLNode::remove_child(XMLNode *n) {
  if(n)
    _children.remove(n);
}

XMLProperty *XMLNode::property(const string &n) {
  if(_prophash.find(n) == _prophash.end())
    return NULL;

  return _prophash[n];
}

XMLProperty *XMLNode::add_property(const string &n, const string &v) {
  XMLProperty *tmp;

  if(_prophash.find(n) != _prophash.end())
    return NULL;

  tmp = new XMLProperty(n, v);

  if(!tmp)
    return NULL;

  _prophash[tmp->name()] = tmp;
  _proplist.insert(_proplist.end(), tmp);

  return tmp;
}

void XMLNode::remove_property(const string &n) {
  if(_prophash.find(n) != _prophash.end()) {
    _proplist.remove(_prophash[n]);                   //memory leak!! Property is not deleted;
    _prophash.erase(n);
  }
}

static XMLNode *readnode(xmlNodePtr node) {
  string name, content;
  xmlNodePtr child;
  XMLNode *tmp;
  xmlAttrPtr attr;

  if(node->name)
    name = (char *) node->name;
  else
    name = string();

  tmp = new XMLNode(name);

  for(attr = node->properties; attr; attr = attr->next) {
    name = (char *) attr->name;
    content = "";
    if(attr->children)
      content = (char *) attr->children->content;
    tmp->add_property(name, content);
  }

  if(node->content)
    tmp->set_content((char *) node->content);
  else
    tmp->set_content(string());

  for(child = node->children; child; child = child->next)
      tmp->add_child(readnode(child));

  return tmp;
}


static void writenode(xmlDocPtr doc, XMLNode *n, xmlNodePtr p, int root ) {
  XMLPropertyList props;
  XMLPropertyIterator curprop;
  XMLNodeList children;
  XMLNodeIterator curchild;
  xmlNodePtr node;

  if(root)
    node = doc->children = xmlNewDocNode(doc, NULL,
        (xmlChar *) n->name().c_str(),
        (xmlChar *) n->content().c_str());
  else
    node = xmlNewChild(p, NULL, (xmlChar *) n->name().c_str(),
      (xmlChar *) n->content().c_str());

  if(n->is_content()) {
    node->type = XML_TEXT_NODE;
    xmlNodeSetContentLen(node, (const xmlChar *) n->content().c_str(),
        n->content().length());
  }

  props = n->properties();
  for(curprop = props.begin(); curprop != props.end(); curprop++)
    xmlSetProp(node, (xmlChar *) (*curprop)->name().c_str(),
    (xmlChar *) (*curprop)->value().c_str());

  children = n->children();
  for(curchild = children.begin(); curchild != children.end(); curchild++)
    writenode(doc, *curchild, node);
}

string XMLEncoder::getUTF8(const string &s) {

  xmlChar * out;
  int ret;
  int size;
  int out_size;
  int temp;
  xmlCharEncodingHandlerPtr handler;
  
  if (s == "") return string("");
  
  handler = xmlFindCharEncodingHandler(_encoding.c_str());
  
  if (!handler) {
    //cerr << "ConvertInput: no encoding handler found for " << _encoding << endl;
    return string("");
  }

    size = (int) s.size() + 1;
    out_size = size * 2 - 1;
    out = (unsigned char *) xmlMalloc((size_t) out_size);

    if (out != 0) {
        temp = size - 1;
        ret = handler->input(out, &out_size, (const xmlChar *)(s.c_str()), &temp);
        if (ret || temp - size + 1) {
          /*
            if (ret) {
	      cerr << "ConvertInput: conversion wasn't successful." << endl;
            } else {
	      cerr << "ConvertInput: conversion wasn't successful. converted: " << temp << "octets." << endl;
            }
          */
            xmlFree(out);
            out = 0;
        } else {
	  out = (unsigned char *) xmlRealloc(out, out_size + 1);
	  out[out_size] = 0;  /*null terminating out */
        }
        return string((char *)out);
    } else {
      //std::cerr << "ConvertInput: no mem" << endl;
      return ("");
    }
}

string XMLEncoder::getEncodedString(const string &s) {
  xmlChar * out;
  int ret;
  int size;
  int out_size;
  int temp;
  xmlCharEncodingHandlerPtr handler;
  
  if (s == "") return string("");
  handler = xmlFindCharEncodingHandler(_encoding.c_str());
  
  if (!handler) {
    //cerr << "ConvertInput: no encoding handler found for " << _encoding << endl;
    return string("");
  }

    size = (int) s.size() + 1;
    out_size = size * 2 - 1;
    out = (unsigned char *) xmlMalloc((size_t) out_size);

    if (out != 0) {
        temp = size - 1;
        ret = handler->output(out, &out_size, (const xmlChar *)(s.c_str()), &temp);
        if (ret || temp - size + 1) {
          /*
            if (ret) {
	      cerr << "ConvertInput: conversion wasn't successful." << endl;
            } else {
	      cerr << "ConvertInput: conversion wasn't successful. converted: " << temp << "octets." << endl;
            }
          */
            xmlFree(out);
            out = 0;
        } else {
	  out = (unsigned char *) xmlRealloc(out, out_size + 1);
	  out[out_size] = 0;  /*null terminating out */
        }
        return string((char *)out);
    } else {
      //std::cerr << "ConvertInput: no mem" << endl;
      return string("");
    }
}




